-- -------------------------------------------------------------------------- --
--         ____ _                ____                                         --
--        / ___| |__   __ _ _ __/ ___|  ___ __ _ _ __  _ __   ___ _ __        --
--       | |   | '_ \ / _` | '__\___ \ / __/ _` | '_ \| '_ \ / _ \ '__|       --
--       | |___| | | | (_| | |   ___) | (_| (_| | | | | | | |  __/ |          --
--        \____|_| |_|\__,_|_|  |____/ \___\__,_|_| |_|_| |_|\___|_|          --
--                                                                            --
-- -------------------------------------------------------------------------- --
--                            CharScanner by kunda                            --
-- -------------------------------------------------------------------------- --
--                                                                            --
-- CharScanner scans Basic Character Data, Talent Specifications, Equipment,  --
-- PvE/PvP Achievements, PvE Boss Kill Statistics and PvP Statistics from     --
-- single target or group and shows this data in a movable frame. Any scanned --
-- unit data is temporarily saved (in memory not in SavedVariables) and can   --
-- be viewed or deleted any time.                                             --
--                                                                            --
-- ONE click and ONE look shows the most important character data!            --
--                                                                            --
-- CharScanner does not rate, score or measure!                               --
-- CharScanner simply displays available ingame information in a nice way.    --
--                                                                            --
-- What you get in ~0.4 Seconds:                                              --
--  - A complete Talent scan (primary and secondary talent specification).    --
--  - A complete Gear scan.                                                   --
--  - A PvE dungeon and raid achievement/statistic scan.                      --
--  - A PvP achievement/statistic scan.                                       --
-- What you get in 50 Seconds if you scan your 25 player group:               --
--    ( 40 Seconds for Achievement/Talent scan restriction - Patch 3.3.5 )    --
--    ( 10 Seconds real scan time )                                           --
--  - All from above.                                                         --
--  - A nice Compare Tool.                                                    --
--                                                                            --
-- -------------------------------------------------------------------------- --
--                                                                            --
--      No SendAddonMessage(), no RealID/BN, no SavedVariables madness!       --
--                                                                            --
-- -------------------------------------------------------------------------- --
--                                                                            --
-- NOTES:                                                                     --
-- - Inspect range IS A MUST!                                                 --
-- - There is NO 'Inspect Glyphs API'! :-(                                    --
-- - Maximum time for one scan is: 1 Second (default).                        --
--  ! Change this value in the options menu if you have a high latency or   ! --
--  ! if you get a 'Scan incomplete!' status message if you scan yourself.  ! --
-- - GroupScan:                                                               --
--    - Raid roster changes between a 'GroupScan'-Button click and end of     --
--      groupscan are discarded!                                              --
--    - GroupScan will stop if:                                               --
--        All group/raid members has been scanned. or                         --
--        You enter combat! or                                                --
--        You click on: 'STOP SCAN'                                           --
-- - Entering combat OR entering/leaving an instance OR zoning (with loading  --
--   screen) will stop any scan.                                              --
-- - Mainrole: (Lex DRUID)                                                    --
--    - DRUID TANK = Thick Hide + Natural Reaction >= 4                       --
-- - LDB is supported.                                                        --
--    But: LibStub, CallbackHandler-1.0 & LibDataBroker-1.1 is not included.  --
--         This is intended and I will not change this!                       --
-- - Scan duration:                                                           --
--  ! Blizzard added a Achievement/Talent scan restriction with Patch 3.3.5 ! --
--  ! Only 6 scans in a row are allowed, than a 10-second break is required ! --
--    Scan duration is mainly latency dependent! My tests:                    --
--    - SINGLESCAN:                                                           --
--      ~0.4 Seconds for one unit.                                            --
--    - GROUPSCAN:                                                            --
--      - A group with 25 player is normally scanned in less than 50 Seconds  --
--        if all members are in inspect range.                                --
--          player  1- 6: ~2.4 Seconds                                        --
--          10-second break: Achievement/Talent scan restriction              --
--          player  7-12: ~2.4 Seconds                                        --
--          10-second break: Achievement/Talent scan restriction              --
--          player 13-18: ~2.4 Seconds                                        --
--          10-second break: Achievement/Talent scan restriction              --
--          player 18-24: ~2.4 Seconds                                        --
--          10-second break: Achievement/Talent scan restriction              --
--          player 24-25: ~2.4 Seconds                                        --
--        40 Seconds for Achievement/Talent scan restriction                  --
--       ~10 Seconds real scan time                                           --
--      =~50 Seconds for a complete 25p-group scan                            --
--                                                                            --
-- -------------------------------------------------------------------------- --
--                                                                            --
-- TODO:                                                                      --
--  - Some frame improvements (undock/resize/move/scale) ... after cata       --
--  - The GearFrame needs some love aka statsdetails ... after cata           --
--  - More Compare (nearly every data) ... after cata                         --
--  - GroupScan: group/other tab, scan missing members only ... after cata    --
--                                                                            --
-- -------------------------------------------------------------------------- --
--                                                                            --
-- INCORRECT DATA:                                                            --
-- - Statistic id 4046/4045 (Anub'arak) bug was fixed with Patch 3.3.3.xxxxx, --
--   but the old 4045 statistic value which is returned from the server is    --
--   still the same (old wrong cumulative 4046-one).                          --
--   -> This means:                                                           --
--      The statistic value returned for id 4045 is NOT 100% CORRECT (!!!)    --
--      (armory too!).                                                        --
--   Look at local function BossStatistic() for the workaround.               --
--                                                                            --
-- -------------------------------------------------------------------------- --
--                                                                            --
-- NEVER EVER:                                                                --
-- - CharScanner will never ever save scanned data to SavedVariables.         --
--   - Most data will change everyday, so it's completely useless to save     --
--     such data to SavedVariables.                                           --
--     The sheer amount of data makes such a feature a no go.                 --
-- - CharScanner will never ever autoscan a character or a group.             --
--   - You must actively initialize a single or group scan by clicking a      --
--     button.                                                                --
--     Triggering INSPECT_READY, INSPECT_TALENT_READY or                      --
--     INSPECT_ACHIEVEMENT_READY without user interaction is very bad and can --
--     cause massive lags.                                                    --
-- - CharScanner will never ever use SendAddonMessage()!!!                    --
--   - I hate SendAddonMessage()!                                             --
--   - EVERYBODY can fake any message that is send via SendAddonMessage().    --
--   - A messages from SendAddonMessage() is ALWAYS! an untrustable source!   --
--   - There is only one trustable source: The WoW server!                    --
-- - CharScanner will never ever use any RealID/BattleNet BN...() function!!! --
--   - NEVER EVER!!!                                                          --
--                                                                            --
-- -------------------------------------------------------------------------- --
--                                                                            --
-- slash commands: /cs - /scan - /charscanner                                 --
--        example: /scan         (CharScanner GUI & SINGLESCAN)               --
--        example: /scan options (CharScanner OPTIONS)                        --
--                                                                            --
-- -------------------------------------------------------------------------- --

-- ---------------------------------------------------------------------------------------------------------------------
CharScanner = CreateFrame("Frame")    -- container
CharScanner_Options = {}              -- SavedVariable options table

local _G                           = _G
local type                         = _G.type
local mod                          = _G.mod
local pairs                        = _G.pairs
local select                       = _G.select
local tonumber                     = _G.tonumber
local math_floor                   = _G.math.floor
local string_find                  = _G.string.find
local string_lower                 = _G.string.lower
local string_sub                   = _G.string.sub
local GetTalentTabInfo             = _G.GetTalentTabInfo
local GetTalentInfo                = _G.GetTalentInfo
local GetTalentLink                = _G.GetTalentLink
local GetStatistic                 = _G.GetStatistic
local GetComparisonStatistic       = _G.GetComparisonStatistic
local GetAchievementComparisonInfo = _G.GetAchievementComparisonInfo
local GetAchievementInfo           = _G.GetAchievementInfo
local GetCategoryNumAchievements   = _G.GetCategoryNumAchievements
local UnitGUID                     = _G.UnitGUID
local GetTime                      = _G.GetTime

local L = CharScanner_Locales         -- localization table
local PVEDUMP = CharScanner_PvE_Data  -- PvE achievement & statistic data table
local PVPDUMP = CharScanner_PvP_Data  -- PvP achievement & statistic data table

local mult2 = 10^(2 or 0)             -- x.xx
local mult4 = 10^(4 or 0)             -- x.xxx

local AchievementIDTable = {}         -- Achievement ID->name table (build after event PLAYER_LOGIN triggered)
local StatisticIDTable = {}           -- Statistic ID->name table (build after event PLAYER_LOGIN triggered)
local inCombat = false                -- false = out of combat | true = in combat
local inCombatGrpChk = false          -- party/raid member check if inCombat invite
local inPartyOrRaid = false           -- false = NOT in party or raid | true = in party or raid
local rosterhack = 0                  -- execute RAID_ROSTER_UPDATE block only if GetNumRaidMembers() has changed
local bugfix                          -- -> workaround for achievementFunctions.selectedCategory bug (Blizzard_AchievementUI.lua)

local SingleScanIsActive = false      -- false = SingleScan is NOT active | true = SingleScan is active
local SingleScanTimerStart = 0        -- Timer (single scan) start (time for one scan)
local GroupScanIsActive = false       -- false = GroupScan is NOT active | true = GroupScan is active
local GroupScanTimerStart = 0         -- Timer (group scan) start (time for whole group scan)
local GroupScanNext = false           --
local GroupScanFirstGUID              -- value for GroupScan: first successful scan guid

local Event_INSPECT_ACHIEVEMENT_READY -- Achievement/Talent scan restriction: checkvalue
local Event_INSPECT_READY             -- Achievement/Talent scan restriction: checkvalue
local RestrictionTime = GetTime()     -- Achievement/Talent scan restriction: latest time event INSPECT_READY or INSPECT_ACHIEVEMENT_READY fired
local RestrictionAchNum = 0           -- Achievement/Talent scan restriction: count INSPECT_ACHIEVEMENT_READY event: 0 to 6
local RestrictionTalNum = 0           -- Achievement/Talent scan restriction: count INSPECT_READY event: 0 to 6
local RestrictionMaxTime = 10         -- Achievement/Talent scan restriction: max time to wait: 10 seconds
local RestrictionCountMax = 6         -- Achievement/Talent scan restriction: max scans in a row: 6
local RestrictionFirstTime            -- Achievement/Talent scan restriction: first time use (to allow scans after login)

local GUI_PvEstats_menu = {}              -- === PvE Achievement/Statistic collapse menu values
local GUI_PvPstats_menu = {}              -- === PvP Achievement/Statistic collapse menu values
local GUI_history_data = {}               -- === History (scanned units) data
local GUI_history_curRows = 25            -- current scroll frame rows
local GUI_history_minRows = 5             -- min. scroll frame rows
local GUI_history_maxRows = 25            -- max. scroll frame rows
local GUI_history_Count = 0               -- count
local GUI_history_SortBy = 1              -- sort by: 1 = class | 2 = name | 3 = time | 4 = mainrole | 5 = talents1 | 6 = talents2 | 7 = ItemLevel | 8 = Achievement | 9 = Boss | 10 = Kills
local GUI_history_SortOrder = 1           -- sort order: 0 = asc | 1 = desc
local GUI_history_RaidL1 = 0              -- raid level1
local GUI_history_RaidL2                  -- raid level2
local GUI_history_RaidL3                  -- raid level3
local GUI_achievementdetail_data = {}     -- === Achievement Detail data
local GUI_achievementdetail_curRows = 25  -- current scroll frame rows
local GUI_achievementdetail_minRows = 5   -- min. scroll frame rows
local GUI_achievementdetail_maxRows = 25  -- max. scroll frame rows
local GUI_achievementdetail_Count = 0     -- count
local GUI_achievementdetail_SortBy = 1    -- sort by: 1 = id | 2 = name
local GUI_achievementdetail_SortOrder = 1 -- sort order: 0 = asc | 1 = desc
local GUI_achievementdetail_Select = 1    -- 1 = ALL | 2 = COMPLETE | 3 = INCOMPLETE
local GUI_achievementdetail_Compare       -- id for Single Achievement Compare
local GUI_achievementdetail_L1            -- level1
local GUI_achievementdetail_L2            -- level2
local GUI_achievementdetail_L3            -- level3
local GUI_PvPachievementdetail_L1         -- PvP level1
local GUI_PvPachievementdetail_L2         -- PvP level2
local GUI_PvPachievementdetail_L3         -- PvP level3
local GUI_ABKsearch_data = {}             -- === ABK (stats) search result data
local GUI_ABKsearch_Count = 0             -- count
local GUI_ABKsearch_SortBy = 2            -- sort by: 1 = standard | 2 = name | 3 = kills
local GUI_ABKsearch_SortOrder = 1         -- sort order: 0 = asc | 1 = desc
local GUI_ABKsearch_curRows = 20          -- current scroll frame rows
local GUI_ABKsearch_minRows = 5           -- min. scroll frame rows
local GUI_ABKsearch_maxRows = 40          -- max. scroll frame rows

local GUIselectedGUID                 -- current shown char
local NumScannedChars = 0             -- number of scanned characters
local playerGUID = ""                 -- player GUID (set after event PLAYER_LOGIN triggered)
local playerFaction = ""              -- player Faction Horde/Alliance (set after event PLAYER_LOGIN triggered)
local IsUnitInpected = false          -- scanned unit: true: "Inspect" | false: "Character" (it's me)
local SCANUNIT                        -- current scan Unit
local SCANGUID                        -- current scan GUID
local SCAN = {}                       -- scanned unit data (key = GUID)
local LatestScanIsComplete            -- true = last scan is complete | false/nil = last scan is NOT complete
local BossSearchQuery                 --

local TalentTreeFrameIsShown          -- for mouseover
local GearFrameIsShown                -- for mouseover
local HistoryFrameIsShown             --
local CompareFrameIsShown             --
local AchievementDetailFrameMode      --

local myMAX_TIER = 7                  -- max. tier
local myMAX_COLUMN = 4                -- Talent Tree column
local myMAX_TALENT_TABS = 3           -- Blizzard default is 3 too
local myMAX_NUM_TALENTS = 28          -- Blizzard default is 28 too

-- magical 180 degree texture cut center rotation
local function rt(H,E,M,P)
	return E,P,E,M,H,P,H,M
end

local Textures = {
	CharScanner       = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-button.tga"}, -- Textures.CharScanner.path
	CharScannerHL     = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-button-highlight.tga"}, -- Textures.CharScannerHL.path
	CharScannerGUI    = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-button-GUI.tga"}, -- Textures.CharScannerGUI.path
	CharScannerIcons  = {path    = "Interface\\Addons\\CharScanner\\CharScanner-images-icons.tga"}, -- Textures.CharScannerIcons.path
	MinimapButton     = {--path    = CharScannerIcons -- Textures.MinimapButton
	                     coords    =    { 96/128, 128/128,   0/128,  32/128},
	                     width     = 32,
	                     height    = 32},
	-- --------------------------------------------------------------------------------
	ButtonHL          = {--path    = CharScannerIcons -- Textures.ButtonHL.path
	                     coords    =    {109/128, 126/128,  37/128,  54/128}},
	SingleScan        = {--path    = CharScannerIcons -- Textures.SingleScan.path
	                     coords    =    {  1/128,  32/128,  55/128,  72/128},
	                     width     = 31,
	                     height    = 17},
	GroupScan         = {--path    = CharScannerIcons -- Textures.GroupScan.path
	                     coords    =    {  1/128,  43/128,  73/128,  90/128},
	                     width     = 42,
	                     height    = 17},
	GroupScanAdd      = {--path    = CharScannerIcons -- Textures.GroupScanAdd.path
	                     coords    =    { 33/128,  64/128,  55/128,  72/128},
	                     width     = 31,
	                     height    = 17},
	GroupScanMissing  = {--path    = CharScannerIcons -- Textures.GroupScanMissing.path
	                     coords    =    { 65/128,  96/128,  55/128,  72/128},
	                     width     = 31,
	                     height    = 17},
	Delete            = {--path    = CharScannerIcons -- Textures.Delete.path
	                     coords    =    { 55/128,  72/128,   1/128,  18/128},
	                     width     = 17,
	                     height    = 17},
	History           = {--path    = CharScannerIcons -- Textures.History.path
	                     coords    =    {  1/128,  18/128,   1/128,  18/128},
	                     width     = 17,
	                     height    = 17},
	Compare           = {--path    = CharScannerIcons -- Textures.Compare.path
	                     coords    =    { 19/128,  36/128,   1/128,  18/128},
	                     width     = 17,
	                     height    = 17},
	Options           = {--path    = CharScannerIcons -- Textures.Options.path
	                     coords    =    { 91/128, 108/128,  37/128,  54/128},
	                     width     = 17,
	                     height    = 17},
	Close             = {--path    = CharScannerIcons -- Textures.Close.path
	                     coords    =    { 73/128,  90/128,   1/128,  18/128},
	                     width     = 17,
	                     height    = 17},
	Expand            = {--path    = CharScannerIcons -- Textures.Expand.path
	                     coords    =    { 37/128,  54/128,   1/128,   18/128},
	                     width     = 17,
	                     height    = 17},
	Collapse          = {--path    = CharScannerIcons -- Textures.Collapse.path
	                     coords    = {rt( 37/128,  54/128,   1/128,  18/128)},
	                     width     = 17,
	                     height    = 17},
	RightSide         = {--path    = CharScannerIcons -- Textures.RightSide.path
	                     coords    =    { 97/128, 110/128,  55/128,  72/128},
	                     coordsOff = {rt( 97/128, 110/128,  55/128,  72/128)},
	                     width     = 13,
	                     height    = 17},
	FreeIcon          = {--path    = CharScannerIcons -- Textures.FreeIcon.path
	                     coords    =    { 55/128,  72/128,  37/128,  54/128},
	                     width     = 17,
	                     height    = 17},
	DockedIcon        = {--path    = CharScannerIcons -- Textures.DockedIcon.path
	                     coords    =    { 73/128,  90/128,  37/128,  54/128},
	                     width     = 17,
	                     height    = 17},
	-- --------------------------------------------------------------------------------
	ResizeUpDown      = {--path    = CharScannerIcons -- Textures.ResizeUpDown.path
	                     coords    =    { 19/128,  36/128,  37/128,  54/128}},
	                     --qsize   = 17 -- sizeGeneral.SliderSize = 17
	ResizeCorner      = {--path    = CharScannerIcons -- Textures.ResizeCorner.path
	                     coordsL   =    { 37/128,  54/128,  37/128,  54/128},
	                     coordsR   = {rt( 37/128,  54/128,  37/128,  54/128)}},
	                     --qsize   = 17 -- sizeGeneral.SliderSize = 17
	-- --------------------------------------------------------------------------------
	PlusMinus         = {--path      = CharScannerIcons -- Textures.PlusMinus.path
	                     coordsPlus  =  { 92/128,  99/128, 113/128, 120/128},
	                     coordsMinus =  { 92/128,  99/128, 121/128, 128/128},
	                     qsize       = 7},
	AchievementDetail = {--path    = CharScannerIcons -- Textures.AchievementDetail.path
	                     coords    =    {112/128, 128/128, 104/128, 128/128},
	                     width     = 8,  -- 16 scaled
	                     height    = 12},-- 24 scaled
	RaidSelect        = {--path    = CharScannerIcons -- Textures.RaidSelect.path
	                     coords    =    { 19/128,  36/128,   1/128,  18/128},
	                     width     = 12, -- 17 scaled
	                     height    = 12},-- 17 scaled
	BossSelect        = {--path    = "Interface\\TargetingFrame\\UI-TargetingFrame-Skull", -- Textures.BossSelect.path
	                     coords    = {3/32, 27/32, 0/32, 32/32},
	                     width     = 9,  -- 24 scaled 1234
	                     height    = 12},-- 32 scaled
	SortArrow         = {--path    = CharScannerIcons -- Textures.SortArrow.path
	                     coords1   =    { 62/128,  69/128, 120/128, 128/128},
	                     coords2   = {rt( 62/128,  69/128, 120/128, 128/128)},
	                     width     = 7,
	                     height    = 8},
	SelectionArrow    = {--path    = CharScannerIcons -- Textures.SelectionArrow.path
	                     coords    =    { 82/128,  91/128, 113/128, 128/128},
	                     width     = 7.2,-- scaled  9  7.2  6
	                     height    = 12},-- scaled 15 12   10
	Trashcan          = {--path    = CharScannerIcons -- Textures.Trashcan.path
	                     coords    =    { 70/128,  81/128, 116/128, 128/128},
	                     width     = 11,
	                     height    = 12},
	-- --------------------------------------------------------------------------------
	ClassIcon         = {path    = "Interface\\WorldStateFrame\\Icons-Classes"}, -- Textures.ClassIcon.path
	                     --coords  = auto
	                     --width   = depends,
	                     --height  = depends},
	RoleIcon          = {[1] = {path = "Interface\\LFGFrame\\LFGRole",            coords = {32/64, 48/64, 0/16, 16/16}}, -- TANK  -- Textures.RoleIcon.path
	                     [2] = {path = "Interface\\LFGFrame\\LFGRole",            coords = {48/64, 64/64, 0/16, 16/16}}, -- HEAL
	                     [3] = {path = "Interface\\LFGFrame\\LFGRole",            coords = {16/64, 32/64, 0/16, 16/16}}, -- DAMAGE
	                     [4] = {path = "Interface\\GossipFrame\\ActiveQuestIcon", coords = {0, 1, 0, 1}},                -- UNKNOWN
	                     [5] = {path = "Interface\\LFGFrame\\LFGRole",            coords = {0/64, 1/64, 0/16, 1/16}},    -- NOTALENTS
	                     sWidth  = 16, -- status
	                     sHeight = 16, -- status
	                     width   = 12, -- 16 scaled
	                     height  = 12},-- 16 scaled
	TalentMiniIcons   = {--path    = auto, -- Textures.TalentMiniIcons.path
	                     coords    = {5/64, 59/64, 5/64, 59/64},
	                     qsize     = 14},
	TalentIconTexture = {--path    = auto -- Textures.TalentIconTexture.path
	                     --coords  = auto
	                     qsize     = 24},
	TalentRankBorder  = {--path    = CharScannerIcons -- Textures.TalentRankBorder.path
	                     coords    =    {100/128, 111/128, 117/128, 128/128},
	                     qsize     = 11},
	GearIconTexture   = {--path    = auto -- Textures.GearIconTexture.path
	                     --coords  = not used - default = {0, 1, 0, 1}
	                     qsize     = 24},
	Enchantment       = {path      = "Interface\\Icons\\Trade_Engraving", -- Textures.Enchantment.path
	                     --coords  = not used - default = {0, 1, 0, 1}
	                     qsize     = 12},
	PvP               = {path      = "Interface\\GossipFrame\\BattleMasterGossipIcon", -- Textures.PvP.path
	                     --coords  = not used - default = {0, 1, 0, 1}
	                     qsize     = 16},
	SearchBoxIcons    = {--path    = CharScannerIcons -- Textures.SearchBoxIcons.path
	                     coordsL   =    { 55/128,  60/128,  19/128,  36/128},
	                     coordsM   =    { 63/128,  64/128,  19/128,  36/128},
	                     coordsR   =    { 67/128,  72/128,  19/128,  36/128},
	                     coordsCLR =    { 73/128,  90/128,  19/128,  37/128},
	                     width     = 17,
	                     height    = 17},
	-- --------------------------------------------------------------------------------
	SliderKnob        = {--path    = CharScannerIcons -- Textures.SliderKnob.path
	                     coords    =    { 50/128,  61/128, 111/128, 128/128},
	                     width     = 11,
	                     height    = 17},
	SliderBG          = {--path    = CharScannerIcons -- Textures.SliderBG.path
	                     coordsL   =    { 62/128,  67/128, 109/128, 115/128},
	                     coordsM   =    { 70/128,  71/128, 109/128, 115/128},
	                     coordsR   =    { 74/128,  79/128, 109/128, 115/128}},
	ScrollFrameUp     = {--path    = CharScannerIcons -- Textures.ScrollFrameUp.path
	                     coords    =    {  1/128,  18/128,  19/128,  36/128},
	                     coordsHL  =    { 19/128,  36/128,  19/128,  36/128},
	                     coordsDIS =    { 37/128,  54/128,  19/128,  36/128}},
	                     --qsize   = 17 -- sizeGeneral.SliderSize = 17
	ScrollFrameDown   = {--path    = CharScannerIcons -- Textures.ScrollFrameDown.path
	                     coords    = {rt(  1/128,  18/128,  19/128,  36/128)},
	                     coordsHL  = {rt( 19/128,  36/128,  19/128,  36/128)},
	                     coordsDIS = {rt( 37/128,  54/128,  19/128,  36/128)}},
	                     --qsize   = 17 -- sizeGeneral.SliderSize = 17
	ScrollFrameKnob   = {--path    = CharScannerIcons -- Textures.ScrollFrameKnob.path
	                     coords    =    {  1/128,  18/128,  37/128,  54/128}},
	                     --qsize   = 17 -- sizeGeneral.SliderSize = 17
}

-- fonts
local fonts = {
	GameFontNormal         = {name = "GameFontNormal"},         -- fonts.GameFontNormal.name
	GameFontNormalSmall    = {name = "GameFontNormalSmall"},    -- fonts.GameFontNormalSmall.name
	GameFontNormalLarge    = {name = "GameFontNormalLarge"},    -- fonts.GameFontNormalLarge.name
	GameFontWhiteSmall     = {name = "GameFontWhiteSmall"},     -- fonts.GameFontWhiteSmall.name
	GameFontHighlight      = {name = "GameFontHighlight"},      -- fonts.GameFontHighlight.name
	GameFontHighlightSmall = {name = "GameFontHighlightSmall"}, -- fonts.GameFontHighlightSmall.name
}
for key, value in pairs(fonts) do
	local _, fontHeight = _G[value.name]:GetFont()
	value.height = math_floor(fontHeight+0.5)
end

-- size General
local sizeGeneral = {}
sizeGeneral.MainWidth    = 450
sizeGeneral.Border       = 1
sizeGeneral.Space        = 2
sizeGeneral.Offset       = 5
sizeGeneral.SliderSize   = 17
sizeGeneral.BarHeight    = fonts.GameFontNormalSmall.height + sizeGeneral.Space -- 10+2=12
sizeGeneral.StatusHeight = sizeGeneral.Offset + fonts.GameFontNormalLarge.height + sizeGeneral.Offset -- 5++5=
sizeGeneral.StatusProgressBarWidth = 300

sizeGeneral.MainButtonHeight = sizeGeneral.Border + sizeGeneral.Space + Textures.SingleScan.height + sizeGeneral.Space + sizeGeneral.Border -- 23 - SingleScan.height is default button height
sizeGeneral.MainHeight       = sizeGeneral.Border + sizeGeneral.Offset + sizeGeneral.MainButtonHeight + sizeGeneral.Offset + sizeGeneral.Border -- 35
sizeGeneral.FreeBarHeight    = sizeGeneral.Border + sizeGeneral.Space + 12 + sizeGeneral.Space + sizeGeneral.Border -- 18

sizeGeneral.TabButtonHeight = sizeGeneral.Border + sizeGeneral.Space + fonts.GameFontNormalSmall.height + sizeGeneral.Space + sizeGeneral.Border -- 16
sizeGeneral.TabFrameHeight  = sizeGeneral.Border + sizeGeneral.Border + sizeGeneral.Space + sizeGeneral.TabButtonHeight -- 1 + 1 + 2 + 16 = 20
sizeGeneral.TabOffset       = 10 -- offset between the tabs -> autocalc | tab1 | TabOffset | tab2 | (TabOffset | tab3 |...) autocalc

-- size BaseCharFrame
local sizeBaseCharFrame_BaseHeight = sizeGeneral.Offset + fonts.GameFontNormal.height + sizeGeneral.Space + fonts.GameFontNormalSmall.height + sizeGeneral.Space + fonts.GameFontNormalSmall.height + sizeGeneral.Offset -- 46

-- size Talent-/TalentTreeFrame and Equipment-/GearFrame
local sizeTalentFrame = {}
local sizeEquipmentFrame = {}

sizeTalentFrame.BaseHeight           = sizeGeneral.Offset + Textures.TalentMiniIcons.qsize + Textures.TalentMiniIcons.qsize + sizeGeneral.Offset -- 5 + 12 + 12 + 5 = 34
sizeTalentFrame.TreeTopTextHeight    = fonts.GameFontWhiteSmall.height + sizeGeneral.Space + fonts.GameFontWhiteSmall.height + sizeGeneral.Space -- 10 + 2 + 10 + 2 == 24
sizeTalentFrame.TreeHeight           = sizeTalentFrame.TreeTopTextHeight + (myMAX_TIER*Textures.TalentIconTexture.qsize) + ((myMAX_TIER-1)*sizeGeneral.Offset) -- 24 + (11*20) + (10*5) = 294
sizeTalentFrame.TreeWidth            = (myMAX_COLUMN*Textures.TalentIconTexture.qsize) + ((myMAX_COLUMN-1)*sizeGeneral.Offset) -- (4*20) + (3*5) = 95

-- make them same size before this vars are used in other vars START
sizeTalentFrame.TreeFrameWidth       = (3*sizeTalentFrame.TreeWidth) + (4*sizeGeneral.TabOffset) -- (3*95) + (4*10) = 325 
sizeEquipmentFrame.GearWidth     = sizeGeneral.TabOffset + Textures.Enchantment.qsize + sizeGeneral.Offset + 30 + sizeGeneral.Offset + (8*Textures.GearIconTexture.qsize) + sizeGeneral.Offset + 30 + sizeGeneral.Offset + Textures.Enchantment.qsize + sizeGeneral.TabOffset
if sizeTalentFrame.TreeFrameWidth > sizeEquipmentFrame.GearWidth then
	sizeEquipmentFrame.GearWidth = sizeTalentFrame.TreeFrameWidth
else
	sizeTalentFrame.TreeFrameWidth = sizeEquipmentFrame.GearWidth
end
-- make them same size before this vars are used in other vars END

sizeTalentFrame.TreeFrameHeight      = sizeGeneral.TabFrameHeight + sizeGeneral.Offset + sizeTalentFrame.TreeHeight + sizeGeneral.TabOffset -- 20 + 5 + 294 + 10 = 329
sizeTalentFrame.Tree1Left            = sizeGeneral.TabOffset -- 10
sizeTalentFrame.Tree2Left            = sizeTalentFrame.Tree1Left + sizeTalentFrame.TreeWidth + sizeGeneral.TabOffset --  10 + 95 + 10 = 115
sizeTalentFrame.Tree3Left            = sizeTalentFrame.Tree2Left + sizeTalentFrame.TreeWidth + sizeGeneral.TabOffset -- 115 + 95 + 10 = 220
sizeTalentFrame.GroupSelectionWidth  = (sizeTalentFrame.TreeFrameWidth/2) - (sizeGeneral.Offset+(sizeGeneral.Offset/2)) -- (325/2) - (5+(5/2)) = 155
sizeTalentFrame.GroupSelection1Left  = sizeGeneral.Offset
sizeTalentFrame.GroupSelection2Left  = sizeGeneral.Offset + sizeTalentFrame.GroupSelectionWidth + sizeGeneral.Offset

sizeEquipmentFrame.BaseHeight    = sizeGeneral.Offset + fonts.GameFontNormalSmall.height + sizeGeneral.Space + sizeGeneral.Space + fonts.GameFontNormalSmall.height + sizeGeneral.Space + sizeGeneral.Space + fonts.GameFontNormalSmall.height + sizeGeneral.Offset -- 48
sizeEquipmentFrame.BaseHeightXXL = sizeGeneral.Offset + fonts.GameFontNormalSmall.height + sizeGeneral.Space + sizeGeneral.Space + fonts.GameFontNormalSmall.height + sizeGeneral.Space + sizeGeneral.Space + fonts.GameFontNormalSmall.height + fonts.GameFontNormalSmall.height + sizeGeneral.Offset -- 58
sizeEquipmentFrame.GearHeight    = sizeGeneral.Offset + (8*Textures.GearIconTexture.qsize) + (Textures.GearIconTexture.qsize/2) + sizeGeneral.Offset + fonts.GameFontNormalSmall.height + sizeGeneral.Offset + Textures.Enchantment.qsize + sizeGeneral.Offset-- 260 -- 5 + (8*25) + 12.5 + 5 + 12 + 5 + 12 + 5
sizeEquipmentFrame.FirstSlotLeft = (sizeEquipmentFrame.GearWidth-(8*Textures.GearIconTexture.qsize))/2 -- 112.5 -- (400-(7*25))/2

-- size Achievement & Statistics
local sizeStatFrame = {}
sizeStatFrame.TabButtonWidth  = 120
sizeStatFrame.TabButton1Left  = (sizeGeneral.MainWidth - sizeStatFrame.TabButtonWidth - sizeStatFrame.TabButtonWidth - sizeGeneral.TabOffset)/2
sizeStatFrame.TabButton2Left  = sizeStatFrame.TabButton1Left + sizeStatFrame.TabButtonWidth + sizeGeneral.TabOffset

sizeStatFrame.OffsetTopSearch = sizeGeneral.Border + sizeGeneral.Space + Textures.SearchBoxIcons.height + sizeGeneral.Space -- 1+2+17+2 = 22
sizeStatFrame.BaseHeight      = sizeGeneral.TabFrameHeight + sizeStatFrame.OffsetTopSearch + (#PVEDUMP*sizeGeneral.BarHeight) + sizeGeneral.Offset -- 20+(3*12)+5 = 63
sizeStatFrame.LevelModifier   = Textures.PlusMinus.qsize + sizeGeneral.Space -- 7+2 = 9
sizeStatFrame.ButtonWidth     = sizeGeneral.MainWidth - sizeGeneral.Offset - sizeGeneral.Space - sizeGeneral.SliderSize - sizeGeneral.Space - sizeGeneral.Border
sizeStatFrame.SearchBoxWidth  = 80

sizeStatFrame.RaidSelectRight        = 0
sizeStatFrame.RaidSelectWidth        = Textures.RaidSelect.width
sizeStatFrame.BossesTotalRight       = sizeStatFrame.RaidSelectRight + sizeStatFrame.RaidSelectWidth + sizeGeneral.Space
sizeStatFrame.BossesTotalWidth       = 50
sizeStatFrame.StatisticRight         = sizeStatFrame.BossesTotalRight + sizeStatFrame.BossesTotalWidth + sizeGeneral.Space
sizeStatFrame.StatisticWidth         = 60
sizeStatFrame.AchievementDetailRight = sizeStatFrame.StatisticRight + sizeStatFrame.StatisticWidth + sizeGeneral.Space
sizeStatFrame.AchievementDetailWidth = Textures.AchievementDetail.width
sizeStatFrame.AchievementRight       = sizeStatFrame.AchievementDetailRight + sizeStatFrame.AchievementDetailWidth + sizeGeneral.Space
sizeStatFrame.AchievementWidth       = 60
sizeStatFrame.AchievementRightTitle  = sizeStatFrame.AchievementDetailRight + sizeStatFrame.AchievementDetailWidth + sizeGeneral.Space -- only title
sizeStatFrame.AchievementWidthTitle  = 80 -- only title

sizeStatFrame.NameLeft        = 0
sizeStatFrame.NameWidth       = sizeStatFrame.ButtonWidth - (sizeGeneral.Space + sizeStatFrame.AchievementWidth + sizeStatFrame.AchievementRight)

sizeStatFrame.NameWidthTitle  = sizeStatFrame.ButtonWidth - (sizeGeneral.Space + sizeStatFrame.AchievementWidthTitle + sizeStatFrame.AchievementRightTitle)
sizeStatFrame.NameWidthSearch = sizeStatFrame.ButtonWidth - (sizeGeneral.Space + sizeStatFrame.BossesTotalWidth + sizeStatFrame.BossesTotalRight)
sizeStatFrame.SearchSortStdLeft    = Textures.SearchBoxIcons.width + sizeGeneral.Space + sizeStatFrame.SearchBoxWidth + sizeGeneral.Space + Textures.SearchBoxIcons.width + sizeGeneral.Space + 40
sizeStatFrame.SearchSortStdWidth   = (sizeStatFrame.ButtonWidth - sizeStatFrame.SearchSortStdLeft - sizeGeneral.Space - sizeStatFrame.BossesTotalWidth - sizeStatFrame.BossesTotalRight - sizeGeneral.Space)/2
sizeStatFrame.SearchSortNameLeft   = sizeStatFrame.SearchSortStdLeft + sizeStatFrame.SearchSortStdWidth + sizeGeneral.Space
sizeStatFrame.SearchSortNameWidth  = sizeStatFrame.SearchSortStdWidth
sizeStatFrame.SearchSortKillsWidth = sizeStatFrame.BossesTotalWidth

sizeStatFrame.NameLeft_L1c    = sizeStatFrame.LevelModifier
sizeStatFrame.NameWidth_L1c   = sizeStatFrame.NameWidth - sizeStatFrame.LevelModifier
sizeStatFrame.NameLeft_L2     = sizeStatFrame.LevelModifier
sizeStatFrame.NameWidth_L2    = sizeStatFrame.NameWidth_L1c + sizeGeneral.Space + sizeStatFrame.AchievementWidth + sizeGeneral.Space + sizeStatFrame.AchievementDetailWidth + sizeGeneral.Space + sizeStatFrame.StatisticWidth
sizeStatFrame.NameLeft_L2c    = sizeStatFrame.LevelModifier * 2
sizeStatFrame.NameWidth_L2c   = sizeStatFrame.NameWidth_L1c - sizeStatFrame.LevelModifier
sizeStatFrame.NameLeft_L3     = sizeStatFrame.LevelModifier * 2
sizeStatFrame.NameWidth_L3    = sizeStatFrame.NameWidth_L2 - sizeStatFrame.LevelModifier
sizeStatFrame.NameLeft_L3c    = sizeStatFrame.LevelModifier * 3
sizeStatFrame.NameWidth_L3c   = sizeStatFrame.NameWidth_L2c - sizeStatFrame.LevelModifier
sizeStatFrame.NameLeft_L4     = sizeStatFrame.LevelModifier * 3
sizeStatFrame.NameWidth_L4    = sizeStatFrame.NameWidth_L3 - sizeStatFrame.LevelModifier

sizeStatFrame.SF_curHeight    = (GUI_ABKsearch_curRows*sizeGeneral.BarHeight) -- (20*12)
sizeStatFrame.SFpvp_curHeight = (GUI_ABKsearch_curRows*sizeGeneral.BarHeight) -- (20*12)
sizeStatFrame.Stats_curHeight = (GUI_ABKsearch_curRows*sizeGeneral.BarHeight) + sizeGeneral.TabFrameHeight + sizeStatFrame.OffsetTopSearch + sizeGeneral.Offset -- (20*12) + 20 + 5
sizeStatFrame.Stats_minHeight = (GUI_ABKsearch_minRows*sizeGeneral.BarHeight) + sizeGeneral.TabFrameHeight + sizeStatFrame.OffsetTopSearch + sizeGeneral.Offset -- (5*12) + 20 + 5
sizeStatFrame.Stats_maxHeight = (GUI_ABKsearch_maxRows*sizeGeneral.BarHeight) + sizeGeneral.TabFrameHeight + sizeStatFrame.OffsetTopSearch + sizeGeneral.Offset -- (40*12) + 20 + 5

sizeStatFrame.PvPHonorableKillsRight    = sizeStatFrame.RaidSelectRight + sizeStatFrame.RaidSelectWidth + sizeGeneral.Space
sizeStatFrame.PvPHonorableKillsWidth    = 50
sizeStatFrame.PvPWinLossPercentRight    = sizeStatFrame.PvPHonorableKillsRight + sizeStatFrame.PvPHonorableKillsWidth + sizeGeneral.Space
sizeStatFrame.PvPWinLossPercentWidth    = 50
sizeStatFrame.PvPWinLossRight           = sizeStatFrame.PvPWinLossPercentRight + sizeStatFrame.PvPWinLossPercentWidth + sizeGeneral.Space
sizeStatFrame.PvPWinLossWidth           = 70
sizeStatFrame.PvPAchievementDetailRight = sizeStatFrame.PvPWinLossRight + sizeStatFrame.PvPWinLossWidth + sizeGeneral.Space
sizeStatFrame.PvPAchievementDetailWidth = Textures.AchievementDetail.width
sizeStatFrame.PvPAchievementRight       = sizeStatFrame.PvPAchievementDetailRight + sizeStatFrame.PvPAchievementDetailWidth + sizeGeneral.Space
sizeStatFrame.PvPAchievementWidth       = 60
sizeStatFrame.PvPAchievementRightTitle  = sizeStatFrame.PvPAchievementDetailRight + sizeStatFrame.PvPAchievementDetailWidth + sizeGeneral.Space -- only title
sizeStatFrame.PvPAchievementWidthTitle  = 80 -- only title
sizeStatFrame.PvPWinLossPerRightTitle   = sizeStatFrame.PvPHonorableKillsRight + sizeStatFrame.PvPHonorableKillsWidth + sizeGeneral.Space -- only title
sizeStatFrame.PvPWinLossPerWidthTitle   = sizeStatFrame.PvPWinLossPercentWidth + sizeGeneral.Space + sizeStatFrame.PvPWinLossWidth -- only title

sizeStatFrame.PvPNameLeft        = 0
sizeStatFrame.PvPNameWidth       = sizeStatFrame.ButtonWidth - (sizeGeneral.Space + sizeStatFrame.PvPAchievementWidth + sizeStatFrame.PvPAchievementRight)

sizeStatFrame.PvPNameWidthTitle  = sizeStatFrame.ButtonWidth - (sizeGeneral.Space + sizeStatFrame.PvPAchievementWidthTitle + sizeStatFrame.PvPAchievementRightTitle)

sizeStatFrame.PvPNameLeft_L1c    = sizeStatFrame.LevelModifier
sizeStatFrame.PvPNameWidth_L1c   = sizeStatFrame.PvPNameWidth - sizeStatFrame.LevelModifier
sizeStatFrame.PvPNameLeft_L2     = sizeStatFrame.LevelModifier
sizeStatFrame.PvPNameWidth_L2    = sizeStatFrame.PvPNameWidth_L1c + sizeGeneral.Space + sizeStatFrame.PvPAchievementWidth + sizeGeneral.Space + sizeStatFrame.PvPAchievementDetailWidth + sizeGeneral.Space + sizeStatFrame.PvPWinLossWidth + sizeGeneral.Space + sizeStatFrame.PvPWinLossPercentWidth
sizeStatFrame.PvPNameLeft_L2c    = sizeStatFrame.LevelModifier * 2
sizeStatFrame.PvPNameWidth_L2c   = sizeStatFrame.PvPNameWidth_L1c - sizeStatFrame.LevelModifier
sizeStatFrame.PvPNameLeft_L3     = sizeStatFrame.LevelModifier * 2
sizeStatFrame.PvPNameWidth_L3    = sizeStatFrame.PvPNameWidth_L2 - sizeStatFrame.LevelModifier
sizeStatFrame.PvPNameLeft_L3c    = sizeStatFrame.LevelModifier * 3
sizeStatFrame.PvPNameWidth_L3c   = sizeStatFrame.PvPNameWidth_L2c - sizeStatFrame.LevelModifier
sizeStatFrame.PvPNameLeft_L4     = sizeStatFrame.LevelModifier * 3
sizeStatFrame.PvPNameWidth_L4    = sizeStatFrame.PvPNameWidth_L3 - sizeStatFrame.LevelModifier

-- size AchievementDetailFrame
local sizeAchievementDetail = {}
sizeAchievementDetail.SortHeight   = 12
sizeAchievementDetail.IdLeft       = 0
sizeAchievementDetail.IdWidth      = 35
sizeAchievementDetail.NameLeft     = sizeAchievementDetail.IdLeft + sizeAchievementDetail.IdWidth + sizeGeneral.Space
sizeAchievementDetail.NameWidth    = 220
sizeAchievementDetail.CompareLeft  = sizeAchievementDetail.NameLeft + sizeAchievementDetail.NameWidth + sizeGeneral.Space
sizeAchievementDetail.CompareWidth = Textures.RaidSelect.width
sizeAchievementDetail.ButtonWidth  = sizeAchievementDetail.CompareLeft + sizeAchievementDetail.CompareWidth
sizeAchievementDetail.Width        = sizeGeneral.Offset + sizeAchievementDetail.ButtonWidth + sizeGeneral.Space + sizeGeneral.SliderSize + sizeGeneral.Space + sizeGeneral.Border
sizeAchievementDetail.SortLine1Top    = sizeGeneral.Offset
sizeAchievementDetail.SortLine1Height = 22
sizeAchievementDetail.SortLine2Top    = sizeAchievementDetail.SortLine1Top + sizeAchievementDetail.SortLine1Height + sizeGeneral.Space
sizeAchievementDetail.SortLine2Height = sizeGeneral.BarHeight
sizeAchievementDetail.SortLineXTop    = sizeAchievementDetail.SortLine2Top + sizeAchievementDetail.SortLine2Height + sizeGeneral.Space
sizeAchievementDetail.Height          = sizeAchievementDetail.SortLineXTop + (GUI_achievementdetail_curRows*sizeGeneral.BarHeight) + sizeGeneral.Offset
sizeAchievementDetail.InfoTexture       = 24
sizeAchievementDetail.InfoTextWidth     = sizeAchievementDetail.Width - (3*sizeGeneral.Offset) - sizeAchievementDetail.InfoTexture

-- size SummaryFrame
local sizeSummaryFrame_BaseHeight = sizeGeneral.Offset + sizeGeneral.BarHeight + sizeGeneral.BarHeight + sizeGeneral.Offset

-- size History/CompareFrame
local sizeCompareFrame = {}
sizeCompareFrame.RaidSelectLeft   = sizeGeneral.Space + sizeGeneral.SliderSize + sizeGeneral.Space + sizeGeneral.Border
sizeCompareFrame.TrashcanLeft     = 0
sizeCompareFrame.TrashcanWidth    = Textures.Trashcan.width
sizeCompareFrame.ArrowLeft        = sizeCompareFrame.TrashcanLeft + sizeCompareFrame.TrashcanWidth + sizeGeneral.Space
sizeCompareFrame.ArrowWidth       = Textures.SelectionArrow.width
sizeCompareFrame.ClassiconLeft    = sizeCompareFrame.ArrowLeft + sizeCompareFrame.ArrowWidth + sizeGeneral.Space
sizeCompareFrame.ClassiconWidth   = sizeGeneral.BarHeight
sizeCompareFrame.NameLeft         = sizeCompareFrame.ClassiconLeft + sizeCompareFrame.ClassiconWidth + sizeGeneral.Space
sizeCompareFrame.NameWidth        = 120
sizeCompareFrame.TimeLeft         = sizeCompareFrame.NameLeft + sizeCompareFrame.NameWidth + sizeGeneral.Space
sizeCompareFrame.TimeWidth        = 40
sizeCompareFrame.MainroleLeft     = sizeCompareFrame.TimeLeft + sizeCompareFrame.TimeWidth + sizeGeneral.Space
sizeCompareFrame.MainroleWidth    = 12
sizeCompareFrame.Talents1Left     = sizeCompareFrame.MainroleLeft + sizeCompareFrame.MainroleWidth + sizeGeneral.Space
sizeCompareFrame.Talents1Width    = 150
sizeCompareFrame.Talents2Left     = sizeCompareFrame.Talents1Left + sizeCompareFrame.Talents1Width + sizeGeneral.Space
sizeCompareFrame.Talents2Width    = 40
sizeCompareFrame.EnchLeft         = sizeCompareFrame.Talents2Left + sizeCompareFrame.Talents2Width + sizeGeneral.Space
sizeCompareFrame.EnchWidth        = 25
sizeCompareFrame.PvPLeft          = sizeCompareFrame.EnchLeft + sizeCompareFrame.EnchWidth + sizeGeneral.Space
sizeCompareFrame.PvPWidth         = 25
sizeCompareFrame.EquipLeft        = sizeCompareFrame.PvPLeft + sizeCompareFrame.PvPWidth + sizeGeneral.Space
sizeCompareFrame.EquipWidth       = 50
sizeCompareFrame.AchievementLeft  = sizeCompareFrame.EquipLeft + sizeCompareFrame.EquipWidth + sizeGeneral.Space
sizeCompareFrame.AchievementWidth = 80
sizeCompareFrame.BossLeft         = sizeCompareFrame.AchievementLeft + sizeCompareFrame.AchievementWidth + sizeGeneral.Space
sizeCompareFrame.BossWidth        = 80
sizeCompareFrame.KillsLeft        = sizeCompareFrame.BossLeft + sizeCompareFrame.BossWidth + sizeGeneral.Space
sizeCompareFrame.KillsWidth       = 50

sizeCompareFrame.SingleAchievementCompareLeft  = sizeCompareFrame.AchievementLeft
sizeCompareFrame.SingleAchievementCompareWidth = sizeCompareFrame.AchievementWidth + sizeGeneral.Space + sizeCompareFrame.BossWidth + sizeGeneral.Space + sizeCompareFrame.KillsWidth

sizeCompareFrame.HistoryButtonWidth = sizeCompareFrame.MainroleLeft + sizeCompareFrame.MainroleWidth
sizeCompareFrame.HistoryWidth       = sizeGeneral.Offset + sizeCompareFrame.HistoryButtonWidth + sizeGeneral.Space + sizeGeneral.SliderSize + sizeGeneral.Space + sizeGeneral.Border
sizeCompareFrame.CompareButtonWidth = sizeCompareFrame.KillsLeft + sizeCompareFrame.KillsWidth
sizeCompareFrame.CompareWidth       = sizeGeneral.Offset + sizeCompareFrame.CompareButtonWidth + sizeGeneral.Space + sizeGeneral.SliderSize + sizeGeneral.Space + sizeGeneral.Border
sizeCompareFrame.SortLine1Top     = sizeGeneral.Offset
sizeCompareFrame.SortLine1Height  = 22
sizeCompareFrame.SortLine2Top     = sizeCompareFrame.SortLine1Top + sizeCompareFrame.SortLine1Height + sizeGeneral.Space
sizeCompareFrame.SortLine2Height  = sizeGeneral.BarHeight
sizeCompareFrame.SortLineXTop     = sizeCompareFrame.SortLine2Top + sizeCompareFrame.SortLine2Height + sizeGeneral.Space
sizeCompareFrame.HistoryHeight    = sizeCompareFrame.SortLineXTop + (GUI_history_curRows*sizeGeneral.BarHeight) + sizeGeneral.Offset

--local DefaultPvEStatsViewData = { -- PvE WotLK default view
--	[1] = {lvl = 1, name1 = "WotLK", index1 = 3, name2 = nil,       index2 = nil, name3 = nil, index3 = nil},
--	[2] = {lvl = 2, name1 = "WotLK", index1 = 3, name2 = "raid10p", index2 = 2,   name3 = nil, index3 = nil},
--	[3] = {lvl = 2, name1 = "WotLK", index1 = 3, name2 = "raid25p", index2 = 3,   name3 = nil, index3 = nil}
--}
local DefaultPvEStatsViewData = { -- PvE Cataclysm default view
	[1] = {lvl = 1, name1 = "Cataclysm", index1 = 4, name2 = nil,        index2 = nil, name3 = nil, index3 = nil},
	[2] = {lvl = 2, name1 = "Cataclysm", index1 = 4, name2 = "dungeons", index2 = 1,   name3 = nil, index3 = nil},
	[3] = {lvl = 2, name1 = "Cataclysm", index1 = 4, name2 = "raids",    index2 = 2,   name3 = nil, index3 = nil}
}

local DefaultPvPStatsViewData = { -- PvP default view
	[1] = {lvl = 1, name1 = "Battlegrounds", index1 = 2, name2 = nil, index2 = nil, name3 = nil, index3 = nil},
	[2] = {lvl = 1, name1 = "Arena",         index1 = 3, name2 = nil, index2 = nil, name3 = nil, index3 = nil},
	[3] = {lvl = 1, name1 = "WorldPvP",      index1 = 4, name2 = nil, index2 = nil, name3 = nil, index3 = nil}
}

local color = { -- e.g.: color.White.hex | color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4]
	DarkGrey            = {rgba = {0.2,  0.2,  0.2,  1    }, hex = "333333"}, -- color.DarkGrey.
	Grey                = {rgba = {0.5,  0.5,  0.5,  1    }, hex = "808080"}, -- color.Grey.
	Red                 = {rgba = {1,    0,    0,    1    }, hex = "ff0000"}, -- color.Red.
	White               = {rgba = {1,    1,    1,    1    }, hex = "ffffff"}, -- color.White.
	Level1              = {rgba = {1,    1,    1,    1    }, hex = "ffffff"}, -- color.Level1.
	Level2              = {rgba = {0.8,  0.8,  0.8,  1    }, hex = "cccccc"}, -- color.Level2.
	Level3              = {rgba = {0.6,  0.6,  0.6,  1    }, hex = "999999"}, -- color.Level3.
	LightYellow         = {rgba = {1,    1,    0.49, 1    }, hex = "ffff7f"}, -- color.LightYellow.
	Yellow              = {rgba = {1,    0.82, 0,    1    }, hex = "ffd100"}, -- color.Yellow.
	Green               = {rgba = {0,    0.75, 0,    1    }, hex = "00bf00"}, -- color.Green.
	Brown               = {rgba = {0.78, 0.52, 0.28, 1    }, hex = "c78547"}, -- color.Brown.
	Firebrick           = {rgba = {0.8,  0.00, 0.00, 1    }, hex = "cd0000"}, -- color.Firebrick.
	Role         = {[1] = {rgba = {0.78, 0.52, 0.28, 1    }, hex = "c78547"}, -- color.Role[1]. - Brown     - TANK
	                [2] = {rgba = {0,    0.75, 0,    1    }, hex = "00bf00"}, -- color.Role[2]. - Green     - HEAL
	                [3] = {rgba = {0.8,  0.00, 0.00, 1    }, hex = "cd0000"}, -- color.Role[3]. - Firebrick - DAMAGE
	                [4] = {rgba = {1,    0.82, 0,    1    }, hex = "ffd100"}},-- color.Role[4]. - Yellow    - UNKNOWN
	FrameBG             = {rgba = {0,    0,    0,    1    }},                 -- color.FrameBG.
	FrameBorder         = {rgba = {0.4,  0.4,  0.4,  1    }},                 -- color.FrameBorder.
	ButtonBG            = {rgba = {0,    0,    0,    1    }},                 -- color.ButtonBG.
	ButtonBGActive      = {rgba = {0.25, 0.25, 0.25, 1    }},                 -- color.ButtonBGActive.
	ButtonHighlight     = {rgba = {1,    1,    1,    0.2  }},                 -- color.ButtonHighlight.
	ButtonHighlightLow  = {rgba = {0.6,  0.6,  0.6,  0.2  }},                 -- color.ButtonHighlightLow.
	ButtonHighlightRed  = {rgba = {0.55, 0,    0,    0.5  }},                 -- color.ButtonHighlightRed.
	ButtonBorder        = {rgba = {0.4,  0.4,  0.4,  1    }},                 -- color.ButtonBorder.
	ButtonBorderDisable = {rgba = {0.2,  0.2,  0.2,  1    }},                 -- color.ButtonBorderDisable.
	ButtonSortBG        = {rgba = {1,    1,    1,    0.1  }},                 -- color.ButtonSortBG.
	ButtonLineMarker    = {rgba = {0.06, 0.06, 0.06, 1    }},                 -- color.ButtonLineMarker.
}

local RoleTextShort = {
	[1] = "|cff"..color.Role[1].hex..L["T"].."|r", -- TANK
	[2] = "|cff"..color.Role[2].hex..L["H"].."|r", -- HEAL
	[3] = "|cff"..color.Role[3].hex..L["D"].."|r", -- DAMAGE
	[4] = "|cff"..color.Role[4].hex..L["?"].."|r"  -- UNKNOWN
}

local classcolors = RAID_CLASS_COLORS
local classes = {-- the classes table is populated with talent data if a class is scanned the first time
	DEATHKNIGHT = {icon = {0.265625, 0.484375, 0.515625, 0.734375}, -- (68/256, 124/256, 132/256, 188/256)
	               roleNum = {[1] = 1,   -- TANK   | Blood
	                          [2] = 3,   -- DAMAGE | Frost
	                          [3] = 3}}, -- DAMAGE | Unholy
	DRUID       = {icon = {0.7578125, 0.9765625, 0.015625, 0.234375}, -- (194/256, 250/256, 4/256, 60/256)
	               roleNum = {[1] = 3,   -- DAMAGE | Balance
	                          [2] = 1,   -- TANK   | FeralCombat --> Thick Hide + Natural Reaction < 4 = DAMAGE
	                          [3] = 2}}, -- HEAL   | Restoration
	HUNTER      = {icon = {0.01953125, 0.23828125, 0.265625, 0.484375}, -- (5/256, 61/256, 68/256, 124/256)
	               roleNum = {[1] = 3,   -- DAMAGE | BeastMastery
	                          [2] = 3,   -- DAMAGE | Marksmanship
	                          [3] = 3}}, -- DAMAGE | Survival
	MAGE        = {icon = {0.265625, 0.484375, 0.015625, 0.234375}, -- (68/256, 124/256, 4/256, 60/256)
	               roleNum = {[1] = 3,   -- DAMAGE | Arcane
	                          [2] = 3,   -- DAMAGE | Fire
	                          [3] = 3}}, -- DAMAGE | Frost
	PALADIN     = {icon = {0.01953125, 0.23828125, 0.515625, 0.734375}, -- (5/256, 61/256, 132/256, 188/256)
	               roleNum = {[1] = 2,   -- HEAL   | Holy
	                          [2] = 1,   -- TANK   | Protection
	                          [3] = 3}}, -- DAMAGE | Combat
	PRIEST      = {icon = {0.51171875, 0.73046875, 0.265625, 0.484375}, -- (131/256, 187/256, 68/256, 124/256)
	               roleNum = {[1] = 2,   -- HEAL   | Discipline
	                          [2] = 2,   -- HEAL   | Holy
	                          [3] = 3}}, -- DAMAGE | Shadow
	ROGUE       = {icon = {0.51171875, 0.73046875, 0.015625, 0.234375}, -- (131/256, 187/256, 4/256, 60/256)
	               roleNum = {[1] = 3,   -- DAMAGE | Assassination
	                          [2] = 3,   -- DAMAGE | Combat
	                          [3] = 3}}, -- DAMAGE | Subtlety
	SHAMAN      = {icon = {0.265625, 0.484375, 0.265625, 0.484375}, -- (68/256, 124/256, 68/256, 124/256)
	               roleNum = {[1] = 3,   -- DAMAGE | ElementalCombat
	                          [2] = 3,   -- DAMAGE | Enhancement
	                          [3] = 2}}, -- HEAL   | Restoration
	WARLOCK     = {icon = {0.7578125, 0.9765625, 0.265625, 0.484375}, -- (194/256, 250/256, 68/256, 124/256)
	               roleNum = {[1] = 3,   -- DAMAGE | Curses
	                          [2] = 3,   -- DAMAGE | Summoning
	                          [3] = 3}}, -- DAMAGE | Destruction
	WARRIOR     = {icon = {0.01953125, 0.23828125, 0.015625, 0.234375}, -- (5/256, 61/256, 4/256, 60/256)
	               roleNum = {[1] = 3,   -- DAMAGE | Arms
	                          [2] = 3,   -- DAMAGE | Fury
	                          [3] = 1}}, -- TANK   | Protection
}

local guildRank = {
	[1] = GUILD_RANK0_DESC, -- 0 | Guild Master
	[2] = GUILD_RANK1_DESC, -- 1 | Officer
	[3] = GUILD_RANK2_DESC, -- 2 | Veteran
	[4] = GUILD_RANK3_DESC, -- 3 | Member
	[5] = GUILD_RANK4_DESC  -- 4 | Initiate
}

-- Equipment
local slots = {
	[1] = "HeadSlot",
	[2] = "NeckSlot",
	[3] = "ShoulderSlot",
	[4] = "BackSlot",
	[5] = "ChestSlot",
	[6] = "WristSlot",
	[7] = "HandsSlot",
	[8] = "WaistSlot",
	[9] = "LegsSlot",
	[10] = "FeetSlot",
	[11] = "Finger0Slot",
	[12] = "Finger1Slot",
	[13] = "Trinket0Slot",
	[14] = "Trinket1Slot",
	[15] = "MainHandSlot",
	[16] = "SecondaryHandSlot",
	[17] = "RangedSlot"	
}
local slotsNum = #slots

local slotButton = {
	[1]  = {id =  1, sid =  1, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Head"},
	[2]  = {id =  2, sid =  2, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Neck"},
	[3]  = {id =  3, sid =  3, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Shoulder"},
	[4]  = {id =  4, sid = 15, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Chest"}, -- Back
	[5]  = {id =  5, sid =  5, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Chest"},
	[6]  = {id = -1, sid = -1, IQ = false, texture="Interface\\PaperDollInfoFrame\\UI-GearManager-LeaveItem-Opaque"}, -- Interface\\Paperdoll\\UI-PaperDoll-Slot-Shirt
	[7]  = {id = -1, sid = -1, IQ = false, texture="Interface\\PaperDollInfoFrame\\UI-GearManager-LeaveItem-Opaque"}, -- Interface\\Paperdoll\\UI-PaperDoll-Slot-Tabart
	[8]  = {id =  6, sid =  9, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Wrists"},
	[9]  = {id = 15, sid = 16, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-MainHand"}, -- middle
	[10] = {id = 16, sid = 17, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-SecondaryHand"}, -- middle
	[11] = {id = 17, sid = 18, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Ranged"}, -- middle
	[12] = {id = 14, sid = 14, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Trinket"},
	[13] = {id = 13, sid = 13, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Trinket"},
	[14] = {id = 12, sid = 12, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Finger"},
	[15] = {id = 11, sid = 11, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Finger"},
	[16] = {id = 10, sid =  8, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Feet"},
	[17] = {id =  9, sid =  7, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Legs"},
	[18] = {id =  8, sid =  6, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Waist"},
	[19] = {id =  7, sid = 10, IQ = true,  texture="Interface\\Paperdoll\\UI-PaperDoll-Slot-Hands"}
}

local itemQuality = {
	[1] = {name = ITEM_QUALITY0_DESC, color = "9d9d9d"}, -- 0 | poor
	[2] = {name = ITEM_QUALITY1_DESC, color = "ffffff"}, -- 1 | common
	[3] = {name = ITEM_QUALITY2_DESC, color = "1eff00"}, -- 2 | uncommon
	[4] = {name = ITEM_QUALITY3_DESC, color = "0070dd"}, -- 3 | rare
	[5] = {name = ITEM_QUALITY4_DESC, color = "a335ee"}, -- 4 | epic
	[6] = {name = ITEM_QUALITY5_DESC, color = "ff8000"}, -- 5 | legendary
	[7] = {name = ITEM_QUALITY6_DESC, color = "e6cc80"}, -- 6 | artifact (not in game)
	[8] = {name = ITEM_ACCOUNTBOUND,  color = "e6cc80"}, -- 7 | account bound
	[9] = {name = UNKNOWN,            color = "666666"}  -- 8 | unknown (item that is not in local cache or currently not availabe)
}
local itemQualityNum = #itemQuality
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
local function Print(...)
	print("|cff"..color.LightYellow.hex.."CharScanner:|r", ...)
end

local function ClassHexColor(class)
	local hex = string.format("%.2x%.2x%.2x", classcolors[class].r*255, classcolors[class].g*255, classcolors[class].b*255)
	return hex or "cccccc"
end

local function cnum(number)
	local a = tostring(number)
	local p = string_find(a, ".", 1, true)
	if p then
		local r = string_sub(a, p+1)
		if string.len(r) == 1 then
			number = number.."0"
		end
	else
		number = number..".00"
	end
	return number -- string
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
local function BuildIDTable()
	--[[ TEST code for future achievement/statistic updates
	for i=1, GetCategoryNumAchievements(-1) do
		local id, name = GetAchievementInfo(-1, i)--local id, name, _, _, _, _, _, description, _, icon, _ = GetAchievementInfo(-1, i)
		if id and name then
			if CharScanner_Options.AchievementTest == nil then CharScanner_Options.AchievementTest = {} end
			if not CharScanner_Options.AchievementTest[id] then
				CharScanner_Options.AchievementTest[id] = name
			end
		end
	end
	for i=1, GetCategoryNumAchievements(-2) do
		local id, name = GetAchievementInfo(-2, i)
		if id and name then
			if CharScanner_Options.StatisticTest == nil then CharScanner_Options.StatisticTest = {} end
			if not CharScanner_Options.StatisticTest[id] then
				CharScanner_Options.StatisticTest[id] = name
			end
		end
	end
	--]]

	-- achievement START
	-- - PvE START
	for i = 1, #PVEDUMP do
		if PVEDUMP[i].achievement then
			for key, value in pairs(PVEDUMP[i].achievement) do
				if value == 9999 then
					AchievementIDTable[4297] = select(2, GetAchievementInfo(4297))
					AchievementIDTable[4298] = select(2, GetAchievementInfo(4298))
				else
					AchievementIDTable[value] = select(2, GetAchievementInfo(value))
				end
			end
		end
		for j = 1, #PVEDUMP[i] do
			if PVEDUMP[i][j].achievement then
				for key, value in pairs(PVEDUMP[i][j].achievement) do
					if value == 9999 then
						AchievementIDTable[4297] = select(2, GetAchievementInfo(4297))
						AchievementIDTable[4298] = select(2, GetAchievementInfo(4298))
					else
						AchievementIDTable[value] = select(2, GetAchievementInfo(value))
					end
				end
			end
			for k = 1, #PVEDUMP[i][j] do
				if PVEDUMP[i][j][k].achievement then
					for key, value in pairs(PVEDUMP[i][j][k].achievement) do
						if value == 9999 then
							AchievementIDTable[4297] = select(2, GetAchievementInfo(4297))
							AchievementIDTable[4298] = select(2, GetAchievementInfo(4298))
						else
							AchievementIDTable[value] = select(2, GetAchievementInfo(value))
						end
					end
				end
			end
		end
	end
	-- - PvE END

	-- - PvP START
	for i = 1, #PVPDUMP do
		if PVPDUMP[i].achievement then
			for key, value in pairs(PVPDUMP[i].achievement) do
				if value.id then -- same id and same title
					AchievementIDTable[value.id] = select(2, GetAchievementInfo(value.id))
				elseif value.Aid then -- different ids and different title
					AchievementIDTable[value.Aid] = select(2, GetAchievementInfo(value.Aid))
					AchievementIDTable[value.Hid] = select(2, GetAchievementInfo(value.Hid))
				end
			end
		end
		for j = 1, #PVPDUMP[i] do
			if PVPDUMP[i][j].achievement then
				for key, value in pairs(PVPDUMP[i][j].achievement) do
					if value.id then -- same id and same title
						AchievementIDTable[value.id] = select(2, GetAchievementInfo(value.id))
					elseif value.Aid then -- different ids and different title
						AchievementIDTable[value.Aid] = select(2, GetAchievementInfo(value.Aid))
						AchievementIDTable[value.Hid] = select(2, GetAchievementInfo(value.Hid))
					end
				end
			end
		end
	end
	-- - PvP END
	-- achievement END

	-- statistic START
	-- PvE START
	for i = 1, #PVEDUMP do
		if PVEDUMP[i].statistic then
			for key, value in pairs(PVEDUMP[i].statistic) do
				StatisticIDTable[value] = select(2, GetAchievementInfo(value))
			end
		end
		for j = 1, #PVEDUMP[i] do
			if PVEDUMP[i][j].statistic then
				for key, value in pairs(PVEDUMP[i][j].statistic) do
					StatisticIDTable[value] = select(2, GetAchievementInfo(value))
				end
			end
			for k = 1, #PVEDUMP[i][j] do
				if PVEDUMP[i][j][k].statistic then
					for key, value in pairs(PVEDUMP[i][j][k].statistic) do
						StatisticIDTable[value] = select(2, GetAchievementInfo(value))
					end
				end
			end
		end
	end
	-- PvE END

	-- PvP START
	for i = 1, #PVPDUMP do
		if PVPDUMP[i].statistic then
			for key, value in pairs(PVPDUMP[i].statistic) do
				if value.matches then
					StatisticIDTable[value.matches] = select(2, GetAchievementInfo(value.matches))
				elseif value.wins then
					StatisticIDTable[value.wins] = select(2, GetAchievementInfo(value.wins))
				elseif value.kills then
					StatisticIDTable[value.kills] = select(2, GetAchievementInfo(value.kills))
				elseif value.misc then
					StatisticIDTable[value.misc] = select(2, GetAchievementInfo(value.misc))
				end
			end
		end
		for j = 1, #PVPDUMP[i] do
			if PVPDUMP[i][j].statistic then
				for key, value in pairs(PVPDUMP[i][j].statistic) do
					if value.matches then
						StatisticIDTable[value.matches] = select(2, GetAchievementInfo(value.matches))
					elseif value.wins then
						StatisticIDTable[value.wins] = select(2, GetAchievementInfo(value.wins))
					elseif value.kills then
						StatisticIDTable[value.kills] = select(2, GetAchievementInfo(value.kills))
					elseif value.misc then
						StatisticIDTable[value.misc] = select(2, GetAchievementInfo(value.misc))
					end
				end
			end
		end
	end
	-- PvP END
	-- statistic END

	--for key, value in pairs(AchievementIDTable) do print(key, value) end -- TEST
	--for key, value in pairs(StatisticIDTable) do print(key, value) end -- TEST
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
local function CreateScanDataByGUID(guid)
	SCAN[guid] = {}
	SCAN[guid].name = nil                      -- scanned unit: base info - name
	SCAN[guid].level = nil                     -- scanned unit: base info - level
	SCAN[guid].race = nil                      -- scanned unit: base info - race
	SCAN[guid].class = nil                     -- scanned unit: base info - class (localized)
	SCAN[guid].classEN = nil                   -- scanned unit: base info - class (english)
	SCAN[guid].faction = nil                   -- scanned unit: base info - faction (english)
	SCAN[guid].guildName = nil                 -- scanned unit: base info - guild name
	SCAN[guid].guildRankName = nil             -- scanned unit: base info - guild rank name
	SCAN[guid].guildRankIndex = nil            -- scanned unit: base info - guild rank index

	SCAN[guid].talentGroups = nil              -- scanned unit: talent - talent groups (count)
	SCAN[guid].activeTalentGroup = nil         -- scanned unit: talent - active talent group
	SCAN[guid].talents = {}                    -- scanned unit: all spend talent points
	SCAN[guid].mainRole = 5                    -- scanned unit: talent - main role | 1 = TANK, 2 = HEAL, 3 = DAMAGE, 4 = UNKNOWN (lvl>9 with all talentpoints unspent), 5 = NOTALENTS (lvl<=9)
	SCAN[guid].talentGroup1name = nil          -- scanned unit: talent - talent group 1 name with most points (eg: Affliction)
	SCAN[guid].talentGroup1role = nil          -- scanned unit: talent - talent group 1 role (class role: eg: PRIEST 2 -> HEAL)
	SCAN[guid].talentGroup1pointsString = nil  -- scanned unit: talent - talent group 1 points (eg: 55/11/6)
	SCAN[guid].talentGroup1pointsUnspent = nil -- scanned unit: talent - talent group 1 points unspent (eg: 0)
	SCAN[guid].talentGroup1icon = nil          -- scanned unit: talent - talent group 1 icon
	SCAN[guid].talentGroup2name = nil          -- scanned unit: talent - talent group 2 name with most points (eg: Destruction)
	SCAN[guid].talentGroup2role = nil          -- scanned unit: talent - talent group 2 role (class role: eg: PRIEST 3 -> DAMAGE)
	SCAN[guid].talentGroup2pointsString = nil  -- scanned unit: talent - talent group 2 points (eg: 3/10/58)
	SCAN[guid].talentGroup2pointsUnspent = nil -- scanned unit: talent - talent group 2 points unspent (eg: 1)
	SCAN[guid].talentGroup2icon = nil          -- scanned unit: talent - talent group 2 icon

	SCAN[guid].gear = {}                       -- scanned unit: itemLink from every gear slot
	--SCAN[guid].gearStats = {}                  -- scanned unit: basic gear stats from every gear slot --GEAR
	SCAN[guid].itemlevelAvg = 0                -- scanned unit: itemlevel - average item level
	SCAN[guid].itemlevelMin = 0                -- scanned unit: itemlevel - lowest item level
	SCAN[guid].itemlevelMax = 0                -- scanned unit: itemlevel - highest item level
	SCAN[guid].itemlevelMiss = 0               -- scanned unit: itemlevel - number of unused Slots
	SCAN[guid].itemlevelQuality = nil          -- scanned unit: itemlevel - quality
	SCAN[guid].gearEnchCount = 0               -- scanned unit: gear - Enchantment count
	SCAN[guid].gearPvPCount = 0                -- scanned unit: gear - PvP item count
	SCAN[guid].gearTwoHandWeapon = nil         -- scanned unit: gear - TwoHandWeapon check

	SCAN[guid].PVEDATA = {}                    -- scanned unit: PvE achievement & statistic data
	SCAN[guid].PvEAchievementTotal = 0         -- scanned unit: PvE summary: achievements - total achievements
	SCAN[guid].PvEAchievementCount = 0         -- scanned unit: PvE summary: achievements - count
	SCAN[guid].PvEStatisticTotal = 0           -- scanned unit: PvE summary: statistics
	SCAN[guid].PvEStatisticCount = 0           -- scanned unit: PvE summary: statistics
	SCAN[guid].PvEStatisticBossKills = 0       -- scanned unit: PvE summary: statistics

	SCAN[guid].PVPDATA = {}                    -- scanned unit: PvP achievement & statistic data
	SCAN[guid].PvPAchievementTotal = 0         -- scanned unit: PvP summary: achievements - total achievements
	SCAN[guid].PvPAchievementCount = 0         -- scanned unit: PvP summary: achievements - count
	SCAN[guid].PvPStatisticTotalKills = 0      -- scanned unit: PvP summary:?
	SCAN[guid].PvPStatisticWins = 0            -- scanned unit: PvP summary:?
	SCAN[guid].PvPStatisticMatches = 0         -- scanned unit: PvP summary:?
	SCAN[guid].PvPStatisticKills = 0           -- scanned unit: PvP summary:?
	SCAN[guid].PvPBattlegroundWins = 0         -- scanned unit: PvP summary:?
	SCAN[guid].PvPBattlegroundMatches = 0      -- scanned unit: PvP summary:?
	SCAN[guid].PvPBattlegroundKills = 0        -- scanned unit: PvP summary:?
	SCAN[guid].PvPArenaWins = 0                -- scanned unit: PvP summary:?
	SCAN[guid].PvPArenaMatches = 0             -- scanned unit: PvP summary:?
	SCAN[guid].PvPArenaKills = 0               -- scanned unit: PvP summary:?

	SCAN[guid].time = nil                      -- scanned unit: scan time (eg.: '22:19')
	SCAN[guid].scanTime = nil                  -- scanned unit: time that elapsed for scan (eg.: 0.362)
end

local function CleanTempData()
	for k, v in pairs(SCAN) do -- clean
		if not v.time then
			SCAN[k] = nil
		end
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
local function BossStatistic(id, inspect)
	-- workaround START
	-- bug       : Incorrect return value from GetComparisonStatistic()/GetStatistic().
	-- workaround: Check related achievement and return 0 or GetComparisonStatistic()/GetStatistic().
	-- NOTE      : Statistic id 4046/4045 (Anub'arak) bug was fixed with Patch 3.3.3.xxxxx, but the old 4045 statistic value which
	--             is returned from the server is still the same (old wrong cumulative 4046-one).
	--             -> This means: The statistic value returned for id 4045 is NOT 100% CORRECT (!!!) (armory too!).
	if id == 4045 then -- Times completed the Trial of the Grand Crusader (10 player)
		local x = GetAchievementComparisonInfo(3918, 1)
		if not x then
			return 0
		end
	end
	-- workaround END

	if inspect then
		return tonumber(GetComparisonStatistic(id)) or 0
	else
		return tonumber(GetStatistic(id)) or 0
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
local function ScrollFrameEnable(frameName)
	_G[frameName.."ScrollBarScrollUpButton"]:Enable()
	_G[frameName.."ScrollBarScrollDownButton"]:Enable()
	_G[frameName.."ScrollBarScrollUpButton"]:SetAlpha(1)
	_G[frameName.."ScrollBarScrollDownButton"]:SetAlpha(1)
	_G[frameName.."ScrollBarThumbTexture"]:SetAlpha(1)
end

local function ScrollFrameDisable(frameName)
	_G[frameName.."ScrollBarScrollUpButton"]:Disable()
	_G[frameName.."ScrollBarScrollDownButton"]:Disable()
	_G[frameName.."ScrollBarScrollUpButton"]:SetAlpha(0.2)
	_G[frameName.."ScrollBarScrollDownButton"]:SetAlpha(0.2)
	_G[frameName.."ScrollBarThumbTexture"]:SetAlpha(0.2)
end

local function ScrollFrameUpdate()
	-- 1 = PvE | 2 = PvP
	local statsHeight = math_floor(CharScanner_MainFrame_Stats:GetHeight() + 0.5)
	for i = 1, 2 do
		-- PvE START ---------------
		local scrollFrame = "CharScanner_MainFrame_PvEStats_ScrollFrame"
		local contentHeight = sizeStatFrame.SF_curHeight
		-- PvE END -----------------
		-- PvP START ---------------
		if i == 2 then
			scrollFrame = "CharScanner_MainFrame_PvPStats_ScrollFrame"
			contentHeight = sizeStatFrame.SFpvp_curHeight
		end
		-- PvP END -----------------
	
		local scrollbarValue = _G[scrollFrame.."ScrollBar"]:GetValue()
	
		if contentHeight > (statsHeight - sizeGeneral.TabFrameHeight - sizeStatFrame.OffsetTopSearch - sizeGeneral.Offset) then
			ScrollFrameEnable(scrollFrame)
		else
			ScrollFrameDisable(scrollFrame)
		end
	
		if scrollbarValue == 0 then
			_G[scrollFrame.."ScrollBarScrollUpButton"]:Disable()
		else
			_G[scrollFrame.."ScrollBarScrollUpButton"]:Enable()
		end
	
		if (scrollbarValue - contentHeight) == 0 then
			_G[scrollFrame.."ScrollBarScrollDownButton"]:Disable()
		else
			_G[scrollFrame.."ScrollBarScrollDownButton"]:Enable()
		end
	end
end

local function ResizeRowUpdate()
	local height = math_floor(CharScanner_MainFrame_Stats:GetHeight() + 0.5)
	local height_opt = (GUI_ABKsearch_curRows*sizeGeneral.BarHeight) + sizeGeneral.TabFrameHeight + sizeStatFrame.OffsetTopSearch + sizeGeneral.Offset -- (20*12) + 20 + 22 + 5
	if height < height_opt then
		local diff = height_opt - height
		local freq = math_floor(diff/12)
		if diff > 12 then
			local big = diff-(freq*12)
			if big > 6 then
				freq = freq + 1
			end
		else
			if diff < 6 then
				freq = freq - 1
			else
				freq = freq + 1
			end
		end
		if freq > 0 then
			GUI_ABKsearch_curRows = GUI_ABKsearch_curRows - freq
			if GUI_ABKsearch_curRows < GUI_ABKsearch_minRows then
				GUI_ABKsearch_curRows = GUI_ABKsearch_minRows
			elseif GUI_ABKsearch_curRows > GUI_ABKsearch_maxRows then
				GUI_ABKsearch_curRows = GUI_ABKsearch_maxRows
			end
			if BossSearchQuery then
				CharScanner:ABKsearch_FillList()
			end
		end
	else
		local diff = height - height_opt
		local freq = math_floor(diff/12)
		if diff > 12 then
			local big = diff-(freq*12)
			if big > 6 then
				freq = freq + 1
			end
		else
			if diff > 6 then
				freq = freq + 1
			else
				freq = freq - 1
			end
		end
		if freq > 0 then
			GUI_ABKsearch_curRows = GUI_ABKsearch_curRows + freq
			if GUI_ABKsearch_curRows < GUI_ABKsearch_minRows then
				GUI_ABKsearch_curRows = GUI_ABKsearch_minRows
			elseif GUI_ABKsearch_curRows > GUI_ABKsearch_maxRows then
				GUI_ABKsearch_curRows = GUI_ABKsearch_maxRows
			end
			if BossSearchQuery then
				CharScanner:ABKsearch_FillList()
			end
		end
	end
end

local function ResizeUpdate()
	CharScanner_MainFrame_Stats:ClearAllPoints()
	--if GearFrameIsShown then
	--	CharScanner_MainFrame_Stats:SetPoint("TOPLEFT", CharScanner_MainFrame_Gear, "BOTTOMLEFT", 0, sizeGeneral.Border)
	--else
		CharScanner_MainFrame_Stats:SetPoint("TOPLEFT", CharScanner_MainFrame_Equipment, "BOTTOMLEFT", 0, sizeGeneral.Border)
	--end
	CharScanner_MainFrame_Stats:SetWidth(sizeGeneral.MainWidth)

	local height = CharScanner_MainFrame_Stats:GetHeight() - sizeGeneral.TabFrameHeight -- height - 20
	CharScanner_MainFrame_PvEStats:SetHeight(height)
	CharScanner_MainFrame_PvPStats:SetHeight(height)

	if BossSearchQuery then
		ResizeRowUpdate()
	end
	ScrollFrameUpdate()
	CharScanner:MainFrame_CalculateHeight()--MHIGH
end

local function ResizeEnd()
	CharScanner_MainFrame_Stats:SetScript("OnUpdate", nil)
	CharScanner_MainFrame_Stats:StopMovingOrSizing()
	CharScanner_MainFrame_Stats:SetResizable(false)
	CharScanner_MainFrame_Stats:ClearAllPoints()
	--if GearFrameIsShown then
	--	CharScanner_MainFrame_Stats:SetPoint("TOPLEFT", CharScanner_MainFrame_Gear, "BOTTOMLEFT", 0, sizeGeneral.Border)
	--else
		CharScanner_MainFrame_Stats:SetPoint("TOPLEFT", CharScanner_MainFrame_Equipment, "BOTTOMLEFT", 0, sizeGeneral.Border)
	--end
	CharScanner_MainFrame_Stats:SetWidth(sizeGeneral.MainWidth)
	ResizeRowUpdate()

	local height = (GUI_ABKsearch_curRows*sizeGeneral.BarHeight) + sizeStatFrame.OffsetTopSearch + sizeGeneral.Offset -- (savedRows*12) + 16 + 22 + 5
	CharScanner_MainFrame_Stats:SetHeight(sizeGeneral.TabFrameHeight + height) -- 20 + height
	CharScanner_MainFrame_PvEStats:SetHeight(height) -- height
	CharScanner_MainFrame_PvPStats:SetHeight(height) -- height

	ScrollFrameUpdate()
	CharScanner_Options.StatsCurRows = GUI_ABKsearch_curRows
	CharScanner:MainFrame_CalculateHeight()--MHIGH
end

local function ResizeStart()
	CharScanner_MainFrame_Stats:SetResizable(true)
	CharScanner_MainFrame_Stats:SetScript("OnUpdate", ResizeUpdate)
	CharScanner_MainFrame_Stats:StartSizing()
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
local function MoveStart(self)
	local frameName = self:GetName()
	if frameName == "CharScanner_MainFrame_Status" or frameName == "CharScanner_MainFrame_Headline_ButtonMover" then
		CharScanner_MainFrame:StartMoving()
	elseif frameName == "CharScanner_OptionsFrame_Status" or frameName == "CharScanner_OptionsFrame_Headline_ButtonMover" then
		CharScanner_OptionsFrame:StartMoving()
	elseif frameName == "CharScanner_ScanButtonFrame_Mover" then
		CharScanner_ScanButtonFrame:StartMoving()
	end
end

local function MoveStop(self)
	local frameName = self:GetName()
	if frameName == "CharScanner_MainFrame_Status" or frameName == "CharScanner_MainFrame_Headline_ButtonMover" then
		CharScanner_MainFrame:StopMovingOrSizing()
		CharScanner:Frame_SavePosition("CharScanner_MainFrame")
	elseif frameName == "CharScanner_OptionsFrame_Status" or frameName == "CharScanner_OptionsFrame_Headline_ButtonMover" then
		CharScanner_OptionsFrame:StopMovingOrSizing()
		CharScanner:Frame_SavePosition("CharScanner_OptionsFrame")
	elseif frameName == "CharScanner_ScanButtonFrame_Mover" then
		CharScanner_ScanButtonFrame:StopMovingOrSizing()
		CharScanner:Frame_SavePosition("CharScanner_ScanButtonFrame")
	end
end

local function MoveFaker(self, button, frame)
	local frameName
	if frame then
		frameName = frame:GetName()
	else
		frameName = self:GetParent():GetName()
	end
	_G[frameName]:StartMoving()
	_G[frameName]:StopMovingOrSizing()
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
local function NodeDisplay(mode, button, state)
	local temptab = GUI_PvEstats_menu
	if mode == "PvP" then
		temptab = GUI_PvPstats_menu
	end

	if state then
		_G[button]:Show()
	else
		_G[button]:Hide()
	end
	for i = 1, #temptab do
		if temptab[i].name == button then
			temptab[i].collapse = state
			break
		end
	end

	local height = 0
	local visibleNode = {}
	local si = 1
	for i = 1, #temptab do
		if temptab[i].collapse then
			if temptab[i].slevel == 2 then
				visibleNode[si] = i
				si = si + 1
			elseif temptab[i].slevel == 3 then
				for j = 1, #temptab do
					if temptab[j].slevel == 2 and temptab[j].index1 == temptab[i].index1 and temptab[j].collapse then
						visibleNode[si] = i
						si = si + 1
					end
				end
			elseif temptab[i].slevel == 4 then

				for j = 1, #temptab do
					if temptab[j].slevel == 2 and temptab[j].index1 == temptab[i].index1 and temptab[j].collapse then--and
						for k = 1, #temptab do
							if temptab[k].slevel == 3 and temptab[k].index2 == temptab[i].index2 and temptab[k].collapse then
								visibleNode[si] = i
								si = si + 1
							end
						end
					end
				end

			end
		end
	end
	for i = 1, #visibleNode do
		height = height + temptab[ visibleNode[i] ].height
	end

	table.wipe(visibleNode)

	if mode == "PvE" then
		sizeStatFrame.SF_curHeight = height + (#PVEDUMP * sizeGeneral.BarHeight) -- height + 36
	else
		sizeStatFrame.SFpvp_curHeight = height + (#PVPDUMP * sizeGeneral.BarHeight) -- height + 48
	end
	ScrollFrameUpdate()
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
local function GearTooltipHandler(self)
	if not self.comparing and IsModifiedClick("COMPAREITEMS") then
		self.shoppingTooltips = {ShoppingTooltip1, ShoppingTooltip2, ShoppingTooltip3}
		GameTooltip_ShowCompareItem()
		self.comparing = true
	elseif self.comparing and not IsModifiedClick("COMPAREITEMS") then
		for _, frame in pairs(self.shoppingTooltips) do
			frame:Hide()
		end
		self.comparing = false
	end
	if IsModifiedClick("DRESSUP") then
		ShowInspectCursor()
	else
		ResetCursor()
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
local MiniOptions_TimerStart
local function MiniOptionsTimer(self)
	if GetTime() >= MiniOptions_TimerStart+3 then
		CharScanner_MiniOptionsFrame:Hide()
		self:SetScript("OnUpdate", nil)
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
local function SlashHandler(msg)
	msg = strtrim(string_lower(msg or ""))
	if msg == "options" then
		CharScanner:Frame_Toggle("CharScanner_OptionsFrame")
	else
		CharScanner:Frame_Toggle("CharScanner_MainFrame", 1)
		if not SingleScanIsActive and not GroupScanIsActive then
			CharScanner:ScanUnit("target")
		end
	end
end
-- -----------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
-- ----------------------------------------
--local function TemplateBorderTEST(frame)
--	local texture = frame:CreateTexture(nil, "BACKGROUND")
--	texture:SetPoint("TOPLEFT", 0, 0)
--	texture:SetPoint("BOTTOMRIGHT", 0, 0)
--	texture:SetTexture(0.5, 0.5, 0.5, 0.2)
--end
-- ----------------------------------------

-- Template BorderTRBL START ----------------------------------------
local function TemplateBorderTRBL(frame, named) -- TRBL = Top-Right-Bottom-Left
	local texture
	if named then
		texture = frame:CreateTexture(frame:GetName().."Border", "BORDER")
	else
		texture = frame:CreateTexture(nil, "BORDER")
	end
	texture:SetPoint("TOPLEFT", sizeGeneral.Border, -sizeGeneral.Border)
	texture:SetPoint("BOTTOMRIGHT", -sizeGeneral.Border, sizeGeneral.Border)
	texture:SetTexture(color.FrameBG.rgba[1], color.FrameBG.rgba[2], color.FrameBG.rgba[3], color.FrameBG.rgba[4])
	texture.Background = frame:CreateTexture(nil, "BACKGROUND")
	texture.Background:SetPoint("TOPLEFT", 0, 0)
	texture.Background:SetPoint("BOTTOMRIGHT", 0, 0)
	texture.Background:SetTexture(color.FrameBorder.rgba[1], color.FrameBorder.rgba[2], color.FrameBorder.rgba[3], color.FrameBorder.rgba[4])
end
-- Template BorderTRBL END ----------------------------------------

-- Template BorderRBL START ----------------------------------------
local function TemplateBorderRBL(frame) -- RBL = Right-Bottom-Left
	local name = frame:GetName()
	local texture = frame:CreateTexture(name.."Border", "BORDER")
	texture:SetPoint("TOPLEFT", sizeGeneral.Border, 0)
	texture:SetPoint("BOTTOMLEFT", 0, sizeGeneral.Border)
	texture:SetPoint("TOPRIGHT", -sizeGeneral.Border, 0)
	texture:SetPoint("BOTTOMRIGHT", -sizeGeneral.Border, -sizeGeneral.Border)
	texture:SetTexture(color.FrameBG.rgba[1], color.FrameBG.rgba[2], color.FrameBG.rgba[3], color.FrameBG.rgba[4])
	texture.Background = frame:CreateTexture(nil, "BACKGROUND")
	texture.Background:SetPoint("TOPLEFT", 0, 0)
	texture.Background:SetPoint("BOTTOMRIGHT", 0, 0)
	texture.Background:SetTexture(color.FrameBorder.rgba[1], color.FrameBorder.rgba[2], color.FrameBorder.rgba[3], color.FrameBorder.rgba[4])
end
-- Template BorderRBL END ----------------------------------------

-- Template BorderTcRBL START ----------------------------------------
local function TemplateBorderTcRBL(frame) -- TcRBL = Topcolor-Right-Bottom-Left
	local name = frame:GetName()
	local texture = frame:CreateTexture(name.."Border", "BORDER")
	texture:SetPoint("TOPLEFT", sizeGeneral.Border, 0)
	texture:SetPoint("BOTTOMLEFT", 0, sizeGeneral.Border)
	texture:SetPoint("TOPRIGHT", -sizeGeneral.Border, 0)
	texture:SetPoint("BOTTOMRIGHT", -sizeGeneral.Border, -sizeGeneral.Border)
	texture:SetTexture(color.FrameBG.rgba[1], color.FrameBG.rgba[2], color.FrameBG.rgba[3], color.FrameBG.rgba[4])
	local texturetop = frame:CreateTexture(name.."BorderTop", "BORDER")
	texturetop:SetPoint("TOPLEFT", sizeGeneral.Border, 0)
	texturetop:SetPoint("TOPRIGHT", -sizeGeneral.Border, -sizeGeneral.Border)
	texturetop:SetHeight(1)
	texturetop:SetTexture(color.DarkGrey.rgba[1], color.DarkGrey.rgba[2], color.DarkGrey.rgba[3], color.DarkGrey.rgba[4])
	texture.Background = frame:CreateTexture(nil, "BACKGROUND")
	texture.Background:SetPoint("TOPLEFT", 0, 0)
	texture.Background:SetPoint("BOTTOMRIGHT", 0, 0)
	texture.Background:SetTexture(color.FrameBorder.rgba[1], color.FrameBorder.rgba[2], color.FrameBorder.rgba[3], color.FrameBorder.rgba[4])
end
-- Template BorderTcRBL END ----------------------------------------

-- Template ScrollFrame START ----------------------------------------
local function TemplateButtonTexture(button, pos)
	local textureNormal = button:CreateTexture(nil, "ARTWORK")
	textureNormal:SetWidth(sizeGeneral.SliderSize)
	textureNormal:SetHeight(sizeGeneral.SliderSize)
	textureNormal:SetPoint("LEFT", 0, 0)
	textureNormal:SetTexture(Textures.CharScannerIcons.path)
	textureNormal:SetTexCoord(unpack(Textures[pos].coords))
	button:SetNormalTexture(textureNormal)

	local texturePush = button:CreateTexture(nil, "ARTWORK")
	texturePush:SetWidth(sizeGeneral.SliderSize)
	texturePush:SetHeight(sizeGeneral.SliderSize)
	texturePush:SetPoint("TOPLEFT", 1, -1)
	texturePush:SetPoint("BOTTOMRIGHT", -1, 1)
	texturePush:SetTexture(Textures.CharScannerIcons.path)
	texturePush:SetTexCoord(unpack(Textures[pos].coords))
	button:SetPushedTexture(texturePush)

	local textureHighlight = button:CreateTexture(nil, "ARTWORK")
	textureHighlight:SetWidth(sizeGeneral.SliderSize)
	textureHighlight:SetHeight(sizeGeneral.SliderSize)
	textureHighlight:SetPoint("LEFT", 0, 0)
	textureHighlight:SetTexture(Textures.CharScannerIcons.path)
	textureHighlight:SetTexCoord(unpack(Textures[pos].coordsHL))
	button:SetHighlightTexture(textureHighlight)

	local textureDisable = button:CreateTexture(nil, "ARTWORK")
	textureDisable:SetWidth(sizeGeneral.SliderSize)
	textureDisable:SetHeight(sizeGeneral.SliderSize)
	textureDisable:SetPoint("LEFT", 0, 0)
	textureDisable:SetTexture(Textures.CharScannerIcons.path)
	textureDisable:SetTexCoord(unpack(Textures[pos].coordsDIS))
	button:SetDisabledTexture(textureDisable)
end

local function TemplateScrollFrame(frame, child)
	local name = frame:GetName()

	frame.offset = 0
	frame:EnableMouseWheel(true)

	local scrollBar = CreateFrame("Slider", name.."ScrollBar", frame)
	scrollBar:SetWidth(sizeGeneral.SliderSize)
	scrollBar:SetHeight(0)
	scrollBar:SetPoint("TOPLEFT", frame, "TOPRIGHT", sizeGeneral.Space, -sizeGeneral.SliderSize)
	scrollBar:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", sizeGeneral.Space, sizeGeneral.SliderSize)
	scrollBar:SetMinMaxValues(0, 0)
	scrollBar:SetValue(0)
	scrollBar:SetOrientation("VERTICAL")
	scrollBar:EnableMouseWheel(true)
	scrollBar:SetScript("OnValueChanged", function(self, value)
		frame:SetVerticalScroll(value)
	end)
	scrollBar:SetScript("OnMouseWheel", function(self, delta)
		ScrollFrameTemplate_OnMouseWheel(frame, delta)
	end)

	local scrollBG = scrollBar:CreateTexture(nil, "BACKGROUND")
	scrollBG:SetAllPoints(scrollBar)
	scrollBG:SetTexture(0, 0, 0, 0)

	local upButton = CreateFrame("Button",  name.."ScrollBarScrollUpButton", scrollBar)
	upButton:SetWidth(sizeGeneral.SliderSize)
	upButton:SetHeight(sizeGeneral.SliderSize)
	upButton:SetPoint("BOTTOM", scrollBar, "TOP", 0, 0)
	upButton:SetScript("OnClick", function()
		scrollBar:SetValue(scrollBar:GetValue() - (scrollBar:GetHeight() / 2))
		PlaySound("UChatScrollButton")
	end)
	TemplateButtonTexture(upButton, "ScrollFrameUp")

	local downButton = CreateFrame("Button",  name.."ScrollBarScrollDownButton", scrollBar)
	downButton:SetWidth(sizeGeneral.SliderSize)
	downButton:SetHeight(sizeGeneral.SliderSize)
	downButton:SetPoint("TOP", scrollBar, "BOTTOM", 0, 0)
	downButton:SetScript("OnClick", function()
		scrollBar:SetValue(scrollBar:GetValue() + (scrollBar:GetHeight() / 2))
		PlaySound("UChatScrollButton")
	end)
	TemplateButtonTexture(downButton, "ScrollFrameDown")

	local thumb = scrollBar:CreateTexture(name.."ScrollBarThumbTexture", "BORDER")
	thumb:SetWidth(sizeGeneral.SliderSize)
	thumb:SetHeight(sizeGeneral.SliderSize)
	thumb:SetTexture(Textures.CharScannerIcons.path)
	thumb:SetTexCoord(unpack(Textures.ScrollFrameKnob.coords))
	scrollBar:SetThumbTexture(thumb)

	frame:SetScript("OnScrollRangeChanged", function(self, xrange, yrange)
		-- ScrollFrame_OnScrollRangeChanged(self, xrange, yrange)
		local scrollbar = _G[self:GetName().."ScrollBar"]
		if not yrange then
			yrange = self:GetVerticalScrollRange()
		end
		local value = scrollbar:GetValue()
		if value > yrange then
			value = yrange
		end
		scrollbar:SetMinMaxValues(0, yrange)
		scrollbar:SetValue(value)
		if floor(yrange) == 0 then
			_G[scrollbar:GetName().."ScrollDownButton"]:Disable()
			_G[scrollbar:GetName().."ScrollUpButton"]:Disable()
			_G[scrollbar:GetName().."ScrollDownButton"]:Show()
			_G[scrollbar:GetName().."ScrollUpButton"]:Show()
			--_G[scrollbar:GetName().."ThumbTexture"]:Hide()
		else
			_G[scrollbar:GetName().."ScrollDownButton"]:Show()
			_G[scrollbar:GetName().."ScrollUpButton"]:Show()
			_G[self:GetName().."ScrollBar"]:Show()
			--_G[scrollbar:GetName().."ThumbTexture"]:Show()
			-- The 0.005 is to account for precision errors
			if yrange - value > 0.005 then
				_G[scrollbar:GetName().."ScrollDownButton"]:Enable()
			else
				_G[scrollbar:GetName().."ScrollDownButton"]:Disable()
			end
		end
	end)
	frame:SetScript("OnMouseWheel", function(self, delta)
		ScrollFrameTemplate_OnMouseWheel(self, delta)
	end)
	frame:SetScript("OnVerticalScroll", function(self, off)
		scrollBar:SetValue(off)
	end)

	if child then
		local childFrame = CreateFrame("Frame",  name.."ScrollChildFrame", frame)
		childFrame:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, 0)
		childFrame:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", 0, 0)
		--childFrame:SetHeight(0)
		--childFrame:SetWidth(0)
		frame:SetScrollChild(childFrame)
	end
end
-- Template ScrollFrame END ----------------------------------------

-- Template TextButton START ----------------------------------------
local function DisableTextButton(name)
	_G[name.."Border"]:SetTexture(0.4, 0.4, 0.4, 1)
	_G[name]:Disable()
end
local function EnableTextButton(name, action)
	local buttoncolor
	if action == 1 then
		bordercolor = {0.73, 0.26, 0.21, 1}
	elseif action == 2 then
		bordercolor = {0.43, 0.32, 0.68, 1}
	elseif action == 3 then
		bordercolor = {0.24, 0.46, 0.21, 1}
	else
		bordercolor = {1, 1, 1, 1}
	end
	_G[name.."Border"]:SetTexture(bordercolor[1], bordercolor[2], bordercolor[3], bordercolor[4])
	_G[name]:Enable()
end

local function TemplateTextButton(button, text, action)
	local buttonName = button:GetName()

	local buttoncolor
	if action == 1 then
		button:SetNormalFontObject("GameFontNormal")
		button:SetDisabledFontObject("GameFontDisable") 
		buttoncolor = {0.38, 0, 0, 1}
		bordercolor = {0.73, 0.26, 0.21, 1}
	elseif action == 2 then
		button:SetNormalFontObject("GameFontWhiteSmall")
		button:SetDisabledFontObject("GameFontDisableSmall")
		buttoncolor = {0, 0, 0.5, 1}
		bordercolor = {0.43, 0.32, 0.68, 1}
	elseif action == 3 then
		button:SetNormalFontObject("GameFontWhiteSmall")
		button:SetDisabledFontObject("GameFontDisableSmall")
		buttoncolor = {0, 0.2, 0, 1}
		bordercolor = {0.24, 0.46, 0.21, 1}
	else
		button:SetNormalFontObject("GameFontNormal")
		button:SetDisabledFontObject("GameFontDisable")
		buttoncolor = {0, 0, 0, 1}
		bordercolor = {1, 1, 1, 1}
	end

	local textureBack = button:CreateTexture(nil, "BORDER")
	textureBack:SetPoint("TOPLEFT", 1, -1)
	textureBack:SetPoint("BOTTOMRIGHT", -1, 1)
	textureBack:SetTexture(0, 0, 0, 1)

	local textureBorder = button:CreateTexture(buttonName.."Border", "BACKGROUND")
	textureBorder:SetPoint("TOPLEFT", 0, 0)
	textureBorder:SetPoint("BOTTOMRIGHT", 0, 0)
	textureBorder:SetTexture(bordercolor[1], bordercolor[2], bordercolor[3], bordercolor[4])

	local textureNormal = button:CreateTexture(nil, "ARTWORK")
	textureNormal:SetPoint("TOPLEFT", 2, -2)
	textureNormal:SetPoint("BOTTOMRIGHT", -2, 2)
	textureNormal:SetTexture(buttoncolor[1], buttoncolor[2], buttoncolor[3], buttoncolor[4])
	button:SetNormalTexture(textureNormal)

	local textureDisabled = button:CreateTexture(nil, "OVERLAY")
	textureDisabled:SetPoint("TOPLEFT", 3, -3)
	textureDisabled:SetPoint("BOTTOMRIGHT", -3, 3)
	textureDisabled:SetTexture(0.6, 0.6, 0.6, 0.2)
	button:SetDisabledTexture(textureDisabled)

	local textureHighlight = button:CreateTexture(nil, "OVERLAY")
	textureHighlight:SetPoint("TOPLEFT", 3, -3)
	textureHighlight:SetPoint("BOTTOMRIGHT", -3, 3)
	textureHighlight:SetTexture(0.6, 0.6, 0.6, 0.2)
	button:SetHighlightTexture(textureHighlight)

	button:SetPushedTextOffset(1, -1)	
	button:SetText(text)
end
-- Template TextButton END ----------------------------------------

-- Template IconButton START ----------------------------------------
local function DisableIconButton(button, bg, which)
	_G[button]:Disable()
	if which then
		_G[button.."BorderColor"]:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
	else
		if bg then
			_G[button.."Background"]:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
		end
		_G[button.."BorderColor"]:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
	end
end
local function EnableIconButton(button, bg, which)
	_G[button]:Enable()
	if which then
		_G[button.."BorderColor"]:SetTexture(0.55, 0, 0, 1)
	else
		if bg then
			_G[button.."Background"]:SetTexture(color.ButtonBGActive.rgba[1], color.ButtonBGActive.rgba[2], color.ButtonBGActive.rgba[3], color.ButtonBGActive.rgba[4])
		end
		_G[button.."BorderColor"]:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])
	end
end

local function TemplateIconButton(button, which, height)
	local cut
	if which == 1 then
		cut = "GroupScanAdd"
	elseif which == 2 then
		cut = "History"
	elseif which == 3 then
		cut = "Compare"
	elseif which == 4 then
		cut = "Delete"
	elseif which == 5 then
		cut = "GroupScan"
	elseif which == 6 then
		cut = "Expand"
	elseif which == 7 then
		cut = "SingleScan"
	elseif which == 8 then
		cut = "Options"
	elseif which == 9 then
		cut = "Close"
	elseif which == 10 then
		cut = "RightSide"
	elseif which == 11 then
		cut = "FreeIcon"
	elseif which == 12 then
		cut = "DockedIcon"
	end

	button:SetWidth(sizeGeneral.Border + sizeGeneral.Space + Textures[cut].width + sizeGeneral.Space + sizeGeneral.Border)
	button:SetHeight(sizeGeneral.MainButtonHeight)

	if height then
		button:SetWidth(sizeGeneral.Border + sizeGeneral.Space + (Textures[cut].width/(Textures[cut].height/height)) + sizeGeneral.Space + sizeGeneral.Border)
		button:SetHeight(sizeGeneral.Border + sizeGeneral.Space + height + sizeGeneral.Space + sizeGeneral.Border)
	end

	local name = button:GetName()
	local texture = button:CreateTexture(name.."Background", "BORDER")
	texture:SetPoint("TOPLEFT", 1, -1)
	texture:SetPoint("BOTTOMRIGHT", -1, 1)
	texture:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
	local textureBackground = button:CreateTexture(name.."BorderColor", "BACKGROUND")
	textureBackground:SetPoint("TOPLEFT", 0, 0)
	textureBackground:SetPoint("BOTTOMRIGHT", 0, 0)
	textureBackground:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])

	local textureNormal = button:CreateTexture(name.."Normal", "ARTWORK")
	textureNormal:SetPoint("TOPLEFT", 3, -3)
	textureNormal:SetPoint("BOTTOMRIGHT", -3, 3)
	textureNormal:SetTexture(Textures.CharScannerIcons.path)
	textureNormal:SetTexCoord(unpack(Textures[cut].coords))
	button:SetNormalTexture(textureNormal)

	local texturePush = button:CreateTexture(name.."Push", "ARTWORK")
	texturePush:SetPoint("TOPLEFT", 4, -4)
	texturePush:SetPoint("BOTTOMRIGHT", -4, 4)
	texturePush:SetTexture(Textures.CharScannerIcons.path)
	texturePush:SetTexCoord(unpack(Textures[cut].coords))
	button:SetPushedTexture(texturePush)

	local textureDisable = button:CreateTexture(name.."Disable", "ARTWORK")
	textureDisable:SetPoint("TOPLEFT", 3, -3)
	textureDisable:SetPoint("BOTTOMRIGHT", -3, 3)
	textureDisable:SetTexture(Textures.CharScannerIcons.path)
	textureDisable:SetTexCoord(unpack(Textures[cut].coords))
	button:SetDisabledTexture(textureDisable)

	local shaderSupported = textureDisable:SetDesaturated(true)
	if not shaderSupported then
		textureDisable:SetVertexColor(0.5, 0.5, 0.5)
	end

	local textureHighlight = button:CreateTexture(nil, "ARTWORK")
	textureHighlight:SetPoint("TOPLEFT", 3, -3)
	textureHighlight:SetPoint("BOTTOMRIGHT", -3, 3)
	textureHighlight:SetTexture(Textures.CharScannerIcons.path)
	textureHighlight:SetTexCoord(unpack(Textures.ButtonHL.coords))
	button:SetHighlightTexture(textureHighlight)
end
-- Template IconButton END ----------------------------------------

-- Template TabButton START ----------------------------------------
local function TemplateTabButton(button, text, buttonwidth)
	local buttonName = button:GetName()

	local texture = button:CreateTexture(buttonName.."Border", "BORDER")
	texture:SetPoint("TOPLEFT", sizeGeneral.Border, -sizeGeneral.Border)
	texture:SetPoint("BOTTOMRIGHT", -sizeGeneral.Border, sizeGeneral.Border)
	texture:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])

	local textureBackground = button:CreateTexture(nil, "BACKGROUND")
	textureBackground:SetPoint("TOPLEFT", 0, 0)
	textureBackground:SetPoint("BOTTOMRIGHT", 0, 0)
	textureBackground:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])

	local textureHighlight = button:CreateTexture(nil, "ARTWORK")
	textureHighlight:SetPoint("TOPLEFT", 3, -3)
	textureHighlight:SetPoint("BOTTOMRIGHT", -3, 3)
	textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
	button:SetHighlightTexture(textureHighlight)

	local buttontext = button:CreateFontString(buttonName.."Text", "OVERLAY", fonts.GameFontNormalSmall.name)
	buttontext:SetText(text)
	buttontext:SetWidth(buttonwidth-10)
	buttontext:SetHeight(fonts.GameFontNormalSmall.height)
	buttontext:SetPoint("CENTER", button, "CENTER", 0, 0)
	buttontext:SetJustifyH("CENTER")
	buttontext:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
end
-- Template TabButton END ----------------------------------------

-- Template PvEStats START ----------------------------------------
local function TemplatePvEStats(button, collapse, level, l1, l2, l3, NOBOSS) --NOBOSS
	local buttonname = button:GetName()
	local collapseWidth = 0
	local nameWidth = sizeStatFrame.NameWidth
	local nameLeft = sizeStatFrame.NameLeft
	local achievementWidth = sizeStatFrame.AchievementWidth
	local col = "White"

	if level == "title" then
		col = "Yellow"
		nameWidth = sizeStatFrame.NameWidthTitle
		achievementWidth = sizeStatFrame.AchievementWidthTitle
	elseif level == "level1" then
		col = "Level1"
		if collapse then
			nameWidth = sizeStatFrame.NameWidth_L1c
			nameLeft = sizeStatFrame.NameLeft_L1c
			collapseWidth = 0
		end
	elseif level == "level2" then
		col = "Level2"
		nameWidth = sizeStatFrame.NameWidth_L2
		nameLeft = sizeStatFrame.NameLeft_L2
		if collapse then
			nameWidth = sizeStatFrame.NameWidth_L2c
			nameLeft = sizeStatFrame.NameLeft_L2c
			collapseWidth = sizeStatFrame.LevelModifier
		end
	elseif level == "level3" then
		col = "Level3"
		nameWidth = sizeStatFrame.NameWidth_L3
		nameLeft = sizeStatFrame.NameLeft_L3
		if collapse then
			nameWidth = sizeStatFrame.NameWidth_L3c
			nameLeft = sizeStatFrame.NameLeft_L3c
			collapseWidth = sizeStatFrame.LevelModifier*2
		end
	elseif level == "level4" then
		nameWidth = sizeStatFrame.NameWidth_L4
		nameLeft = sizeStatFrame.NameLeft_L4
	elseif level == "search" then
		nameWidth = sizeStatFrame.NameWidthSearch
	end

	button:SetFrameLevel( CharScanner_MainFrame_Stats:GetFrameLevel()+1 )
	button:SetWidth(sizeStatFrame.ButtonWidth)
	button:SetHeight(sizeGeneral.BarHeight)

	local nameButton = button:CreateFontString(buttonname.."Name", "OVERLAY", fonts.GameFontNormalSmall.name)
	nameButton:SetWidth(nameWidth)
	nameButton:SetHeight(fonts.GameFontNormalSmall.height)
	nameButton:SetPoint("LEFT", nameLeft, 0)
	nameButton:SetJustifyH("LEFT")
	nameButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])

	local bossestotalButton = button:CreateFontString(buttonname.."BossesTotal", "OVERLAY", fonts.GameFontNormalSmall.name)
	bossestotalButton:SetWidth(sizeStatFrame.BossesTotalWidth)
	bossestotalButton:SetHeight(fonts.GameFontNormalSmall.height)
	bossestotalButton:SetPoint("RIGHT", -sizeStatFrame.BossesTotalRight, 0)
	bossestotalButton:SetJustifyH("RIGHT")
	bossestotalButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])

	if collapse then
		local collapseButton = button:CreateTexture(buttonname.."Collapse", "ARTWORK")
		collapseButton:SetWidth(Textures.PlusMinus.qsize)
		collapseButton:SetHeight(Textures.PlusMinus.qsize)
		collapseButton:SetPoint("LEFT", collapseWidth, 0)
		collapseButton:SetTexture(Textures.CharScannerIcons.path)
		collapseButton:SetTexCoord(unpack(Textures.PlusMinus.coordsPlus))
		if NOBOSS then --NOBOSS
			collapseButton:SetAlpha(0.2)
		end

		local achievementButton = CreateFrame("Button", buttonname.."AchievementDetail", button)
		achievementButton:SetWidth(achievementWidth + sizeStatFrame.AchievementDetailWidth + sizeGeneral.Space)
		achievementButton:SetHeight(sizeGeneral.BarHeight)
		achievementButton:SetPoint("RIGHT", -sizeStatFrame.AchievementDetailRight, 0)
		achievementButton:SetFrameLevel( CharScanner_MainFrame_Stats:GetFrameLevel()+2 )

		local achievementButtonHighlight = achievementButton:CreateTexture(buttonname.."AchievementDetailHighlight", "ARTWORK")
		achievementButtonHighlight:SetPoint("TOPLEFT", button, "TOPRIGHT", -(sizeStatFrame.AchievementRight+achievementWidth), 0)
		achievementButtonHighlight:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", -sizeStatFrame.AchievementRight, 0)
		achievementButtonHighlight:SetTexture(0, 0, 0, 0)

		local achievementButtonIcon = achievementButton:CreateTexture(buttonname.."AchievementDetailIcon", "ARTWORK")
		achievementButtonIcon:SetWidth(Textures.AchievementDetail.width)
		achievementButtonIcon:SetHeight(Textures.AchievementDetail.height)
		achievementButtonIcon:SetPoint("RIGHT", 0, 0)
		achievementButtonIcon:SetTexture(Textures.CharScannerIcons.path)
		achievementButtonIcon:SetTexCoord(unpack(Textures.AchievementDetail.coords))
		achievementButtonIcon:SetAlpha(0.25) 

		local achievementButtonText = achievementButton:CreateFontString(buttonname.."Achievement", "OVERLAY", fonts.GameFontNormalSmall.name)
		achievementButtonText:SetWidth(achievementWidth)
		achievementButtonText:SetHeight(fonts.GameFontNormalSmall.height)
		achievementButtonText:SetPoint("RIGHT", achievementButtonIcon, "LEFT", -sizeGeneral.Space, 0)
		achievementButtonText:SetJustifyH("RIGHT")
		achievementButtonText:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])

		achievementButton:SetScript("OnEnter", function() CharScanner:AchievementDetail_OnEnter("PvE", buttonname, l1, l2, l3) end)
		achievementButton:SetScript("OnLeave", function() CharScanner:AchievementDetail_OnLeave("PvE", buttonname, l1, l2, l3) end)
		achievementButton:SetScript("OnClick", function() CharScanner:AchievementDetail_OnClick("PvE", buttonname, l1, l2, l3) end)

		local RaidSelectButton = CreateFrame("Button", buttonname.."RaidSelect", button)
		RaidSelectButton:SetWidth(Textures.RaidSelect.width)
		RaidSelectButton:SetHeight(Textures.RaidSelect.height)
		RaidSelectButton:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.RaidSelectRight, 0)
		RaidSelectButton:SetFrameLevel( CharScanner_MainFrame_Stats:GetFrameLevel()+2 )
		local RaidSelectButtonTexture = RaidSelectButton:CreateTexture(nil, "ARTWORK")
		RaidSelectButtonTexture:SetWidth(Textures.RaidSelect.width)
		RaidSelectButtonTexture:SetHeight(Textures.RaidSelect.height)
		RaidSelectButtonTexture:SetPoint("LEFT", 0, 0)
		RaidSelectButtonTexture:SetTexture(Textures.CharScannerIcons.path)
		RaidSelectButtonTexture:SetTexCoord(unpack(Textures.RaidSelect.coords))
		RaidSelectButtonTexture:SetAlpha(0.25) 

		RaidSelectButton:SetScript("OnEnter", function() RaidSelectButtonTexture:SetAlpha(1) end)
		RaidSelectButton:SetScript("OnLeave", function() RaidSelectButtonTexture:SetAlpha(0.25) end)
		RaidSelectButton:SetScript("OnClick", function() CharScanner:CompareFrameShow() CharScanner:History_GUISort("RAID", GUI_history_SortBy, l1, l2, l3) end)
	end

	if level == "title" or level == "summary" then
		local achievementButton = button:CreateFontString(buttonname.."Achievement", "OVERLAY", fonts.GameFontNormalSmall.name)
		achievementButton:SetWidth(achievementWidth)
		achievementButton:SetHeight(fonts.GameFontNormalSmall.height)
		achievementButton:SetPoint("RIGHT", -sizeStatFrame.AchievementRight, 0)
		achievementButton:SetJustifyH("RIGHT")
		achievementButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])
	end
	if collapse or level == "title" or level == "summary" then
		local statisticButton = button:CreateFontString(buttonname.."Statistic", "OVERLAY", fonts.GameFontNormalSmall.name)
		statisticButton:SetWidth(sizeStatFrame.StatisticWidth)
		statisticButton:SetHeight(fonts.GameFontNormalSmall.height)
		statisticButton:SetPoint("RIGHT", -sizeStatFrame.StatisticRight, 0)
		statisticButton:SetJustifyH("RIGHT")
		statisticButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])
	end

	if level ~= "title" and level ~= "summary" then
		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", 0, 0)
		textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
		textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
		button:SetHighlightTexture(textureHighlight)
		local texture = button:CreateTexture(nil, "BACKGROUND")
		texture:SetPoint("TOPLEFT", nameLeft, -5)
		texture:SetPoint("BOTTOMRIGHT", -sizeStatFrame.BossesTotalRight, 5)
		texture:SetTexture(color.ButtonLineMarker.rgba[1], color.ButtonLineMarker.rgba[2], color.ButtonLineMarker.rgba[3], color.ButtonLineMarker.rgba[4])
	end
--[[ TEST
	if _G[buttonname.."Name"] then
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "LEFT", nameLeft, 0)
		texture:SetPoint("RIGHT", button, "LEFT", nameWidth+nameLeft, 0)
		texture:SetTexture(0.22, 0.52, 0.22, 1)
	end
	if _G[buttonname.."Achievement"] then
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.AchievementRight+achievementWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.AchievementRight, 0)
		texture:SetTexture(0.22, 0.22, 0.82, 1)
	end
	if _G[buttonname.."AchievementDetail"] then
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.AchievementDetailRight+sizeStatFrame.AchievementDetailWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.AchievementDetailRight, 0)
		texture:SetTexture(0.32, 0.22, 0.22, 1)
	end
	if _G[buttonname.."Statistic"] then
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.StatisticRight+sizeStatFrame.StatisticWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.StatisticRight, 0)
		texture:SetTexture(0.42, 0.22, 0.22, 1)
	end
	if _G[buttonname.."BossesTotal"] then
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.BossesTotalRight+sizeStatFrame.BossesTotalWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.BossesTotalRight, 0)
		texture:SetTexture(0.42, 0.12, 0.12, 1)
	end
	if _G[buttonname.."RaidSelect"] then
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.RaidSelectRight+sizeStatFrame.RaidSelectWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.RaidSelectRight, 0)
		texture:SetTexture(0.22, 0.72, 0.82, 1)
	end
--]]
end
-- Template PvEStats END ----------------------------------------

-- Template PvPStats START ----------------------------------------
local function TemplatePvPStats(button, collapse, level, l1, l2, l3)
	local buttonname = button:GetName()
	local collapseWidth = 0
	local nameWidth = sizeStatFrame.PvPNameWidth
	local nameLeft = sizeStatFrame.PvPNameLeft
	local achievementWidth = sizeStatFrame.PvPAchievementWidth
	local winlossWidth = sizeStatFrame.PvPWinLossWidth
	local winlosspercentWidth = sizeStatFrame.PvPWinLossPercentWidth
	local honorablekillsRight = sizeStatFrame.PvPHonorableKillsRight
	local col = "White"

	if level == "title" then
		col = "Yellow"
		nameWidth = sizeStatFrame.PvPNameWidthTitle
		achievementWidth = sizeStatFrame.PvPAchievementWidthTitle
		winlossWidth = 0
		winlosspercentWidth = sizeStatFrame.PvPWinLossPerWidthTitle
	elseif level == "level1" then
		col = "Level1"
		if collapse then
			nameWidth = sizeStatFrame.PvPNameWidth_L1c
			nameLeft = sizeStatFrame.PvPNameLeft_L1c
			collapseWidth = 0
		end
	elseif level == "level2" then
		col = "Level2"
		nameWidth = sizeStatFrame.PvPNameWidth_L2c + sizeStatFrame.LevelModifier
		nameLeft = sizeStatFrame.PvPNameLeft_L2c - sizeStatFrame.LevelModifier
		if collapse then
			nameWidth = sizeStatFrame.PvPNameWidth_L2c
			nameLeft = sizeStatFrame.PvPNameLeft_L2c
			collapseWidth = sizeStatFrame.LevelModifier
		end
	elseif level == "level3" then
		col = "Level3"
		honorablekillsRight = honorablekillsRight + 5 -- space for arena rating
		nameWidth = sizeStatFrame.PvPNameWidth_L3
		nameLeft = sizeStatFrame.PvPNameLeft_L3
		if collapse then
			nameWidth = sizeStatFrame.PvPNameWidth_L3c
			nameLeft = sizeStatFrame.PvPNameLeft_L3c
			collapseWidth = sizeStatFrame.LevelModifier*2
		end
	elseif level == "level4" then
		nameWidth = sizeStatFrame.PvPNameWidth_L4
		nameLeft = sizeStatFrame.vNameLeft_L4
	end

	button:SetFrameLevel( CharScanner_MainFrame_Stats:GetFrameLevel()+1 )
	button:SetWidth(sizeStatFrame.ButtonWidth)
	button:SetHeight(sizeGeneral.BarHeight)

	-- Collapse Button START
	if collapse then
		local collapseButton = button:CreateTexture(buttonname.."Collapse", "ARTWORK")
		collapseButton:SetWidth(Textures.PlusMinus.qsize)
		collapseButton:SetHeight(Textures.PlusMinus.qsize)
		collapseButton:SetPoint("LEFT", collapseWidth, 0)
		collapseButton:SetTexture(Textures.CharScannerIcons.path)
		collapseButton:SetTexCoord(unpack(Textures.PlusMinus.coordsPlus))
	end
	-- Collapse Button END

	-- Name Button START
	local nameButton = button:CreateFontString(buttonname.."Name", "OVERLAY", fonts.GameFontNormalSmall.name)
	nameButton:SetWidth(nameWidth)
	nameButton:SetHeight(fonts.GameFontNormalSmall.height)
	nameButton:SetPoint("LEFT", nameLeft, 0)
	nameButton:SetJustifyH("LEFT")
	nameButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])
	-- Name Button END

	-- Achievement Button START
	if level ~= "title" and level ~= "summary" and level ~= "level3" then
		local achievementButton = CreateFrame("Button", buttonname.."AchievementDetail", button)
		achievementButton:SetWidth(achievementWidth + sizeStatFrame.PvPAchievementDetailWidth + sizeGeneral.Space)
		achievementButton:SetHeight(sizeGeneral.BarHeight)
		achievementButton:SetPoint("RIGHT", -sizeStatFrame.PvPAchievementDetailRight, 0)
		achievementButton:SetFrameLevel( CharScanner_MainFrame_Stats:GetFrameLevel()+2 )

		local achievementButtonHighlight = achievementButton:CreateTexture(buttonname.."AchievementDetailHighlight", "ARTWORK")
		achievementButtonHighlight:SetPoint("TOPLEFT", button, "TOPRIGHT", -(sizeStatFrame.PvPAchievementRight+achievementWidth), 0)
		achievementButtonHighlight:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", -sizeStatFrame.PvPAchievementRight, 0)
		achievementButtonHighlight:SetTexture(0, 0, 0, 0)

		local achievementButtonIcon = achievementButton:CreateTexture(buttonname.."AchievementDetailIcon", "ARTWORK")
		achievementButtonIcon:SetWidth(Textures.AchievementDetail.width)
		achievementButtonIcon:SetHeight(Textures.AchievementDetail.height)
		achievementButtonIcon:SetPoint("RIGHT", 0, 0)
		achievementButtonIcon:SetTexture(Textures.CharScannerIcons.path)
		achievementButtonIcon:SetTexCoord(unpack(Textures.AchievementDetail.coords))
		achievementButtonIcon:SetAlpha(0.25) 

		local achievementButtonText = achievementButton:CreateFontString(buttonname.."Achievement", "OVERLAY", fonts.GameFontNormalSmall.name)
		achievementButtonText:SetWidth(achievementWidth)
		achievementButtonText:SetHeight(fonts.GameFontNormalSmall.height)
		achievementButtonText:SetPoint("RIGHT", achievementButtonIcon, "LEFT", -sizeGeneral.Space, 0)
		achievementButtonText:SetJustifyH("RIGHT")
		achievementButtonText:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])

		achievementButton:SetScript("OnEnter", function() CharScanner:AchievementDetail_OnEnter("PvP", buttonname, l1, l2, l3) end)
		achievementButton:SetScript("OnLeave", function() CharScanner:AchievementDetail_OnLeave("PvP", buttonname, l1, l2, l3) end)
		achievementButton:SetScript("OnClick", function() CharScanner:AchievementDetail_OnClick("PvP", buttonname, l1, l2, l3) end)

		local RaidSelectButton = CreateFrame("Button", buttonname.."RaidSelect", button)
		RaidSelectButton:SetWidth(Textures.RaidSelect.width)
		RaidSelectButton:SetHeight(Textures.RaidSelect.height)
		RaidSelectButton:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.RaidSelectRight, 0)
		RaidSelectButton:SetFrameLevel( CharScanner_MainFrame_Stats:GetFrameLevel()+2 )
		local RaidSelectButtonTexture = RaidSelectButton:CreateTexture(nil, "ARTWORK")
		RaidSelectButtonTexture:SetWidth(Textures.RaidSelect.width)
		RaidSelectButtonTexture:SetHeight(Textures.RaidSelect.height)
		RaidSelectButtonTexture:SetPoint("LEFT", 0, 0)
		RaidSelectButtonTexture:SetTexture(Textures.CharScannerIcons.path)
		RaidSelectButtonTexture:SetTexCoord(unpack(Textures.RaidSelect.coords))
		RaidSelectButtonTexture:SetAlpha(0) -- 0.25

		--RaidSelectButton:SetScript("OnEnter", function() RaidSelectButtonTexture:SetAlpha(1) end)
		--RaidSelectButton:SetScript("OnLeave", function() RaidSelectButtonTexture:SetAlpha(0.25) end)
		--RaidSelectButton:SetScript("OnClick", function() CharScanner:CompareFrameShow() CharScanner:History_GUISort("PVP", GUI_history_SortBy, l1, l2, l3) end)
	end
	-- Achievement Button END

	if level ~= "level3" then
		-- WinLoss Button START
		local winlossButton = button:CreateFontString(buttonname.."WinLoss", "OVERLAY", fonts.GameFontNormalSmall.name)
		winlossButton:SetWidth(winlossWidth)
		winlossButton:SetHeight(fonts.GameFontNormalSmall.height)
		winlossButton:SetPoint("RIGHT", -sizeStatFrame.PvPWinLossRight, 0)
		winlossButton:SetJustifyH("RIGHT")
		winlossButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])
		-- WinLoss Button END

		-- WinLossPercent Button START
		local winlosspercentButton = button:CreateFontString(buttonname.."WinLossPercent", "OVERLAY", fonts.GameFontNormalSmall.name)
		winlosspercentButton:SetWidth(winlosspercentWidth)
		winlosspercentButton:SetHeight(fonts.GameFontNormalSmall.height)
		winlosspercentButton:SetPoint("RIGHT", -sizeStatFrame.PvPWinLossPercentRight, 0)
		winlosspercentButton:SetJustifyH("RIGHT")
		winlosspercentButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])
		-- WinLossPercent Button END
	end

	-- HonorableKills Button START
	local honorablekillsButton = button:CreateFontString(buttonname.."HonorableKills", "OVERLAY", fonts.GameFontNormalSmall.name)
	honorablekillsButton:SetWidth(sizeStatFrame.PvPHonorableKillsWidth)
	honorablekillsButton:SetHeight(fonts.GameFontNormalSmall.height)
	honorablekillsButton:SetPoint("RIGHT", -honorablekillsRight, 0)
	honorablekillsButton:SetJustifyH("RIGHT")
	honorablekillsButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])
	-- HonorableKills Button END

	if level == "title" or level == "summary" then
		local achievementButton = button:CreateFontString(buttonname.."Achievement", "OVERLAY", fonts.GameFontNormalSmall.name)
		achievementButton:SetWidth(achievementWidth)
		achievementButton:SetHeight(fonts.GameFontNormalSmall.height)
		achievementButton:SetPoint("RIGHT", -sizeStatFrame.PvPAchievementRight, 0)
		achievementButton:SetJustifyH("RIGHT")
		achievementButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])
	end

	if level ~= "title" and level ~= "summary" then
		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", 0, 0)
		textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
		textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
		button:SetHighlightTexture(textureHighlight)
		local texture = button:CreateTexture(nil, "BACKGROUND")
		texture:SetPoint("TOPLEFT", nameLeft, -5)
		texture:SetPoint("BOTTOMRIGHT", -sizeStatFrame.PvPHonorableKillsRight, 5)
		texture:SetTexture(color.ButtonLineMarker.rgba[1], color.ButtonLineMarker.rgba[2], color.ButtonLineMarker.rgba[3], color.ButtonLineMarker.rgba[4])
	end
--[[ TEST
	if _G[buttonname.."Name"] then
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "LEFT", nameLeft, 0)
		texture:SetPoint("RIGHT", button, "LEFT", nameWidth+nameLeft, 0)
		texture:SetTexture(0.22, 0.42, 0.22, 1)
	end
	if _G[buttonname.."Achievement"] then
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.PvPAchievementRight+achievementWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.PvPAchievementRight, 0)
		texture:SetTexture(0.42, 0.42, 0.42, 1)
	end
	if _G[buttonname.."AchievementDetail"] then
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.PvPAchievementDetailRight+sizeStatFrame.PvPAchievementDetailWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.PvPAchievementDetailRight, 0)
		texture:SetTexture(0.22, 0.32, 0.22, 1)
	end
	if _G[buttonname.."WinLoss"] then
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.PvPWinLossRight+winlossWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.PvPWinLossRight, 0)
		texture:SetTexture(0.82, 0.22, 0.32, 1)
	end
	if _G[buttonname.."WinLossPercent"] then
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.PvPWinLossPercentRight+winlosspercentWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.PvPWinLossPercentRight, 0)
		texture:SetTexture(0.22, 0.22, 0.62, 1)
	end
	if _G[buttonname.."HonorableKills"] then
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.PvPHonorableKillsRight+sizeStatFrame.PvPHonorableKillsWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.PvPHonorableKillsRight, 0)
		texture:SetTexture(0.12, 0.22, 0.22, 1)
	end
	if _G[buttonname.."RaidSelect"] then
		local texture = button:CreateTexture(nil, "ARTWORK")
		texture:SetPoint("LEFT", button, "RIGHT", -(sizeStatFrame.RaidSelectRight+sizeStatFrame.RaidSelectWidth), 0)
		texture:SetPoint("RIGHT", button, "RIGHT", -sizeStatFrame.RaidSelectRight, 0)
		texture:SetTexture(0.52, 0.02, 0.22, 1)
	end
--]]
end
-- Template PvPStats END ----------------------------------------

-- Template CompareSelect START ----------------------------------------
local function TemplateCompareSelect(button, level, height)
	local name = button:GetName()
	local col = "White"
	local buttonLeft = 0

	if level == "title" then
		buttonLeft = sizeGeneral.Offset
		col = "Yellow"
	elseif level == "level1" then
		buttonLeft = 0
		col = "Level1"
	elseif level == "level2" then
		buttonLeft = sizeStatFrame.LevelModifier
		col = "Level2"
	elseif level == "level3" then
		buttonLeft = sizeStatFrame.LevelModifier * 2
		col = "Level3"
	end
	button:SetHeight(height)
	
	local nameButton = button:CreateFontString(name.."Name", "OVERLAY", fonts.GameFontNormalSmall.name)
	nameButton:SetHeight(fonts.GameFontNormalSmall.height)
	nameButton:SetPoint("LEFT", buttonLeft, 0)
	nameButton:SetJustifyH("LEFT")
	nameButton:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])

	if level == "title" then
		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", 0, 0)
		textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
		textureHighlight:SetTexture(color.ButtonHighlightLow.rgba[1], color.ButtonHighlightLow.rgba[2], color.ButtonHighlightLow.rgba[3], color.ButtonHighlightLow.rgba[4])
		button:SetHighlightTexture(textureHighlight)
	else
		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", 0, 0)
		textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
		textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
		button:SetHighlightTexture(textureHighlight)
	end
end
-- Template CompareSelect END ----------------------------------------

-- Template HistoryListSortButton START ----------------------------------------
local function TemplateHistoryListSortButton(button, text, pos, width)
	local name = button:GetName()

	local textureBackground = button:CreateTexture(nil, "BACKGROUND")
	textureBackground:SetPoint("TOPLEFT", 0, 0)
	textureBackground:SetPoint("BOTTOMRIGHT", 0, 0)
	textureBackground:SetTexture(color.ButtonSortBG.rgba[1], color.ButtonSortBG.rgba[2], color.ButtonSortBG.rgba[3], color.ButtonSortBG.rgba[4])

	local buttontext
	if text then
		if pos == "right" then
			buttontext = button:CreateFontString(name.."Text", "OVERLAY", fonts.GameFontNormalSmall.name)
			buttontext:SetText(text)
			local fw = buttontext:GetStringWidth()
			if fw >= width then
				buttontext:SetWidth( width-10 )
			end
			buttontext:SetHeight(fonts.GameFontNormalSmall.height)
			buttontext:SetPoint("RIGHT", button, "RIGHT", 0, 0)
			buttontext:SetJustifyH("RIGHT")
		elseif pos == "left" then
			buttontext = button:CreateFontString(name.."Text", "OVERLAY", fonts.GameFontNormalSmall.name)
			buttontext:SetText(text)
			local fw = buttontext:GetStringWidth()
			if fw >= width then
				buttontext:SetWidth( width-10 )
			end
			buttontext:SetHeight(fonts.GameFontNormalSmall.height)
			buttontext:SetPoint("LEFT", button, "LEFT", 0, 0)
			buttontext:SetJustifyH("LEFT")
		else
			buttontext = button:CreateFontString(name.."Text", "OVERLAY", fonts.GameFontNormalSmall.name)
			buttontext:SetText(text)
			local fw = buttontext:GetStringWidth()
			if fw >= width then
				buttontext:SetWidth( width-10 )
			end
			buttontext:SetHeight(fonts.GameFontNormalSmall.height)
			buttontext:SetPoint("CENTER", button, "CENTER", 0, 0)
			buttontext:SetJustifyH("CENTER")
		end
	end

	local texture = button:CreateTexture(name.."Arrow", "BACKGROUND")
	texture:SetWidth(Textures.SortArrow.width)
	texture:SetHeight(Textures.SortArrow.height)
	texture:SetTexture(Textures.CharScannerIcons.path)
	if text then
		if pos == "right" then
			texture:SetPoint("RIGHT", buttontext, "LEFT", -1, 0)
		elseif pos == "left" then
			texture:SetPoint("LEFT", buttontext, "RIGHT", 1, 0)
		else
			texture:SetPoint("RIGHT", buttontext, "LEFT", -1, 0)
		end
	else
		texture:SetPoint("CENTER", button, "CENTER", 0, 0)
	end

	local textureHighlight = button:CreateTexture(nil, "ARTWORK")
	textureHighlight:SetPoint("TOPLEFT", 0, 0)
	textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
	textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
	button:SetHighlightTexture(textureHighlight)
end
-- Template HistoryListSortButton END ----------------------------------------

-- Template HistoryList START ----------------------------------------
local function TemplateHistoryList(button, buttonid, isSort)
	local name = button:GetName()

	if isSort then
		local sort_icon = CreateFrame("Button", name.."Button_Icon", button)
		TemplateHistoryListSortButton(sort_icon, nil, nil)
		sort_icon:SetWidth(sizeGeneral.BarHeight)
		sort_icon:SetHeight(sizeGeneral.BarHeight)
		sort_icon:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.ClassiconLeft, 0)
		sort_icon:SetScript("OnClick", function() CharScanner:History_GUISort("NORMAL", 1) end)

		local sort_name = CreateFrame("Button", name.."Button_Name", button)
		TemplateHistoryListSortButton(sort_name, NAME, "left", sizeCompareFrame.NameWidth)
		sort_name:SetWidth(sizeCompareFrame.NameWidth)
		sort_name:SetHeight(sizeGeneral.BarHeight)
		sort_name:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.NameLeft, 0)
		sort_name:SetScript("OnClick", function() CharScanner:History_GUISort("NORMAL", 2) end)

		local sort_time = CreateFrame("Button", name.."Button_Time", button)
		TemplateHistoryListSortButton(sort_time, L["Time"], "left", sizeCompareFrame.TimeWidth)
		sort_time:SetWidth(sizeCompareFrame.TimeWidth)
		sort_time:SetHeight(sizeGeneral.BarHeight)
		sort_time:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.TimeLeft, 0)
		sort_time:SetScript("OnClick", function() CharScanner:History_GUISort("NORMAL", 3) end)

		local sort_mainrole = CreateFrame("Button", name.."Button_MainRole", button)
		TemplateHistoryListSortButton(sort_mainrole, nil, nil)
		sort_mainrole:SetWidth(sizeCompareFrame.MainroleWidth)
		sort_mainrole:SetHeight(sizeGeneral.BarHeight)
		sort_mainrole:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.MainroleLeft, 0)
		sort_mainrole:SetScript("OnClick", function() CharScanner:History_GUISort("NORMAL", 4) end)

		local frame = CreateFrame("Frame", name.."_CompareFrame", button)
		frame:SetToplevel(true)

		local sort_talents1 = CreateFrame("Button", name.."Button_Talents1", frame)
		TemplateHistoryListSortButton(sort_talents1, L["T1 (Current Talents)"], "left", sizeCompareFrame.Talents1Width)
		sort_talents1:SetWidth(sizeCompareFrame.Talents1Width)
		sort_talents1:SetHeight(sizeGeneral.BarHeight)
		sort_talents1:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.Talents1Left, 0)
		sort_talents1:SetScript("OnClick", function() CharScanner:History_GUISort("NORMAL", 5) end)

		local sort_talents2 = CreateFrame("Button", name.."Button_Talents2", frame)
		TemplateHistoryListSortButton(sort_talents2, L["T2"], "left", sizeCompareFrame.Talents2Width)
		sort_talents2:SetWidth(sizeCompareFrame.Talents2Width)
		sort_talents2:SetHeight(sizeGeneral.BarHeight)
		sort_talents2:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.Talents2Left, 0)
		sort_talents2:SetScript("OnClick", function() CharScanner:History_GUISort("NORMAL", 6) end)

		local sort_enchcount = CreateFrame("Button", name.."Button_EnchantmentCount", frame)
		TemplateHistoryListSortButton(sort_enchcount, L["E"], "right", sizeCompareFrame.EnchWidth)
		sort_enchcount:SetWidth(sizeCompareFrame.EnchWidth)
		sort_enchcount:SetHeight(sizeGeneral.BarHeight)
		sort_enchcount:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.EnchLeft, 0)
		sort_enchcount:SetScript("OnClick", function() CharScanner:History_GUISort("NORMAL", 7) end)

		local sort_pvpcount = CreateFrame("Button", name.."Button_PvPCount", frame)
		TemplateHistoryListSortButton(sort_pvpcount, L["P"], "right", sizeCompareFrame.PvPWidth)
		sort_pvpcount:SetWidth(sizeCompareFrame.PvPWidth)
		sort_pvpcount:SetHeight(sizeGeneral.BarHeight)
		sort_pvpcount:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.PvPLeft, 0)
		sort_pvpcount:SetScript("OnClick", function() CharScanner:History_GUISort("NORMAL", 8) end)

		local sort_itemlevel = CreateFrame("Button", name.."Button_ItemLevel", frame)
		TemplateHistoryListSortButton(sort_itemlevel, L["iLvl"], "right", sizeCompareFrame.EquipWidth)
		sort_itemlevel:SetWidth(sizeCompareFrame.EquipWidth)
		sort_itemlevel:SetHeight(sizeGeneral.BarHeight)
		sort_itemlevel:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.EquipLeft, 0)
		sort_itemlevel:SetScript("OnClick", function() CharScanner:History_GUISort("NORMAL", 9) end)

		local sort_achievement = CreateFrame("Button", name.."Button_Achievement", frame)
		TemplateHistoryListSortButton(sort_achievement, L["Achievements"], "right", sizeCompareFrame.AchievementWidth)
		sort_achievement:SetWidth(sizeCompareFrame.AchievementWidth)
		sort_achievement:SetHeight(sizeGeneral.BarHeight)
		sort_achievement:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.AchievementLeft, 0)
		sort_achievement:SetScript("OnClick", function() CharScanner:History_GUISort("NORMAL", 10) end)

		local sort_boss = CreateFrame("Button", name.."Button_Boss", frame)
		TemplateHistoryListSortButton(sort_boss, L["Bosses"], "right", sizeCompareFrame.BossWidth)
		sort_boss:SetWidth(sizeCompareFrame.BossWidth)
		sort_boss:SetHeight(sizeGeneral.BarHeight)
		sort_boss:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.BossLeft, 0)
		sort_boss:SetScript("OnClick", function() CharScanner:History_GUISort("NORMAL", 11) end)

		local sort_kills = CreateFrame("Button", name.."Button_Kills", frame)
		TemplateHistoryListSortButton(sort_kills, L["Kills"], "right", sizeCompareFrame.KillsWidth)
		sort_kills:SetWidth(sizeCompareFrame.KillsWidth)
		sort_kills:SetHeight(sizeGeneral.BarHeight)
		sort_kills:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.KillsLeft, 0)
		sort_kills:SetScript("OnClick", function() CharScanner:History_GUISort("NORMAL", 12) end)

		local sort_SingleAchievementCompare = CreateFrame("Button", name.."Button_SingleAchievementCompare", frame)
		TemplateHistoryListSortButton(sort_SingleAchievementCompare, L["Single Achievement Compare"], "left", sizeCompareFrame.SingleAchievementCompareWidth)
		sort_SingleAchievementCompare:SetWidth(sizeCompareFrame.SingleAchievementCompareWidth)
		sort_SingleAchievementCompare:SetHeight(sizeGeneral.BarHeight)
		sort_SingleAchievementCompare:SetPoint("LEFT", "CharScanner_HistoryFrame_Sort", "LEFT", sizeCompareFrame.SingleAchievementCompareLeft, 0)
		sort_SingleAchievementCompare:SetScript("OnClick", function() CharScanner:History_GUISort("NORMAL", 13) end)
		sort_SingleAchievementCompare:Hide()

		-- Raid Selection START
		local button = CreateFrame("Button", name.."_RaidSelectMainButton", frame)
		TemplateBorderTRBL(button)
		TemplateCompareSelect(button, "title", sizeCompareFrame.SortLine1Height)
		button:SetPoint("TOPRIGHT", CharScanner_HistoryFrame, "TOPRIGHT", -sizeCompareFrame.RaidSelectLeft, -sizeGeneral.Offset)
		_G[name.."_RaidSelectMainButtonName"]:SetText(L["Total"])
		button:SetScript("OnClick",
			function()
				if _G[name.."_RaidSelectFrame"]:IsShown() then
					_G[name.."_RaidSelectFrame"]:Hide()
					_G[name.."_RaidSelectMainButton_Expand"]:SetTexCoord(unpack(Textures.Expand.coords))
				else
					_G[name.."_RaidSelectFrame"]:Show()
					_G[name.."_RaidSelectMainButton_Expand"]:SetTexCoord(unpack(Textures.Collapse.coords))
				end
			end)

		local buttontexture = button:CreateTexture(name.."_RaidSelectMainButton_Expand", "ARTWORK")
		buttontexture:SetHeight(16)
		buttontexture:SetWidth(16)
		buttontexture:SetPoint("RIGHT", button, "RIGHT", -sizeGeneral.Offset, 0)
		buttontexture:SetTexture(Textures.CharScannerIcons.path)
		buttontexture:SetTexCoord(unpack(Textures.Expand.coords))

		local historyraidselect = CreateFrame("Frame", name.."_RaidSelectFrame", frame)
		TemplateBorderTRBL(historyraidselect)
		historyraidselect:SetToplevel(true)
		historyraidselect:EnableMouse(true)
		historyraidselect:SetPoint("TOPLEFT", button, "BOTTOMLEFT", 0, sizeGeneral.Border)
		historyraidselect:Hide()

		local bnumber = 1
		local textwidth = 0
		local strw = 0

		local buttonframe = CreateFrame("Button", name.."_RaidSelect_Button"..bnumber, historyraidselect)
		TemplateCompareSelect(buttonframe, "level1", sizeGeneral.BarHeight)
		buttonframe:SetPoint("TOPLEFT", name.."_RaidSelectFrame", "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
		buttonframe:SetScript("OnClick", function() CharScanner:History_GUISort("RAID", GUI_history_SortBy, 0) end)
		_G[name.."_RaidSelect_Button"..bnumber.."Name"]:SetText(L["Total"])
		strw = _G[name.."_RaidSelect_Button"..bnumber.."Name"]:GetStringWidth()
		if strw > textwidth then textwidth = strw end
		bnumber = bnumber + 1

		for i = 1, #PVEDUMP do
			if PVEDUMP[i].name then
				local buttonframe = CreateFrame("Button", name.."_RaidSelect_Button"..bnumber, historyraidselect)
				TemplateCompareSelect(buttonframe, "level1", sizeGeneral.BarHeight)
				buttonframe:SetPoint("TOPLEFT", name.."_RaidSelect_Button"..bnumber-1, "BOTTOMLEFT", 0, 0)
				buttonframe:SetScript("OnClick", function() CharScanner:History_GUISort("RAID", GUI_history_SortBy, i) end)
				_G[name.."_RaidSelect_Button"..bnumber.."Name"]:SetText(L[PVEDUMP[i].name])
				strw = _G[name.."_RaidSelect_Button"..bnumber.."Name"]:GetStringWidth()
				if strw > textwidth then textwidth = strw end
				bnumber = bnumber + 1
			end

			for j = 1, #PVEDUMP[i] do

				if PVEDUMP[i][j].name then
					local buttonframe = CreateFrame("Button", name.."_RaidSelect_Button"..bnumber, historyraidselect)
					TemplateCompareSelect(buttonframe, "level2", sizeGeneral.BarHeight)
					buttonframe:SetPoint("TOPLEFT", name.."_RaidSelect_Button"..bnumber-1, "BOTTOMLEFT", 0, 0)
					buttonframe:SetScript("OnClick", function() CharScanner:History_GUISort("RAID", GUI_history_SortBy, i, j) end)
					_G[name.."_RaidSelect_Button"..bnumber.."Name"]:SetText(L[PVEDUMP[i][j].name])
					strw = _G[name.."_RaidSelect_Button"..bnumber.."Name"]:GetStringWidth()
					if strw > textwidth then textwidth = strw end
					bnumber = bnumber + 1
				end

				for k = 1, #PVEDUMP[i][j] do
					if PVEDUMP[i][j][k].name then
						local buttonframe = CreateFrame("Button", name.."_RaidSelect_Button"..bnumber, historyraidselect)
						TemplateCompareSelect(buttonframe, "level3", sizeGeneral.BarHeight)
						buttonframe:SetPoint("TOPLEFT", name.."_RaidSelect_Button"..bnumber-1, "BOTTOMLEFT", 0, 0)
						buttonframe:SetScript("OnClick", function() CharScanner:History_GUISort("RAID", GUI_history_SortBy, i, j, k) end)
						_G[name.."_RaidSelect_Button"..bnumber.."Name"]:SetText(L[PVEDUMP[i][j][k].name].." ("..L[PVEDUMP[i][j].name]..")")
						strw = _G[name.."_RaidSelect_Button"..bnumber.."Name"]:GetStringWidth()
						if strw > textwidth then textwidth = strw end
						bnumber = bnumber + 1
					end
				end

			end

		end

		historyraidselect:SetHeight(((bnumber-1)*sizeGeneral.BarHeight) + (2*sizeGeneral.Offset))
		
		-- set width
		textwidth = textwidth + 20

		CharScanner_HistoryFrame_Sort_RaidSelectMainButton:SetWidth(textwidth + (2*sizeGeneral.Offset))
		CharScanner_HistoryFrame_Sort_RaidSelectMainButtonName:SetWidth(textwidth - (2*sizeGeneral.Offset) - 16)
		CharScanner_HistoryFrame_Sort_RaidSelectFrame:SetWidth(textwidth + (2*sizeGeneral.Offset))

		for i = 1, bnumber-1 do
			_G[name.."_RaidSelect_Button"..i]:SetWidth(textwidth)
			_G[name.."_RaidSelect_Button"..i.."Name"]:SetWidth(textwidth)
		end
		-- set width

		-- auto hide
		local function OnLeave()
			if not CharScanner_HistoryFrame_Sort_RaidSelectMainButton:IsMouseOver() and not CharScanner_HistoryFrame_Sort_RaidSelectFrame:IsMouseOver() then
				CharScanner_HistoryFrame_Sort_RaidSelectFrame:Hide()
				CharScanner_HistoryFrame_Sort_RaidSelectMainButton_Expand:SetTexCoord(unpack(Textures.Expand.coords))
			end
		end
		CharScanner_HistoryFrame_Sort_RaidSelectMainButton:SetScript("OnLeave", OnLeave)
		CharScanner_HistoryFrame_Sort_RaidSelectMainButton:SetScript("OnHide", function() CharScanner_HistoryFrame_Sort_RaidSelectFrame:Hide() end) -- for esc close
		CharScanner_HistoryFrame_Sort_RaidSelectFrame:SetScript("OnLeave", OnLeave)
		-- auto hide
		-- Raid Selection END
	else
		local deleteButton = CreateFrame("Button", nil, button)
		deleteButton:SetWidth(Textures.Trashcan.width)
		deleteButton:SetHeight(Textures.Trashcan.height)
		deleteButton:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.TrashcanLeft, 0)
		local deleteButtonTexture = deleteButton:CreateTexture(nil, "ARTWORK")
		deleteButtonTexture:SetPoint("LEFT", 0, 0)
		deleteButtonTexture:SetWidth(Textures.Trashcan.width)
		deleteButtonTexture:SetHeight(Textures.Trashcan.height)
		deleteButtonTexture:SetTexture(Textures.CharScannerIcons.path)
		deleteButtonTexture:SetTexCoord(unpack(Textures.Trashcan.coords))
		deleteButtonTexture:SetAlpha(0.5)
		deleteButton:SetScript("OnEnter", function() deleteButtonTexture:SetAlpha(1) end)
		deleteButton:SetScript("OnLeave", function() deleteButtonTexture:SetAlpha(0.5) end)
		deleteButton:SetScript("OnClick", function() CharScanner:DeleteCharacter(button.guid, buttonid) end)

		local texture = button:CreateTexture(name.."Arrow", "BACKGROUND")
		texture:SetWidth(Textures.SelectionArrow.width)
		texture:SetHeight(Textures.SelectionArrow.height)
		texture:SetTexture(Textures.CharScannerIcons.path)
		texture:SetTexCoord(unpack(Textures.SelectionArrow.coords))
		texture:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.ArrowLeft, 0)
		texture:Hide()

		local iconButton = button:CreateTexture(name.."Icon", "ARTWORK")
		iconButton:SetWidth(sizeGeneral.BarHeight)
		iconButton:SetHeight(sizeGeneral.BarHeight)
		iconButton:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.ClassiconLeft, 0)
		iconButton:SetTexture(Textures.ClassIcon.path)

		local nameButton = button:CreateFontString(name.."Name", "ARTWORK", fonts.GameFontNormalSmall.name)
		nameButton:SetWidth(sizeCompareFrame.NameWidth)
		nameButton:SetHeight(fonts.GameFontNormalSmall.height)
		nameButton:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.NameLeft, 0)
		nameButton:SetJustifyH("LEFT")
		nameButton:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])

		local nameTime = button:CreateFontString(name.."Time", "ARTWORK", fonts.GameFontNormalSmall.name)
		nameTime:SetWidth(sizeCompareFrame.TimeWidth)
		nameTime:SetHeight(fonts.GameFontWhiteSmall.height)
		nameTime:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.TimeLeft, 0)
		nameTime:SetJustifyH("LEFT")
		nameTime:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])

		local nameMainRole = button:CreateTexture(name.."MainRole", "ARTWORK")
		nameMainRole:SetWidth(sizeCompareFrame.MainroleWidth)
		nameMainRole:SetHeight(sizeCompareFrame.MainroleWidth)
		nameMainRole:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.MainroleLeft, 0)

		local frame = CreateFrame("Frame", name.."_CompareFrame", button)
		frame:SetToplevel(true)

		local nameTalents1 = frame:CreateFontString(name.."Talents1", "ARTWORK", fonts.GameFontNormalSmall.name)
		nameTalents1:SetWidth(sizeCompareFrame.Talents1Width)
		nameTalents1:SetHeight(fonts.GameFontWhiteSmall.height)
		nameTalents1:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.Talents1Left, 0)
		nameTalents1:SetJustifyH("LEFT")
		nameTalents1:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])

		local nameTalents2 = frame:CreateFontString(name.."Talents2", "ARTWORK", fonts.GameFontNormalSmall.name)
		nameTalents2:SetWidth(sizeCompareFrame.Talents2Width)
		nameTalents2:SetHeight(fonts.GameFontWhiteSmall.height)
		nameTalents2:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.Talents2Left, 0)
		nameTalents2:SetJustifyH("LEFT")
		nameTalents2:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])

		local nameEnchantmentCount = frame:CreateFontString(name.."EnchantmentCount", "ARTWORK", fonts.GameFontNormalSmall.name)
		nameEnchantmentCount:SetWidth(sizeCompareFrame.EnchWidth)
		nameEnchantmentCount:SetHeight(fonts.GameFontWhiteSmall.height)
		nameEnchantmentCount:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.EnchLeft, 0)
		nameEnchantmentCount:SetJustifyH("RIGHT")
		nameEnchantmentCount:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])

		local namePvPCount = frame:CreateFontString(name.."PvPCount", "ARTWORK", fonts.GameFontNormalSmall.name)
		namePvPCount:SetWidth(sizeCompareFrame.PvPWidth)
		namePvPCount:SetHeight(fonts.GameFontWhiteSmall.height)
		namePvPCount:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.PvPLeft, 0)
		namePvPCount:SetJustifyH("RIGHT")
		namePvPCount:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])

		local nameItemLevel = frame:CreateFontString(name.."ItemLevel", "ARTWORK", fonts.GameFontNormalSmall.name)
		nameItemLevel:SetWidth(sizeCompareFrame.EquipWidth)
		nameItemLevel:SetHeight(fonts.GameFontWhiteSmall.height)
		nameItemLevel:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.EquipLeft, 0)
		nameItemLevel:SetJustifyH("RIGHT")
		nameItemLevel:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])

		local nameAchievement = frame:CreateFontString(name.."Achievement", "ARTWORK", fonts.GameFontNormalSmall.name)
		nameAchievement:SetWidth(sizeCompareFrame.AchievementWidth)
		nameAchievement:SetHeight(fonts.GameFontWhiteSmall.height)
		nameAchievement:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.AchievementLeft, 0)
		nameAchievement:SetJustifyH("RIGHT")
		nameAchievement:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])

		local nameBoss = frame:CreateFontString(name.."Boss", "ARTWORK", fonts.GameFontNormalSmall.name)
		nameBoss:SetWidth(sizeCompareFrame.BossWidth)
		nameBoss:SetHeight(fonts.GameFontWhiteSmall.height)
		nameBoss:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.BossLeft, 0)
		nameBoss:SetJustifyH("RIGHT")
		nameBoss:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])

		local nameKills = frame:CreateFontString(name.."Kills", "ARTWORK", fonts.GameFontNormalSmall.name)
		nameKills:SetWidth(sizeCompareFrame.KillsWidth)
		nameKills:SetHeight(fonts.GameFontWhiteSmall.height)
		nameKills:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.KillsLeft, 0)
		nameKills:SetJustifyH("RIGHT")
		nameKills:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])

		local nameSingleAchievementCompare = frame:CreateFontString(name.."SingleAchievementCompare", "ARTWORK", fonts.GameFontNormalSmall.name)
		nameSingleAchievementCompare:SetWidth(sizeCompareFrame.SingleAchievementCompareWidth)
		nameSingleAchievementCompare:SetHeight(fonts.GameFontWhiteSmall.height)
		nameSingleAchievementCompare:SetPoint("LEFT", button, "LEFT", sizeCompareFrame.SingleAchievementCompareLeft, 0)
		nameSingleAchievementCompare:SetJustifyH("LEFT")
		nameSingleAchievementCompare:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
		nameSingleAchievementCompare:Hide()

		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", button, "TOPLEFT", sizeCompareFrame.ArrowLeft, 0)--sizeCompareFrame.ClassiconLeft, 0)
		textureHighlight:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 0, 0)
		textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
		button:SetHighlightTexture(textureHighlight)
		local texture = button:CreateTexture(nil, "BACKGROUND")
		texture:SetPoint("TOPLEFT", button, "TOPLEFT", sizeCompareFrame.NameLeft, -5)
		texture:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 0, 5)
		texture:SetTexture(color.ButtonLineMarker.rgba[1], color.ButtonLineMarker.rgba[2], color.ButtonLineMarker.rgba[3], color.ButtonLineMarker.rgba[4])

		button:SetScript("OnClick", function() CharScanner:PreShowSavedData(button.guid, true) end)
	end
end
-- Template HistoryList END ----------------------------------------

-- Template AchievementDetailList START ----------------------------------------
local function TemplateAchievementDetailList(button, isSort)
	local buttonname = button:GetName()

	if isSort then
		-- AchievementDetailSelect START
		local selection = CreateFrame("Button", buttonname.."_AchievementDetailSelectMainButton", button)
		TemplateBorderTRBL(selection)
		TemplateCompareSelect(selection, "title", sizeAchievementDetail.SortLine1Height)
		selection:SetPoint("TOPLEFT", CharScanner_AchievementDetailFrame_Sort, "TOPLEFT", 0, sizeAchievementDetail.SortLine1Height+sizeGeneral.Space)
		_G[buttonname.."_AchievementDetailSelectMainButtonName"]:SetText(ALL)
		_G[buttonname.."_AchievementDetailSelectMainButtonName"]:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
		selection:SetScript("OnClick",
			function()
				if _G[buttonname.."_AchievementDetailSelectFrame"]:IsShown() then
					_G[buttonname.."_AchievementDetailSelectFrame"]:Hide()
					_G[buttonname.."_AchievementDetailSelectMainButton_Expand"]:SetTexCoord(unpack(Textures.Expand.coords))
				else
					_G[buttonname.."_AchievementDetailSelectFrame"]:Show()
					_G[buttonname.."_AchievementDetailSelectMainButton_Expand"]:SetTexCoord(unpack(Textures.Collapse.coords))
				end
			end)

		local buttontexture = selection:CreateTexture(buttonname.."_AchievementDetailSelectMainButton_Expand", "ARTWORK")
		buttontexture:SetHeight(16)
		buttontexture:SetWidth(16)
		buttontexture:SetPoint("RIGHT", selection, "RIGHT", -sizeGeneral.Offset, 0)
		buttontexture:SetTexture(Textures.CharScannerIcons.path)
		buttontexture:SetTexCoord(unpack(Textures.Expand.coords))

		local selectionFrame = CreateFrame("Frame", buttonname.."_AchievementDetailSelectFrame", selection)
		TemplateBorderTRBL(selectionFrame)
		selectionFrame:SetToplevel(true)
		selectionFrame:EnableMouse(true)
		selectionFrame:SetPoint("TOPLEFT", selection, "BOTTOMLEFT", 0, sizeGeneral.Border)
		selectionFrame:Hide()

		local textwidth = 0
		local strw = 0

		local buttonAll = CreateFrame("Button", buttonname.."_AchievementDetailSelect_Button1", selectionFrame)
		TemplateCompareSelect(buttonAll, "level1", sizeGeneral.BarHeight)
		buttonAll:SetPoint("TOPLEFT", buttonname.."_AchievementDetailSelectFrame", "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
		buttonAll:SetScript("OnClick", function() CharScanner:AchievementDetail_Update(AchievementDetailFrameMode, nil, nil, nil, 1) end)
		_G[buttonname.."_AchievementDetailSelect_Button1Name"]:SetText(ALL)
		_G[buttonname.."_AchievementDetailSelect_Button1Name"]:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
		strw = _G[buttonname.."_AchievementDetailSelect_Button1Name"]:GetStringWidth()
		if strw > textwidth then textwidth = strw end

		local buttonComplete = CreateFrame("Button", buttonname.."_AchievementDetailSelect_Button2", selectionFrame)
		TemplateCompareSelect(buttonComplete, "level1", sizeGeneral.BarHeight)
		buttonComplete:SetPoint("TOPLEFT", buttonAll, "BOTTOMLEFT", 0, 0)
		buttonComplete:SetScript("OnClick", function() CharScanner:AchievementDetail_Update(AchievementDetailFrameMode, nil, nil, nil, 2) end)
		_G[buttonname.."_AchievementDetailSelect_Button2Name"]:SetText(COMPLETE)
		_G[buttonname.."_AchievementDetailSelect_Button2Name"]:SetTextColor(color.Green.rgba[1], color.Green.rgba[2], color.Green.rgba[3], color.Green.rgba[4])
		strw = _G[buttonname.."_AchievementDetailSelect_Button2Name"]:GetStringWidth()
		if strw > textwidth then textwidth = strw end

		local buttonIncomplete = CreateFrame("Button", buttonname.."_AchievementDetailSelect_Button3", selectionFrame)
		TemplateCompareSelect(buttonIncomplete, "level1", sizeGeneral.BarHeight)
		buttonIncomplete:SetPoint("TOPLEFT", buttonComplete, "BOTTOMLEFT", 0, 0)
		buttonIncomplete:SetScript("OnClick", function() CharScanner:AchievementDetail_Update(AchievementDetailFrameMode, nil, nil, nil, 3) end)
		_G[buttonname.."_AchievementDetailSelect_Button3Name"]:SetText(INCOMPLETE)
		_G[buttonname.."_AchievementDetailSelect_Button3Name"]:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
		strw = _G[buttonname.."_AchievementDetailSelect_Button3Name"]:GetStringWidth()
		if strw > textwidth then textwidth = strw end

		selectionFrame:SetHeight((3*sizeGeneral.BarHeight) + (2*sizeGeneral.Offset))
		
		-- set width
		textwidth = textwidth + 20

		_G[buttonname.."_AchievementDetailSelectMainButton"]:SetWidth(textwidth + (2*sizeGeneral.Offset))
		_G[buttonname.."_AchievementDetailSelectMainButtonName"]:SetWidth(textwidth + (2*sizeGeneral.Offset))
		_G[buttonname.."_AchievementDetailSelectFrame"]:SetWidth(textwidth + (2*sizeGeneral.Offset))

		for i = 1, 3 do
			_G[buttonname.."_AchievementDetailSelect_Button"..i]:SetWidth(textwidth)
			_G[buttonname.."_AchievementDetailSelect_Button"..i.."Name"]:SetWidth(textwidth)
		end
		-- set width
		-- auto hide
		local function OnLeave()
			if not selection:IsMouseOver() and not selectionFrame:IsMouseOver() then
				_G[buttonname.."_AchievementDetailSelectFrame"]:Hide()
				_G[buttonname.."_AchievementDetailSelectMainButton_Expand"]:SetTexCoord(unpack(Textures.Expand.coords))
			end
		end
		selection:SetScript("OnLeave", OnLeave)
		selection:SetScript("OnHide", function() _G[buttonname.."_AchievementDetailSelectFrame"]:Hide() end) -- for esc close
		selectionFrame:SetScript("OnLeave", OnLeave)
		-- auto hide
		-- AchievementDetailSelect END

		local sort_id = CreateFrame("Button", buttonname.."Button_Id", button)
		TemplateHistoryListSortButton(sort_id, "ID", "right", sizeAchievementDetail.IdWidth)
		sort_id:SetWidth(sizeAchievementDetail.IdWidth)
		sort_id:SetHeight(sizeGeneral.BarHeight)
		sort_id:SetPoint("LEFT", "CharScanner_AchievementDetailFrame_Sort", "LEFT", sizeAchievementDetail.IdLeft, 0)
		sort_id:SetScript("OnClick", function() CharScanner:AchievementDetail_GUISort(1) end)

		local sort_name = CreateFrame("Button", buttonname.."Button_Name", button)
		TemplateHistoryListSortButton(sort_name, NAME, "left", sizeAchievementDetail.NameWidth + sizeGeneral.Space + sizeAchievementDetail.CompareWidth)
		sort_name:SetWidth(sizeAchievementDetail.NameWidth + sizeGeneral.Space + sizeAchievementDetail.CompareWidth)
		sort_name:SetHeight(sizeGeneral.BarHeight)
		sort_name:SetPoint("LEFT", "CharScanner_AchievementDetailFrame_Sort", "LEFT", sizeAchievementDetail.NameLeft, 0)
		sort_name:SetScript("OnClick", function() CharScanner:AchievementDetail_GUISort(2) end)
	else
		local id = button:CreateFontString(buttonname.."Id", "ARTWORK", fonts.GameFontNormalSmall.name)
		id:SetWidth(sizeAchievementDetail.IdWidth)
		id:SetHeight(fonts.GameFontNormalSmall.height)
		id:SetPoint("LEFT", button, "LEFT", sizeAchievementDetail.IdLeft, 0)
		id:SetJustifyH("RIGHT")
		id:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])

		local name = button:CreateFontString(buttonname.."Name", "ARTWORK", fonts.GameFontNormalSmall.name)
		name:SetWidth(sizeAchievementDetail.NameWidth)
		name:SetHeight(fonts.GameFontNormalSmall.height)
		name:SetPoint("LEFT", button, "LEFT", sizeAchievementDetail.NameLeft, 0)
		name:SetJustifyH("LEFT")

		local compareButton = CreateFrame("Button", buttonname.."CompareButton", button)
		compareButton:SetWidth(Textures.RaidSelect.width)
		compareButton:SetHeight(Textures.RaidSelect.height)
		compareButton:SetPoint("LEFT", button, "LEFT", sizeAchievementDetail.CompareLeft, 0)
		compareButton:SetFrameLevel( CharScanner_AchievementDetailFrame:GetFrameLevel()+2 )
		local compareTexture = compareButton:CreateTexture(nil, "ARTWORK")
		compareTexture:SetWidth(Textures.RaidSelect.width)
		compareTexture:SetHeight(Textures.RaidSelect.height)
		compareTexture:SetPoint("LEFT", 0, 0)
		compareTexture:SetTexture(Textures.CharScannerIcons.path)
		compareTexture:SetTexCoord(unpack(Textures.RaidSelect.coords))
		compareTexture:SetAlpha(0.25) 
		compareButton:SetScript("OnEnter", function() compareTexture:SetAlpha(1) end)
		compareButton:SetScript("OnLeave", function() compareTexture:SetAlpha(0.25) end)
		compareButton:SetScript("OnClick", function() CharScanner:CompareFrameShow() CharScanner:History_GUISort("ACHIEVEMENT", GUI_history_SortBy, button.achievementID) end)

		local textureHighlight = button:CreateTexture(nil, "ARTWORK")
		textureHighlight:SetPoint("TOPLEFT", button, "TOPLEFT", 0, 0)
		textureHighlight:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 0, 0)
		textureHighlight:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
		button:SetHighlightTexture(textureHighlight)

		local info = CreateFrame("Button", buttonname.."Info", button)
		TemplateBorderTRBL(info)
		info:SetWidth(sizeAchievementDetail.Width)
		info:SetPoint("TOPLEFT", CharScanner_AchievementDetailFrame, "BOTTOMLEFT", 0, -(sizeGeneral.StatusHeight-(2*sizeGeneral.Border)))
		info:Hide()
		info.title = info:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
		info.title:SetWidth(sizeAchievementDetail.Width-sizeGeneral.Offset-sizeGeneral.Offset)
		info.title:SetPoint("TOPLEFT", info, "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
		info.title:SetJustifyH("LEFT")
		info.texture = info:CreateTexture(nil, "ARTWORK")
		info.texture:SetWidth(sizeAchievementDetail.InfoTexture)
		info.texture:SetHeight(sizeAchievementDetail.InfoTexture)
		info.texture:SetPoint("TOPLEFT", info.title, "BOTTOMLEFT", 0, -sizeGeneral.Offset)
		info.text = info:CreateFontString(buttonname.."InfoText", "OVERLAY", fonts.GameFontWhiteSmall.name)
		info.text:SetWidth(sizeAchievementDetail.InfoTextWidth)
		info.text:SetPoint("TOPLEFT", info.texture, "TOPRIGHT", sizeGeneral.Offset, 0)
		info.text:SetJustifyH("LEFT")

		button:SetScript("OnEnter", function()
			local _, _, _, _, _, _, _, description, _, icon = GetAchievementInfo(button.achievementID)
			info.title:SetText(AchievementIDTable[button.achievementID])
			info.text:SetText(description)
			info.texture:SetTexture(icon)
			local textheight = info.text:GetStringHeight()
			if textheight < sizeAchievementDetail.InfoTexture then
				textheight = sizeAchievementDetail.InfoTexture
			end
			local height = sizeGeneral.Offset + info.title:GetStringHeight() + sizeGeneral.Offset + textheight + sizeGeneral.Offset
			info:SetHeight(height)
			info:Show()
		end)
		button:SetScript("OnLeave", function() info:Hide() end)
	end
end
-- Template AchievementDetailList END ----------------------------------------

-- Template SearchList START ----------------------------------------
local function TemplateSearchList(button, isSort)
	local buttonname = button:GetName()

	if isSort then
		local sort_standard = CreateFrame("Button", buttonname.."Button_Standard", button)
		TemplateHistoryListSortButton(sort_standard, L["Standard"], "left", sizeStatFrame.SearchSortStdWidth)
		sort_standard:SetWidth(sizeStatFrame.SearchSortStdWidth)
		sort_standard:SetHeight(sizeGeneral.BarHeight)
		sort_standard:SetPoint("LEFT", "CharScanner_MainFrame_PvEStats_Search_Sort", "LEFT", sizeStatFrame.SearchSortStdLeft, 0)
		sort_standard:SetScript("OnClick", function() CharScanner:ABKsearch_GUISort(1) end)

		local sort_name = CreateFrame("Button", buttonname.."Button_Name", button)
		TemplateHistoryListSortButton(sort_name, NAME, "left", sizeStatFrame.SearchSortNameWidth)
		sort_name:SetWidth(sizeStatFrame.SearchSortNameWidth)
		sort_name:SetHeight(sizeGeneral.BarHeight)
		sort_name:SetPoint("LEFT", "CharScanner_MainFrame_PvEStats_Search_Sort", "LEFT", sizeStatFrame.SearchSortNameLeft, 0)
		sort_name:SetScript("OnClick", function() CharScanner:ABKsearch_GUISort(2) end)

		local sort_kills = CreateFrame("Button", buttonname.."Button_Kills", button)
		TemplateHistoryListSortButton(sort_kills, L["Kills"], "right", sizeStatFrame.SearchSortKillsWidth)
		sort_kills:SetWidth(sizeStatFrame.SearchSortKillsWidth)
		sort_kills:SetHeight(sizeGeneral.BarHeight)
		sort_kills:SetPoint("RIGHT", "CharScanner_MainFrame_PvEStats_Search_Sort", "RIGHT", -sizeStatFrame.BossesTotalRight, 0)
		sort_kills:SetScript("OnClick", function() CharScanner:ABKsearch_GUISort(3) end)

		--local sorttext = button:CreateFontString(nil, "OVERLAY", fonts.GameFontNormalSmall.name)
		--sorttext:SetPoint("RIGHT", sort_standard, "LEFT", -sizeGeneral.Space, 0)
		--sorttext:SetJustifyH("RIGHT")
		--sorttext:SetText("Sort:")
		--sorttext:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
	end
end
-- Template SearchList END ----------------------------------------

-- Template TalentButton START ----------------------------------------
local function TemplateTalentButton(button)
	button:SetWidth(Textures.TalentIconTexture.qsize)
	button:SetHeight(Textures.TalentIconTexture.qsize)
	local name = button:GetName()

	local border = button:CreateTexture(name.."Border", "BACKGROUND")
	border:SetWidth(Textures.TalentIconTexture.qsize)
	border:SetHeight(Textures.TalentIconTexture.qsize)
	border:SetAllPoints(button)
	border:SetTexture(0.2, 0.2, 0.2, 1)

	local texture = button:CreateTexture(name.."IconTexture", "BORDER")
	texture:SetWidth(Textures.TalentIconTexture.qsize-2)
	texture:SetHeight(Textures.TalentIconTexture.qsize-2)
	texture:SetPoint("TOPLEFT", 1, -1)
	texture:SetTexCoord(5/64, 59/64, 5/64, 59/64)

	local rankborder = button:CreateTexture(name.."RankBorder", "ARTWORK")
	rankborder:SetWidth(Textures.TalentRankBorder.qsize)
	rankborder:SetHeight(Textures.TalentRankBorder.qsize)
	rankborder:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 2, -2)
	rankborder:SetTexture(Textures.CharScannerIcons.path)
	rankborder:SetTexCoord(unpack(Textures.TalentRankBorder.coords))

	local ranktext = button:CreateFontString(name.."RankNumber", "OVERLAY", fonts.GameFontNormalSmall.name)
	ranktext:SetWidth(12)
	ranktext:SetHeight(12)
	ranktext:SetPoint("CENTER", rankborder, "CENTER", 0, 0)
	ranktext:SetJustifyH("CENTER")
end
-- Template TalentButton END ----------------------------------------

-- Template GearButton START ----------------------------------------
local function TemplateGearButton(button, pos)
	button:SetWidth(Textures.GearIconTexture.qsize)
	button:SetHeight(Textures.GearIconTexture.qsize)
	local name = button:GetName()

	local border = button:CreateTexture(name.."Border", "BACKGROUND")
	border:SetWidth(Textures.GearIconTexture.qsize)
	border:SetHeight(Textures.GearIconTexture.qsize)
	border:SetAllPoints(button)
	border:SetTexture(0.2, 0.2, 0.2, 1)

	local texture = button:CreateTexture(name.."IconTexture", "BORDER")
	texture:SetWidth(Textures.GearIconTexture.qsize-2)
	texture:SetHeight(Textures.GearIconTexture.qsize-2)
	texture:SetPoint("TOPLEFT", 1, -1)
	texture:SetTexCoord(5/64, 59/64, 5/64, 59/64)

	local text = button:CreateFontString(name.."Count", "OVERLAY", fonts.GameFontNormalSmall.name)
	if pos == 1 then -- LEFT
		text:SetPoint("RIGHT", texture, "LEFT", -sizeGeneral.Offset, 0)
	elseif pos == 2 then -- RIGHT
		text:SetPoint("LEFT", texture, "RIGHT", sizeGeneral.Offset, 0)
	elseif pos == 3 then -- BOTTOM
		text:SetPoint("TOP", texture, "BOTTOM", 0, -sizeGeneral.Offset)
	end

	local enchant = button:CreateTexture(name.."EnchTexture", "BORDER")
	enchant:SetWidth(Textures.Enchantment.qsize)
	enchant:SetHeight(Textures.Enchantment.qsize)
	if pos == 1 then -- LEFT
		enchant:SetPoint("RIGHT", text, "LEFT", -sizeGeneral.Offset, 0)
	elseif pos == 2 then -- RIGHT
		enchant:SetPoint("LEFT", text, "RIGHT", sizeGeneral.Offset, 0)
	elseif pos == 3 then -- BOTTOM
		enchant:SetPoint("TOP", text, "BOTTOM", 0, -sizeGeneral.Offset)
	end
	enchant:SetTexture(Textures.Enchantment.path)

	local pvp = button:CreateTexture(name.."PvPTexture", "BORDER")
	pvp:SetWidth(Textures.PvP.qsize)
	pvp:SetHeight(Textures.PvP.qsize)
	if pos == 1 then -- LEFT
		pvp:SetPoint("LEFT", texture, "RIGHT", -1, 0)
	elseif pos == 2 then -- RIGHT
		pvp:SetPoint("RIGHT", texture, "LEFT", 1, 0)
	elseif pos == 3 then -- BOTTOM
		pvp:SetPoint("BOTTOM", texture, "TOP", 0, -1)
	end
	pvp:SetTexture(Textures.PvP.path)
end
-- Template GearButton END ----------------------------------------

-- Template CheckButton START ----------------------------------------
local function DisableCheckButton(name, texture)
	if texture then
		_G[name.."Texture"]:SetAlpha(0.2)
	else
		_G[name.."Text"]:SetTextColor(0.5, 0.5, 0.5)
	end
	_G[name.."Border"]:SetTexture(0.4, 0.4, 0.4, 1)
	_G[name]:Disable()
end
local function EnableCheckButton(name, texture)
	if texture then
		_G[name.."Texture"]:SetAlpha(1)
	else
		_G[name.."Text"]:SetTextColor(1, 1, 1)
	end
	_G[name.."Border"]:SetTexture(0.8, 0.2, 0.2, 1)
	_G[name]:Enable()
end

local function TemplateCheckButton(button, size, space, text)
	local buttonName = button:GetName()
	local textureBorder = button:CreateTexture(buttonName.."Border", "BACKGROUND")
	textureBorder:SetWidth( size )
	textureBorder:SetHeight( size )
	textureBorder:SetPoint("LEFT", 0, 0)
	textureBorder:SetTexture(0.4, 0.4, 0.4, 1)

	local textureBackground = button:CreateTexture(nil, "BORDER")
	textureBackground:SetPoint("TOPLEFT", textureBorder, "TOPLEFT", 1, -1)
	textureBackground:SetPoint("BOTTOMRIGHT", textureBorder, "BOTTOMRIGHT", -1, 1)
	textureBackground:SetTexture(0, 0, 0, 1)

	local textureNormal = button:CreateTexture(nil, "ARTWORK")
	textureNormal:SetPoint("TOPLEFT", textureBorder, "TOPLEFT", 1, -1)
	textureNormal:SetPoint("BOTTOMRIGHT", textureBorder, "BOTTOMRIGHT", -1, 1)
	textureNormal:SetTexture(0, 0, 0, 1)
	button:SetNormalTexture(textureNormal)

	local texturePush = button:CreateTexture(nil, "ARTWORK")
	texturePush:SetPoint("TOPLEFT", textureBorder, "TOPLEFT", 4, -4)
	texturePush:SetPoint("BOTTOMRIGHT", textureBorder, "BOTTOMRIGHT", -4, 4)
	texturePush:SetTexture(0.4, 0.4, 0.4, 0.5)
	button:SetPushedTexture(texturePush)

	local textureDisabled = button:CreateTexture(nil, "ARTWORK")
	textureDisabled:SetPoint("TOPLEFT", textureBorder, "TOPLEFT", 3, -3)
	textureDisabled:SetPoint("BOTTOMRIGHT", textureBorder, "BOTTOMRIGHT", -3, 3)
	textureDisabled:SetTexture(0.4, 0.4, 0.4, 0.5)
	button:SetDisabledTexture(textureDisabled)

	local textureChecked = button:CreateTexture(nil, "ARTWORK")
	textureChecked:SetWidth( size )
	textureChecked:SetHeight( size )
	textureChecked:SetPoint("LEFT", 0, 0)
	textureChecked:SetTexture("Interface\\Buttons\\UI-CheckBox-Check")
	button:SetCheckedTexture(textureChecked)

	local buttontext = button:CreateFontString(buttonName.."Text", "OVERLAY", fonts.GameFontNormalSmall.name)
	buttontext:SetHeight( size )
	buttontext:SetPoint("LEFT", textureNormal, "RIGHT", space, 0)
	buttontext:SetJustifyH("LEFT")
	buttontext:SetText(text)
	buttontext:SetTextColor(1, 1, 1, 1)

	button:SetWidth(size + space + buttontext:GetStringWidth() + space)
	button:SetHeight(size)

	local textureHighlight = button:CreateTexture(nil, "OVERLAY")
	textureHighlight:SetPoint("TOPLEFT", button, "TOPLEFT", 0, 0)
	textureHighlight:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 0, 0)
	textureHighlight:SetTexture(1, 1, 1, 0.1)
	textureHighlight:Hide()

	button:SetScript("OnEnter", function() textureHighlight:Show() end)
	button:SetScript("OnLeave", function() textureHighlight:Hide() end)
end
-- Template CheckButton END ----------------------------------------

-- Template Slider START ----------------------------------------
local function TemplateSlider(slider, width, step, minVal, maxVal, curVal, func, pro, scale)
	if not scale then scale = 1 end

	slider:SetWidth(width)
	slider:SetHeight(sizeGeneral.SliderSize)
	slider:SetValueStep(step) 
	slider:SetMinMaxValues(minVal, maxVal)
	slider:SetValue(curVal*scale)
	slider:SetOrientation("HORIZONTAL")

	local textMin = slider:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	textMin:SetPoint("TOP", slider, "BOTTOM", 0, -2)
	textMin:SetPoint("LEFT", slider, "LEFT", 0, 0)
	textMin:SetJustifyH("CENTER")
	textMin:SetTextColor(0.6, 0.6, 0.6, 1)
	--if pro then
	--	textMin:SetText((minVal/scale)..pro)
	--else
		textMin:SetText((minVal/scale))
	--end
	local textMax = slider:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	textMax:SetPoint("TOP", slider, "BOTTOM", 0, -2)
	textMax:SetPoint("RIGHT", slider, "RIGHT", 0, 0)
	textMax:SetJustifyH("CENTER")
	textMax:SetTextColor(0.6, 0.6, 0.6, 1)
	--if pro then
	--	textMax:SetText((maxVal/scale)..pro)
	--else
		textMax:SetText((maxVal/scale))
	--end
	local textCur = slider:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	textCur:SetPoint("TOP", slider, "BOTTOM", 0, -2)
	textCur:SetPoint("CENTER", slider, "CENTER", 0, 0)
	textCur:SetJustifyH("CENTER")
	textCur:SetTextColor(1, 1, 1, 1)
	if pro then
		textCur:SetText(curVal..pro)
	else
		textCur:SetText(curVal)
	end

	local sliderBGL = slider:CreateTexture(nil, "BACKGROUND")
	sliderBGL:SetWidth(5)
	sliderBGL:SetHeight(6)
	sliderBGL:SetPoint("LEFT", slider, "LEFT", 0, 0)
	sliderBGL:SetTexture(Textures.CharScannerIcons.path)
	sliderBGL:SetTexCoord(unpack(Textures.SliderBG.coordsL))
	local sliderBGM = slider:CreateTexture(nil, "BACKGROUND")
	sliderBGM:SetWidth(width-5-5)
	sliderBGM:SetHeight(6)
	sliderBGM:SetPoint("LEFT", sliderBGL, "RIGHT", 0, 0)
	sliderBGM:SetTexture(Textures.CharScannerIcons.path)
	sliderBGM:SetTexCoord(unpack(Textures.SliderBG.coordsM))
	local sliderBGR = slider:CreateTexture(nil, "BACKGROUND")
	sliderBGR:SetWidth(5)
	sliderBGR:SetHeight(6)
	sliderBGR:SetPoint("LEFT", sliderBGM, "RIGHT", 0, 0)
	sliderBGR:SetTexture(Textures.CharScannerIcons.path)
	sliderBGR:SetTexCoord(unpack(Textures.SliderBG.coordsR))

	local thumb = slider:CreateTexture(nil, "BORDER")
	thumb:SetWidth(Textures.SliderKnob.width)
	thumb:SetHeight(Textures.SliderKnob.height)
	thumb:SetTexture(Textures.CharScannerIcons.path)
	thumb:SetTexCoord(unpack(Textures.SliderKnob.coords))
	slider:SetThumbTexture(thumb)

	slider:SetScript("OnValueChanged", function(self, value)
		value = value/scale
		if pro then
			textCur:SetText(value..pro)
		else
			textCur:SetText(value)
		end
		if func then
			func(self, value)
		end
	end)
end
-- Template Slider END ----------------------------------------
-- ---------------------------------------------------------------------------------------------------------------------



-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:InitializeOptions()
	SlashCmdList["CHARSCANNER"] = SlashHandler
	SLASH_CHARSCANNER1 = "/cs"
	SLASH_CHARSCANNER2 = "/scan"
	SLASH_CHARSCANNER3 = "/charscanner"

	if CharScanner_Options.version == nil then
		CharScanner_Options.version = 4
	end

	if CharScanner_Options.version == 1 then
		table.wipe(CharScanner_Options)
		CharScanner_Options.version = 2
		Print("Options reset.")
	end

	if CharScanner_Options.version == 2 then
		CharScanner_Options.version = 3
		CharScanner_Options.scanButtonTrans = nil
	end

	if CharScanner_Options.version == 3 then
		CharScanner_Options.MainFrame_relP = nil
		CharScanner_Options.MainFrame_posX = nil
		CharScanner_Options.MainFrame_posY = nil
		CharScanner_Options.ScanButton_relP = nil
		CharScanner_Options.ScanButton_posX = nil
		CharScanner_Options.ScanButton_posY = nil
		Print("MainFrame position reset.")
		if CharScanner_Options.scanButton then
			CharScanner_Options.scanButtonLock = false
			Print("SCAN-Button position reset! -> "..L["click & move"].." + "..L["Lock SCAN-Button"])
		end
		CharScanner_Options.version = 4
	end

	if CharScanner_Options.scanButton       == nil then CharScanner_Options.scanButton       = false end
	if CharScanner_Options.scanButtonLock   == nil then CharScanner_Options.scanButtonLock   = false end
	if CharScanner_Options.scanButtonSize   == nil then CharScanner_Options.scanButtonSize   = 24    end
	if CharScanner_Options.chatoutput       == nil then CharScanner_Options.chatoutput       = false end
	if CharScanner_Options.chatoutputGrp    == nil then CharScanner_Options.chatoutputGrp    = false end
	if CharScanner_Options.scanDuration     == nil then CharScanner_Options.scanDuration     = 1     end
	if CharScanner_Options.MinimapButton    == nil then CharScanner_Options.MinimapButton    = false end
	if CharScanner_Options.MinimapButtonPos == nil then CharScanner_Options.MinimapButtonPos = -90   end
	if CharScanner_Options.StatsCurRows == nil then
		CharScanner_Options.StatsCurRows = GUI_ABKsearch_curRows
	else
		GUI_ABKsearch_curRows = CharScanner_Options.StatsCurRows
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:LDBcheck()
	if LibStub and LibStub:GetLibrary("CallbackHandler-1.0", true) and LibStub:GetLibrary("LibDataBroker-1.1", true) then
		LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("CharScanner", {
			type = "launcher",
			icon = Textures.CharScannerGUI.path,
			OnClick = function(self, button)
				CharScanner:MiniOptionsToggle(button, 1)
			end,
		})
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:CreateInterfaceOptions()
	local options = CreateFrame("FRAME", "CharScanner_InterfaceOptions")
	options.name = "CharScanner"

	local title = options:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalLarge.name)
	title:SetPoint("TOPLEFT", 16, -16)
	title:SetText("CharScanner")
	title:SetJustifyH("LEFT")
	title:SetJustifyV("TOP")

	CreateFrame("Button", "CharScanner_InterfaceOptions_GUI", CharScanner_InterfaceOptions, "UIPanelButtonTemplate2")
	CharScanner_InterfaceOptions_GUI:SetPoint("TOPLEFT", title, "BOTTOMLEFT", 0, -10)
	CharScanner_InterfaceOptions_GUI:SetWidth(150)
	CharScanner_InterfaceOptions_GUI:SetHeight(24)
	CharScanner_InterfaceOptions_GUIText:SetText(L["Open GUI"])
	CharScanner_InterfaceOptions_GUI:SetScript("OnClick", function(self)
		InterfaceOptionsFrame_Show() -- hide InterfaceOptionsFrame
		HideUIPanel(GameMenuFrame) -- hide GameMenuFrame
		CharScanner:Frame_Toggle("CharScanner_MainFrame") -- show GUI
	end)

	local slashcommandText = options:CreateFontString(nil, "ARTWORK", fonts.GameFontWhiteSmall.name)
	slashcommandText:SetPoint("LEFT", CharScanner_InterfaceOptions_GUI, "RIGHT", 10, 0)
	slashcommandText:SetText("(|cff"..color.LightYellow.hex.."/cs|r - |cff"..color.LightYellow.hex.."/scan|r - |cff"..color.LightYellow.hex.."/charscanner|r)")
	slashcommandText:SetNonSpaceWrap(true)

	CreateFrame("Button", "CharScanner_InterfaceOptions_OPTIONS", CharScanner_InterfaceOptions, "UIPanelButtonTemplate2")
	CharScanner_InterfaceOptions_OPTIONS:SetPoint("TOPLEFT", CharScanner_InterfaceOptions_GUI, "BOTTOMLEFT", 0, -10)
	CharScanner_InterfaceOptions_OPTIONS:SetWidth(150)
	CharScanner_InterfaceOptions_OPTIONS:SetHeight(24)
	CharScanner_InterfaceOptions_OPTIONSText:SetText(L["Open Options"])
	CharScanner_InterfaceOptions_OPTIONS:SetScript("OnClick", function(self)
		InterfaceOptionsFrame_Show() -- hide InterfaceOptionsFrame
		HideUIPanel(GameMenuFrame) -- hide GameMenuFrame
		CharScanner:Frame_Toggle("CharScanner_OptionsFrame") -- show OPTIONS
	end)

	InterfaceOptions_AddCategory(options)
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
function CharScanner:InitializeFrame()
	-- initial achievement & statistics collapse menu appearance
	for i = 1, #DefaultPvEStatsViewData do
		CharScanner:NodePvE_Click("PvE", DefaultPvEStatsViewData[i].lvl,
		                      DefaultPvEStatsViewData[i].name1, DefaultPvEStatsViewData[i].index1,
		                      DefaultPvEStatsViewData[i].name2, DefaultPvEStatsViewData[i].index2,
		                      DefaultPvEStatsViewData[i].name3, DefaultPvEStatsViewData[i].index3)
	end
	for i = 1, #DefaultPvPStatsViewData do
		CharScanner:NodePvP_Click("PvP", DefaultPvPStatsViewData[i].lvl,
		                      DefaultPvPStatsViewData[i].name1, DefaultPvPStatsViewData[i].index1,
		                      DefaultPvPStatsViewData[i].name2, DefaultPvPStatsViewData[i].index2,
		                      DefaultPvPStatsViewData[i].name3, DefaultPvPStatsViewData[i].index3, true)
	end
	-- initial stats height
	local height = (GUI_ABKsearch_curRows*sizeGeneral.BarHeight) + sizeStatFrame.OffsetTopSearch + sizeGeneral.Offset -- (savedRows*12) + 16 + 22 + 5
	CharScanner_MainFrame_Stats:SetHeight(sizeGeneral.TabFrameHeight + height) -- 20 + height
	CharScanner_MainFrame_PvEStats:SetHeight(height) -- height
	CharScanner_MainFrame_PvPStats:SetHeight(height) -- height
	ScrollFrameUpdate()
	-- initialize PvE/PvP tab
	CharScanner:PvEPvPSwitchFrame_Selection_OnClick("PvE")
	-- initialize sort buttons: History/Compare, Achievement Detail, ABKsearch
	CharScanner:History_SortButtonUpdate()
	CharScanner:AchievementDetail_SortButtonUpdate()
	CharScanner:ABKsearch_SortButtonUpdate()
	-- initialize ScanButton (force user to lock)
	if CharScanner_Options.scanButton and not CharScanner_Options.scanButtonLock then
		CharScanner:Frame_Toggle("CharScanner_ScanButtonFrame", 1)
	end
	--CharScanner:ScanFrameToggle(1) -- MENUDEV
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:Frame_Toggle(frameName, show)
	if show then
		_G[frameName]:Show()
	else
		if _G[frameName]:IsShown() then
			_G[frameName]:Hide()
		else
			_G[frameName]:Show()
		end
	end
end

function CharScanner:Frame_SetupPosition(frameName)
	if CharScanner_Options.pos and CharScanner_Options.pos[frameName.."_posX"] then
		_G[frameName]:ClearAllPoints()
		_G[frameName]:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", CharScanner_Options.pos[frameName.."_posX"], CharScanner_Options.pos[frameName.."_posY"])
	else
		if frameName == "CharScanner_ScanButtonFrame" then
			_G[frameName]:ClearAllPoints()
			_G[frameName]:SetPoint("TOP", UIParent, "TOP", 0, -100)
		elseif frameName == "CharScanner_MainFrame" then
			_G[frameName]:ClearAllPoints()
			_G[frameName]:SetPoint("TOPRIGHT", UIParent, "CENTER", -20, 150)
		elseif frameName == "CharScanner_OptionsFrame" then
			_G[frameName]:ClearAllPoints()
			_G[frameName]:SetPoint("TOPLEFT", UIParent, "CENTER", 20, 150)
		else
			_G[frameName]:ClearAllPoints()
			_G[frameName]:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
		end
	end
end

function CharScanner:Frame_SavePosition(frameName)
	if not CharScanner_Options.pos then CharScanner_Options.pos = {} end
	CharScanner_Options.pos[frameName.."_posX"] = _G[frameName]:GetLeft()
	CharScanner_Options.pos[frameName.."_posY"] = _G[frameName]:GetTop()
end

function CharScanner:Frame_MoverInfo(frameName, action, where)
	if not action then
		_G[frameName.."_MoverInfoTop"]:Hide()
		_G[frameName.."_MoverInfoBottom"]:Hide()
		return
	end

	if action == "MOVE" then
		_G[frameName.."_MoverInfoTopText"]:SetText(L["click & move"])
		_G[frameName.."_MoverInfoBottomText"]:SetText(L["click & move"])
	else
		_G[frameName.."_MoverInfoTopText"]:SetText(L["click & resize"])
		_G[frameName.."_MoverInfoBottomText"]:SetText(L["click & resize"])
	end

	if where == "TOP" then
		_G[frameName.."_MoverInfoTop"]:Show()
	else
		_G[frameName.."_MoverInfoBottom"]:Show()
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:ScanFrameShow()
	CharScanner_MainFrame_Base:Show()
	CharScanner_MainFrame_Talent:Show()
	CharScanner_MainFrame_Equipment:Show()
	CharScanner_MainFrame_Stats:Show()
	CharScanner_MainFrame_Summary:Show()
	CharScanner_MainFrame_Status:Show()
	CharScanner_MainFrame_Headline_ButtonCollapseNormal:SetTexCoord(unpack(Textures.Collapse.coords))
	CharScanner_MainFrame_Headline_ButtonCollapsePush:SetTexCoord(unpack(Textures.Collapse.coords))
	if NumScannedChars == 0 then
		DisableIconButton("CharScanner_MainFrame_Headline_ButtonCollapse", 1)
		DisableIconButton("CharScanner_MainFrame_Headline_ButtonHistory")
		DisableIconButton("CharScanner_MainFrame_Headline_ButtonCompare")
		HistoryFrameIsShown = false
		CompareFrameIsShown = false
	else
		EnableIconButton("CharScanner_MainFrame_Headline_ButtonCollapse", 1)
		EnableIconButton("CharScanner_MainFrame_Headline_ButtonHistory")
		EnableIconButton("CharScanner_MainFrame_Headline_ButtonCompare")
	end
end

function CharScanner:ScanFrameHide()
	CharScanner_MainFrame_Base:Hide()
	CharScanner_MainFrame_Talent:Hide()
	CharScanner_MainFrame_Equipment:Hide()
	CharScanner_MainFrame_Stats:Hide()
	CharScanner_MainFrame_Summary:Hide()
	CharScanner_MainFrame_Status:Hide()
	CharScanner_HistoryFrame:Hide()
	CharScanner_MainFrame_Headline_ButtonCollapseBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
	DisableIconButton("CharScanner_MainFrame_Headline_ButtonHistory", 1)
	DisableIconButton("CharScanner_MainFrame_Headline_ButtonCompare", 1)
	CharScanner_MainFrame_Headline_ButtonCollapseNormal:SetTexCoord(unpack(Textures.Expand.coords))
	CharScanner_MainFrame_Headline_ButtonCollapsePush:SetTexCoord(unpack(Textures.Expand.coords))
	HistoryFrameIsShown = false
	CompareFrameIsShown = false
	if NumScannedChars == 0 then
		DisableIconButton("CharScanner_MainFrame_Headline_ButtonCollapse")
	else
		EnableIconButton("CharScanner_MainFrame_Headline_ButtonCollapse")
	end
end

function CharScanner:ScanFrameToggle(show, hide)
	if show then
		CharScanner:ScanFrameShow()
		CharScanner:MainFrame_CalculateHeight()--MHIGH
		return
	end
	if hide then
		CharScanner:ScanFrameHide()
		CharScanner:MainFrame_CalculateHeight(1)--MHIGH
		return
	end
	if CharScanner_MainFrame_Base:IsShown() then
		CharScanner:ScanFrameHide()
		CharScanner:MainFrame_CalculateHeight(1)--MHIGH
	else
		CharScanner:ScanFrameShow()
		CharScanner:MainFrame_CalculateHeight()--MHIGH
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:StatusFrame_CheckPosition()
	if CharScanner_MainFrame_Base:IsShown() then
		CharScanner_MainFrame_Status:SetPoint("TOPLEFT", CharScanner_MainFrame_Summary, "BOTTOMLEFT", 0, sizeGeneral.Border)
	else
		CharScanner_MainFrame_Status:SetPoint("TOPLEFT", CharScanner_MainFrame_Headline, "BOTTOMLEFT", 0, sizeGeneral.Border)
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:DeleteButtonToggle(show)
	if show then
		EnableIconButton("CharScanner_MainFrame_Headline_ButtonDelete", nil, 1)
	else
		DisableIconButton("CharScanner_MainFrame_Headline_ButtonDelete", nil, 1)
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:SingleScanShow()
	EnableIconButton("CharScanner_MainFrame_Headline_ButtonScan", nil)
	CharScanner_MiniOptionsFrame_SingleScan:Enable()
	CharScanner_MiniOptionsFrame_SingleScanText:SetTextColor(color.Yellow.rgba[1], color.Yellow.rgba[2], color.Yellow.rgba[3], color.Yellow.rgba[4])
end

function CharScanner:SingleScanHide()
	DisableIconButton("CharScanner_MainFrame_Headline_ButtonScan", nil)
	CharScanner_MiniOptionsFrame_SingleScan:Disable()
	CharScanner_MiniOptionsFrame_SingleScanText:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:GroupScanToggle()
	if not GroupScanIsActive then
		if inPartyOrRaid then
			CharScanner:GroupScanShow()
		else
			CharScanner:GroupScanHide()
		end
	end
end

function CharScanner:GroupScanShow()
	if inPartyOrRaid then
		EnableIconButton("CharScanner_MainFrame_Headline_ButtonGroupScan", nil)
		CharScanner_MiniOptionsFrame_GroupScanAdd:Enable()
		CharScanner_MiniOptionsFrame_GroupScanAddText:SetTextColor(color.Yellow.rgba[1], color.Yellow.rgba[2], color.Yellow.rgba[3], color.Yellow.rgba[4])
		CharScanner_MiniOptionsFrame_GroupScan:Enable()
		CharScanner_MiniOptionsFrame_GroupScanText:SetTextColor(color.Yellow.rgba[1], color.Yellow.rgba[2], color.Yellow.rgba[3], color.Yellow.rgba[4])
	end
end

function CharScanner:GroupScanHide()
	DisableIconButton("CharScanner_MainFrame_Headline_ButtonGroupScan", nil)
	CharScanner_MiniOptionsFrame_GroupScanAdd:Disable()
	CharScanner_MiniOptionsFrame_GroupScanAddText:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
	CharScanner_MiniOptionsFrame_GroupScan:Disable()
	CharScanner_MiniOptionsFrame_GroupScanText:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:HistoryFrameToggle()
	if CharScanner_HistoryFrame:IsShown() then
		if CompareFrameIsShown then
			HistoryFrameIsShown = true
			CompareFrameIsShown = false
			CharScanner:HistoryCompareReorganize()
		else
			HistoryFrameIsShown = false
			CharScanner_MainFrame_Headline_ButtonHistoryBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
			CharScanner_MainFrame_Headline_ButtonCompareBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
			CharScanner_HistoryFrame:Hide()
		end
	else
		HistoryFrameIsShown = true
		CharScanner_HistoryFrame:Show()
		CharScanner:HistoryCompareReorganize()
	end
end

function CharScanner:HistoryFrameShow()
	HistoryFrameIsShown = true
	CompareFrameIsShown = false
	CharScanner_HistoryFrame:Show()
	CharScanner:HistoryCompareReorganize()
end

function CharScanner:CompareFrameToggle()
	if CharScanner_HistoryFrame:IsShown() then
		if HistoryFrameIsShown then
			CompareFrameIsShown = true
			HistoryFrameIsShown = false
			CharScanner:HistoryCompareReorganize(true)
		else
			CompareFrameIsShown = false
			CharScanner_MainFrame_Headline_ButtonHistoryBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
			CharScanner_MainFrame_Headline_ButtonCompareBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
			CharScanner_HistoryFrame:Hide()
		end
	else
		CompareFrameIsShown = true
		CharScanner_HistoryFrame:Show()
		CharScanner:HistoryCompareReorganize(true)
	end
end

function CharScanner:CompareFrameShow()
	HistoryFrameIsShown = false
	CompareFrameIsShown = true
	CharScanner_HistoryFrame:Show()
	CharScanner:HistoryCompareReorganize(true)
end

function CharScanner:HistoryCompareReorganize(show)
	if show then
		CharScanner_MainFrame_Headline_ButtonHistoryBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
		CharScanner_MainFrame_Headline_ButtonCompareBackground:SetTexture(color.ButtonBGActive.rgba[1], color.ButtonBGActive.rgba[2], color.ButtonBGActive.rgba[3], color.ButtonBGActive.rgba[4])
		CharScanner_HistoryFrame:SetWidth(sizeCompareFrame.CompareWidth)
		CharScanner_HistoryFrame_Sort:SetWidth(sizeCompareFrame.CompareWidth)
		CharScanner_HistoryFrame_Status:SetWidth(sizeCompareFrame.CompareWidth)
		for i = 1, GUI_history_maxRows do
			_G["CharScanner_HistoryFrame_Button"..i]:SetWidth(sizeCompareFrame.CompareButtonWidth)
			_G["CharScanner_HistoryFrame_Button"..i.."_CompareFrame"]:Show()
		end
		CharScanner_HistoryFrame:Hide()
		CharScanner_HistoryFrame:Show()
		CharScanner_HistoryFrame_Sort_CompareFrame:Show()
	else
		CharScanner_MainFrame_Headline_ButtonHistoryBackground:SetTexture(color.ButtonBGActive.rgba[1], color.ButtonBGActive.rgba[2], color.ButtonBGActive.rgba[3], color.ButtonBGActive.rgba[4])
		CharScanner_MainFrame_Headline_ButtonCompareBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
		CharScanner_HistoryFrame:SetWidth(sizeCompareFrame.HistoryWidth)
		CharScanner_HistoryFrame_Sort:SetWidth(sizeCompareFrame.HistoryWidth)
		CharScanner_HistoryFrame_Status:SetWidth(sizeCompareFrame.HistoryWidth)
		for i = 1, GUI_history_maxRows do
			_G["CharScanner_HistoryFrame_Button"..i]:SetWidth(sizeCompareFrame.HistoryButtonWidth)
			_G["CharScanner_HistoryFrame_Button"..i.."_CompareFrame"]:Hide()
		end
		CharScanner_HistoryFrame_Sort_CompareFrame:Hide()
	end
end

function CharScanner:HistoryCompareFrameHide()
	HistoryFrameIsShown = false
	CompareFrameIsShown = false
	CharScanner_MainFrame_Headline_ButtonHistoryBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
	CharScanner_MainFrame_Headline_ButtonCompareBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
	CharScanner_HistoryFrame:Hide()
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:PvEPvPSwitchFrame_Selection_OnClick(mode)
	local button1name = "CharScanner_MainFrame_PvEPvPSwitchFrame_Selection_Button1"
	local button2name = "CharScanner_MainFrame_PvEPvPSwitchFrame_Selection_Button2"
	if mode == "PvE" then
		_G[button1name.."Text"]:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
		_G[button2name.."Text"]:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])

		_G[button1name]:Disable()
		local border1 = _G[button1name.."Border"]
		border1:SetPoint("TOPLEFT"     ,sizeGeneral.Border, -sizeGeneral.Border)
		border1:SetPoint("BOTTOMLEFT"  ,sizeGeneral.Border, 0)
		border1:SetPoint("TOPRIGHT"    ,-sizeGeneral.Border, -sizeGeneral.Border)
		border1:SetPoint("BOTTOMRIGHT" ,-sizeGeneral.Border, 0)

		_G[button2name]:Enable()
		local border2 = _G[button2name.."Border"]
		border2:SetPoint("TOPLEFT"     ,sizeGeneral.Border, -sizeGeneral.Border)
		border2:SetPoint("BOTTOMLEFT"  ,0, sizeGeneral.Border)
		border2:SetPoint("TOPRIGHT"    ,-sizeGeneral.Border, -sizeGeneral.Border)
		border2:SetPoint("BOTTOMRIGHT" ,-sizeGeneral.Border, sizeGeneral.Border)

		CharScanner_MainFrame_PvEStats:Show()
		CharScanner_MainFrame_PvPStats:Hide()
		CharScanner_MainFrame_SummaryButton1PvE:Show()
		CharScanner_MainFrame_SummaryButton2PvE:Show()
		CharScanner_MainFrame_SummaryButton1PvP:Hide()
		CharScanner_MainFrame_SummaryButton2PvP:Hide()
		AchievementDetailFrameMode = "PvE"
		if CharScanner_MainFrame_PvEStats_Title.AchievementDetailShow then
			CharScanner:AchievementDetail_Update("PvE", GUI_achievementdetail_L1, GUI_achievementdetail_L2, GUI_achievementdetail_L3)
		else
			CharScanner:AchievementDetail_Hide(AchievementDetailFrameMode)
		end
	else
		_G[button1name.."Text"]:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
		_G[button2name.."Text"]:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])

		_G[button1name]:Enable()
		local border1 = _G[button1name.."Border"]
		border1:SetPoint("TOPLEFT"     ,sizeGeneral.Border, -sizeGeneral.Border)
		border1:SetPoint("BOTTOMLEFT"  ,0, sizeGeneral.Border)
		border1:SetPoint("TOPRIGHT"    ,-sizeGeneral.Border, -sizeGeneral.Border)
		border1:SetPoint("BOTTOMRIGHT" ,-sizeGeneral.Border, sizeGeneral.Border)

		_G[button2name]:Disable()
		local border2 = _G[button2name.."Border"]
		border2:SetPoint("TOPLEFT"     ,sizeGeneral.Border, -sizeGeneral.Border)
		border2:SetPoint("BOTTOMLEFT"  ,sizeGeneral.Border, 0)
		border2:SetPoint("TOPRIGHT"    ,-sizeGeneral.Border, -sizeGeneral.Border)
		border2:SetPoint("BOTTOMRIGHT" ,-sizeGeneral.Border, 0)

		CharScanner_MainFrame_PvEStats:Hide()
		CharScanner_MainFrame_PvPStats:Show()
		CharScanner_MainFrame_SummaryButton1PvE:Hide()
		CharScanner_MainFrame_SummaryButton2PvE:Hide()
		CharScanner_MainFrame_SummaryButton1PvP:Show()
		CharScanner_MainFrame_SummaryButton2PvP:Show()
		AchievementDetailFrameMode = "PvP"
		if CharScanner_MainFrame_PvPStats_Title.AchievementDetailShow then
			CharScanner:AchievementDetail_Update("PvP", GUI_PvPachievementdetail_L1, GUI_PvPachievementdetail_L2, GUI_PvPachievementdetail_L3)
		else
			CharScanner:AchievementDetail_Hide(AchievementDetailFrameMode)
		end
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:AchievementDetail_OnEnter(mode, buttonname, l1, l2, l3)
	if LatestScanIsComplete then
		local button = _G["CharScanner_MainFrame_"..mode.."Stats_Title"]
		if button.AchievementClicked ~= buttonname then
			CharScanner:AchievementDetail_Highlight(buttonname, true)
			CharScanner:AchievementDetail_Update(mode, l1, l2, l3)
		end
	end
end

function CharScanner:AchievementDetail_OnLeave(mode, buttonname, l1, l2, l3)
	local button = _G["CharScanner_MainFrame_"..mode.."Stats_Title"]
	if button.AchievementClicked ~= buttonname then -- XFIX
		CharScanner:AchievementDetail_Highlight(buttonname, false) -- XFIX
	end -- XFIX
	if LatestScanIsComplete then
		if button.AchievementClicked ~= buttonname then
			--CharScanner:AchievementDetail_Highlight(buttonname, false) -- XFIX
			CharScanner_AchievementDetailFrame:Hide()
			table.wipe(GUI_achievementdetail_data) -- TEST TODO
		end
		if button.AchievementClicked ~= "" then
			CharScanner:AchievementDetail_Update(mode, button.AchievementL1, button.AchievementL2, button.AchievementL3)
		end
	end
end

function CharScanner:AchievementDetail_OnClick(mode, buttonname, l1, l2, l3)
	if LatestScanIsComplete then
		local button = _G["CharScanner_MainFrame_"..mode.."Stats_Title"]
		if button.AchievementClicked == "" then
			button.AchievementClicked = buttonname
			button.AchievementL1 = l1
			button.AchievementL2 = l2
			button.AchievementL3 = l3
			CharScanner:AchievementDetail_Highlight(buttonname, true)
			button.AchievementDetailShow = true
			CharScanner:AchievementDetail_Update(mode, l1, l2, l3)
		elseif button.AchievementClicked == buttonname then
			button.AchievementClicked = ""
			button.AchievementL1 = nil
			button.AchievementL2 = nil
			button.AchievementL3 = nil
			CharScanner:AchievementDetail_Highlight(buttonname, false)
			button.AchievementDetailShow = false
			CharScanner_AchievementDetailFrame:Hide()
		elseif  button.AchievementClicked ~= buttonname then
			CharScanner:AchievementDetail_Highlight(button.AchievementClicked, false)
			button.AchievementClicked = buttonname
			button.AchievementL1 = l1
			button.AchievementL2 = l2
			button.AchievementL3 = l3
			CharScanner:AchievementDetail_Highlight(buttonname, true)
			button.AchievementDetailShow = true
			CharScanner:AchievementDetail_Update(mode, l1, l2, l3)
		end
	end
end

function CharScanner:AchievementDetail_Highlight(buttonname, show)
	if show then
		_G[buttonname.."AchievementDetailIcon"]:SetAlpha(1)
		_G[buttonname.."AchievementDetailHighlight"]:SetTexture(color.ButtonHighlight.rgba[1], color.ButtonHighlight.rgba[2], color.ButtonHighlight.rgba[3], color.ButtonHighlight.rgba[4])
	else
		_G[buttonname.."AchievementDetailIcon"]:SetAlpha(0.25)
		_G[buttonname.."AchievementDetailHighlight"]:SetTexture(0, 0, 0, 0)
	end
end

--function CharScanner:AchievementDetail_Show(mode)
--	CharScanner_AchievementDetailFrame:Show()
--end

function CharScanner:AchievementDetail_Hide(mode)
	local button = _G["CharScanner_MainFrame_"..mode.."Stats_Title"]
	if button.AchievementClicked ~= "" then
		CharScanner:AchievementDetail_Highlight(button.AchievementClicked, false)
		button.AchievementClicked = ""
	end
	button.AchievementDetailShow = false

	CharScanner_AchievementDetailFrame:Hide()
end

function CharScanner:AchievementDetail_HideAll()
	if CharScanner_MainFrame_PvEStats_Title.AchievementClicked ~= "" then
		CharScanner:AchievementDetail_Highlight(CharScanner_MainFrame_PvEStats_Title.AchievementClicked, false)
		CharScanner_MainFrame_PvEStats_Title.AchievementClicked = ""
	end
	CharScanner_MainFrame_PvEStats_Title.AchievementDetailShow = false

	if CharScanner_MainFrame_PvPStats_Title.AchievementClicked ~= "" then
		CharScanner:AchievementDetail_Highlight(CharScanner_MainFrame_PvPStats_Title.AchievementClicked, false)
		CharScanner_MainFrame_PvPStats_Title.AchievementClicked = ""
	end
	CharScanner_MainFrame_PvPStats_Title.AchievementDetailShow = false

	CharScanner_AchievementDetailFrame:Hide()
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:GearFrame_OnClick()
	if GearFrameIsShown then
		CharScanner_MainFrame_Gear:Hide()
		CharScanner:BorderPolice("TRBL", "CharScanner_MainFrame_Equipment")
		CharScanner_MainFrame_Equipment_ButtonCollapseBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
		CharScanner_MainFrame_Equipment_ButtonCollapseNormal:SetTexCoord(unpack(Textures.RightSide.coords))
		CharScanner_MainFrame_Equipment_ButtonCollapsePush:SetTexCoord(unpack(Textures.RightSide.coords))
		GearFrameIsShown = false
	else
		CharScanner_MainFrame_Gear:Show()
		CharScanner:BorderPolice("TBL", "CharScanner_MainFrame_Equipment")
		CharScanner_MainFrame_Equipment_ButtonCollapseBackground:SetTexture(color.ButtonBGActive.rgba[1], color.ButtonBGActive.rgba[2], color.ButtonBGActive.rgba[3], color.ButtonBGActive.rgba[4])
		CharScanner_MainFrame_Equipment_ButtonCollapseNormal:SetTexCoord(unpack(Textures.RightSide.coordsOff))
		CharScanner_MainFrame_Equipment_ButtonCollapsePush:SetTexCoord(unpack(Textures.RightSide.coordsOff))
		GearFrameIsShown = true
	end
end

function CharScanner:GearFrame_OnEnter()
	if not GearFrameIsShown then
		CharScanner_MainFrame_Gear:Show()
		CharScanner:BorderPolice("TBL", "CharScanner_MainFrame_Equipment")
		CharScanner_MainFrame_Equipment:SetFrameLevel( CharScanner_MainFrame_Gear:GetFrameLevel()+10 )
	else
		CharScanner_MainFrame_Gear:Hide() -- for SetToplevel
		CharScanner_MainFrame_Gear:Show()
	end
end

function CharScanner:GearFrame_OnLeave()
	if not GearFrameIsShown then
		CharScanner_MainFrame_Gear:Hide()
		CharScanner:BorderPolice("TRBL", "CharScanner_MainFrame_Equipment")
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:BorderPolice(style, frameName)
	--             T             --
	--        -----------        --
	--      L |         | R      --
	--        -----------        --
	--             B             --
	local border = _G[frameName.."Border"]
	border:ClearAllPoints()
	if style == "TBL" then
	-- TBL:   -----------        --
	--        |                  --
	--        -----------        --
		border:SetPoint("TOPLEFT",     sizeGeneral.Border, -sizeGeneral.Border)
		border:SetPoint("BOTTOMLEFT",  sizeGeneral.Border, sizeGeneral.Border)
		border:SetPoint("TOPRIGHT",    0, -sizeGeneral.Border)
		border:SetPoint("BOTTOMRIGHT", 0, sizeGeneral.Border)
	elseif style == "TRBL" then
	-- TRBL:  -----------        --
	--        |         |        --
	--        -----------        --
		border:SetPoint("TOPLEFT",     sizeGeneral.Border, -sizeGeneral.Border)
		border:SetPoint("BOTTOMRIGHT", -sizeGeneral.Border, sizeGeneral.Border)
	elseif style == "TRB" then
	-- TRB:   -----------        --
	--                  |        --
	--        -----------        --
		border:SetPoint("TOPLEFT",     0, -sizeGeneral.Border)
		border:SetPoint("BOTTOMLEFT",  0, sizeGeneral.Border)
		border:SetPoint("TOPRIGHT",    -sizeGeneral.Border, -sizeGeneral.Border)
		border:SetPoint("BOTTOMRIGHT", -sizeGeneral.Border, sizeGeneral.Border)
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:TalentTreeFrame_OnClick()
	CharScanner_MainFrame_TalentTree_GroupSelection_Button2:EnableMouse(true) -- fast MouseOver Button2 hackfix
	if TalentTreeFrameIsShown then
		CharScanner_MainFrame_TalentTree:Hide()
		CharScanner_MainFrame_Talent_ButtonCollapseBackground:SetTexture(color.ButtonBG.rgba[1], color.ButtonBG.rgba[2], color.ButtonBG.rgba[3], color.ButtonBG.rgba[4])
		CharScanner_MainFrame_Talent_ButtonCollapseNormal:SetTexCoord(unpack(Textures.RightSide.coords))
		CharScanner_MainFrame_Talent_ButtonCollapsePush:SetTexCoord(unpack(Textures.RightSide.coords))
		TalentTreeFrameIsShown = false
	else
		CharScanner_MainFrame_TalentTree:Show()
		CharScanner_MainFrame_Talent_ButtonCollapseBackground:SetTexture(color.ButtonBGActive.rgba[1], color.ButtonBGActive.rgba[2], color.ButtonBGActive.rgba[3], color.ButtonBGActive.rgba[4])
		CharScanner_MainFrame_Talent_ButtonCollapseNormal:SetTexCoord(unpack(Textures.RightSide.coordsOff))
		CharScanner_MainFrame_Talent_ButtonCollapsePush:SetTexCoord(unpack(Textures.RightSide.coordsOff))
		TalentTreeFrameIsShown = true
	end
end

function CharScanner:TalentTreeFrame_OnEnter()
	if not TalentTreeFrameIsShown then
		CharScanner_MainFrame_TalentTree:Show()
		CharScanner_MainFrame_TalentTree_GroupSelection_Button2:EnableMouse(false) -- fast MouseOver Button2 hackfix
	else
		CharScanner_MainFrame_TalentTree:Hide() -- for SetToplevel
		CharScanner_MainFrame_TalentTree:Show()
	end
end

function CharScanner:TalentTreeFrame_OnLeave()
	if not TalentTreeFrameIsShown then
		CharScanner_MainFrame_TalentTree:Hide()
		CharScanner_MainFrame_TalentTree_GroupSelection_Button2:EnableMouse(true) -- fast MouseOver Button2 hackfix
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:ProgressDisplay(scantype, progress, restriction)
	if progress then
		if scantype == "SINGLE" then
			CharScanner_MainFrame_Status_ProgressText:SetPoint("TOP", CharScanner_MainFrame_Status, "TOP", 0, -7)
			CharScanner_MainFrame_Status_ProgressDoneBG:SetPoint("TOP", CharScanner_MainFrame_Status, "TOP", 0, -7)
			CharScanner_MainFrame_Status_ProgressButton:Hide()
		else
			CharScanner_MainFrame_Status_ProgressText:SetPoint("TOP", CharScanner_MainFrame_Status_ProgressButton, "BOTTOM", 0, -2)
			CharScanner_MainFrame_Status_ProgressDoneBG:SetPoint("TOP", CharScanner_MainFrame_Status_ProgressButton, "BOTTOM", 0, -2)
			CharScanner_MainFrame_Status_ProgressButton:Show()
		end
		CharScanner_MainFrame_Status_ProgressText:Show()
		CharScanner_MainFrame_Status_ProgressDoneBG:Show()
		CharScanner_MainFrame_Status_ProgressDone:Show()
		CharScanner_MainFrame_Status_Text:Hide()
	else
		CharScanner_MainFrame_Status_ProgressButton:Hide()
		CharScanner_MainFrame_Status_ProgressText:Hide()
		CharScanner_MainFrame_Status_ProgressDoneBG:Hide()
		CharScanner_MainFrame_Status_ProgressDone:Hide()
		CharScanner_MainFrame_Status_Text:Show()
	end
	if restriction then
		CharScanner_MainFrame_Status_ThrottleText1:Show()
		CharScanner_MainFrame_Status_ThrottleText2:Show()
		CharScanner_MainFrame_Status_ThrottleDoneBG:Show()
		CharScanner_MainFrame_Status_ThrottleDone:Show()
	else
		CharScanner_MainFrame_Status_ThrottleText1:Hide()
		CharScanner_MainFrame_Status_ThrottleText2:Hide()
		CharScanner_MainFrame_Status_ThrottleDoneBG:Hide()
		CharScanner_MainFrame_Status_ThrottleDone:Hide()
	end

	if scantype == "SINGLE" then
		CharScanner_MainFrame_Status:SetHeight(sizeGeneral.StatusHeight)
	else
		if progress then
			CharScanner_MainFrame_Status:SetHeight(46)
			if restriction then
				CharScanner_MainFrame_Status:SetHeight(81)
			end
		else
			CharScanner_MainFrame_Status:SetHeight(sizeGeneral.StatusHeight)
		end
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:MiniOptionsToggle(button, pos)
	if button == "RightButton" then
		if CharScanner_MiniOptionsFrame:IsShown() then
			if CharScanner_MiniOptionsFrame.position and CharScanner_MiniOptionsFrame.position ~= pos then
				if pos == 1 then
					local xpos, ypos = GetCursorPosition()
					CharScanner_MiniOptionsFrame:ClearAllPoints()
					CharScanner_MiniOptionsFrame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", xpos, ypos)
				else
					CharScanner_MiniOptionsFrame:ClearAllPoints()
					CharScanner_MiniOptionsFrame:SetPoint("TOP", CharScanner_MinimapButton, "BOTTOM", 0, 0)	
				end
				CharScanner_MiniOptionsFrame.position = pos
				MiniOptions_TimerStart = GetTime()
				return
			end

			CharScanner_MiniOptionsFrame:Hide()
			CharScanner_MiniOptionsFrame:SetScript("OnUpdate", nil)
		else
			if pos == 1 then
				local xpos, ypos = GetCursorPosition()
				CharScanner_MiniOptionsFrame:ClearAllPoints()
				CharScanner_MiniOptionsFrame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", xpos, ypos)
			else
				CharScanner_MiniOptionsFrame:ClearAllPoints()
				CharScanner_MiniOptionsFrame:SetPoint("TOP", CharScanner_MinimapButton, "BOTTOM", 0, 0)	
			end

			CharScanner_MiniOptionsFrame.position = pos

			CharScanner_MiniOptionsFrame:Show()
			MiniOptions_TimerStart = GetTime()
			CharScanner_MiniOptionsFrame:SetScript("OnUpdate", MiniOptionsTimer)
		end
	else
		CharScanner_MiniOptionsFrame:Hide()
		CharScanner_MiniOptionsFrame:SetScript("OnUpdate", nil)
		CharScanner:Frame_Toggle("CharScanner_MainFrame")
	end
end
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:MainFrame_Hide()
	ResizeEnd() -- STOP resize
	PlaySound("igQuestListClose")
	CharScanner_InterfaceOptions_GUI:Enable()
	EnableTextButton("CharScanner_OptionsFrame_OpenGUI", 1)
	EnableTextButton("CharScanner_ScanButtonFrame_GUI", 1)
end

function CharScanner:MainFrame_Show()
	PlaySound("igQuestListOpen")
	CharScanner_InterfaceOptions_GUI:Disable()
	DisableTextButton("CharScanner_OptionsFrame_OpenGUI")
	DisableTextButton("CharScanner_ScanButtonFrame_GUI")
	CharScanner:Frame_SetupPosition("CharScanner_MainFrame")
	CharScanner:History_Update()
end
-- -----------------------------------------------------------------------------


-- -----------------------------------------------------------------------------
--[[ TODO
function CharScanner:AchievementDetail_FreeBarToggle()
	if CharScanner_AchievementDetailFrame_Headline:IsShown() then
		CharScanner_AchievementDetailFrame_Headline:Hide()
		CharScanner_AchievementDetailFrame_ButtonX:SetParent("CharScanner_AchievementDetailFrame")
		CharScanner_AchievementDetailFrame_ButtonFree:SetParent("CharScanner_AchievementDetailFrame")
		
		CharScanner_AchievementDetailFrame:SetParent("CharScanner_MainFrame_Stats")
		CharScanner_AchievementDetailFrame:ClearAllPoints()
		CharScanner_AchievementDetailFrame:SetPoint("TOPLEFT", CharScanner_MainFrame_Base, "TOPRIGHT", -sizeGeneral.Border, 0)
		CharScanner_AchievementDetailFrame:SetMovable(false)
		CharScanner_AchievementDetailFrame:SetToplevel(false)
		CharScanner_AchievementDetailFrame:SetClampedToScreen(false)
	else
		CharScanner_AchievementDetailFrame_Headline:Show()
		CharScanner_AchievementDetailFrame_ButtonX:SetParent("CharScanner_AchievementDetailFrame_Headline")
		CharScanner_AchievementDetailFrame_ButtonFree:SetParent("CharScanner_AchievementDetailFrame_Headline")
		
		local top  = CharScanner_AchievementDetailFrame:GetTop()
		local left = CharScanner_AchievementDetailFrame:GetLeft()
		
		CharScanner_AchievementDetailFrame:SetParent(UIParent)
		CharScanner_AchievementDetailFrame:ClearAllPoints()
		CharScanner_AchievementDetailFrame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", left+5, top)
		CharScanner_AchievementDetailFrame:SetMovable(true)
		CharScanner_AchievementDetailFrame:SetToplevel(true)
		CharScanner_AchievementDetailFrame:SetClampedToScreen(true)
	end
end
--]]
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function CharScanner:BossSearch()
	BossSearchQuery = CharScanner_MainFrame_PvEStats_Search:GetText()
	if BossSearchQuery == "" then
		BossSearchQuery = nil
	end
	if BossSearchQuery then
		CharScanner_MainFrame_PvEStats_SearchClear:Enable()
	else
		CharScanner_MainFrame_PvEStats_SearchClear:Disable()
	end
	CharScanner:ABKsearch_Update()
end

function CharScanner:BossSearchClear()
	CharScanner_MainFrame_PvEStats_Search:SetText("")
	BossSearchQuery = nil
	CharScanner_MainFrame_PvEStats_Search:ClearFocus()
	CharScanner:ABKsearch_Update()
end
-- -----------------------------------------------------------------------------
-- ---------------------------------------------------------------------------------------------------------------------


-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:MainFrame_CalculateHeight(basic)--MHIGH
	if basic then
		CharScanner_MainFrame:SetHeight(sizeGeneral.MainHeight)
		CharScanner_MainFrame:SetClampRectInsets(0, 0, 0, 0)
	else
		local height =
		sizeGeneral.MainHeight + -- Headline
		(sizeBaseCharFrame_BaseHeight - sizeGeneral.Border) + -- Base Char Data
		(sizeTalentFrame.BaseHeight - sizeGeneral.Border) + -- Talent
		(CharScanner_MainFrame_Equipment:GetHeight() - sizeGeneral.Border) + -- Equipment
		(CharScanner_MainFrame_Stats:GetHeight() - sizeGeneral.Border) + -- Stats
		(sizeSummaryFrame_BaseHeight - sizeGeneral.Border) + -- Summary
		(sizeGeneral.StatusHeight - sizeGeneral.Border); -- Status

		CharScanner:Frame_SetupPosition("CharScanner_MainFrame")

		CharScanner_MainFrame:SetHeight(height)
		CharScanner_MainFrame:SetClampRectInsets(0, 0, 0, (height-sizeGeneral.MainHeight))
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:CreateMainFrame()
	-- Main Frame START
	local mainframe = CreateFrame("Frame", "CharScanner_MainFrame", UIParent)
	mainframe:EnableMouse(true)
	mainframe:SetMovable(true)
	mainframe:SetToplevel(true)
	mainframe:SetClampedToScreen(true)
	mainframe:SetWidth(sizeGeneral.MainWidth)
	CharScanner:MainFrame_CalculateHeight(1)--MHIGH
	mainframe:Hide()
	-- mainframe:SetScript at end of function
	-- Main Frame END
	

	-- Headline Frame START
	local headline = CreateFrame("Frame", "CharScanner_MainFrame_Headline", CharScanner_MainFrame)
	TemplateBorderTRBL(headline)
	headline:SetWidth(sizeGeneral.MainWidth)
	headline:SetHeight(sizeGeneral.MainHeight)
	headline:SetPoint("TOPLEFT", CharScanner_MainFrame, "TOPLEFT", 0, 0)
	headline:EnableMouse(true)

	local headlineButtonScan = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonScan", headline)
	TemplateIconButton(headlineButtonScan, 7)
	headlineButtonScan:SetPoint("LEFT", headline, "LEFT", sizeGeneral.Offset, 0)
	headlineButtonScan:SetScript("OnClick", function() CharScanner:ScanUnit("target") end)

	local headlineButtonGroupScan = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonGroupScan", headline)
	TemplateIconButton(headlineButtonGroupScan, 1)
	headlineButtonGroupScan:SetPoint("TOPLEFT", headlineButtonScan, "TOPRIGHT", sizeGeneral.Offset, 0)
	headlineButtonGroupScan:Disable()
	CharScanner_MainFrame_Headline_ButtonGroupScanBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])

	local headlineButtonGroupScan2 = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonGroupScan2", headline)
	TemplateIconButton(headlineButtonGroupScan2, 5)
	headlineButtonGroupScan2:SetPoint("BOTTOM", headlineButtonGroupScan, "TOP", 0, 0)
	CharScanner_MainFrame_Headline_ButtonGroupScan2BorderColor:SetTexture(color.ButtonBorder.rgba[1], color.ButtonBorder.rgba[2], color.ButtonBorder.rgba[3], color.ButtonBorder.rgba[4])
	headlineButtonGroupScan2:Hide()

	headlineButtonGroupScan:SetScript("OnClick", function() headlineButtonGroupScan2:Hide() CharScanner:ScanGroup(false) end)
	headlineButtonGroupScan:SetScript("OnEnter",
		function()
			headlineButtonGroupScan2:Show()
		end)
	headlineButtonGroupScan:SetScript("OnLeave",
		function()
			if not headlineButtonGroupScan2:IsMouseOver() then
				headlineButtonGroupScan2:Hide()
			end
		end)

	headlineButtonGroupScan2:SetScript("OnClick", function() headlineButtonGroupScan2:Hide() CharScanner:ScanGroup(true) end)
	headlineButtonGroupScan2:SetScript("OnLeave",
		function()
			if not headlineButtonGroupScan:IsMouseOver() then
				headlineButtonGroupScan2:Hide()
			end
		end)

	local headlineButtonCollapse = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonCollapse", headline)
	TemplateIconButton(headlineButtonCollapse, 6)
	headlineButtonCollapse:SetPoint("TOPLEFT", headlineButtonGroupScan, "TOPRIGHT", sizeGeneral.Offset, 0)
	headlineButtonCollapse:Disable()
	CharScanner_MainFrame_Headline_ButtonCollapseBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
	headlineButtonCollapse:SetScript("OnClick", function() CharScanner:ScanFrameToggle() end)

	local headlineButtonHistory = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonHistory", headline)
	TemplateIconButton(headlineButtonHistory, 2)
	headlineButtonHistory:SetPoint("LEFT", headlineButtonCollapse, "RIGHT", sizeGeneral.Offset, 0)
	headlineButtonHistory:Disable()
	CharScanner_MainFrame_Headline_ButtonHistoryBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
	headlineButtonHistory:SetScript("OnClick", function() CharScanner:HistoryFrameToggle() end)

	local headlineButtonCompare = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonCompare", headline)
	TemplateIconButton(headlineButtonCompare, 3)
	headlineButtonCompare:SetPoint("LEFT", headlineButtonHistory, "RIGHT", sizeGeneral.Offset, 0)
	headlineButtonCompare:Disable()
	CharScanner_MainFrame_Headline_ButtonCompareBorderColor:SetTexture(color.ButtonBorderDisable.rgba[1], color.ButtonBorderDisable.rgba[2], color.ButtonBorderDisable.rgba[3], color.ButtonBorderDisable.rgba[4])
	headlineButtonCompare:SetScript("OnClick", function() CharScanner:CompareFrameToggle() end)

	local headlineButtonX = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonX", headline)
	TemplateIconButton(headlineButtonX, 9)
	headlineButtonX:SetPoint("RIGHT", headline, "RIGHT", -sizeGeneral.Offset, 0)
	headlineButtonX:SetScript("OnClick", function() CharScanner:Frame_Toggle("CharScanner_MainFrame") end)

	local headlineButtonOptions = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonOptions", headline)
	TemplateIconButton(headlineButtonOptions, 8)
	headlineButtonOptions:SetPoint("RIGHT", headlineButtonX, "LEFT", -sizeGeneral.Offset, 0)
	headlineButtonOptions:SetScript("OnClick", function() CharScanner:Frame_Toggle("CharScanner_OptionsFrame", 1) end)

	local headlineButtonDelete = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonDelete", headline)
	TemplateIconButton(headlineButtonDelete, 4)
	headlineButtonDelete:SetPoint("RIGHT", headlineButtonOptions, "LEFT", -sizeGeneral.Offset, 0)
	DisableIconButton("CharScanner_MainFrame_Headline_ButtonDelete", nil, 1)
	headlineButtonDelete:SetScript("OnClick", function() CharScanner:DeleteCharacter() end)

	local headlineButtonMover = CreateFrame("Button", "CharScanner_MainFrame_Headline_ButtonMover", headline)
	headlineButtonMover:SetWidth(22)
	headlineButtonMover:SetHeight(22)
	headlineButtonMover:SetPoint("TOPLEFT", headlineButtonCompare, "TOPRIGHT", sizeGeneral.Offset, 0)
	headlineButtonMover:SetPoint("BOTTOMRIGHT", headlineButtonDelete, "BOTTOMLEFT", -sizeGeneral.Offset, 0)
	local textureMover = headlineButtonMover:CreateTexture(nil, "BACKGROUND")
	textureMover:SetWidth(1)
	textureMover:SetHeight(1)
	textureMover:SetPoint("TOPLEFT", headlineButtonMover, "TOPLEFT", 0, 0)
	textureMover:SetPoint("BOTTOMRIGHT", headlineButtonMover, "BOTTOMRIGHT", 0, 0)
	textureMover:SetTexture(1, 1, 1, 0)
	headlineButtonMover:SetScript("OnEnter", function() textureMover:SetTexture(1, 1, 1, 0.075) CharScanner:Frame_MoverInfo("CharScanner_MainFrame", "MOVE", "TOP") end)
	headlineButtonMover:SetScript("OnLeave", function() textureMover:SetTexture(1, 1, 1, 0) CharScanner:Frame_MoverInfo("CharScanner_MainFrame") end)
	headlineButtonMover:SetScript("OnMouseDown", MoveStart)
	headlineButtonMover:SetScript("OnMouseUp", MoveStop)
	-- Headline Frame END



	-- Base Character Information Frame START
	local base = CreateFrame("Frame", "CharScanner_MainFrame_Base", mainframe)
	TemplateBorderTRBL(base)
	base:SetWidth(sizeGeneral.MainWidth)
	base:SetHeight(sizeBaseCharFrame_BaseHeight)
	base:SetPoint("TOPLEFT", headline, "BOTTOMLEFT", 0, sizeGeneral.Border)
	base:EnableMouse(true)
	base:SetScript("OnShow", function() CharScanner:StatusFrame_CheckPosition() end)
	base:Hide()

	local baseClassIcon = base:CreateTexture("CharScanner_MainFrame_Base_ClassIcon", "ARTWORK")
	baseClassIcon:SetWidth(28)
	baseClassIcon:SetHeight(28)
	baseClassIcon:SetPoint("LEFT", sizeGeneral.Offset, 0)
	baseClassIcon:SetTexture(Textures.ClassIcon.path)

	local baseFactionIcon = base:CreateTexture("CharScanner_MainFrame_Base_FactionIcon", "ARTWORK")
	baseFactionIcon:SetWidth(18)
	baseFactionIcon:SetHeight(18)
	baseFactionIcon:SetPoint("LEFT", baseClassIcon, "RIGHT", 0, 0)

	local baseName = base:CreateFontString("CharScanner_MainFrame_Base_Name", "ARTWORK", fonts.GameFontNormal.name)
	baseName:SetPoint("TOPLEFT", base, "TOPLEFT", sizeGeneral.Offset+28+18+sizeGeneral.Space, -sizeGeneral.Offset)
	baseName:SetJustifyH("LEFT")
	baseName:SetTextColor(1, 1, 1, 1)
	local baseGUID = base:CreateFontString("CharScanner_MainFrame_Base_GUID", "ARTWORK", fonts.GameFontNormalSmall.name)
	baseGUID:SetPoint("BOTTOM", baseName, "BOTTOM")
	baseGUID:SetPoint("RIGHT", base, "RIGHT", -sizeGeneral.Offset, 0)
	baseGUID:SetJustifyH("LEFT")
	baseGUID:SetTextColor(color.ButtonBGActive.rgba[1], color.ButtonBGActive.rgba[2], color.ButtonBGActive.rgba[3], color.ButtonBGActive.rgba[4])
	local baseLevel = base:CreateFontString("CharScanner_MainFrame_Base_Level", "ARTWORK", fonts.GameFontNormalSmall.name)
	baseLevel:SetPoint("TOPLEFT", baseName, "BOTTOMLEFT", 0, -sizeGeneral.Space)
	baseLevel:SetJustifyH("LEFT")
	baseLevel:SetTextColor(1, 1, 1, 1)
	local baseGuild = base:CreateFontString("CharScanner_MainFrame_Base_Guild", "ARTWORK", fonts.GameFontNormalSmall.name)
	baseGuild:SetPoint("TOPLEFT", baseLevel, "BOTTOMLEFT", 0, -sizeGeneral.Space)
	baseGuild:SetJustifyH("LEFT")
	baseGuild:SetTextColor(1, 1, 1, 1)
	-- Base Character Information Frame END



	-- Talent Frame START
	local talent = CreateFrame("Frame", "CharScanner_MainFrame_Talent", mainframe)
	TemplateBorderTRBL(talent)
	talent:SetWidth(sizeGeneral.MainWidth)
	talent:SetHeight(sizeTalentFrame.BaseHeight)
	talent:SetPoint("TOPLEFT", base, "BOTTOMLEFT", 0, sizeGeneral.Border)
	talent:EnableMouse(true)
	talent:Hide()

	local talentTitle = talent:CreateFontString("CharScanner_MainFrame_Talent_Title", "ARTWORK", fonts.GameFontNormalSmall.name)
	talentTitle:SetPoint("TOPLEFT", talent, "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
	talentTitle:SetJustifyH("LEFT")
	talentTitle:SetText(TALENTS..":")

	local talentMainRole = talent:CreateTexture("CharScanner_MainFrame_Talent_TalentMainRole", "ARTWORK")
	talentMainRole:SetWidth(Textures.TalentMiniIcons.qsize*2)
	talentMainRole:SetHeight(Textures.TalentMiniIcons.qsize*2)
	talentMainRole:SetPoint("TOPLEFT", talent, "TOPLEFT", sizeGeneral.Offset+talentTitle:GetStringWidth()+sizeGeneral.Offset, -sizeGeneral.Offset)

	local talentMainRoleActive = talent:CreateTexture("CharScanner_MainFrame_Talent_TalentMainRoleActive", "ARTWORK")
	talentMainRoleActive:SetWidth(Textures.SelectionArrow.width)
	talentMainRoleActive:SetHeight(Textures.SelectionArrow.height)
	talentMainRoleActive:SetTexture(Textures.CharScannerIcons.path)
	talentMainRoleActive:SetTexCoord(unpack(Textures.SelectionArrow.coords))

	local talentGroup1Icon = talent:CreateTexture("CharScanner_MainFrame_Talent_TalentGroup1Icon", "ARTWORK")
	talentGroup1Icon:SetWidth(Textures.TalentMiniIcons.qsize)
	talentGroup1Icon:SetHeight(Textures.TalentMiniIcons.qsize)
	talentGroup1Icon:SetPoint("TOPLEFT", talentMainRole, "TOPRIGHT", sizeGeneral.Offset+Textures.SelectionArrow.width+sizeGeneral.Offset, 0)
	talentGroup1Icon:SetTexCoord(unpack(Textures.TalentMiniIcons.coords))
	local talentGroup1Text = talent:CreateFontString("CharScanner_MainFrame_Talent_TalentGroup1Text", "ARTWORK", fonts.GameFontNormalSmall.name)
	talentGroup1Text:SetPoint("BOTTOM", talentTitle, "BOTTOM", 0, 0)
	talentGroup1Text:SetPoint("LEFT", talentGroup1Icon, "RIGHT", sizeGeneral.Offset, 0)
	talentGroup1Text:SetJustifyH("LEFT")

	local talentGroup2Icon = talent:CreateTexture("CharScanner_MainFrame_Talent_TalentGroup2Icon", "ARTWORK")
	talentGroup2Icon:SetWidth(Textures.TalentMiniIcons.qsize)
	talentGroup2Icon:SetHeight(Textures.TalentMiniIcons.qsize)
	talentGroup2Icon:SetPoint("TOPLEFT", talentGroup1Icon, "BOTTOMLEFT", 0, 0)
	talentGroup2Icon:SetTexCoord(unpack(Textures.TalentMiniIcons.coords))
	local talentGroup2Text = talent:CreateFontString("CharScanner_MainFrame_Talent_TalentGroup2Text", "ARTWORK", fonts.GameFontNormalSmall.name)
	talentGroup2Text:SetPoint("LEFT", talentGroup2Icon, "RIGHT", sizeGeneral.Offset, 0)
	talentGroup2Text:SetJustifyH("LEFT")

	local talentButtonCollapse = CreateFrame("Button", "CharScanner_MainFrame_Talent_ButtonCollapse", talent)
	TemplateIconButton(talentButtonCollapse, 10)
	talentButtonCollapse:SetPoint("RIGHT", talent, "RIGHT", 0, 0)
	talentButtonCollapse:SetScript("OnClick", function() CharScanner:TalentTreeFrame_OnClick() end)
	talentButtonCollapse:SetScript("OnEnter", function() CharScanner:TalentTreeFrame_OnEnter() end)
	talentButtonCollapse:SetScript("OnLeave", function() CharScanner:TalentTreeFrame_OnLeave() end)

	-- - TalentTree Frame START
	local TalentTree = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree", talent)
	TemplateBorderTRBL(TalentTree)
	--TalentTree:SetFrameLevel( CharScanner_MainFrame_Talent:GetFrameLevel() )
	TalentTree:SetWidth(sizeTalentFrame.TreeFrameWidth)
	TalentTree:SetHeight(sizeTalentFrame.TreeFrameHeight)
	TalentTree:SetPoint("TOPLEFT", talent, "TOPRIGHT", -sizeGeneral.Border, sizeGeneral.TabFrameHeight-sizeGeneral.Border)
	TalentTree:EnableMouse(true)
	TalentTree:SetToplevel(true)
	TalentTree:Hide()

	local GroupSelection = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree_GroupSelection", TalentTree)
	TemplateBorderTRBL(GroupSelection)
	GroupSelection:SetWidth(sizeTalentFrame.TreeFrameWidth)
	GroupSelection:SetHeight(sizeGeneral.TabFrameHeight)
	GroupSelection:SetPoint("TOPLEFT", TalentTree, "TOPLEFT", 0, 0)

	local GroupSelection_Button1 = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_GroupSelection_Button1", GroupSelection)
	TemplateTabButton(GroupSelection_Button1, TALENT_SPEC_PRIMARY, sizeTalentFrame.GroupSelectionWidth)
	GroupSelection_Button1:SetWidth(sizeTalentFrame.GroupSelectionWidth)
	GroupSelection_Button1:SetHeight(sizeGeneral.TabButtonHeight)
	GroupSelection_Button1:SetPoint("BOTTOMLEFT", GroupSelection, "BOTTOMLEFT", sizeTalentFrame.GroupSelection1Left, 0)
	GroupSelection_Button1:Hide()

	local GroupSelection_Button2 = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_GroupSelection_Button2", GroupSelection)
	TemplateTabButton(GroupSelection_Button2, TALENT_SPEC_SECONDARY, sizeTalentFrame.GroupSelectionWidth)
	GroupSelection_Button2:SetWidth(sizeTalentFrame.GroupSelectionWidth)
	GroupSelection_Button2:SetHeight(sizeGeneral.TabButtonHeight)
	GroupSelection_Button2:SetPoint("BOTTOMLEFT", GroupSelection, "BOTTOMLEFT", sizeTalentFrame.GroupSelection2Left, 0)
	GroupSelection_Button2:Hide()

	local TalentTree1 = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree1", TalentTree)
	TalentTree1:SetWidth(sizeTalentFrame.TreeWidth)
	TalentTree1:SetHeight(sizeTalentFrame.TreeHeight)
	TalentTree1:SetPoint("TOPLEFT", GroupSelection, "BOTTOMLEFT", sizeTalentFrame.Tree1Left, -sizeGeneral.Offset)
	local TalentTree1Name = TalentTree1:CreateFontString("CharScanner_MainFrame_TalentTree1Name", "ARTWORK", fonts.GameFontWhiteSmall.name)
	TalentTree1Name:SetPoint("TOP", TalentTree1, "TOP", 0, 0)
	TalentTree1Name:SetJustifyH("CENTER")
	local TalentTree1Points = TalentTree1:CreateFontString("CharScanner_MainFrame_TalentTree1Points", "ARTWORK", fonts.GameFontWhiteSmall.name)
	TalentTree1Points:SetPoint("TOP", TalentTree1Name, "BOTTOM", 0, -sizeGeneral.Space)
	TalentTree1Points:SetJustifyH("CENTER")

	local TalentTree2 = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree2", TalentTree)
	TalentTree2:SetWidth(sizeTalentFrame.TreeWidth)
	TalentTree2:SetHeight(sizeTalentFrame.TreeHeight)
	TalentTree2:SetPoint("TOPLEFT", GroupSelection, "BOTTOMLEFT", sizeTalentFrame.Tree2Left, -sizeGeneral.Offset)
	local TalentTree2Name = TalentTree2:CreateFontString("CharScanner_MainFrame_TalentTree2Name", "ARTWORK", fonts.GameFontWhiteSmall.name)
	TalentTree2Name:SetPoint("TOP", TalentTree2, "TOP", 0, 0)
	TalentTree2Name:SetJustifyH("CENTER")
	local TalentTree2Points = TalentTree2:CreateFontString("CharScanner_MainFrame_TalentTree2Points", "ARTWORK", fonts.GameFontWhiteSmall.name)
	TalentTree2Points:SetPoint("TOP", TalentTree2Name, "BOTTOM", 0, -sizeGeneral.Space)
	TalentTree2Points:SetJustifyH("CENTER")

	local TalentTree3 = CreateFrame("Frame", "CharScanner_MainFrame_TalentTree3", TalentTree)
	TalentTree3:SetWidth(sizeTalentFrame.TreeWidth)
	TalentTree3:SetHeight(sizeTalentFrame.TreeHeight)
	TalentTree3:SetPoint("TOPLEFT", GroupSelection, "BOTTOMLEFT", sizeTalentFrame.Tree3Left, -sizeGeneral.Offset)
	local TalentTree3Name = TalentTree3:CreateFontString("CharScanner_MainFrame_TalentTree3Name", "ARTWORK", fonts.GameFontWhiteSmall.name)
	TalentTree3Name:SetPoint("TOP", TalentTree3, "TOP", 0, 0)
	TalentTree3Name:SetJustifyH("CENTER")
	local TalentTree3Points = TalentTree3:CreateFontString("CharScanner_MainFrame_TalentTree3Points", "ARTWORK", fonts.GameFontWhiteSmall.name)
	TalentTree3Points:SetPoint("TOP", TalentTree3Name, "BOTTOM", 0, -sizeGeneral.Space)
	TalentTree3Points:SetJustifyH("CENTER")

	local tier = 1
	local column = 1
	for tab = 1, myMAX_TALENT_TABS do
		for talent = 1, myMAX_NUM_TALENTS do
			local button
			if talent == 1 then
				tier = 1
				column = 1
				button = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column, TalentTree)
				button:SetPoint("TOPLEFT", "CharScanner_MainFrame_TalentTree"..tab, "TOPLEFT", 0, -((2*fonts.GameFontWhiteSmall.height)+(2*sizeGeneral.Space))) -- -(12*-2) font height
			elseif mod(talent, myMAX_COLUMN) == 1 then
				tier = tier + 1
				column = 1
				button = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column, TalentTree)
				button:SetPoint("TOP", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..(tier-1).."_"..(column), "BOTTOM", 0, -sizeGeneral.Offset)
			else
				column = column + 1
				button = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column, TalentTree)
				button:SetPoint("LEFT", "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..(column-1), "RIGHT", sizeGeneral.Offset, 0)
			end
			TemplateTalentButton(button)
		end
	end

	local TalentTreeButtonX = CreateFrame("Button", "CharScanner_MainFrame_TalentTree_ButtonX", TalentTree)
	TemplateIconButton(TalentTreeButtonX, 9, 12)
	TalentTreeButtonX:SetPoint("BOTTOMRIGHT", TalentTree, "TOPRIGHT", 0, -sizeGeneral.Border)
	TalentTreeButtonX:SetScript("OnClick", function() CharScanner:TalentTreeFrame_OnClick() end)
	-- - TalentTree Frame END
	-- Talent Frame END



	-- Equipment Frame START
	local equipment = CreateFrame("Frame", "CharScanner_MainFrame_Equipment", mainframe)
	TemplateBorderTRBL(equipment, 1)
	equipment:SetWidth(sizeGeneral.MainWidth)
	equipment:SetHeight(sizeEquipmentFrame.BaseHeight)
	equipment:SetPoint("TOPLEFT", talent, "BOTTOMLEFT", 0, sizeGeneral.Border)
	equipment:EnableMouse(true)
	equipment:Hide()

	local equipmentTitle = equipment:CreateFontString("CharScanner_MainFrame_Equipment_Title", "ARTWORK", fonts.GameFontNormalSmall.name)
	equipmentTitle:SetPoint("TOPLEFT", equipment, "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
	equipmentTitle:SetJustifyH("LEFT")
	local equipmentText1 = equipment:CreateFontString("CharScanner_MainFrame_Equipment_Text1", "ARTWORK", fonts.GameFontNormalSmall.name)
	equipmentText1:SetPoint("LEFT", equipmentTitle, "RIGHT", sizeGeneral.Offset, 0)
	equipmentText1:SetJustifyH("LEFT")
	equipmentText1:SetTextColor(1, 1, 1, 1)
	local equipmentText2 = equipment:CreateFontString("CharScanner_MainFrame_Equipment_Text2", "ARTWORK", fonts.GameFontNormalSmall.name)
	equipmentText2:SetPoint("TOPLEFT", equipmentText1, "BOTTOMLEFT", 0, -(sizeGeneral.Space+sizeGeneral.Space))
	equipmentText2:SetJustifyH("LEFT")
	equipmentText2:SetTextColor(1, 1, 1, 1)
	local equipmentText3 = equipment:CreateFontString("CharScanner_MainFrame_Equipment_Text3", "ARTWORK", fonts.GameFontNormalSmall.name)
	equipmentText3:SetPoint("TOPLEFT", equipmentText2, "BOTTOMLEFT", 0, -(sizeGeneral.Space+sizeGeneral.Space))
	equipmentText3:SetJustifyH("LEFT")

	local equipmentButtonCollapse = CreateFrame("Button", "CharScanner_MainFrame_Equipment_ButtonCollapse", equipment)
	TemplateIconButton(equipmentButtonCollapse, 10)
	equipmentButtonCollapse:SetPoint("RIGHT", equipment, "RIGHT", 0, 0)
	equipmentButtonCollapse:SetScript("OnClick", function() CharScanner:GearFrame_OnClick() end)
	equipmentButtonCollapse:SetScript("OnEnter", function() CharScanner:GearFrame_OnEnter() end)
	equipmentButtonCollapse:SetScript("OnLeave", function() CharScanner:GearFrame_OnLeave() end)

	-- - Gear Frame START
	local gear = CreateFrame("Frame", "CharScanner_MainFrame_Gear", equipment)
	TemplateBorderTRBL(gear, 1)
	--gear:SetFrameLevel( CharScanner_MainFrame_Equipment:GetFrameLevel() )
	gear:SetWidth(sizeEquipmentFrame.GearWidth)
	gear:SetHeight(sizeEquipmentFrame.GearHeight)
	gear:SetPoint("TOPLEFT", equipment, "TOPRIGHT", -sizeGeneral.Border, 0)
	gear:EnableMouse(true)
	gear:SetToplevel(true)
	gear:Hide()

	for i = 1, #slotButton do
		local button = CreateFrame("Button", "CharScanner_MainFrame_Gear_Slot"..i, gear)
		if i == 1 then
			TemplateGearButton(button, 1)
			button:SetPoint("TOPLEFT", gear, "TOPLEFT", sizeEquipmentFrame.FirstSlotLeft, -sizeGeneral.Offset)
		else
			if i >= 2 and i <= 8 then
				TemplateGearButton(button, 1)
				button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-1), "BOTTOMLEFT", 0, 0)
			else
				if i == 9 then
					TemplateGearButton(button, 3)
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-1), "TOPRIGHT", Textures.GearIconTexture.qsize+(Textures.GearIconTexture.qsize/2), -(Textures.GearIconTexture.qsize/2))
				else
					if i >= 10 and i <= 11 then
						TemplateGearButton(button, 3)
						button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-1), "TOPRIGHT", 0, 0)
					else
						if i == 12 then
							TemplateGearButton(button, 2)
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-1), "TOPRIGHT", Textures.GearIconTexture.qsize+(Textures.GearIconTexture.qsize/2), (Textures.GearIconTexture.qsize/2))
						else
							TemplateGearButton(button, 2)
							button:SetPoint("BOTTOMLEFT", "CharScanner_MainFrame_Gear_Slot"..(i-1), "TOPLEFT", 0, 0)
						end
					end
				end
			end
		end
		_G["CharScanner_MainFrame_Gear_Slot"..i.."IconTexture"]:SetTexture(slotButton[i].texture)
		_G["CharScanner_MainFrame_Gear_Slot"..i.."Count"]:Hide()
		_G["CharScanner_MainFrame_Gear_Slot"..i.."EnchTexture"]:Hide()
		_G["CharScanner_MainFrame_Gear_Slot"..i.."PvPTexture"]:Hide()

		if not slotButton[i].IQ then -- disabled slots
			_G["CharScanner_MainFrame_Gear_Slot"..i.."Border"]:SetTexture(0, 0, 0, 0)
			_G["CharScanner_MainFrame_Gear_Slot"..i.."IconTexture"]:SetAlpha(0.4)
		end
	end

	local gearButtonX = CreateFrame("Button", "CharScanner_MainFrame_Gear_ButtonX", gear)
	TemplateIconButton(gearButtonX, 9, 12)
	gearButtonX:SetPoint("BOTTOMRIGHT", gear, "TOPRIGHT", 0, -sizeGeneral.Border)
	gearButtonX:SetScript("OnClick", function() CharScanner:GearFrame_OnClick() end)
	-- - Gear Frame END
	-- Equipment Frame END



	-- PvE / PvP Achievement & Statistics Frame START
	local statsframe = CreateFrame("Frame", "CharScanner_MainFrame_Stats", mainframe)
	TemplateBorderTRBL(statsframe)
	statsframe:SetWidth(sizeGeneral.MainWidth)
	statsframe:SetHeight(sizeStatFrame.BaseHeight)
	statsframe:SetMinResize(sizeGeneral.MainWidth, sizeStatFrame.Stats_minHeight) -- (minWidth, minHeight)
	statsframe:SetMaxResize(sizeGeneral.MainWidth, sizeStatFrame.Stats_maxHeight) -- (maxWidth, maxHeight)
	statsframe:SetPoint("TOPLEFT", equipment, "BOTTOMLEFT", 0, sizeGeneral.Border)
	statsframe:EnableMouse(true)
	statsframe:Hide()

	-- - PvE / PvP Switch Frame START
	local pvepvpSwitch = CreateFrame("Frame", "CharScanner_MainFrame_PvEPvPSwitchFrame", statsframe)
	TemplateBorderTRBL(pvepvpSwitch)
	pvepvpSwitch:SetWidth(sizeGeneral.MainWidth)
	pvepvpSwitch:SetHeight(sizeGeneral.TabFrameHeight)
	pvepvpSwitch:SetPoint("TOPLEFT", statsframe, "TOPLEFT", 0, 0)
	pvepvpSwitch:EnableMouse(true)

	local statsTitle = pvepvpSwitch:CreateFontString("CharScanner_MainFrame_Stats_TopTitle", "ARTWORK", fonts.GameFontNormalSmall.name)
	statsTitle:SetPoint("TOPLEFT", pvepvpSwitch, "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
	statsTitle:SetJustifyH("LEFT")
	statsTitle:SetText(L["Statistics"]..":")

	local selection_Button1 = CreateFrame("Button", "CharScanner_MainFrame_PvEPvPSwitchFrame_Selection_Button1", pvepvpSwitch)
	TemplateTabButton(selection_Button1, L["PvE"], sizeStatFrame.TabButtonWidth)
	selection_Button1:SetWidth(sizeStatFrame.TabButtonWidth)
	selection_Button1:SetHeight(sizeGeneral.TabButtonHeight)
	selection_Button1:SetPoint("BOTTOMLEFT", pvepvpSwitch, "BOTTOMLEFT", sizeStatFrame.TabButton1Left, 0)
	selection_Button1:SetScript("OnClick", function() CharScanner:PvEPvPSwitchFrame_Selection_OnClick("PvE") end)

	local selection_Button2 = CreateFrame("Button", "CharScanner_MainFrame_PvEPvPSwitchFrame_Selection_Button2", pvepvpSwitch)
	TemplateTabButton(selection_Button2, L["PvP"], sizeStatFrame.TabButtonWidth)
	selection_Button2:SetWidth(sizeStatFrame.TabButtonWidth)
	selection_Button2:SetHeight(sizeGeneral.TabButtonHeight)
	selection_Button2:SetPoint("BOTTOMLEFT", pvepvpSwitch, "BOTTOMLEFT", sizeStatFrame.TabButton2Left, 0)
	selection_Button2:SetScript("OnClick", function() CharScanner:PvEPvPSwitchFrame_Selection_OnClick("PvP") end)
	-- - PvE / PvP Switch Frame END

	local statsPvEframe = CreateFrame("Frame", "CharScanner_MainFrame_PvEStats", statsframe)
	statsPvEframe:SetWidth(sizeGeneral.MainWidth)
	statsPvEframe:SetHeight(sizeStatFrame.BaseHeight)
	statsPvEframe:SetPoint("TOPLEFT", pvepvpSwitch, "BOTTOMLEFT", 0, 0)
	statsPvEframe:Hide()
	local statsPvPframe = CreateFrame("Frame", "CharScanner_MainFrame_PvPStats", statsframe)
	statsPvPframe:SetWidth(sizeGeneral.MainWidth)
	statsPvPframe:SetHeight(sizeStatFrame.BaseHeight)
	statsPvPframe:SetPoint("TOPLEFT", pvepvpSwitch, "BOTTOMLEFT", 0, 0)
	statsPvPframe:Hide()

	-- - PvE START
	-- - - PvE search box START
	local searchBoxIcon = statsPvEframe:CreateTexture(nil, "ARTWORK")
	searchBoxIcon:SetWidth(Textures.SearchBoxIcons.width) -- 17
	searchBoxIcon:SetHeight(Textures.SearchBoxIcons.height) --17 
	searchBoxIcon:SetPoint("TOPLEFT", statsPvEframe, "TOPLEFT", sizeGeneral.Offset, -(sizeGeneral.Space+sizeGeneral.Border))
	searchBoxIcon:SetTexture("Interface\\Minimap\\Tracking\\None")
	searchBoxIcon:SetTexCoord(3/32, 31/32, 1/32, 29/32)

	local searchBox = CreateFrame("EditBox", "CharScanner_MainFrame_PvEStats_Search", statsPvEframe, "InputBoxTemplate")
	searchBox:SetWidth(80)
	searchBox:SetHeight(Textures.SearchBoxIcons.height)
	searchBox:SetPoint("LEFT", searchBoxIcon, "RIGHT", sizeGeneral.Space, 0)
	searchBox:SetAutoFocus(false)

	CharScanner_MainFrame_PvEStats_SearchLeft:SetWidth(5)
	CharScanner_MainFrame_PvEStats_SearchLeft:SetHeight(Textures.SearchBoxIcons.height)
	CharScanner_MainFrame_PvEStats_SearchLeft:SetPoint("LEFT", 0, 0)
	CharScanner_MainFrame_PvEStats_SearchLeft:SetTexture(Textures.CharScannerIcons.path)
	CharScanner_MainFrame_PvEStats_SearchLeft:SetTexCoord(unpack(Textures.SearchBoxIcons.coordsL))

	CharScanner_MainFrame_PvEStats_SearchRight:SetWidth(5)
	CharScanner_MainFrame_PvEStats_SearchRight:SetHeight(Textures.SearchBoxIcons.height)
	CharScanner_MainFrame_PvEStats_SearchRight:SetPoint("RIGHT", 0, 0)
	CharScanner_MainFrame_PvEStats_SearchRight:SetTexture(Textures.CharScannerIcons.path)
	CharScanner_MainFrame_PvEStats_SearchRight:SetTexCoord(unpack(Textures.SearchBoxIcons.coordsR))

	CharScanner_MainFrame_PvEStats_SearchMiddle:SetWidth(1)
	CharScanner_MainFrame_PvEStats_SearchMiddle:SetHeight(Textures.SearchBoxIcons.height)
	CharScanner_MainFrame_PvEStats_SearchMiddle:SetPoint("LEFT", CharScanner_MainFrame_PvEStats_SearchLeft, "RIGHT", 0, 0)
	CharScanner_MainFrame_PvEStats_SearchMiddle:SetPoint("RIGHT", CharScanner_MainFrame_PvEStats_SearchRight, "LEFT", 0, 0)
	CharScanner_MainFrame_PvEStats_SearchMiddle:SetTexture(Textures.CharScannerIcons.path)
	CharScanner_MainFrame_PvEStats_SearchMiddle:SetTexCoord(unpack(Textures.SearchBoxIcons.coordsM))

	searchBox:SetScript("OnShow",         function(self) self:SetAutoFocus(false) end)
	searchBox:SetScript("OnEnterPressed", function(self) self:ClearFocus() end)
	searchBox:SetScript("OnTextChanged",  function() CharScanner:BossSearch() end)

	local searchclear = CreateFrame("Button", "CharScanner_MainFrame_PvEStats_SearchClear", searchBox)
	searchclear:SetWidth(Textures.SearchBoxIcons.width)
	searchclear:SetHeight(Textures.SearchBoxIcons.height)
	searchclear:SetPoint("LEFT", searchBox, "RIGHT", sizeGeneral.Space, 0)

	local textureNormal = searchclear:CreateTexture(nil, "ARTWORK")
	textureNormal:SetPoint("TOPLEFT", 0, 0)
	textureNormal:SetPoint("BOTTOMRIGHT", 0, 0)
	textureNormal:SetAlpha(0.85)
	textureNormal:SetTexture(Textures.CharScannerIcons.path)
	textureNormal:SetTexCoord(unpack(Textures.SearchBoxIcons.coordsCLR))
	searchclear:SetNormalTexture(textureNormal)

	local textureDisable = searchclear:CreateTexture(nil, "ARTWORK")
	textureDisable:SetPoint("TOPLEFT", 0, 0)
	textureDisable:SetPoint("BOTTOMRIGHT", 0, 0)
	textureDisable:SetTexture(Textures.CharScannerIcons.path)
	textureDisable:SetTexCoord(unpack(Textures.SearchBoxIcons.coordsCLR))
	searchclear:SetDisabledTexture(textureDisable)

	local shaderSupported = textureDisable:SetDesaturated(true)
	if not shaderSupported then
		textureDisable:SetVertexColor(0.5, 0.5, 0.5)
	end

	searchclear:Disable()
	searchclear:SetScript("OnClick", function()
		CharScanner:BossSearchClear()
	end)
	searchclear:SetScript("OnMouseDown", function()
		textureNormal:SetPoint("TOPLEFT", 1, -1)
		textureNormal:SetPoint("BOTTOMRIGHT", -1, 1)
	end)
	searchclear:SetScript("OnMouseUp", function()
		textureNormal:SetPoint("TOPLEFT", 0, 0)
		textureNormal:SetPoint("BOTTOMRIGHT", 0, 0)
	end)
	searchclear:SetScript("OnEnter", function()
		textureNormal:SetAlpha(1)
	end)
	searchclear:SetScript("OnLeave", function()
		textureNormal:SetAlpha(0.85)
	end)
	-- - - PvE search box END

	local button = CreateFrame("Button", "CharScanner_MainFrame_PvEStats_Title", statsPvEframe)
	TemplatePvEStats(button, false, "title")
	button:SetPoint("BOTTOMLEFT", searchBoxIcon, "BOTTOMLEFT", 0, 0)
	button.AchievementClicked = ""
	button.AchievementDetailShow = false

	-- - - PvE scrollframe normal START
	local statsScrollFrame = CreateFrame("ScrollFrame", "CharScanner_MainFrame_PvEStats_ScrollFrame", statsPvEframe)
	TemplateScrollFrame(statsScrollFrame, true)
	statsScrollFrame:SetPoint("TOPLEFT", button, "BOTTOMLEFT", 0, -sizeGeneral.Space)
	statsScrollFrame:SetPoint("BOTTOMRIGHT", statsPvEframe, "BOTTOMRIGHT", -(sizeGeneral.Space+sizeGeneral.SliderSize+sizeGeneral.Space+sizeGeneral.Border), sizeGeneral.Offset) -- -(2+17+2+1), 5
	statsScrollFrame:SetScript("OnVerticalScroll", function(self, value)
		CharScanner_MainFrame_PvEStats_ScrollFrameScrollBar:SetValue(value)
		CharScanner_MainFrame_PvEStats_ScrollFrame.offset = value / sizeGeneral.BarHeight--math_floor((value / sizeGeneral.BarHeight) + 0.5)
		_G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_Title", "BOTTOMLEFT", 0, value-2)
		ScrollFrameUpdate()
	end)
	-- - - PvE scrollframe normal END

	-- - - PvE scrollframe search Sort START
	local searchScrollFramesort = CreateFrame("Frame", "CharScanner_MainFrame_PvEStats_Search_Sort", statsPvEframe)
	searchScrollFramesort:SetWidth(sizeStatFrame.ButtonWidth)
	searchScrollFramesort:SetHeight(sizeGeneral.BarHeight)
	searchScrollFramesort:SetPoint("BOTTOMLEFT", searchBoxIcon, "BOTTOMLEFT", 0, 0)
	searchScrollFramesort:Hide()
	TemplateSearchList(searchScrollFramesort, true)
	-- - - PvE scrollframe search Sort END

	-- - - PvE scrollframe search START
	local searchScrollFrame = CreateFrame("ScrollFrame", "CharScanner_MainFrame_PvEStats_Search_ScrollFrame", statsPvEframe)
	TemplateScrollFrame(searchScrollFrame)
	searchScrollFrame:SetPoint("TOPLEFT", button, "BOTTOMLEFT", 0, -sizeGeneral.Space)
	searchScrollFrame:SetPoint("BOTTOMRIGHT", statsPvEframe, "BOTTOMRIGHT", -(sizeGeneral.Space+sizeGeneral.SliderSize+sizeGeneral.Space+sizeGeneral.Border), sizeGeneral.Offset) -- -(2+17+2+1), 5
	searchScrollFrame:SetScript("OnVerticalScroll", function(self, value) CharScanner:ABKsearch_OnVerticalScroll(value) end)
	searchScrollFrame:Hide()
	for i = 1, GUI_ABKsearch_maxRows do
		local button = CreateFrame("Button", "CharScanner_MainFrame_PvEStats_Search_ScrollFrame_Button"..i, searchScrollFrame)
		TemplatePvEStats(button, false, "search", i)
		button:SetWidth(sizeStatFrame.ButtonWidth)
		button:SetHeight(sizeGeneral.BarHeight)
		if i == 1 then
			button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_Search_ScrollFrame", "TOPLEFT", 0, 0)
		else
			button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_Search_ScrollFrame_Button"..(i-1), "BOTTOMLEFT", 0, 0)
		end
	end
	-- - - PvE scrollframe search END

	local TEMPpvestats = {}
	-- - - PvE level 1 START
	for i = 1, #PVEDUMP do
		local button = CreateFrame("Button", "CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name, CharScanner_MainFrame_PvEStats_ScrollFrameScrollChildFrame)
		TemplatePvEStats(button, true, "level1", i)
		if i == 1 then
			button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_Title", "BOTTOMLEFT", 0, -sizeGeneral.Space)
		else
			button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i-1].name, "BOTTOMLEFT", 0, 0)
		end
		button:SetScript("OnClick", function() CharScanner:NodePvE_Click("PvE", 1, PVEDUMP[i].name, i) end)
		--_G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name.."Name"]:SetText("StatsLevel1_"..PVEDUMP[i].name) -- MENUDEV

		local n = "CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name
		TEMPpvestats[n] = {}
		TEMPpvestats[n].slevel = 1
		TEMPpvestats[n].index1 = i
		TEMPpvestats[n].index2 = 0
		TEMPpvestats[n].index3 = 0
		TEMPpvestats[n].height = sizeGeneral.BarHeight
		TEMPpvestats[n].collapse = false
	end
	-- - - PvE level 1 END

	-- - - PvE level 2 START
	for i = 1, #PVEDUMP do
		if PVEDUMP[i].statistic then
			local height = 0
			local statsLvL2 = CreateFrame("Frame", "CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name, _G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name])
			statsLvL2:SetWidth(sizeGeneral.MainWidth)
			statsLvL2:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name, "BOTTOMLEFT", 0, 0)
			for j = 1, #PVEDUMP[i].statistic do
				local button = CreateFrame("Button", "CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..j, statsLvL2)
				TemplatePvEStats(button, false, "level2")
				if j == 1 then
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name, "TOPLEFT", 0, 0)
				else
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..(j-1), "BOTTOMLEFT", 0, 0)
				end
				--_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..j.."Name"]:SetText("StatsLevel2_"..PVEDUMP[i].name..j) -- MENUDEV
				height = height + 1
			end
			statsLvL2:SetHeight(height*sizeGeneral.BarHeight)
			statsLvL2:Hide()

			local n = "CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name
			TEMPpvestats[n] = {}
			TEMPpvestats[n].slevel = 2
			TEMPpvestats[n].index1 = i
			TEMPpvestats[n].index2 = 0
			TEMPpvestats[n].index3 = 0
			TEMPpvestats[n].height = height*sizeGeneral.BarHeight
			TEMPpvestats[n].collapse = false
		else
			local height = 0
			local statsLvL2 = CreateFrame("Frame", "CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name, _G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name])
			statsLvL2:SetWidth(sizeGeneral.MainWidth)
			statsLvL2:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name, "BOTTOMLEFT", 0, 0)
			for j = 1, #PVEDUMP[i] do
				local button = CreateFrame("Button", "CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name, statsLvL2)
				TemplatePvEStats(button, true, "level2", i, j)
				if j == 1 then
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name, "TOPLEFT", 0, 0)
				else
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j-1].name, "BOTTOMLEFT", 0, 0)
				end
				button:SetScript("OnClick", function() CharScanner:NodePvE_Click("PvE", 2, PVEDUMP[i].name, i, PVEDUMP[i][j].name, j) end)
				--_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name.."Name"]:SetText("StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name) -- MENUDEV
				height = height + 1
			end
			statsLvL2:SetHeight(height*sizeGeneral.BarHeight)
			statsLvL2:Hide()

			local n = "CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name
			TEMPpvestats[n] = {}
			TEMPpvestats[n].slevel = 2
			TEMPpvestats[n].index1 = i
			TEMPpvestats[n].index2 = 0
			TEMPpvestats[n].index3 = 0
			TEMPpvestats[n].height = height*sizeGeneral.BarHeight
			TEMPpvestats[n].collapse = false
		end
	end
	-- - - PvE level 2 END

	-- - - PvE level 3 START
	for i = 1, #PVEDUMP do
		if not PVEDUMP[i].statistic then
			for j = 1, #PVEDUMP[i] do

				if PVEDUMP[i][j].statistic then
					local height = 0
					local statsLvL3 = CreateFrame("Frame", "CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name, _G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name])
					statsLvL3:SetWidth(sizeGeneral.MainWidth)
					statsLvL3:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name, "BOTTOMLEFT", 0, 0)
					for k = 1, #PVEDUMP[i][j].statistic do
						local button = CreateFrame("Button", "CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k, statsLvL3)
						TemplatePvEStats(button, false, "level3")
						if k == 1 then
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name, "TOPLEFT", 0, 0)
						else
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..(k-1), "BOTTOMLEFT", 0, 0)
						end
						--_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."Name"]:SetText("StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k) -- MENUDEV
						height = height + 1
					end
					statsLvL3:SetHeight(height*sizeGeneral.BarHeight)
					statsLvL3:Hide()

					local n = "CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name
					TEMPpvestats[n] = {}
					TEMPpvestats[n].slevel = 3
					TEMPpvestats[n].index1 = i
					TEMPpvestats[n].index2 = j
					TEMPpvestats[n].index3 = 0
					TEMPpvestats[n].height = height*sizeGeneral.BarHeight
					TEMPpvestats[n].collapse = false
				else
					local height = 0
					local statsLvL3 = CreateFrame("Frame", "CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name, _G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name])
					statsLvL3:SetWidth(sizeGeneral.MainWidth)
					statsLvL3:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name, "BOTTOMLEFT", 0, 0)
					for k = 1, #PVEDUMP[i][j] do
						local button = CreateFrame("Button", "CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k, statsLvL3)
						if PVEDUMP[i][j][k].noboss then --NOBOSS
							TemplatePvEStats(button, true, "level3", i, j, k, "NOBOSS")
						else
							TemplatePvEStats(button, true, "level3", i, j, k)
						end
						if k == 1 then
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name, "TOPLEFT", 0, 0)
						else
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..(k-1), "BOTTOMLEFT", 0, 0)
						end
						if not PVEDUMP[i][j][k].noboss then --NOBOSS
							button:SetScript("OnClick", function() CharScanner:NodePvE_Click("PvE", 3, PVEDUMP[i].name, i, PVEDUMP[i][j].name, j, PVEDUMP[i][j][k].name, k) end)
						end
						--_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."Name"]:SetText("StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k) -- MENUDEV
						height = height + 1
					end
					statsLvL3:SetHeight(height*sizeGeneral.BarHeight)
					statsLvL3:Hide()

					local n = "CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name
					TEMPpvestats[n] = {}
					TEMPpvestats[n].slevel = 3
					TEMPpvestats[n].index1 = i
					TEMPpvestats[n].index2 = j
					TEMPpvestats[n].index3 = 0
					TEMPpvestats[n].height = height*sizeGeneral.BarHeight
					TEMPpvestats[n].collapse = false
				end

			end
		end
	end
	-- - - PvE level 3 END

	-- - - PvE level 4 START
	for i = 1, #PVEDUMP do
		if not PVEDUMP[i].statistic then
			for j = 1, #PVEDUMP[i] do

				if not PVEDUMP[i][j].statistic then
					for k = 1, #PVEDUMP[i][j] do

						if PVEDUMP[i][j][k].statistic then
							local height = 0
							local statsLvL4 = CreateFrame("Frame", "CharScanner_MainFrame_PvEStats_StatsLevel4_"..PVEDUMP[i].name..PVEDUMP[i][j].name..PVEDUMP[i][j][k].name, _G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name])
							statsLvL4:SetWidth(sizeGeneral.MainWidth)
							statsLvL4:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k, "BOTTOMLEFT", 0, 0)
							for l = 1, #PVEDUMP[i][j][k].statistic do
								local button = CreateFrame("Button", "CharScanner_MainFrame_PvEStats_StatsLevel4_"..PVEDUMP[i].name..PVEDUMP[i][j].name..PVEDUMP[i][j][k].name..l, statsLvL4)
								TemplatePvEStats(button, false, "level4")
								if l == 1 then
									button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel4_"..PVEDUMP[i].name..PVEDUMP[i][j].name..PVEDUMP[i][j][k].name, "TOPLEFT", 0, 0)
								else
									button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvEStats_StatsLevel4_"..PVEDUMP[i].name..PVEDUMP[i][j].name..PVEDUMP[i][j][k].name..(l-1), "BOTTOMLEFT", 0, 0)
								end
								--_G["CharScanner_MainFrame_PvEStats_StatsLevel4_"..PVEDUMP[i].name..PVEDUMP[i][j].name..PVEDUMP[i][j][k].name..l.."Name"]:SetText("StatsLevel4_"..PVEDUMP[i].name..PVEDUMP[i][j].name..PVEDUMP[i][j][k].name..l) -- MENUDEV
								height = height + 1
							end
							statsLvL4:SetHeight(height*sizeGeneral.BarHeight)
							statsLvL4:Hide()

							local n = "CharScanner_MainFrame_PvEStats_StatsLevel4_"..PVEDUMP[i].name..PVEDUMP[i][j].name..PVEDUMP[i][j][k].name
							TEMPpvestats[n] = {}
							TEMPpvestats[n].slevel = 4
							TEMPpvestats[n].index1 = i
							TEMPpvestats[n].index2 = j
							TEMPpvestats[n].index3 = k
							TEMPpvestats[n].height = height*sizeGeneral.BarHeight
							TEMPpvestats[n].collapse = false
						end

					end
				end

			end
		end
	end
	-- - - PvE level 4 END

	-- - - PvE stats sort START
	local function SortedPairs(t)
		local ntab = {}
		for k,v in pairs(t) do
			table.insert(ntab,k)
		end
		table.sort(ntab,
			function(a,b)
				if t[a].slevel == t[b].slevel then
					if t[a].index1 == t[b].index1 then
						if t[a].index2 == t[b].index2 then
							if t[a].index3 == t[b].index3 then
								if a < b then
									return true
						  	end
							elseif t[a].index3 < t[b].index3 then
								return true
							end
						elseif t[a].index2 < t[b].index2 then
							return true
						end
					elseif t[a].index1 < t[b].index1 then
						return true
					end
				elseif t[a].slevel < t[b].slevel then
					return true
				end
			end
		)
		local i = 0
		local function _f(_s,_v)
			i = i + 1
			local k = ntab[i]
			if (k) then
				return k,t[k]
			end
		end
		return _f,nil,nil
	end

	local ni = 1
	for key, value in SortedPairs(TEMPpvestats) do
		GUI_PvEstats_menu[ni] = {}
		GUI_PvEstats_menu[ni].name     = key
		GUI_PvEstats_menu[ni].slevel    = value.slevel
		GUI_PvEstats_menu[ni].index1   = value.index1
		GUI_PvEstats_menu[ni].index2   = value.index2
		GUI_PvEstats_menu[ni].index3   = value.index3
		GUI_PvEstats_menu[ni].height   = value.height
		GUI_PvEstats_menu[ni].collapse = value.collapse
		ni = ni + 1
	end
	table.wipe(TEMPpvestats)
	-- - - PvE stats sort END
	-- - PvE END

	-- - PvP START
	local button = CreateFrame("Button", "CharScanner_MainFrame_PvPStats_Title", statsPvPframe)
	TemplatePvPStats(button, false, "title")
	button:SetPoint("BOTTOMLEFT", searchBoxIcon, "BOTTOMLEFT", 0, 0)
	button.AchievementClicked = ""
	button.AchievementDetailShow = false

	-- - - PvP scrollframe normal START
	local pvpScrollFrame = CreateFrame("ScrollFrame", "CharScanner_MainFrame_PvPStats_ScrollFrame", statsPvPframe)
	TemplateScrollFrame(pvpScrollFrame, true)
	pvpScrollFrame:SetPoint("TOPLEFT", button, "BOTTOMLEFT", 0, -sizeGeneral.Space)
	pvpScrollFrame:SetPoint("BOTTOMRIGHT", statsPvPframe, "BOTTOMRIGHT", -(sizeGeneral.Space+sizeGeneral.SliderSize+sizeGeneral.Space+sizeGeneral.Border), sizeGeneral.Offset) -- -(2+17+2+1), 5
	pvpScrollFrame:SetScript("OnVerticalScroll", function(self, value)
		CharScanner_MainFrame_PvPStats_ScrollFrameScrollBar:SetValue(value)
		CharScanner_MainFrame_PvPStats_ScrollFrame.offset = value / sizeGeneral.BarHeight--math_floor((value / sizeGeneral.BarHeight) + 0.5)
		_G["CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[1].name]:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvPStats_Title", "BOTTOMLEFT", 0, value-2)
		ScrollFrameUpdate()
	end)
	-- - - PvP scrollframe normal END

	local TEMPpvpstats = {}
	-- - - PvP level 1 START
	for i = 1, #PVPDUMP do
		local button = CreateFrame("Button", "CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[i].name, CharScanner_MainFrame_PvPStats_ScrollFrameScrollChildFrame)
		if PVPDUMP[i].statisticHide then
			TemplatePvPStats(button, false, "level1", i)
		else
			TemplatePvPStats(button, true, "level1", i)
		end
		if i == 1 then
			button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvPStats_Title", "BOTTOMLEFT", 0, -sizeGeneral.Space)
		else
			button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[i-1].name, "BOTTOMLEFT", 0, 0)
		end
		if not PVPDUMP[i].statisticHide then
			button:SetScript("OnClick", function() CharScanner:NodePvP_Click("PvP", 1, PVPDUMP[i].name, i) end)
		end
		--_G["CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[i].name.."Name"]:SetText("level1".."StatsLevel1_"..PVPDUMP[i].name) -- TEST

		local n = "CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[i].name
		TEMPpvpstats[n] = {}
		TEMPpvpstats[n].slevel = 1
		TEMPpvpstats[n].index1 = i
		TEMPpvpstats[n].index2 = 0
		TEMPpvpstats[n].index3 = 0
		TEMPpvpstats[n].height = sizeGeneral.BarHeight
		TEMPpvpstats[n].collapse = false
	end
	-- - - PvP level 1 END

	-- - - PvP level 2 START
	for i = 1, #PVPDUMP do
		if PVPDUMP[i].statistic then
			local height = 0
			local statsLvL2 = CreateFrame("Frame", "CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name, _G["CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[i].name])
			statsLvL2:SetWidth(sizeGeneral.MainWidth)
			statsLvL2:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[i].name, "BOTTOMLEFT", 0, 0)
			for j = 1, PVPDUMP[i].statisticTotal do
				local button = CreateFrame("Button", "CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..j, statsLvL2)
				TemplatePvPStats(button, false, "level2")
				if j == 1 then
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name, "TOPLEFT", 0, 0)
				else
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..(j-1), "BOTTOMLEFT", 0, 0)
				end
				--_G["CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..j.."Name"]:SetText("level2".."StatsLevel2_"..PVPDUMP[i].name..j) -- TEST
				height = height + 1
			end
			statsLvL2:SetHeight(height*sizeGeneral.BarHeight)
			statsLvL2:Hide()

			local n = "CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name
			TEMPpvpstats[n] = {}
			TEMPpvpstats[n].slevel = 2
			TEMPpvpstats[n].index1 = i
			TEMPpvpstats[n].index2 = 0
			TEMPpvpstats[n].index3 = 0
			TEMPpvpstats[n].height = height*sizeGeneral.BarHeight
			TEMPpvpstats[n].collapse = false
		else
			local height = 0
			local statsLvL2 = CreateFrame("Frame", "CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name, _G["CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[i].name])
			statsLvL2:SetWidth(sizeGeneral.MainWidth)
			statsLvL2:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[i].name, "BOTTOMLEFT", 0, 0)
			for j = 1, #PVPDUMP[i] do
				local button = CreateFrame("Button", "CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..PVPDUMP[i][j].name, statsLvL2)
				if PVPDUMP[i][j].statisticHide then
					TemplatePvPStats(button, false, "level2", i, j)
				else
					TemplatePvPStats(button, true, "level2", i, j)
				end
				if j == 1 then
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name, "TOPLEFT", 0, 0)
				else
					button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..PVPDUMP[i][j-1].name, "BOTTOMLEFT", 0, 0)
				end
				if not PVPDUMP[i][j].statisticHide then
					button:SetScript("OnClick", function() CharScanner:NodePvP_Click("PvP", 2, PVPDUMP[i].name, i, PVPDUMP[i][j].name, j) end)
				end
				--_G["CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..PVPDUMP[i][j].name.."Name"]:SetText("level2".."StatsLevel2_"..PVPDUMP[i].name..PVPDUMP[i][j].name) -- TEST
				height = height + 1
			end
			statsLvL2:SetHeight(height*sizeGeneral.BarHeight)
			statsLvL2:Hide()

			local n = "CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name
			TEMPpvpstats[n] = {}
			TEMPpvpstats[n].slevel = 2
			TEMPpvpstats[n].index1 = i
			TEMPpvpstats[n].index2 = 0
			TEMPpvpstats[n].index3 = 0
			TEMPpvpstats[n].height = height*sizeGeneral.BarHeight
			TEMPpvpstats[n].collapse = false
		end
	end
	-- - - PvP level 2 END

	-- - - PvP level 3 START
	for i = 1, #PVPDUMP do
		if not PVPDUMP[i].statistic then
			for j = 1, #PVPDUMP[i] do

				if PVPDUMP[i][j].statistic then
					local height = 0
					local statsLvL3 = CreateFrame("Frame", "CharScanner_MainFrame_PvPStats_StatsLevel3_"..PVPDUMP[i].name..PVPDUMP[i][j].name, _G["CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name])
					statsLvL3:SetWidth(sizeGeneral.MainWidth)
					statsLvL3:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..PVPDUMP[i][j].name, "BOTTOMLEFT", 0, 0)
					for k = 1, PVPDUMP[i][j].statisticTotal do
						local button = CreateFrame("Button", "CharScanner_MainFrame_PvPStats_StatsLevel3_"..PVPDUMP[i].name..PVPDUMP[i][j].name..k, statsLvL3)
						TemplatePvPStats(button, false, "level3")
						if k == 1 then
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvPStats_StatsLevel3_"..PVPDUMP[i].name..PVPDUMP[i][j].name, "TOPLEFT", 0, 0)
						else
							button:SetPoint("TOPLEFT", "CharScanner_MainFrame_PvPStats_StatsLevel3_"..PVPDUMP[i].name..PVPDUMP[i][j].name..(k-1), "BOTTOMLEFT", 0, 0)
						end
						--_G["CharScanner_MainFrame_PvPStats_StatsLevel3_"..PVPDUMP[i].name..PVPDUMP[i][j].name..k.."Name"]:SetText("level3".."StatsLevel3_"..PVPDUMP[i].name..PVPDUMP[i][j].name..k) -- TEST
						height = height + 1
					end
					statsLvL3:SetHeight(height*sizeGeneral.BarHeight)
					statsLvL3:Hide()

					local n = "CharScanner_MainFrame_PvPStats_StatsLevel3_"..PVPDUMP[i].name..PVPDUMP[i][j].name
					TEMPpvpstats[n] = {}
					TEMPpvpstats[n].slevel = 3
					TEMPpvpstats[n].index1 = i
					TEMPpvpstats[n].index2 = j
					TEMPpvpstats[n].index3 = 0
					TEMPpvpstats[n].height = height*sizeGeneral.BarHeight
					TEMPpvpstats[n].collapse = false
				end

			end
		end
	end
	-- - - PvP level 3 END

	-- - - PvP stats sort START
	local ni = 1
	for key, value in SortedPairs(TEMPpvpstats) do
		GUI_PvPstats_menu[ni] = {}
		GUI_PvPstats_menu[ni].name     = key
		GUI_PvPstats_menu[ni].slevel   = value.slevel
		GUI_PvPstats_menu[ni].index1   = value.index1
		GUI_PvPstats_menu[ni].index2   = value.index2
		GUI_PvPstats_menu[ni].index3   = value.index3
		GUI_PvPstats_menu[ni].height   = value.height
		GUI_PvPstats_menu[ni].collapse = value.collapse
		ni = ni + 1
	end
	table.wipe(TEMPpvpstats)
	-- - - PvP stats sort END
	-- - PvP END
	-- PvE / PvP Achievement & Statistics Frame END



	-- PvE AchievementDetailFrame START
	local achievementDetail = CreateFrame("Frame", "CharScanner_AchievementDetailFrame", statsframe)
	TemplateBorderTRBL(achievementDetail)
	achievementDetail:SetWidth(sizeAchievementDetail.Width)
	achievementDetail:SetHeight(sizeAchievementDetail.Height)
	achievementDetail:SetPoint("TOPLEFT", base, "TOPRIGHT", -sizeGeneral.Border, 0)
	achievementDetail:EnableMouse(true)
	achievementDetail:SetToplevel(true)
--	achievementDetail:SetFrameLevel( CharScanner_MainFrame_Stats:GetFrameLevel()+3 )
	achievementDetail:Hide()

	-- - PvE AchievementDetail headline START
	local headline = CreateFrame("Frame", "CharScanner_AchievementDetailFrame_Headline", achievementDetail)
	TemplateBorderTRBL(headline)
	headline:SetWidth(sizeAchievementDetail.Width)
	headline:SetHeight(sizeGeneral.FreeBarHeight)
	headline:SetPoint("BOTTOMLEFT", achievementDetail, "TOPLEFT", 0, -sizeGeneral.Border)
	headline:EnableMouse(true)
	headline:Hide()

	local achievementDetailButtonX = CreateFrame("Button", "CharScanner_AchievementDetailFrame_ButtonX", achievementDetail)
	TemplateIconButton(achievementDetailButtonX, 9, 12)
	achievementDetailButtonX:SetPoint("BOTTOMRIGHT", achievementDetail, "TOPRIGHT", 0, -sizeGeneral.Border)
	achievementDetailButtonX:SetScript("OnClick", function() CharScanner:AchievementDetail_Hide(AchievementDetailFrameMode) end)
	--local achievementDetailButtonFree = CreateFrame("Button", "CharScanner_AchievementDetailFrame_ButtonFree", achievementDetail)
	--TemplateIconButton(achievementDetailButtonFree, 11, 12)
	--achievementDetailButtonFree:SetPoint("TOPRIGHT", achievementDetailButtonX, "TOPLEFT", 0, 0)
	--achievementDetailButtonFree:SetScript("OnClick", function() CharScanner:AchievementDetail_FreeBarToggle() end)
	-- - PvE AchievementDetail headline END

	-- - PvE AchievementDetail Sort START
	local achievementDetailsort = CreateFrame("Frame", "CharScanner_AchievementDetailFrame_Sort", achievementDetail)
	achievementDetailsort:SetWidth(sizeAchievementDetail.ButtonWidth)
	achievementDetailsort:SetHeight(sizeGeneral.BarHeight)
	achievementDetailsort:SetPoint("TOPLEFT", achievementDetail, "TOPLEFT", sizeGeneral.Offset, -sizeAchievementDetail.SortLine2Top)
	TemplateAchievementDetailList(achievementDetailsort, true)
	-- - PvE AchievementDetail Sort END

	-- - PvE AchievementDetail ScrollFrame START
	local achievementDetailScrollFrame = CreateFrame("ScrollFrame", "CharScanner_AchievementDetailFrameScrollFrame", achievementDetail)
	TemplateScrollFrame(achievementDetailScrollFrame)
	achievementDetailScrollFrame:SetPoint("TOPLEFT", achievementDetail, "TOPLEFT", sizeGeneral.Offset, -sizeAchievementDetail.SortLineXTop)
	achievementDetailScrollFrame:SetPoint("BOTTOMRIGHT", achievementDetail, "BOTTOMRIGHT", -(sizeGeneral.Space+sizeGeneral.SliderSize+sizeGeneral.Space+sizeGeneral.Border), sizeGeneral.Offset) -- -(2+17+2+1), 5
	achievementDetailScrollFrame:SetScript("OnVerticalScroll", function(self, value) CharScanner:AchievementDetail_OnVerticalScroll(value) end)
	for i = 1, GUI_achievementdetail_maxRows do
		local button = CreateFrame("Button", "CharScanner_AchievementDetailFrame_Button"..i, achievementDetail)
		TemplateAchievementDetailList(button)
		button:SetWidth(sizeAchievementDetail.ButtonWidth)
		button:SetHeight(sizeGeneral.BarHeight)
		if i == 1 then
			button:SetPoint("TOPLEFT", "CharScanner_AchievementDetailFrameScrollFrame", "TOPLEFT", 0, 0)
		else
			button:SetPoint("TOPLEFT", "CharScanner_AchievementDetailFrame_Button"..(i-1), "BOTTOMLEFT", 0, 0)
		end
	end
	-- - PvE AchievementDetail ScrollFrame END

	-- - PvE AchievementDetail StatusFrame START
	local achievementDetailStatus = CreateFrame("Frame", "CharScanner_AchievementDetailFrame_Status", achievementDetail)
	TemplateBorderTRBL(achievementDetailStatus)
	achievementDetailStatus:SetWidth(sizeAchievementDetail.Width)
	achievementDetailStatus:SetHeight(sizeGeneral.StatusHeight)
	achievementDetailStatus:SetPoint("TOPLEFT", achievementDetail, "BOTTOMLEFT", 0, sizeGeneral.Border)
	achievementDetailStatus:EnableMouse(true)

	local achievementDetailStatusText = achievementDetailStatus:CreateFontString("CharScanner_AchievementDetailFrame_Status_Text", "ARTWORK", fonts.GameFontNormalSmall.name)
	achievementDetailStatusText:SetPoint("CENTER", achievementDetailStatus, "CENTER", 0, 0)
	achievementDetailStatusText:SetTextColor(1, 1, 1, 1)
	-- - PvE AchievementDetail StatusFrame END
	-- PvE AchievementDetailFrame END



	-- Summary Frame START
	local summary = CreateFrame("Frame", "CharScanner_MainFrame_Summary", mainframe)
	TemplateBorderTcRBL(summary)
	summary:SetWidth(sizeGeneral.MainWidth)
	summary:SetHeight(sizeSummaryFrame_BaseHeight)
	summary:SetPoint("TOPLEFT", statsframe, "BOTTOMLEFT", 0, sizeGeneral.Border)
	summary:EnableMouse(true)
	summary:Hide()

	-- - resize START
	local resizeL = CreateFrame("Button", nil, summary)
	resizeL:SetPoint("TOPLEFT", summary, "TOPLEFT", 3, -3)
	resizeL:SetWidth(sizeGeneral.SliderSize)
	resizeL:SetHeight(sizeGeneral.SliderSize)
	local textureNormal = resizeL:CreateTexture(nil, "ARTWORK")
	textureNormal:SetPoint("TOPLEFT", 0, 0)
	textureNormal:SetPoint("BOTTOMRIGHT", 0, 0)
	textureNormal:SetTexture(Textures.CharScannerIcons.path)
	textureNormal:SetTexCoord(unpack(Textures.ResizeUpDown.coords))
	resizeL:SetNormalTexture(textureNormal)
	local texturePush = resizeL:CreateTexture(nil, "ARTWORK")
	texturePush:SetPoint("TOPLEFT", 1, -1)
	texturePush:SetPoint("BOTTOMRIGHT", -1, 1)
	texturePush:SetTexture(Textures.CharScannerIcons.path)
	texturePush:SetTexCoord(unpack(Textures.ResizeUpDown.coords))
	resizeL:SetPushedTexture(texturePush)
	local textureHighlight = resizeL:CreateTexture(nil, "ARTWORK")
	textureHighlight:SetPoint("TOPLEFT", 0, 0)
	textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
	textureHighlight:SetTexture(Textures.CharScannerIcons.path)
	textureHighlight:SetTexCoord(unpack(Textures.ButtonHL.coords))
	resizeL:SetHighlightTexture(textureHighlight)

	resizeL:SetScript("OnMouseDown", ResizeStart)
	resizeL:SetScript("OnMouseUp", ResizeEnd)
	resizeL:SetScript("OnEnter", function() CharScanner:Frame_MoverInfo("CharScanner_MainFrame", "RESIZE", "BOTTOM") end)
	resizeL:SetScript("OnLeave", function() CharScanner:Frame_MoverInfo("CharScanner_MainFrame") end)

	local resizeR = CreateFrame("Button", nil, summary)
	resizeR:SetPoint("TOPRIGHT", summary, "TOPRIGHT", -3, -3)
	resizeR:SetWidth(sizeGeneral.SliderSize)
	resizeR:SetHeight(sizeGeneral.SliderSize)
	local textureNormal = resizeR:CreateTexture(nil, "ARTWORK")
	textureNormal:SetPoint("TOPLEFT", 0, 0)
	textureNormal:SetPoint("BOTTOMRIGHT", 0, 0)
	textureNormal:SetTexture(Textures.CharScannerIcons.path)
	textureNormal:SetTexCoord(unpack(Textures.ResizeUpDown.coords))
	resizeR:SetNormalTexture(textureNormal)
	local texturePush = resizeR:CreateTexture(nil, "ARTWORK")
	texturePush:SetPoint("TOPLEFT", 1, -1)
	texturePush:SetPoint("BOTTOMRIGHT", -1, 1)
	texturePush:SetTexture(Textures.CharScannerIcons.path)
	texturePush:SetTexCoord(unpack(Textures.ResizeUpDown.coords))
	resizeR:SetPushedTexture(texturePush)
	local textureHighlight = resizeR:CreateTexture(nil, "ARTWORK")
	textureHighlight:SetPoint("TOPLEFT", 0, 0)
	textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
	textureHighlight:SetTexture(Textures.CharScannerIcons.path)
	textureHighlight:SetTexCoord(unpack(Textures.ButtonHL.coords))
	resizeR:SetHighlightTexture(textureHighlight)

	resizeR:SetScript("OnMouseDown", ResizeStart)
	resizeR:SetScript("OnMouseUp", ResizeEnd)
	resizeR:SetScript("OnEnter", function() CharScanner:Frame_MoverInfo("CharScanner_MainFrame", "RESIZE", "BOTTOM") end)
	resizeR:SetScript("OnLeave", function() CharScanner:Frame_MoverInfo("CharScanner_MainFrame") end)
	-- - resize END

	local button1pve = CreateFrame("Button", "CharScanner_MainFrame_SummaryButton1PvE", summary)
	TemplatePvEStats(button1pve, false, "summary")
	button1pve:SetPoint("TOPLEFT", summary, "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
	local button2pve = CreateFrame("Button", "CharScanner_MainFrame_SummaryButton2PvE", summary)
	TemplatePvEStats(button2pve, false, "summary")
	button2pve:SetPoint("TOPLEFT", button1pve, "BOTTOMLEFT", 0, 0)

	local button1pvp = CreateFrame("Button", "CharScanner_MainFrame_SummaryButton1PvP", summary)
	TemplatePvPStats(button1pvp, false, "summary")
	button1pvp:SetPoint("TOPLEFT", summary, "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
	button1pvp:Hide()
	local button2pvp = CreateFrame("Button", "CharScanner_MainFrame_SummaryButton2PvP", summary)
	TemplatePvPStats(button2pvp, false, "summary")
	button2pvp:SetPoint("TOPLEFT", button1pvp, "BOTTOMLEFT", 0, 0)
	button2pvp:Hide()
	-- Summary Frame END



	-- Status Frame START
	local status = CreateFrame("Frame", "CharScanner_MainFrame_Status", mainframe)
	TemplateBorderTRBL(status)
	status:SetWidth(sizeGeneral.MainWidth)
	status:SetHeight(sizeGeneral.StatusHeight)
	status:SetPoint("TOPLEFT", summary, "BOTTOMLEFT", 0, sizeGeneral.Border)
	status:EnableMouse(true)
	status:SetScript("OnShow", function() CharScanner:StatusFrame_CheckPosition() end)
	status:SetScript("OnEnter", function() CharScanner:Frame_MoverInfo("CharScanner_MainFrame", "MOVE", "BOTTOM") end)
	status:SetScript("OnLeave", function() CharScanner:Frame_MoverInfo("CharScanner_MainFrame") end)
	status:SetScript("OnMouseDown", MoveStart)
	status:SetScript("OnMouseUp", MoveStop)
	status:Hide()

	local statusText = status:CreateFontString("CharScanner_MainFrame_Status_Text", "ARTWORK", fonts.GameFontNormalSmall.name)
	statusText:SetPoint("CENTER", status, "CENTER", 0, 0)
	statusText:SetTextColor(1, 1, 1, 1)

	local progressButton = CreateFrame("Button", "CharScanner_MainFrame_Status_ProgressButton", status)
	TemplateTextButton(progressButton, L["STOP SCAN"], 1)
	progressButton:SetWidth(150)
	progressButton:SetHeight(22)
	progressButton:SetPoint("TOP", status, "TOP", 0, -5)
	progressButton:SetScript("OnClick", function() CharScanner:StopGroupScan() end)
	progressButton:Hide()
	local progressText = status:CreateFontString("CharScanner_MainFrame_Status_ProgressText", "ARTWORK", fonts.GameFontNormalSmall.name)
	progressText:SetPoint("TOP", progressButton, "BOTTOM", 0, -2)
	progressText:SetJustifyH("CENTER")
	progressText:SetWidth(100)
	progressText:SetText("-/-")
	progressText:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
	progressText:Hide()
	local progressBackground = status:CreateTexture("CharScanner_MainFrame_Status_ProgressDoneBG", "BORDER")
	progressBackground:SetPoint("TOP", progressButton, "BOTTOM", 0, -2)
	progressBackground:SetWidth(sizeGeneral.StatusProgressBarWidth)
	progressBackground:SetHeight(12)
	progressBackground:SetTexture(color.DarkGrey.rgba[1], color.DarkGrey.rgba[2], color.DarkGrey.rgba[3], color.DarkGrey.rgba[4])
	progressBackground:Hide()
	local progressDone = status:CreateTexture("CharScanner_MainFrame_Status_ProgressDone", "ARTWORK")
	progressDone:SetPoint("LEFT", progressBackground, "LEFT", 0, 0)
	progressDone:SetWidth(1)
	progressDone:SetHeight(12)
	progressDone:SetBlendMode("BLEND")
	progressDone:SetTexture(color.Green.rgba[1], color.Green.rgba[2], color.Green.rgba[3], color.Green.rgba[4])
	progressDone:Hide()

	local throttleText1 = status:CreateFontString("CharScanner_MainFrame_Status_ThrottleText1", "ARTWORK", fonts.GameFontNormalSmall.name)
	throttleText1:SetPoint("TOP", progressBackground, "BOTTOM", 0, -5)
	throttleText1:SetJustifyH("CENTER")
	throttleText1:SetWidth(sizeGeneral.MainWidth-20)
	throttleText1:SetText(L["Achievement/Talent scan restriction from Blizzard:"])
	throttleText1:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
	throttleText1:Hide()
	local throttleText2 = status:CreateFontString("CharScanner_MainFrame_Status_ThrottleText2", "ARTWORK", fonts.GameFontNormalSmall.name)
	throttleText2:SetPoint("TOP", throttleText1, "BOTTOM", 0, -2)
	throttleText2:SetJustifyH("CENTER")
	throttleText2:SetWidth(sizeGeneral.MainWidth-20)
	throttleText2:SetText( format(L["Please wait %.2f seconds."], RestrictionMaxTime) )
	throttleText2:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
	throttleText2:Hide()
	local throttleBackground = status:CreateTexture("CharScanner_MainFrame_Status_ThrottleDoneBG", "BORDER")
	throttleBackground:SetPoint("TOP", throttleText2, "BOTTOM", 0, -2)
	throttleBackground:SetWidth(sizeGeneral.StatusProgressBarWidth)
	throttleBackground:SetHeight(6)
	throttleBackground:SetTexture(color.DarkGrey.rgba[1], color.DarkGrey.rgba[2], color.DarkGrey.rgba[3], color.DarkGrey.rgba[4])
	throttleBackground:Hide()
	local throttleDone = status:CreateTexture("CharScanner_MainFrame_Status_ThrottleDone", "ARTWORK")
	throttleDone:SetPoint("LEFT", throttleBackground, "LEFT", 0, 0)
	throttleDone:SetWidth(1)
	throttleDone:SetHeight(6)
	throttleDone:SetBlendMode("BLEND")
	throttleDone:SetTexture(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
	throttleDone:Hide()
	-- Status Frame END



	-- Mover Info START
	local moverinfo = CreateFrame("Frame", "CharScanner_MainFrame_MoverInfoTop", mainframe)
	moverinfo:SetWidth(150)
	moverinfo:SetHeight(14)
	moverinfo:SetPoint("BOTTOM", CharScanner_MainFrame, "TOP", 0, 0)
	moverinfo:EnableMouse(true)
	moverinfo:Hide()
	local texture = moverinfo:CreateTexture(nil, "BACKGROUND")
	texture:SetAllPoints(moverinfo)
	texture:SetTexture(0, 0, 0, 0.5)
	local text = moverinfo:CreateFontString("CharScanner_MainFrame_MoverInfoTopText", "ARTWORK", fonts.GameFontNormalSmall.name)
	text:SetPoint("CENTER", moverinfo, "CENTER", 0, 0)
	text:SetJustifyH("CENTER")
	text:SetTextColor(0.6, 0.6, 0.6, 1)

	local moverinfo = CreateFrame("Frame", "CharScanner_MainFrame_MoverInfoBottom", mainframe)
	moverinfo:SetWidth(150)
	moverinfo:SetHeight(14)
	moverinfo:SetPoint("TOP", CharScanner_MainFrame_Status, "BOTTOM", 0, 0)
	moverinfo:EnableMouse(true)
	moverinfo:Hide()
	local texture = moverinfo:CreateTexture(nil, "BACKGROUND")
	texture:SetAllPoints(moverinfo)
	texture:SetTexture(0, 0, 0, 0.5)
	local text = moverinfo:CreateFontString("CharScanner_MainFrame_MoverInfoBottomText", "ARTWORK", fonts.GameFontNormalSmall.name)
	text:SetPoint("CENTER", moverinfo, "CENTER", 0, 0)
	text:SetJustifyH("CENTER")
	text:SetTextColor(0.6, 0.6, 0.6, 1)
	-- Mover Info END



	-- History Frame START
	local history = CreateFrame("Frame", "CharScanner_HistoryFrame", mainframe)
	TemplateBorderTRBL(history)
	history:SetWidth(sizeCompareFrame.HistoryWidth)
	history:SetHeight(sizeCompareFrame.HistoryHeight)
	history:SetPoint("TOPLEFT", base, "TOPLEFT", -(sizeCompareFrame.HistoryWidth-sizeGeneral.Border), 0)
	history:EnableMouse(true)
	history:SetMovable(true)
	history:SetToplevel(true)
	history:Hide()

	local historyButtonX = CreateFrame("Button", "CharScanner_HistoryFrame_ButtonX", history)
	TemplateIconButton(historyButtonX, 9, 12)
	historyButtonX:SetPoint("BOTTOMRIGHT", history, "TOPRIGHT", 0, -sizeGeneral.Border)
	historyButtonX:SetScript("OnClick", function() CharScanner:HistoryCompareFrameHide() end)

	-- - History Sort START
	local historysort = CreateFrame("Frame", "CharScanner_HistoryFrame_Sort", history)
	historysort:SetWidth(sizeCompareFrame.HistoryWidth)
	historysort:SetHeight(sizeGeneral.BarHeight)
	historysort:SetPoint("TOPLEFT", history, "TOPLEFT", sizeGeneral.Offset, -sizeCompareFrame.SortLine2Top)
	TemplateHistoryList(historysort, nil, true)
	-- - History Sort END

	-- - History ScrollFrame START
	local historyScrollFrame = CreateFrame("ScrollFrame", "CharScanner_HistoryFrameScrollFrame", history)
	TemplateScrollFrame(historyScrollFrame)
	historyScrollFrame:SetPoint("TOPLEFT", history, "TOPLEFT", sizeGeneral.Offset, -sizeCompareFrame.SortLineXTop)
	historyScrollFrame:SetPoint("BOTTOMRIGHT", history, "BOTTOMRIGHT", -(sizeGeneral.Space+sizeGeneral.SliderSize+sizeGeneral.Space+sizeGeneral.Border), sizeGeneral.Offset) -- -(2+17+2+1), 5
	historyScrollFrame:SetScript("OnVerticalScroll", function(self, value) CharScanner:History_OnVerticalScroll(value) end)
	for i = 1, GUI_history_maxRows do
		local button = CreateFrame("Button", "CharScanner_HistoryFrame_Button"..i, history)
		TemplateHistoryList(button, i)
		button:SetWidth(sizeCompareFrame.HistoryButtonWidth)
		button:SetHeight(sizeGeneral.BarHeight)
		if i == 1 then
			button:SetPoint("TOPLEFT", "CharScanner_HistoryFrameScrollFrame", "TOPLEFT", 0, 0)
		else
			button:SetPoint("TOPLEFT", "CharScanner_HistoryFrame_Button"..(i-1), "BOTTOMLEFT", 0, 0)
		end
	end
	-- - History ScrollFrame END

	-- - History StatusFrame START
	local historyStatus = CreateFrame("Frame", "CharScanner_HistoryFrame_Status", history)
	TemplateBorderTRBL(historyStatus)
	historyStatus:SetWidth(sizeCompareFrame.HistoryWidth)
	historyStatus:SetHeight(sizeGeneral.StatusHeight)
	historyStatus:SetPoint("TOPLEFT", history, "BOTTOMLEFT", 0, sizeGeneral.Border)
	historyStatus:EnableMouse(true)

	local faketexture = historyStatus:CreateTexture(nil, "ARTWORK")
	faketexture:SetWidth(100)
	faketexture:SetHeight(sizeGeneral.StatusHeight)
	faketexture:SetPoint("TOPLEFT", historyStatus, "TOPLEFT", (sizeCompareFrame.HistoryWidth-100)/2, 0)
	faketexture:SetTexture(0, 0, 0, 0)
	local historyStatusText = historyStatus:CreateFontString("CharScanner_HistoryFrame_Status_Text", "ARTWORK", fonts.GameFontNormalSmall.name)
	historyStatusText:SetPoint("CENTER", faketexture, "CENTER", 0, 0)
	historyStatusText:SetTextColor(1, 1, 1, 1)
	local texture = CharScanner_HistoryFrame_Sort_CompareFrame:CreateTexture(nil, "ARTWORK")
	texture:SetPoint("TOPLEFT", historyStatus, "TOPLEFT", sizeCompareFrame.Talents1Left, -sizeGeneral.Space)
	texture:SetPoint("BOTTOMRIGHT", historyStatus, "TOPLEFT", sizeCompareFrame.Talents2Left+sizeCompareFrame.Talents1Width, -(sizeGeneral.StatusHeight-sizeGeneral.Space))
	texture:SetTexture(1, 1, 1, 0.075)

	-- 4 = 1 = TANK, 2 = HEAL, 3 = DAMAGE, 4 = UNKNOWN (lvl>9 with all talentpoints unspent)
	for i = 1, 4 do
		local name = "CharScanner_HistoryFrame_Status_Role"..i
		local t = CharScanner_HistoryFrame_Sort_CompareFrame:CreateTexture(name.."Texture", "ARTWORK")
		t:SetWidth(Textures.RoleIcon.sWidth)
		t:SetHeight(Textures.RoleIcon.sHeight)
		t:SetTexture(Textures.RoleIcon[i].path)
		t:SetTexCoord(Textures.RoleIcon[i].coords[1], Textures.RoleIcon[i].coords[2], Textures.RoleIcon[i].coords[3], Textures.RoleIcon[i].coords[4])
		t:SetAlpha(0.25)
		local txt = CharScanner_HistoryFrame_Sort_CompareFrame:CreateFontString(name.."Text", "OVERLAY", fonts.GameFontNormal.name)
		txt:SetText("- (-)")
		txt:SetJustifyH("LEFT")
		txt:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])

		if i == 1 then
			t:SetPoint("LEFT", texture, "LEFT", (2*sizeGeneral.Offset), 0)
			txt:SetPoint("LEFT", t, "RIGHT", sizeGeneral.Space, 0)
		elseif i == 2 then
			t:SetPoint("LEFT", _G["CharScanner_HistoryFrame_Status_Role"..(i-1).."Text"], "RIGHT", (2*sizeGeneral.Offset), 0)
			txt:SetPoint("LEFT", t, "RIGHT", sizeGeneral.Space, 0)
		elseif i == 3 then
			t:SetPoint("LEFT", _G["CharScanner_HistoryFrame_Status_Role"..(i-1).."Text"], "RIGHT", (2*sizeGeneral.Offset), 0)
			txt:SetPoint("LEFT", t, "RIGHT", sizeGeneral.Space, 0)
		else
			txt:SetPoint("RIGHT", texture, "RIGHT", -(2*sizeGeneral.Offset), 0)
			t:SetPoint("RIGHT", txt, "LEFT", -sizeGeneral.Space, 0)
		end
	end
	-- - History StatusFrame END
	-- History Frame END

	-- OnUpdate-UIParent Buttons START
	CharScanner.OnUpdateSingleScan  = CreateFrame("Button", nil, UIParent)
	CharScanner.OnUpdateGroupScan   = CreateFrame("Button", nil, UIParent)
	CharScanner.OnUpdateRestriction = CreateFrame("Button", nil, UIParent)
	-- OnUpdate-UIParent Buttons END

	-- Main Frame SetSript START
	mainframe:SetScript("OnShow", function() CharScanner:MainFrame_Show() end)
	mainframe:SetScript("OnHide", function() CharScanner:MainFrame_Hide() end)
	-- Main Frame SetSript END
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:CreateOptionsFrame()
	if CharScanner_OptionsFrame then return end

	-- Main Frame START
	local optionframe = CreateFrame("Frame", "CharScanner_OptionsFrame", UIParent)
	TemplateBorderTRBL(optionframe)
	optionframe:EnableMouse(true)
	optionframe:SetMovable(true)
	optionframe:SetToplevel(true)
	optionframe:SetClampedToScreen(true)
	optionframe:SetWidth(350)
	optionframe:SetScript("OnShow", function()
		CharScanner:Frame_SetupPosition("CharScanner_OptionsFrame")
		MoveFaker(nil, nil, optionframe)
		DisableIconButton("CharScanner_MainFrame_Headline_ButtonOptions")
		DisableTextButton("CharScanner_ScanButtonFrame_OPTIONS")
		CharScanner_InterfaceOptions_OPTIONS:Disable()
		CharScanner_MiniOptionsFrame_Options:Disable()
		CharScanner_MiniOptionsFrame_OptionsText:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
	end)
	optionframe:SetScript("OnHide", function()
		EnableIconButton("CharScanner_MainFrame_Headline_ButtonOptions")
		EnableTextButton("CharScanner_ScanButtonFrame_OPTIONS", 1)
		CharScanner_InterfaceOptions_OPTIONS:Enable()
		CharScanner_MiniOptionsFrame_Options:Enable()
		CharScanner_MiniOptionsFrame_OptionsText:SetTextColor(color.Yellow.rgba[1], color.Yellow.rgba[2], color.Yellow.rgba[3], color.Yellow.rgba[4])
	end)
	optionframe:Hide()
	-- Main Frame END

	-- Headline Frame START
	local headline = CreateFrame("Frame", "CharScanner_OptionsFrame_Headline", optionframe)
	TemplateBorderTRBL(headline)
	headline:SetWidth(350)
	headline:SetHeight(sizeGeneral.MainHeight)
	headline:SetPoint("TOPLEFT", CharScanner_OptionsFrame, "TOPLEFT", 0, 0)
	headline:EnableMouse(true)
	headline:SetToplevel(true)
	headline:SetScript("OnMouseDown", MoveFaker)

	local text = headline:CreateFontString(nil, "ARTWORK", fonts.GameFontNormal.name)
	text:SetPoint("LEFT", headline, "LEFT", 10, 0)
	text:SetJustifyH("LEFT")
	text:SetText("CharScanner "..L["Options"])

	local headlineButtonX = CreateFrame("Button", "CharScanner_OptionsFrame_Headline_ButtonX", headline)
	TemplateIconButton(headlineButtonX, 9)
	headlineButtonX:SetPoint("RIGHT", headline, "RIGHT", -sizeGeneral.Offset, 0)
	headlineButtonX:SetScript("OnClick", function() CharScanner:Frame_Toggle("CharScanner_OptionsFrame") end)

	local headlineButtonMover = CreateFrame("Button", "CharScanner_OptionsFrame_Headline_ButtonMover", headline)
	headlineButtonMover:SetWidth(22)
	headlineButtonMover:SetHeight(22)
	headlineButtonMover:SetPoint("TOPLEFT", headline, "TOPLEFT", sizeGeneral.Offset, 0)
	headlineButtonMover:SetPoint("BOTTOMRIGHT", headlineButtonX, "BOTTOMLEFT", -sizeGeneral.Offset, 0)
	local textureMover = headlineButtonMover:CreateTexture(nil, "BACKGROUND")
	textureMover:SetWidth(1)
	textureMover:SetHeight(1)
	textureMover:SetPoint("TOPLEFT", headlineButtonMover, "TOPLEFT", 0, 0)
	textureMover:SetPoint("BOTTOMRIGHT", headlineButtonMover, "BOTTOMRIGHT", 0, 0)
	textureMover:SetTexture(1, 1, 1, 0)
	headlineButtonMover:SetScript("OnEnter", function() textureMover:SetTexture(1, 1, 1, 0.075) CharScanner:Frame_MoverInfo("CharScanner_OptionsFrame", "MOVE", "TOP") end)
	headlineButtonMover:SetScript("OnLeave", function() textureMover:SetTexture(1, 1, 1, 0) CharScanner:Frame_MoverInfo("CharScanner_OptionsFrame") end)
	headlineButtonMover:SetScript("OnMouseDown", MoveStart)
	headlineButtonMover:SetScript("OnMouseUp", MoveStop)
	-- Headline Frame END

	-- Base Frame START
	local base = CreateFrame("Frame", "CharScanner_OptionsFrame_Base", optionframe)
	TemplateBorderTRBL(base)
	base:SetWidth(350)
	base:SetHeight(480)
	base:SetPoint("TOPLEFT", headline, "BOTTOMLEFT", 0, sizeGeneral.Border)
	base:EnableMouse(true)
	base:SetToplevel(true)
	base:SetScript("OnMouseDown", MoveFaker)
	-- Base Frame END

	-- Status Frame START
	local status = CreateFrame("Frame", "CharScanner_OptionsFrame_Status", optionframe)
	TemplateBorderTRBL(status)
	status:SetWidth(350)
	status:SetHeight(sizeGeneral.StatusHeight)
	status:SetPoint("TOPLEFT", base, "BOTTOMLEFT", 0, sizeGeneral.Border)
	status:EnableMouse(true)
	status:SetToplevel(true)
	status:SetScript("OnEnter", function() CharScanner:Frame_MoverInfo("CharScanner_OptionsFrame", "MOVE", "BOTTOM") end)
	status:SetScript("OnLeave", function() CharScanner:Frame_MoverInfo("CharScanner_OptionsFrame") end)
	status:SetScript("OnMouseDown", MoveStart)
	status:SetScript("OnMouseUp", MoveStop)
	-- Status Frame END

	-- Mover Info START
	local moverinfo = CreateFrame("Frame", "CharScanner_OptionsFrame_MoverInfoTop", optionframe)
	moverinfo:SetWidth(150)
	moverinfo:SetHeight(14)
	moverinfo:SetPoint("BOTTOM", CharScanner_OptionsFrame, "TOP", 0, 0)
	moverinfo:EnableMouse(true)
	moverinfo:Hide()
	local texture = moverinfo:CreateTexture(nil, "BACKGROUND")
	texture:SetAllPoints(moverinfo)
	texture:SetTexture(0, 0, 0, 0.5)
	local text = moverinfo:CreateFontString("CharScanner_OptionsFrame_MoverInfoTopText", "ARTWORK", fonts.GameFontNormalSmall.name)
	text:SetPoint("CENTER", moverinfo, "CENTER", 0, 0)
	text:SetJustifyH("CENTER")
	text:SetTextColor(0.6, 0.6, 0.6, 1)

	local moverinfo = CreateFrame("Frame", "CharScanner_OptionsFrame_MoverInfoBottom", optionframe)
	moverinfo:SetWidth(150)
	moverinfo:SetHeight(14)
	moverinfo:SetPoint("TOP", CharScanner_OptionsFrame_Status, "BOTTOM", 0, 0)
	moverinfo:EnableMouse(true)
	moverinfo:Hide()
	local texture = moverinfo:CreateTexture(nil, "BACKGROUND")
	texture:SetAllPoints(moverinfo)
	texture:SetTexture(0, 0, 0, 0.5)
	local text = moverinfo:CreateFontString("CharScanner_OptionsFrame_MoverInfoBottomText", "ARTWORK", fonts.GameFontNormalSmall.name)
	text:SetPoint("CENTER", moverinfo, "CENTER", 0, 0)
	text:SetJustifyH("CENTER")
	text:SetTextColor(0.6, 0.6, 0.6, 1)
	-- Mover Info END

	local height = headline:GetHeight() + (base:GetHeight()-sizeGeneral.Border) + (status:GetHeight()-sizeGeneral.Border)
	optionframe:SetHeight(height)
	optionframe:SetClampRectInsets(0, 0, 0, height-sizeGeneral.MainHeight)

	-- Options START
	local opengui = CreateFrame("Button", "CharScanner_OptionsFrame_OpenGUI", base)
	TemplateTextButton(opengui, L["Open GUI"], 1)
	opengui:SetWidth(150)
	opengui:SetHeight(22)
	opengui:SetPoint("TOP", base, "TOP", 0, -4)
	opengui:SetScript("OnClick", function()
		CharScanner:Frame_Toggle("CharScanner_MainFrame", 1)
		MoveFaker(nil, nil, optionframe)
	end)

	-- - scanbutton START
	local text = base:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	text:SetPoint("LEFT", base, "LEFT", 10, -10)
	text:SetPoint("TOP", opengui, "BOTTOM", 0, -10)
	text:SetJustifyH("LEFT")
	text:SetText(L["SCAN-Button"]..":")

	local scanButton = CreateFrame("CheckButton", "CharScanner_OptionsFrame_ScanButton", base)
	TemplateCheckButton(scanButton, 16, 4, L["Show SCAN-Button"])
	scanButton:SetPoint("LEFT", base, "LEFT", 10, 0)
	scanButton:SetPoint("TOP", text, "BOTTOM", 0, -10)
	scanButton:SetChecked(CharScanner_Options.scanButton)
	EnableCheckButton("CharScanner_OptionsFrame_ScanButton")

	local scanButtonInfo = base:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	scanButtonInfo:SetWidth(310)
	scanButtonInfo:SetPoint("LEFT", scanButton, "LEFT", 16+4, 0)
	scanButtonInfo:SetPoint("TOP", scanButton, "BOTTOM", 0, -5)
	scanButtonInfo:SetJustifyH("LEFT")
	scanButtonInfo:SetText(L["buttonshow_desc"])
	scanButtonInfo:SetTextColor(0.6, 0.6, 0.6, 1)

	local scanButtonLock = CreateFrame("CheckButton", "CharScanner_OptionsFrame_ScanButtonLock", base)
	TemplateCheckButton(scanButtonLock, 16, 4, L["Lock SCAN-Button"])
	scanButtonLock:SetPoint("TOP", scanButtonInfo, "BOTTOM", 0, -10)
	scanButtonLock:SetPoint("LEFT", scanButton, "LEFT", 16+4, 0)
	scanButtonLock:SetChecked(CharScanner_Options.scanButtonLock)
	if CharScanner_Options.scanButton then
		EnableCheckButton("CharScanner_OptionsFrame_ScanButtonLock")
	else
		DisableCheckButton("CharScanner_OptionsFrame_ScanButtonLock")
	end

	scanButton:SetScript("OnClick", function(self)
		CharScanner_Options.scanButton = not CharScanner_Options.scanButton
		if CharScanner_Options.scanButton then
			CharScanner_ScanButtonFrame:Show()
			if CharScanner_Options.scanButtonLock then
				CharScanner_ScanButtonFrame_Mover:Hide()
			else
				CharScanner_ScanButtonFrame:Show()
				CharScanner_ScanButtonFrame_Mover:Show()
			end
			EnableCheckButton("CharScanner_OptionsFrame_ScanButtonLock")
		else
			CharScanner_ScanButtonFrame:Hide()
			CharScanner_Options.scanButtonLock = false
			scanButtonLock:SetChecked(false)
			DisableCheckButton("CharScanner_OptionsFrame_ScanButtonLock")
		end
	end)

	scanButtonLock:SetScript("OnClick", function(self)
		CharScanner_Options.scanButtonLock = not CharScanner_Options.scanButtonLock
		if CharScanner_Options.scanButtonLock then
			if UnitIsPlayer("target") then
				CharScanner_ScanButtonFrame:Show()
			else
				CharScanner_ScanButtonFrame:Hide()
			end
			CharScanner_ScanButtonFrame_Mover:Hide()
		else
			CharScanner_ScanButtonFrame:Show()
			CharScanner_ScanButtonFrame_Mover:Show()
		end
	end)
	-- - scanbutton END

	-- - scanduration START
	local textScanDuration = base:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	textScanDuration:SetHeight(20)
	textScanDuration:SetPoint("LEFT", base, "LEFT", 10, 0)
	textScanDuration:SetPoint("TOP", scanButtonLock, "BOTTOM", 0, -20)
	textScanDuration:SetJustifyH("LEFT")
	textScanDuration:SetText(L["Maximum duration for one scan"]..":")

	local scanDurationSlider = CreateFrame("Slider", "CharScanner_OptionsFrame_ScanDuration", base)
	TemplateSlider(scanDurationSlider, 150, 1, 10, 20, CharScanner_Options.scanDuration,
	function(self, value)
		CharScanner_Options.scanDuration = value
	end,
	nil,
	10)
	scanDurationSlider:SetPoint("LEFT", textScanDuration, "LEFT", 0, 0)
	scanDurationSlider:SetPoint("TOP", textScanDuration, "BOTTOM", 0, -5)

	local textScanDurationInfo = base:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	textScanDurationInfo:SetWidth(330)
	textScanDurationInfo:SetPoint("LEFT", scanDurationSlider, "LEFT", 0, 0)
	textScanDurationInfo:SetPoint("TOP", scanDurationSlider, "BOTTOM", 0, -20)
	textScanDurationInfo:SetJustifyH("LEFT")
	textScanDurationInfo:SetText(L["scanduration_desc"])
	textScanDurationInfo:SetTextColor(0.6, 0.6, 0.6, 1)
	-- - scanduration END

	-- - chatoutput START
	local text = base:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	text:SetPoint("LEFT", base, "LEFT", 10, 0)
	text:SetPoint("TOP", textScanDurationInfo, "BOTTOM", 0, -20)
	text:SetJustifyH("LEFT")
	text:SetText(L["Chat output"]..":")

	local chatOutput = CreateFrame("CheckButton", "CharScanner_OptionsFrame_ChatOutput", base)
	TemplateCheckButton(chatOutput, 16, 4, L["Chat output"])
	chatOutput:SetPoint("LEFT", base, "LEFT", 10, 0)
	chatOutput:SetPoint("TOP", text, "BOTTOM", 0, -10)
	chatOutput:SetChecked(CharScanner_Options.chatoutput)
	EnableCheckButton("CharScanner_OptionsFrame_ChatOutput")

	local chatOutputInfo = base:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	chatOutputInfo:SetWidth(310)
	chatOutputInfo:SetPoint("LEFT", chatOutput, "LEFT", 16+4, 0)
	chatOutputInfo:SetPoint("TOP", chatOutput, "BOTTOM", 0, -5)
	chatOutputInfo:SetJustifyH("LEFT")
	chatOutputInfo:SetText(L["chat_desc"])
	chatOutputInfo:SetTextColor(0.6, 0.6, 0.6, 1)

	local chatOutputGrp = CreateFrame("CheckButton", "CharScanner_OptionsFrame_ChatOutputGrp", base)
	TemplateCheckButton(chatOutputGrp, 16, 4, L["GroupScan: Show only one chat message"])
	chatOutputGrp:SetPoint("LEFT", chatOutput, "LEFT", 16+4, 0)
	chatOutputGrp:SetPoint("TOP", chatOutputInfo, "BOTTOM", 0, -10)
	chatOutputGrp:SetChecked(CharScanner_Options.chatoutputGrp)
	if CharScanner_Options.chatoutput then
		EnableCheckButton("CharScanner_OptionsFrame_ChatOutputGrp")
	else
		DisableCheckButton("CharScanner_OptionsFrame_ChatOutputGrp")
	end

	local chatOutputGrpInfo = base:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	chatOutputGrpInfo:SetWidth(290)
	chatOutputGrpInfo:SetPoint("LEFT", chatOutputGrp, "LEFT", 16+4, 0)
	chatOutputGrpInfo:SetPoint("TOP", chatOutputGrp, "BOTTOM", 0, -5)
	chatOutputGrpInfo:SetJustifyH("LEFT")
	chatOutputGrpInfo:SetText(L["chatgrp_desc"])
	chatOutputGrpInfo:SetTextColor(0.6, 0.6, 0.6, 1)

	chatOutput:SetScript("OnClick", function(self)
		CharScanner_Options.chatoutput = not CharScanner_Options.chatoutput
		if CharScanner_Options.chatoutput then
			EnableCheckButton("CharScanner_OptionsFrame_ChatOutputGrp")
		else
			CharScanner_Options.chatoutputGrp = false
			chatOutputGrp:SetChecked(false)
			DisableCheckButton("CharScanner_OptionsFrame_ChatOutputGrp")
		end
	end)

	chatOutputGrp:SetScript("OnClick", function(self)
		CharScanner_Options.chatoutputGrp = not CharScanner_Options.chatoutputGrp
	end)
	-- - chatoutput END

	-- - minimap button START
	local text = base:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	text:SetPoint("LEFT", base, "LEFT", 10, 0)
	text:SetPoint("TOP", chatOutputGrpInfo, "BOTTOM", 0, -20)
	text:SetJustifyH("LEFT")
	text:SetText(L["Minimap-Button"]..":")

	local minimapButton = CreateFrame("CheckButton", "CharScanner_OptionsFrame_MinimapButton", base)
	TemplateCheckButton(minimapButton, 16, 4, L["Minimap-Button"])
	minimapButton:SetPoint("LEFT", base, "LEFT", 10, 0)
	minimapButton:SetPoint("TOP", text, "BOTTOM", 0, -10)
	minimapButton:SetChecked(CharScanner_Options.MinimapButton)
	EnableCheckButton("CharScanner_OptionsFrame_MinimapButton")

	local minimapButtonInfo = base:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	minimapButtonInfo:SetWidth(310)
	minimapButtonInfo:SetPoint("LEFT", minimapButton, "LEFT", 16+4, 0)
	minimapButtonInfo:SetPoint("TOP", minimapButton, "BOTTOM", 0, -5)
	minimapButtonInfo:SetJustifyH("LEFT")
	minimapButtonInfo:SetText(L["minimap_desc"])
	minimapButtonInfo:SetTextColor(0.6, 0.6, 0.6, 1)

	minimapButton:SetScript("OnClick", function(self)
		CharScanner_Options.MinimapButton = not CharScanner_Options.MinimapButton
		CharScanner:CreateMinimapButton()
	end)
	-- - minimap button END
	-- Options END
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:CreateScanButtonFrame()
	if CharScanner_ScanButtonFrame then return end

	-- scan button START
	local scanbutton = CreateFrame("Frame", "CharScanner_ScanButtonFrame", UIParent)
	scanbutton:EnableMouse(true)
	scanbutton:SetMovable(true)
	scanbutton:SetToplevel(true)
	scanbutton:SetClampedToScreen(true)
	scanbutton:Hide()
	scanbutton:SetWidth(CharScanner_Options.scanButtonSize)
	scanbutton:SetHeight(CharScanner_Options.scanButtonSize)
	local scanbuttonButton = CreateFrame("Button", nil, scanbutton)
	scanbuttonButton:SetPoint("TOPLEFT", 0, 0)
	scanbuttonButton:SetPoint("BOTTOMRIGHT", 0, 0)
	scanbuttonButton:SetScript("OnClick", function()
		if CharScanner_Options.scanButton and CharScanner_Options.scanButtonLock then
			CharScanner_MainFrame:Show()
			CharScanner:ScanUnit("target")
		end
	end)
	local texture = scanbuttonButton:CreateTexture(nil, "ARTWORK")
	texture:SetPoint("TOPLEFT", 0, 0)
	texture:SetPoint("BOTTOMRIGHT", 0, 0)
	texture:SetTexture(Textures.CharScanner.path)
	scanbuttonButton:SetNormalTexture(texture)
	local texturePush = scanbuttonButton:CreateTexture(nil, "ARTWORK")
	texturePush:SetPoint("TOPLEFT", 1, -1)
	texturePush:SetPoint("BOTTOMRIGHT", -1, 1)
	texturePush:SetTexture(Textures.CharScanner.path)
	scanbuttonButton:SetPushedTexture(texturePush)
	local textureHighlight = scanbuttonButton:CreateTexture(nil, "ARTWORK")
	textureHighlight:SetPoint("TOPLEFT", 0, 0)
	textureHighlight:SetPoint("BOTTOMRIGHT", 0, 0)
	textureHighlight:SetTexture(Textures.CharScannerHL.path)
	scanbuttonButton:SetHighlightTexture(textureHighlight)
	-- scan button END

	-- mover START
	local mover = CreateFrame("Frame", "CharScanner_ScanButtonFrame_Mover", scanbutton)
	mover:EnableMouse(true)
	mover:SetToplevel(true)
	mover:SetWidth(330)
	mover:SetHeight(80)
	mover:SetPoint("TOP", scanbutton, "TOP", 0, 32)
	local moverBGtop = mover:CreateTexture(nil, "ARTWORK")
	moverBGtop:SetPoint("TOPLEFT", mover, "TOPLEFT", 0, 0)
	moverBGtop:SetPoint("BOTTOMRIGHT", mover, "TOPRIGHT", 0, -28)
	moverBGtop:SetTexture(0, 0, 0, 0.8)
	local moverBGbottom = mover:CreateTexture(nil, "ARTWORK")
	moverBGbottom:SetPoint("TOPLEFT", mover, "TOPLEFT", 0, -84)
	moverBGbottom:SetPoint("BOTTOMRIGHT", 0, 0)
	moverBGbottom:SetTexture(0, 0, 0, 0.8)
	mover:SetScript("OnEnter", function() moverBGtop:SetTexture(0, 0, 0, 0.5) moverBGbottom:SetTexture(0, 0, 0, 0.5) end)
	mover:SetScript("OnLeave", function() moverBGtop:SetTexture(0, 0, 0, 0.8) moverBGbottom:SetTexture(0, 0, 0, 0.8) end)
	mover:SetScript("OnMouseDown", MoveStart)
	mover:SetScript("OnMouseUp", MoveStop)

	local text = mover:CreateFontString(nil, "ARTWORK", fonts.GameFontNormal.name)
	text:SetPoint("BOTTOM", mover, "TOP", 0, 4)
	text:SetJustifyH("CENTER")
	text:SetWidth(330)
	text:SetText(L["SCAN-Button"]..":")
	text:SetTextColor(1, 0.82, 0, 1)

	local scanbuttonSize = CreateFrame("Slider", "CharScanner_ScanButtonFrame_ButtonScale", mover)
	TemplateSlider(scanbuttonSize, 200, 2, 10, 48, CharScanner_Options.scanButtonSize,
	function(self, value)
		CharScanner_Options.scanButtonSize = value
		CharScanner_ScanButtonFrame:SetWidth(value)
		CharScanner_ScanButtonFrame:SetHeight(value)
	end,
	"px",
	1)
	scanbuttonSize:SetPoint("TOP", mover, "TOP", 0, -88)

	local text = mover:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	text:SetPoint("TOP", scanbuttonSize, "BOTTOM", 0, -10)
	text:SetJustifyH("CENTER")
	text:SetWidth(330)
	text:SetText("\n"..L["click & move"].."\n ")
	text:SetTextColor(0.6, 0.6, 0.6, 1)
	local text2 = mover:CreateFontString(nil, "ARTWORK", fonts.GameFontNormalSmall.name)
	text2:SetPoint("CENTER", moverBGtop, "CENTER", 0, 0)
	text2:SetJustifyH("CENTER")
	text2:SetWidth(330)
	text2:SetText("\n"..L["click & move"].."\n ")
	text2:SetTextColor(0.6, 0.6, 0.6, 1)

	local scanbuttonLock = CreateFrame("Button", "CharScanner_ScanButtonFrame_ButtonLock", mover)
	TemplateTextButton(scanbuttonLock, L["Lock SCAN-Button"], 1)
	scanbuttonLock:SetWidth(175)
	scanbuttonLock:SetHeight(22)
	scanbuttonLock:SetPoint("TOP", text, "BOTTOM", 0, -4)
	scanbuttonLock:SetScript("OnClick", function()
		CharScanner_Options.scanButtonLock = true
		CharScanner_ScanButtonFrame_Mover:Hide()
		if UnitIsPlayer("target") then
			CharScanner_ScanButtonFrame:Show()
		else
			CharScanner_ScanButtonFrame:Hide()
		end
		CharScanner_OptionsFrame_ScanButtonLock:SetChecked(CharScanner_Options.scanButtonLock)
	end)

	local scanbuttonGUI = CreateFrame("Button", "CharScanner_ScanButtonFrame_GUI", mover)
	TemplateTextButton(scanbuttonGUI, L["Open GUI"], 1)
	scanbuttonGUI:SetWidth(150)
	scanbuttonGUI:SetHeight(22)
	scanbuttonGUI:SetPoint("TOP", scanbuttonLock, "BOTTOM", 0, -20)
	scanbuttonGUI:SetPoint("LEFT", mover, "LEFT", 10, 0)
	scanbuttonGUI:SetScript("OnClick", function()
		CharScanner:Frame_Toggle("CharScanner_MainFrame")
	end)

	local scanbuttonOPTIONS = CreateFrame("Button", "CharScanner_ScanButtonFrame_OPTIONS", mover)
	TemplateTextButton(scanbuttonOPTIONS, L["Open Options"], 1)
	scanbuttonOPTIONS:SetWidth(150)
	scanbuttonOPTIONS:SetHeight(22)
	scanbuttonOPTIONS:SetPoint("LEFT", scanbuttonGUI, "RIGHT", 10, 0)
	scanbuttonOPTIONS:SetScript("OnClick", function()
		CharScanner:Frame_Toggle("CharScanner_OptionsFrame")
	end)

	mover:SetHeight( 28   + 56 +   4 + 16 + 30 + 4 + fonts.GameFontNormalSmall.height + 4 + 22 + 20 + 22 + 4 )
	-- mover END

	if CharScanner_Options.scanButtonLock then
		mover:Hide()
	end
	CharScanner:Frame_SetupPosition("CharScanner_ScanButtonFrame")
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:CreateMinimapButton()
	if not CharScanner_Options.MinimapButton then
		if CharScanner_MinimapButton then
			CharScanner_MinimapButton:Hide()
		end
		return
	else
		if CharScanner_MinimapButton then
			CharScanner_MinimapButton:Show()
			return
		end
	end

	if CharScanner_MinimapButton then return end

	local function MoveMinimapButton()
		local xpos
		local ypos
		local minimapShape = GetMinimapShape and GetMinimapShape() or "ROUND"
		if minimapShape == "SQUARE" then
			xpos = 110 * cos(CharScanner_Options.MinimapButtonPos or 0)
			ypos = 110 * sin(CharScanner_Options.MinimapButtonPos or 0)
			xpos = math.max(-82, math.min(xpos, 84))
			ypos = math.max(-86, math.min(ypos, 82))
		else
			xpos = 80 * cos(CharScanner_Options.MinimapButtonPos or 0)
			ypos = 80 * sin(CharScanner_Options.MinimapButtonPos or 0)
		end
		CharScanner_MinimapButton:SetPoint("TOPLEFT", Minimap, "TOPLEFT", 52-xpos, ypos-52)
	end

	local function DragMinimapButton()
		local xpos, ypos = GetCursorPosition()
		local xmin, ymin = Minimap:GetLeft() or 400, Minimap:GetBottom() or 400
		local scale = Minimap:GetEffectiveScale()
		xpos = xmin-xpos/scale+70
		ypos = ypos/scale-ymin-70
		CharScanner_Options.MinimapButtonPos = math.deg(math.atan2(ypos, xpos))
		MoveMinimapButton()
	end

	local MinimapButton = CreateFrame("Button", "CharScanner_MinimapButton", Minimap)
	MinimapButton:EnableMouse(true)
	MinimapButton:SetMovable(true)
	MinimapButton:SetToplevel(true)
	MinimapButton:SetWidth(Textures.MinimapButton.width)
	MinimapButton:SetHeight(Textures.MinimapButton.height)
	MinimapButton:SetPoint("TOPLEFT")
	MinimapButton:SetFrameStrata("LOW")
	MinimapButton:RegisterForClicks("AnyUp")
	MinimapButton:RegisterForDrag("LeftButton")

	MinimapButton.NormalTexture = MinimapButton:CreateTexture(nil, "ARTWORK")
	MinimapButton.NormalTexture:SetPoint("TOPLEFT", 0, 0)
	MinimapButton.NormalTexture:SetPoint("BOTTOMRIGHT", 0, 0)
	MinimapButton.NormalTexture:SetTexture(Textures.CharScannerIcons.path)
	MinimapButton.NormalTexture:SetTexCoord(Textures.MinimapButton.coords[1], Textures.MinimapButton.coords[2], Textures.MinimapButton.coords[3], Textures.MinimapButton.coords[4])
	MinimapButton:SetNormalTexture(MinimapButton.NormalTexture)

	MinimapButton.PushedTexture = MinimapButton:CreateTexture(nil, "ARTWORK")
	MinimapButton.PushedTexture:SetPoint("TOPLEFT", 1, -1)
	MinimapButton.PushedTexture:SetPoint("BOTTOMRIGHT", -1, 1)
	MinimapButton.PushedTexture:SetTexture(Textures.CharScannerIcons.path)
	MinimapButton.PushedTexture:SetTexCoord(Textures.MinimapButton.coords[1], Textures.MinimapButton.coords[2], Textures.MinimapButton.coords[3], Textures.MinimapButton.coords[4])
	MinimapButton:SetPushedTexture(MinimapButton.PushedTexture)

	MinimapButton.HighlightTexture = MinimapButton:CreateTexture(nil, "ARTWORK")
	MinimapButton.HighlightTexture:SetPoint("TOPLEFT", 0, 0)
	MinimapButton.HighlightTexture:SetPoint("BOTTOMRIGHT", 0, 0)
	MinimapButton.HighlightTexture:SetTexture("Interface\\Minimap\\UI-Minimap-ZoomButton-Highlight")
	MinimapButton:SetHighlightTexture(MinimapButton.HighlightTexture)

	MinimapButton:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
		GameTooltip:AddLine("CharScanner", color.Yellow.rgba[1], color.Yellow.rgba[2], color.Yellow.rgba[3], color.Yellow.rgba[4])
		GameTooltip:Show()
	end)
	MinimapButton:SetScript("OnLeave", function(self) GameTooltip:Hide() end)
	MinimapButton:SetScript("OnClick", function(self, button) CharScanner:MiniOptionsToggle(button, 2) end)
	MinimapButton:SetScript("OnDragStart", function(self) self:LockHighlight() self:SetScript("OnUpdate", DragMinimapButton) end)
	MinimapButton:SetScript("OnDragStop", function(self) self:SetScript("OnUpdate", nil) self:UnlockHighlight() end)

	MoveMinimapButton()
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:CreateMiniOptionsFrame()
	if CharScanner_MiniOptionsFrame then return end

	local buttondata = {}
	buttondata[1] = "SingleScan"
	buttondata[2] = "GroupScanAdd"
	buttondata[3] = "GroupScan"
	buttondata[4] = "Options"

	local maxiconwidth = 0
	local maxtextwidth = 0

	local iw = 0
	for i = 1, #buttondata do
		iw = Textures[ buttondata[i] ].width
		if iw > maxiconwidth then
			maxiconwidth = iw
		end
	end
	
	local buttonheight = Textures.SingleScan.height
	local frameheight = sizeGeneral.Offset + (#buttondata*buttonheight) + sizeGeneral.Offset

	local miniOptionsFrame = CreateFrame("Frame", "CharScanner_MiniOptionsFrame", UIParent)
	TemplateBorderTRBL(miniOptionsFrame)
	miniOptionsFrame:EnableMouse(true)
	miniOptionsFrame:SetMovable(true)
	miniOptionsFrame:SetToplevel(true)
	miniOptionsFrame:SetClampedToScreen(true)
	miniOptionsFrame:SetFrameStrata("DIALOG")
	miniOptionsFrame:SetHeight(frameheight)
	miniOptionsFrame:Hide()

	for i = 1, #buttondata do
		local buttonname = "CharScanner_MiniOptionsFrame_"..buttondata[i]

		local button = CreateFrame("Button", buttonname, miniOptionsFrame)
		button:SetHeight(buttonheight)
		if i == 1 then
			button:SetPoint("TOPLEFT", miniOptionsFrame, "TOPLEFT", sizeGeneral.Offset, -sizeGeneral.Offset)
		else
			button:SetPoint("TOPLEFT", "CharScanner_MiniOptionsFrame_"..buttondata[(i-1)], "BOTTOMLEFT", 0, 0)
		end

		button.text = button:CreateFontString(buttonname.."Text", "OVERLAY", fonts.GameFontNormalSmall.name)
		button.text:SetHeight(fonts.GameFontNormalSmall.height)
		button.text:SetPoint("LEFT", button, "LEFT", maxiconwidth+sizeGeneral.Space, 0)
		button.text:SetJustifyH("LEFT")
		button.text:SetText(L[ buttondata[i] ])
		local tw = button.text:GetStringWidth()
		if tw > maxtextwidth then
			maxtextwidth = tw
		end
		--button.text:SetWidth(tw) -- TEST
	end

	local framewidth = sizeGeneral.Offset + maxiconwidth + sizeGeneral.Offset + maxtextwidth + sizeGeneral.Offset
	miniOptionsFrame:SetWidth(framewidth)

	for i = 1, #buttondata do
		local button = _G[ "CharScanner_MiniOptionsFrame_".. buttondata[i] ]

		button:SetWidth(framewidth-(2*sizeGeneral.Offset))
		
		local bottomright = (framewidth-(2*sizeGeneral.Offset)) - Textures[ buttondata[i] ].width

		button.textureNormal = button:CreateTexture(nil, "ARTWORK")
		button.textureNormal:SetPoint("TOPLEFT", 2, -2)
		button.textureNormal:SetPoint("BOTTOMRIGHT", -(bottomright + 2), 2)
		button.textureNormal:SetTexture(Textures.CharScannerIcons.path)
		button.textureNormal:SetTexCoord(Textures[ buttondata[i] ].coords[1], Textures[ buttondata[i] ].coords[2], Textures[ buttondata[i] ].coords[3], Textures[ buttondata[i] ].coords[4])
		button:SetNormalTexture(button.textureNormal)

		button.texturePush = button:CreateTexture(nil, "ARTWORK")
		button.texturePush:SetPoint("TOPLEFT", 3, -3)
		button.texturePush:SetPoint("BOTTOMRIGHT", -(bottomright + 3), 3)
		button.texturePush:SetTexture(Textures.CharScannerIcons.path)
		button.texturePush:SetTexCoord(Textures[ buttondata[i] ].coords[1], Textures[ buttondata[i] ].coords[2], Textures[ buttondata[i] ].coords[3], Textures[ buttondata[i] ].coords[4])
		button:SetPushedTexture(button.texturePush)

		button.textureDisable = button:CreateTexture(nil, "ARTWORK")
		button.textureDisable:SetPoint("TOPLEFT", 2, -2)
		button.textureDisable:SetPoint("BOTTOMRIGHT", -(bottomright + 2), 2)
		button.textureDisable:SetTexture(Textures.CharScannerIcons.path)
		button.textureDisable:SetTexCoord(Textures[ buttondata[i] ].coords[1], Textures[ buttondata[i] ].coords[2], Textures[ buttondata[i] ].coords[3], Textures[ buttondata[i] ].coords[4])
		button:SetDisabledTexture(button.textureDisable)

		local shaderSupported = button.textureDisable:SetDesaturated(true)
		if not shaderSupported then
			button.textureDisable:SetVertexColor(0.5, 0.5, 0.5)
		end

		button.Highlight = button:CreateTexture(nil, "ARTWORK")
		button.Highlight:SetPoint("TOPLEFT", button, "TOPLEFT", 0, 0)
		button.Highlight:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 0, 0)
		button.Highlight:SetTexture(0.2, 0.2, 0.2, 1)
		button:SetHighlightTexture(button.Highlight)

		if (buttondata[i] == "GroupScanAdd" or buttondata[i] == "GroupScan") and (not inPartyOrRaid or GroupScanIsActive) then
			button:Disable()
			button.text:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
		end
		if buttondata[i] == "SingleScan" and SingleScanIsActive then
			button:Disable()
			button.text:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
		end

		if i == 1 then
			button:SetScript("OnClick", function()
				CharScanner_MiniOptionsFrame:SetScript("OnUpdate", nil)
				CharScanner_MainFrame:Show()
				CharScanner:ScanUnit("target")
				CharScanner_MiniOptionsFrame:Hide()
			end)
		elseif i == 2 then
			button:SetScript("OnClick", function()
				CharScanner_MiniOptionsFrame:SetScript("OnUpdate", nil)
				CharScanner_MainFrame:Show()
				CharScanner:ScanGroup(false)
				CharScanner_MiniOptionsFrame:Hide()
			end)
		elseif i == 3 then
			button:SetScript("OnClick", function()
				CharScanner_MiniOptionsFrame:SetScript("OnUpdate", nil)
				CharScanner_MainFrame:Show()
				CharScanner:ScanGroup(true)
				CharScanner_MiniOptionsFrame:Hide()
			end)
		elseif i == 4 then
			button:SetScript("OnClick", function()
				CharScanner_MiniOptionsFrame:SetScript("OnUpdate", nil)
				CharScanner:Frame_Toggle("CharScanner_OptionsFrame", 1)
				CharScanner_MiniOptionsFrame:Hide()
			end)
		end

		button:SetScript("OnEnter", function() MiniOptions_TimerStart = GetTime() end)
		button:SetScript("OnLeave", function() MiniOptions_TimerStart = GetTime() end)
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:NodePvE_Click(mode, lvl, name1, index1, name2, index2, name3, index3)
	--print("lvl:", lvl, "index1:", index1, "index2:", index2, "index3:", index3) -- MENUDEV
	local baseName = "CharScanner_MainFrame_PvEStats"
	if lvl == 1 then
	-- level 1 START
		if _G[baseName.."_StatsLevel2_"..name1]:IsVisible() then
			_G[baseName.."_StatsLevel1_"..name1.."Collapse"]:SetTexCoord(unpack(Textures.PlusMinus.coordsPlus))

			NodeDisplay(mode, baseName.."_StatsLevel2_"..name1, false)

			if index1 == 1 then
				_G[baseName.."_StatsLevel1_"..PVEDUMP[2].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel1_"..PVEDUMP[1].name, "BOTTOMLEFT", 0, 0)
			elseif index1 > 1 and index1 < #PVEDUMP then
				_G[baseName.."_StatsLevel1_"..PVEDUMP[index1+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel1_"..PVEDUMP[index1].name, "BOTTOMLEFT", 0, 0)
			end
		else
			_G[baseName.."_StatsLevel1_"..name1.."Collapse"]:SetTexCoord(unpack(Textures.PlusMinus.coordsMinus))
			
			NodeDisplay(mode, baseName.."_StatsLevel2_"..name1, true)

			if index1 == 1 then
				_G[baseName.."_StatsLevel1_"..PVEDUMP[2].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel2_"..name1, "BOTTOMLEFT", 0, 0)
			elseif index1 > 1 and index1 < #PVEDUMP then
				_G[baseName.."_StatsLevel1_"..PVEDUMP[index1+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel2_"..name1, "BOTTOMLEFT", 0, 0)
			end
		end
	-- level 1 END
	elseif lvl == 2 then
	-- level 2 START
		if _G[baseName.."_StatsLevel3_"..name1..name2]:IsVisible() then
			_G[baseName.."_StatsLevel2_"..name1..name2.."Collapse"]:SetTexCoord(unpack(Textures.PlusMinus.coordsPlus))

			NodeDisplay(mode, baseName.."_StatsLevel3_"..name1..name2, false)

			if index2 == 1 then
				_G[baseName.."_StatsLevel2_"..PVEDUMP[index1].name..PVEDUMP[index1][2].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel2_"..PVEDUMP[index1].name..PVEDUMP[index1][1].name, "BOTTOMLEFT", 0, 0)
			elseif index2 > 1 and index2 < #PVEDUMP[index1] then
				_G[baseName.."_StatsLevel2_"..PVEDUMP[index1].name..PVEDUMP[index1][index2+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel2_"..PVEDUMP[index1].name..PVEDUMP[index1][index2].name, "BOTTOMLEFT", 0, 0)
			end
		else
			_G[baseName.."_StatsLevel2_"..name1..name2.."Collapse"]:SetTexCoord(unpack(Textures.PlusMinus.coordsMinus))

			NodeDisplay(mode, baseName.."_StatsLevel3_"..name1..name2, true)

			if index2 == 1 then
				_G[baseName.."_StatsLevel2_"..PVEDUMP[index1].name..PVEDUMP[index1][2].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel3_"..PVEDUMP[index1].name..name2, "BOTTOMLEFT", 0, 0)
			elseif index2 > 1 and index2 < #PVEDUMP[index1] then
				_G[baseName.."_StatsLevel2_"..PVEDUMP[index1].name..PVEDUMP[index1][index2+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel3_"..PVEDUMP[index1].name..name2..#PVEDUMP[index1][index2], "BOTTOMLEFT", 0, 0)
				local c1 = #PVEDUMP[index1][index2]
				local c2 = #PVEDUMP[index1][index2][c1].statistic
				if _G[baseName.."_StatsLevel4_"..PVEDUMP[index1].name..PVEDUMP[index1][index2].name..PVEDUMP[index1][index2][c1].name..c2]:IsVisible() then
					_G[baseName.."_StatsLevel2_"..PVEDUMP[index1].name..PVEDUMP[index1][index2+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel4_"..PVEDUMP[index1].name..PVEDUMP[index1][index2].name..PVEDUMP[index1][index2][c1].name..c2, "BOTTOMLEFT", 0, 0)
				end
			end
		end
	-- level 2 END
	elseif lvl == 3 then
	-- level 3 START
		if _G[baseName.."_StatsLevel4_"..name1..name2..name3]:IsVisible() then
			_G[baseName.."_StatsLevel3_"..name1..name2..index3.."Collapse"]:SetTexCoord(unpack(Textures.PlusMinus.coordsPlus))

			NodeDisplay(mode, baseName.."_StatsLevel4_"..name1..name2..name3, false)

			if index3 == 1 then
				_G[baseName.."_StatsLevel3_"..PVEDUMP[index1].name..PVEDUMP[index1][index2].name.."2"]:SetPoint("TOPLEFT", baseName.."_StatsLevel3_"..PVEDUMP[index1].name..PVEDUMP[index1][index2].name.."1", "BOTTOMLEFT", 0, 0)
			elseif index3 > 1 and index3 < #PVEDUMP[index1][index2] then
				_G[baseName.."_StatsLevel3_"..PVEDUMP[index1].name..PVEDUMP[index1][index2].name..(index3+1)]:SetPoint("TOPLEFT", baseName.."_StatsLevel3_"..PVEDUMP[index1].name..PVEDUMP[index1][index2].name..index3, "BOTTOMLEFT", 0, 0)
			elseif index3 == #PVEDUMP[index1][index2] and index2 < #PVEDUMP[index1] then
					_G[baseName.."_StatsLevel2_"..PVEDUMP[index1].name..PVEDUMP[index1][index2+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel3_"..PVEDUMP[index1].name..PVEDUMP[index1][index2].name..#PVEDUMP[index1][index2], "BOTTOMLEFT", 0, 0)
			end
		else
			_G[baseName.."_StatsLevel3_"..name1..name2..index3.."Collapse"]:SetTexCoord(unpack(Textures.PlusMinus.coordsMinus))

			NodeDisplay(mode, baseName.."_StatsLevel4_"..name1..name2..name3, true)

			if index3 == 1 then
				_G[baseName.."_StatsLevel3_"..PVEDUMP[index1].name..PVEDUMP[index1][index2].name.."2"]:SetPoint("TOPLEFT", baseName.."_StatsLevel4_"..PVEDUMP[index1].name..PVEDUMP[index1][index2].name..PVEDUMP[index1][index2][1].name, "BOTTOMLEFT", 0, 0)
			elseif index3 > 1 and index3 < #PVEDUMP[index1][index2] then
				_G[baseName.."_StatsLevel3_"..PVEDUMP[index1].name..PVEDUMP[index1][index2].name..(index3+1)]:SetPoint("TOPLEFT", baseName.."_StatsLevel4_"..PVEDUMP[index1].name..PVEDUMP[index1][index2].name..PVEDUMP[index1][index2][index3].name..#PVEDUMP[index1][index2][index3].statistic, "BOTTOMLEFT", 0, 0)
			elseif index3 == #PVEDUMP[index1][index2] and index2 < #PVEDUMP[index1] then
				_G[baseName.."_StatsLevel2_"..PVEDUMP[index1].name..PVEDUMP[index1][index2+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel4_"..PVEDUMP[index1].name..PVEDUMP[index1][index2].name..PVEDUMP[index1][index2][index3].name..#PVEDUMP[index1][index2][index3].statistic, "BOTTOMLEFT", 0, 0)
			end
		end
	-- level 3 END
	end

	if index1 == 3 and PVEDUMP[index1+1] and PVEDUMP[index1+1].name then -- MENUDEV
		local c1 = #PVEDUMP[index1]
		local c2 = #PVEDUMP[index1][c1]
		local c3 = #PVEDUMP[index1][c1][c2].statistic
		if _G[baseName.."_StatsLevel4_"..PVEDUMP[index1].name..PVEDUMP[index1][c1].name..PVEDUMP[index1][c1][c2].name..c3]:IsVisible() then
			--print("MENUDEV lvl4", "_StatsLevel4_"..PVEDUMP[index1].name..PVEDUMP[index1][c1].name..PVEDUMP[index1][c1][c2].name..c3) -- MENUDEV
			_G[baseName.."_StatsLevel1_"..PVEDUMP[index1+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel4_"..PVEDUMP[index1].name..PVEDUMP[index1][c1].name..PVEDUMP[index1][c1][c2].name..c3, "BOTTOMLEFT", 0, 0)
		elseif _G[baseName.."_StatsLevel3_"..PVEDUMP[index1].name..PVEDUMP[index1][c1].name]:IsVisible() then
			--print("MENUDEV lvl3", "_StatsLevel3_"..PVEDUMP[index1].name..PVEDUMP[index1][c1].name..c2) -- MENUDEV
			_G[baseName.."_StatsLevel1_"..PVEDUMP[index1+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel3_"..PVEDUMP[index1].name..PVEDUMP[index1][c1].name..c2, "BOTTOMLEFT", 0, 0)
		elseif _G[baseName.."_StatsLevel2_"..PVEDUMP[index1].name..PVEDUMP[index1][c1].name]:IsVisible() then
			--print("MENUDEV lvl2", "_StatsLevel2_"..PVEDUMP[index1].name..PVEDUMP[index1][c1].name) -- MENUDEV
			_G[baseName.."_StatsLevel1_"..PVEDUMP[index1+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel2_"..PVEDUMP[index1].name..PVEDUMP[index1][c1].name, "BOTTOMLEFT", 0, 0)
		else
			--print("MENUDEV lvl1", "_StatsLevel1_"..PVEDUMP[index1].name) -- MENUDEV
			_G[baseName.."_StatsLevel1_"..PVEDUMP[index1+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel1_"..PVEDUMP[index1].name, "BOTTOMLEFT", 0, 0)
		end
	end -- MENUDEV

end

function CharScanner:NodePvP_Click(mode, lvl, name1, index1, name2, index2, name3, index3, init)
	local baseName = "CharScanner_MainFrame_PvPStats"
	if lvl == 1 then
	-- level 1 START
		local childButton    = baseName.."_StatsLevel2_"..name1
		local collapseButton = baseName.."_StatsLevel1_"..name1.."Collapse"

		if _G[childButton]:IsVisible() then
			_G[collapseButton]:SetTexCoord(unpack(Textures.PlusMinus.coordsPlus))

			NodeDisplay(mode, childButton, false)

			if index1 == 1 then
				if PVPDUMP[2] and #PVPDUMP[2].name then
					_G[baseName.."_StatsLevel1_"..PVPDUMP[2].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel1_"..PVPDUMP[1].name, "BOTTOMLEFT", 0, 0)
				end
			elseif index1 > 1 and index1 < #PVPDUMP then
				_G[baseName.."_StatsLevel1_"..PVPDUMP[index1+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel1_"..PVPDUMP[index1].name, "BOTTOMLEFT", 0, 0)
			end
		else
			_G[collapseButton]:SetTexCoord(unpack(Textures.PlusMinus.coordsMinus))
			
			NodeDisplay(mode, childButton, true)

			if index1 == 1 then
				if PVPDUMP[2] and #PVPDUMP[2].name then
					_G[baseName.."_StatsLevel1_"..PVPDUMP[2].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel2_"..name1, "BOTTOMLEFT", 0, 0)
				end
			elseif index1 > 1 and index1 < #PVPDUMP then
				_G[baseName.."_StatsLevel1_"..PVPDUMP[index1+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel2_"..name1, "BOTTOMLEFT", 0, 0)
			end
		end

		if PVPDUMP[index1].statistic then return end
		if init then return end

		local c1 = #PVPDUMP[index1]
		local c2 = PVPDUMP[index1][c1].statisticTotal
		local AnchorIfLastChildCollapsed = baseName.."_StatsLevel3_"..PVPDUMP[index1].name..PVPDUMP[index1][c1].name..c2
		local AnchorIfLastCollapsed      = baseName.."_StatsLevel2_"..PVPDUMP[index1].name..PVPDUMP[index1][c1].name
		local AnchorIfLastNotCollapsed   = baseName.."_StatsLevel1_"..PVPDUMP[index1].name

		if PVPDUMP[index1+1] then
			if _G[AnchorIfLastChildCollapsed] and _G[AnchorIfLastChildCollapsed]:IsVisible() then
				local toAnchor = baseName.."_StatsLevel1_"..PVPDUMP[index1+1].name
				if _G[toAnchor] then
					_G[toAnchor]:SetPoint("TOPLEFT", AnchorIfLastChildCollapsed, "BOTTOMLEFT", 0, 0)
				end
				return
			elseif _G[AnchorIfLastCollapsed]:IsVisible() then
				local toAnchor = baseName.."_StatsLevel1_"..PVPDUMP[index1+1].name
				if _G[toAnchor] then
					_G[toAnchor]:SetPoint("TOPLEFT", AnchorIfLastCollapsed, "BOTTOMLEFT", 0, 0)
				end
				return
			else
				local toAnchor = baseName.."_StatsLevel1_"..PVPDUMP[index1+1].name
				if _G[toAnchor] then
					_G[toAnchor]:SetPoint("TOPLEFT", AnchorIfLastNotCollapsed, "BOTTOMLEFT", 0, 0)
				end
				return
			end
		end

	-- level 1 END
	elseif lvl == 2 then
	-- level 2 START
		local childButton    = baseName.."_StatsLevel3_"..name1..name2
		local collapseButton = baseName.."_StatsLevel2_"..name1..name2.."Collapse"

		if _G[childButton]:IsVisible() then
			_G[collapseButton]:SetTexCoord(unpack(Textures.PlusMinus.coordsPlus))

			NodeDisplay(mode, childButton, false)

			if index2 == 1 then
				if PVPDUMP[index1][2] and PVPDUMP[index1][2].statisticTotal > 1 then
					_G[baseName.."_StatsLevel2_"..PVPDUMP[index1].name..PVPDUMP[index1][2].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel2_"..PVPDUMP[index1].name..PVPDUMP[index1][1].name, "BOTTOMLEFT", 0, 0)
				end
			elseif index2 > 1 and index2 < #PVPDUMP[index1] then
				_G[baseName.."_StatsLevel2_"..PVPDUMP[index1].name..PVPDUMP[index1][index2+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel2_"..PVPDUMP[index1].name..PVPDUMP[index1][index2].name, "BOTTOMLEFT", 0, 0)
			end
		else
			_G[collapseButton]:SetTexCoord(unpack(Textures.PlusMinus.coordsMinus))

			NodeDisplay(mode, childButton, true)

			if index2 == 1 then
				if PVPDUMP[index1][2] and PVPDUMP[index1][2].statisticTotal > 1 then
					_G[baseName.."_StatsLevel2_"..PVPDUMP[index1].name..PVPDUMP[index1][2].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel3_"..PVPDUMP[index1].name..name2..PVPDUMP[index1][index2].statisticTotal, "BOTTOMLEFT", 0, 0)
				end
			elseif index2 > 1 and index2 < #PVPDUMP[index1] then
				_G[baseName.."_StatsLevel2_"..PVPDUMP[index1].name..PVPDUMP[index1][index2+1].name]:SetPoint("TOPLEFT", baseName.."_StatsLevel3_"..PVPDUMP[index1].name..name2..PVPDUMP[index1][index2].statisticTotal, "BOTTOMLEFT", 0, 0)
			end
		end

		local c1 = #PVPDUMP[index1]
		local c2 = PVPDUMP[index1][c1].statisticTotal
		local AnchorIfLastCollapsed    = baseName.."_StatsLevel3_"..PVPDUMP[index1].name..PVPDUMP[index1][c1].name..c2
		local AnchorIfLastNotCollapsed = baseName.."_StatsLevel2_"..PVPDUMP[index1].name..PVPDUMP[index1][c1].name

		if PVPDUMP[index1][c1] and PVPDUMP[index1][c1].statisticTotal > 1 then
			if _G[AnchorIfLastCollapsed]:IsVisible() then
				local toAnchor = baseName.."_StatsLevel1_"..PVPDUMP[index1+1].name
				if _G[toAnchor] then
					_G[toAnchor]:SetPoint("TOPLEFT", AnchorIfLastCollapsed, "BOTTOMLEFT", 0, 0)
				end
			else
				local toAnchor = baseName.."_StatsLevel1_"..PVPDUMP[index1+1].name
				if _G[toAnchor] then
					_G[toAnchor]:SetPoint("TOPLEFT", AnchorIfLastNotCollapsed, "BOTTOMLEFT", 0, 0)
				end
			end
		end
	-- level 2 END
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ABKsearch_Update(guid)
	CharScanner:ABKsearch_GUICreate(guid)
	CharScanner:ABKsearch_FillList()
end

function CharScanner:ABKsearch_GUICreate(guid)
	table.wipe(GUI_ABKsearch_data)
	GUI_ABKsearch_Count = 0

	if not guid then
		guid = GUIselectedGUID
	end

	if BossSearchQuery then

		CharScanner_MainFrame_PvEStats_ScrollFrame:Hide()
		CharScanner_MainFrame_PvEStats_Search_ScrollFrame:Show()
		CharScanner_MainFrame_PvEStats_Search_Sort:Show()
		CharScanner_MainFrame_PvEStats_TitleAchievement:Hide()
		CharScanner_MainFrame_PvEStats_TitleStatistic:Hide()
		CharScanner_MainFrame_PvEStats_TitleBossesTotal:Hide()

		CharScanner_MainFrame_SummaryButton1PvEAchievement:Hide()
		CharScanner_MainFrame_SummaryButton2PvEAchievement:Hide()
		CharScanner_MainFrame_SummaryButton1PvEStatistic:Hide()
		CharScanner_MainFrame_SummaryButton2PvEStatistic:Hide()
		--CharScanner_MainFrame_SummaryButton1PvEBossesTotal:Hide()

		local lowerSearchQuery = string_lower(BossSearchQuery)
		local x = 1
		local totalCount = 0

		for i = 1, #PVEDUMP do
			if PVEDUMP[i].statistic then
				for j = 1, #PVEDUMP[i].statistic do
					if string_find(string_lower(StatisticIDTable[ PVEDUMP[i].statistic[j] ]), lowerSearchQuery, 1, true) then
						GUI_ABKsearch_data[x] = {}
						GUI_ABKsearch_data[x].Num = x
						GUI_ABKsearch_data[x].Name = StatisticIDTable[ PVEDUMP[i].statistic[j] ]
						GUI_ABKsearch_data[x].BossesTotal = SCAN[guid].PVEDATA[i].statisticBossEach[j]
						totalCount = totalCount + SCAN[guid].PVEDATA[i].statisticBossEach[j]
						x = x + 1
					end
				end
			else
				for j = 1, #PVEDUMP[i] do
					if PVEDUMP[i][j].statistic then
						for k = 1, #PVEDUMP[i][j].statistic do
							if string_find(string_lower(StatisticIDTable[ PVEDUMP[i][j].statistic[k] ]), lowerSearchQuery, 1, true) then
								GUI_ABKsearch_data[x] = {}
								GUI_ABKsearch_data[x].Num = x
								GUI_ABKsearch_data[x].Name = StatisticIDTable[ PVEDUMP[i][j].statistic[k] ]
								GUI_ABKsearch_data[x].BossesTotal = SCAN[guid].PVEDATA[i][j].statisticBossEach[k]
								totalCount = totalCount + SCAN[guid].PVEDATA[i][j].statisticBossEach[k]
								x = x + 1
							end
						end
					else
						for k = 1, #PVEDUMP[i][j] do
							if PVEDUMP[i][j][k].statistic then
								for l = 1, #PVEDUMP[i][j][k].statistic do
									if string_find(string_lower(StatisticIDTable[ PVEDUMP[i][j][k].statistic[l] ]), lowerSearchQuery, 1, true) then
										GUI_ABKsearch_data[x] = {}
										GUI_ABKsearch_data[x].Num = x
										GUI_ABKsearch_data[x].Name = StatisticIDTable[ PVEDUMP[i][j][k].statistic[l] ]
										GUI_ABKsearch_data[x].BossesTotal = SCAN[guid].PVEDATA[i][j][k].statisticBossEach[l]
										totalCount = totalCount + SCAN[guid].PVEDATA[i][j][k].statisticBossEach[l]
										x = x + 1
									end
								end
							end
						end
					end
				end
			end
		end
		GUI_ABKsearch_Count = #GUI_ABKsearch_data

		local sortfunc
		if GUI_ABKsearch_SortBy == 1 then -- 1 = standard
			if GUI_ABKsearch_SortOrder == 1 then
				sortfunc = function(a, b)
					if a.Num < b.Num then
						return true
					end
				end
			else
				sortfunc = function(a, b)
					if a.Num > b.Num then
						return true
					end
				end
			end
		elseif GUI_ABKsearch_SortBy == 2 then -- 2 = name
			if GUI_ABKsearch_SortOrder == 1 then
				sortfunc = function(a, b)
					if a.Name < b.Name then
						return true
					end
				end
			else
				sortfunc = function(a, b)
					if a.Name > b.Name then
						return true
					end
				end
			end
		elseif GUI_ABKsearch_SortBy == 3 then -- 3 = kills
			if GUI_ABKsearch_SortOrder == 1 then
				sortfunc = function(a, b)
					if a.BossesTotal > b.BossesTotal then
						return true
					end
				end
			else
				sortfunc = function(a, b)
					if a.BossesTotal < b.BossesTotal then
						return true
					end
				end
			end
		end
		table.sort(GUI_ABKsearch_data, sortfunc)

		CharScanner_MainFrame_SummaryButton1PvEBossesTotal:SetText(totalCount)
	else
		CharScanner_MainFrame_PvEStats_ScrollFrame:Show()
		CharScanner_MainFrame_PvEStats_Search_ScrollFrame:Hide()
		CharScanner_MainFrame_PvEStats_Search_Sort:Hide()
		CharScanner_MainFrame_PvEStats_TitleAchievement:Show()
		CharScanner_MainFrame_PvEStats_TitleStatistic:Show()
		CharScanner_MainFrame_PvEStats_TitleBossesTotal:Show()

		CharScanner_MainFrame_SummaryButton1PvEAchievement:Show()
		CharScanner_MainFrame_SummaryButton2PvEAchievement:Show()
		CharScanner_MainFrame_SummaryButton1PvEStatistic:Show()
		CharScanner_MainFrame_SummaryButton2PvEStatistic:Show()
		--CharScanner_MainFrame_SummaryButton1PvEBossesTotal:Show()

		if guid and SCAN[guid] then
			SCAN[guid].DataStep = 0
			CharScanner:ShowData(4, GUIselectedGUID, false)
			SCAN[guid].DataStep = 0
		end
	end
end

function CharScanner:ABKsearch_GUISort(sortBy)
	if GUI_ABKsearch_SortBy == sortBy then
		if GUI_ABKsearch_SortOrder == 1 then
			GUI_ABKsearch_SortOrder = 0
		else
			GUI_ABKsearch_SortOrder = 1
		end
	else
		GUI_ABKsearch_SortOrder = 1
	end
	GUI_ABKsearch_SortBy = sortBy
	CharScanner:ABKsearch_SortButtonUpdate()
	CharScanner:ABKsearch_GUICreate()
	CharScanner:ABKsearch_FillList()
end

function CharScanner:ABKsearch_SortButtonUpdate()
	local arrow
	for i=1, 3 do
		if i == 1 then -- 1 = standard
			arrow = "CharScanner_MainFrame_PvEStats_Search_SortButton_StandardArrow"
		elseif i == 2 then -- 2 = name
			arrow = "CharScanner_MainFrame_PvEStats_Search_SortButton_NameArrow"
		elseif i == 3 then -- 3 = kills
			arrow = "CharScanner_MainFrame_PvEStats_Search_SortButton_KillsArrow"
		end
		if GUI_ABKsearch_SortBy == i then
			if GUI_ABKsearch_SortOrder == 1 then _G[arrow]:SetTexCoord(unpack(Textures.SortArrow.coords1)) end
			if GUI_ABKsearch_SortOrder == 0 then _G[arrow]:SetTexCoord(unpack(Textures.SortArrow.coords2)) end
		else
			_G[arrow]:SetTexCoord(0, 0, 0, 0)
		end
	end
end

function CharScanner:ABKsearch_OnVerticalScroll(value)
	CharScanner_MainFrame_PvEStats_Search_ScrollFrameScrollBar:SetValue(value)
	CharScanner_MainFrame_PvEStats_Search_ScrollFrame.offset = math_floor((value / sizeGeneral.BarHeight) + 0.5)
	CharScanner:ABKsearch_FillList()
end

function CharScanner:ABKsearch_FillList()
	local frameName = "CharScanner_MainFrame_PvEStats_Search_ScrollFrame"
	local frame = _G[frameName]
	local scrollBar = _G[frameName.."ScrollBar"]
	local scrollUpButton = _G[frameName.."ScrollBarScrollUpButton"]
	local scrollDownButton = _G[frameName.."ScrollBarScrollDownButton"]
	local scrollFrameHeight = 0
	local numToDisplay = GUI_ABKsearch_curRows
	local valueStep = 12

	if GUI_ABKsearch_Count < numToDisplay then
		numToDisplay = GUI_ABKsearch_Count
	end

	if GUI_ABKsearch_Count > 0 then
		scrollFrameHeight = (GUI_ABKsearch_Count - numToDisplay) * valueStep
		if scrollFrameHeight < 0 then
			scrollFrameHeight = 0
		end
	end
	scrollBar:SetMinMaxValues(0, scrollFrameHeight)
	scrollBar:SetValueStep(valueStep)

	if GUI_ABKsearch_Count <= GUI_ABKsearch_curRows then
		ScrollFrameDisable(frameName)
	else
		ScrollFrameEnable(frameName)
	end

	if scrollBar:GetValue() == 0 then
		scrollUpButton:Disable()
	else
		scrollUpButton:Enable()
	end
	if ((scrollBar:GetValue() - scrollFrameHeight) == 0) then
		scrollDownButton:Disable()
	else
		scrollDownButton:Enable()
	end

	for i=1, GUI_ABKsearch_maxRows do
		_G["CharScanner_MainFrame_PvEStats_Search_ScrollFrame_Button"..i]:Hide()
	end

	local j = 1
	local sliderPos = CharScanner_MainFrame_PvEStats_Search_ScrollFrame.offset
	for i=sliderPos+1, sliderPos+numToDisplay do
		if sliderPos < GUI_ABKsearch_Count then
			if GUI_ABKsearch_data[i].BossesTotal > 0 then
				_G["CharScanner_MainFrame_PvEStats_Search_ScrollFrame_Button"..j.."Name"]:SetTextColor(color.Green.rgba[1], color.Green.rgba[2], color.Green.rgba[3], color.Green.rgba[4])
			else
				_G["CharScanner_MainFrame_PvEStats_Search_ScrollFrame_Button"..j.."Name"]:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
			end
			_G["CharScanner_MainFrame_PvEStats_Search_ScrollFrame_Button"..j.."Name"]:SetText(GUI_ABKsearch_data[i].Name)
			_G["CharScanner_MainFrame_PvEStats_Search_ScrollFrame_Button"..j.."BossesTotal"]:SetText(GUI_ABKsearch_data[i].BossesTotal)
			_G["CharScanner_MainFrame_PvEStats_Search_ScrollFrame_Button"..j]:Show()
			j = j + 1
		end
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:History_Update()
	CharScanner:History_GUICreate()
	CharScanner:History_FillList()
	CharScanner:History_StatusUpdate()
end

function CharScanner:History_GUICreate()
	if not GroupScanIsActive and not SingleScanIsActive then -- XFIX
		CleanTempData()
	end -- XFIX

	table.wipe(GUI_history_data)

	local x = 1
	for k, v in pairs(SCAN) do
		if v.time then -- XFIX
			GUI_history_data[x] = {}
			GUI_history_data[x].guid = k
			GUI_history_data[x].classEN = v.classEN
			GUI_history_data[x].class = v.class
			GUI_history_data[x].name = v.name
			GUI_history_data[x].time = v.time
			GUI_history_data[x].mainrole = v.mainRole

			-- talents START
			if v.activeTalentGroup then

				if v.activeTalentGroup == 1 then

					if v.talentGroup1name then
						local unspent = ""
						if v.talentGroup1pointsUnspent > 0 then
							unspent = "|cff"..color.Red.hex.."("..v.talentGroup1pointsUnspent..")|r"
						end

						local text = RoleTextShort[v.talentGroup1role].." |cff"..color.White.hex..v.talentGroup1name.." "..v.talentGroup1pointsString.."|r"
						local num = v.talentGroup1role..v.talentGroup1name..v.talentGroup1pointsString
						if v.talentGroup1pointsString == "0/0/0" then
							text = RoleTextShort[v.talentGroup1role].." "..unspent
							num = v.talentGroup1role..v.talentGroup1pointsUnspent
						else
							if unspent ~= "" then
								text = text.." "..unspent
							end
						end
						GUI_history_data[x].talents1 = text
						GUI_history_data[x].talents1Num = num
					else
						GUI_history_data[x].talents1 = ""
						GUI_history_data[x].talents1Num = "zzz"
					end
					if v.talentGroup2name then
						local unspent = ""
						if v.talentGroup2pointsUnspent > 0 then
							unspent = "|cff"..color.Red.hex.."("..v.talentGroup2pointsUnspent..")|r"
						end

						local text = RoleTextShort[v.talentGroup2role]--.." |cff"..col..v.talentGroup2name.." "..v.talentGroup2pointsString.."|r"
						local num = v.talentGroup2role..v.talentGroup2name..v.talentGroup2pointsString
						if v.talentGroup2pointsString == "0/0/0" then
							text = RoleTextShort[v.talentGroup2role].." "..unspent
							num = v.talentGroup2role..v.talentGroup2pointsUnspent
						else
							if unspent ~= "" then
								text = text.." "..unspent
							end
						end
						GUI_history_data[x].talents2 = text
						GUI_history_data[x].talents2Num = num
					else
						GUI_history_data[x].talents2 = ""
						GUI_history_data[x].talents2Num = "zzz"
					end

				elseif v.activeTalentGroup == 2 then

					if v.talentGroup2name then
						local unspent = ""
						if v.talentGroup2pointsUnspent > 0 then
							unspent = "|cff"..color.Red.hex.."("..v.talentGroup2pointsUnspent..")|r"
						end

						local text = RoleTextShort[v.talentGroup2role].." |cff"..color.White.hex..v.talentGroup2name.." "..v.talentGroup2pointsString.."|r"
						local num = v.talentGroup2role..v.talentGroup2name..v.talentGroup2pointsString
						if v.talentGroup2pointsString == "0/0/0" then
							text = RoleTextShort[v.talentGroup2role].." "..unspent
							num = v.talentGroup2role..v.talentGroup2pointsUnspent
						else
							if unspent ~= "" then
								text = text.." "..unspent
							end
						end
						GUI_history_data[x].talents1 = text
						GUI_history_data[x].talents1Num = num
					else
						GUI_history_data[x].talents1 = ""
						GUI_history_data[x].talents1Num = "zzz"
					end
					if v.talentGroup1name then
						local unspent = ""
						if v.talentGroup1pointsUnspent > 0 then
							unspent = "|cff"..color.Red.hex.."("..v.talentGroup1pointsUnspent..")|r"
						end

						local text = RoleTextShort[v.talentGroup1role]--.." |cff"..col..v.talentGroup1name.." "..v.talentGroup1pointsString.."|r"
						local num = v.talentGroup1role..v.talentGroup1name..v.talentGroup1pointsString
						if v.talentGroup1pointsString == "0/0/0" then
							text = RoleTextShort[v.talentGroup1role].." "..unspent
							num = v.talentGroup1role..v.talentGroup1pointsUnspent
						else
							if unspent ~= "" then
								text = text.." "..unspent
							end
						end
						GUI_history_data[x].talents2 = text
						GUI_history_data[x].talents2Num = num
					else
						GUI_history_data[x].talents2 = ""
						GUI_history_data[x].talents2Num = "zzz"
					end

				end

			else
				GUI_history_data[x].talents1 = ""
				GUI_history_data[x].talents1Num = "zzz"
				GUI_history_data[x].talents2 = ""
				GUI_history_data[x].talents2Num = "zzz"
			end
			-- talents END

			GUI_history_data[x].enchcount = v.gearEnchCount
			GUI_history_data[x].pvpcount = v.gearPvPCount
			GUI_history_data[x].itemlevel = cnum(v.itemlevelAvg)
			GUI_history_data[x].itemlevelNum = v.itemlevelAvg

			-- raid selection START
			local col
			if GUI_history_RaidL3 then
				local achievement = math_floor((v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].achievementCount/PVEDUMP[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].achievementTotal*100) * mult2 + 0.5) / mult2
				if PVEDUMP[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].achievementTotal == v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].achievementCount then
					col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
				end
				GUI_history_data[x].achievement = col..v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].achievementCount.."|r / |cff"..color.Level3.hex..PVEDUMP[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].achievementTotal.."|r"
				GUI_history_data[x].achievementNum = achievement
				local boss = math_floor((v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticBoss/PVEDUMP[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticTotal*100) * mult2 + 0.5) / mult2
				if PVEDUMP[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticTotal == v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticBoss then
					col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
				end
				GUI_history_data[x].boss = col..v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticBoss.."|r / |cff"..color.Level3.hex..PVEDUMP[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticTotal.."|r"
				GUI_history_data[x].bossNum = boss
				GUI_history_data[x].kills = v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2][GUI_history_RaidL3].statisticBossAll
			elseif GUI_history_RaidL2 then
				local achievement = math_floor((v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2].achievementCount/PVEDUMP[GUI_history_RaidL1][GUI_history_RaidL2].achievementTotal*100) * mult2 + 0.5) / mult2
				if PVEDUMP[GUI_history_RaidL1][GUI_history_RaidL2].achievementTotal == v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2].achievementCount then
					col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
				end
				GUI_history_data[x].achievement = col..v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2].achievementCount.."|r / |cff"..color.Level2.hex..PVEDUMP[GUI_history_RaidL1][GUI_history_RaidL2].achievementTotal.."|r"
				GUI_history_data[x].achievementNum = achievement
				local boss = math_floor((v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2].statisticBoss/PVEDUMP[GUI_history_RaidL1][GUI_history_RaidL2].statisticTotal*100) * mult2 + 0.5) / mult2
				if PVEDUMP[GUI_history_RaidL1][GUI_history_RaidL2].statisticTotal == v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2].statisticBoss then
					col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
				end
				GUI_history_data[x].boss = col..v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2].statisticBoss.."|r / |cff"..color.Level2.hex..PVEDUMP[GUI_history_RaidL1][GUI_history_RaidL2].statisticTotal.."|r"
				GUI_history_data[x].bossNum = boss
				GUI_history_data[x].kills = v.PVEDATA[GUI_history_RaidL1][GUI_history_RaidL2].statisticBossAll
			else
				if GUI_history_RaidL1 == 0 then -- Total
					local achievement = math_floor((v.PvEAchievementCount/v.PvEAchievementTotal*100) * mult2 + 0.5) / mult2
					GUI_history_data[x].achievement = "|cff"..color.LightYellow.hex..cnum(achievement).."%|r"
					GUI_history_data[x].achievementNum = achievement
					local boss = math_floor((v.PvEStatisticCount/v.PvEStatisticTotal*100) * mult2 + 0.5) / mult2
					GUI_history_data[x].boss = "|cff"..color.LightYellow.hex..cnum(boss).."%|r"
					GUI_history_data[x].bossNum = boss
					GUI_history_data[x].kills = v.PvEStatisticBossKills
				else
					local achievement = math_floor((v.PVEDATA[GUI_history_RaidL1].achievementCount/PVEDUMP[GUI_history_RaidL1].achievementTotal*100) * mult2 + 0.5) / mult2
					if PVEDUMP[GUI_history_RaidL1].achievementTotal == v.PVEDATA[GUI_history_RaidL1].achievementCount then
						col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
					end
					GUI_history_data[x].achievement = col..v.PVEDATA[GUI_history_RaidL1].achievementCount.."|r / |cff"..color.Level1.hex..PVEDUMP[GUI_history_RaidL1].achievementTotal.."|r"
					GUI_history_data[x].achievementNum = achievement
					local boss = math_floor((v.PVEDATA[GUI_history_RaidL1].statisticBoss/PVEDUMP[GUI_history_RaidL1].statisticTotal*100) * mult2 + 0.5) / mult2
					if PVEDUMP[GUI_history_RaidL1].statisticTotal == v.PVEDATA[GUI_history_RaidL1].statisticBoss then
						col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
					end
					GUI_history_data[x].boss = col..v.PVEDATA[GUI_history_RaidL1].statisticBoss.."|r / |cff"..color.Level1.hex..PVEDUMP[GUI_history_RaidL1].statisticTotal.."|r"
					GUI_history_data[x].bossNum = boss
					GUI_history_data[x].kills = v.PVEDATA[GUI_history_RaidL1].statisticBossAll
				end
			end
			-- raid selection END

			-- single achievement compare START
			if GUI_achievementdetail_Compare then
				if AchievementDetailFrameMode == "PvE" then
					if SCAN[k].PVEDATA.achievementID and SCAN[k].PVEDATA.achievementID[ GUI_achievementdetail_Compare ] then
						col="|cff"..color.Green.hex
					else
						col="|cff"..color.Red.hex
					end
					GUI_history_data[x].achievementCompare = col..AchievementIDTable[ GUI_achievementdetail_Compare ].."|r"
				elseif AchievementDetailFrameMode == "PvP" then
					if SCAN[k].PVPDATA.achievementID and SCAN[k].PVPDATA.achievementID[ GUI_achievementdetail_Compare ] then
						col="|cff"..color.Green.hex
					else
						col="|cff"..color.Red.hex
					end
					GUI_history_data[x].achievementCompare = col..AchievementIDTable[ GUI_achievementdetail_Compare ].."|r"
				end
			else
				GUI_history_data[x].achievementCompare = ""
			end
			-- single achievement compare END
			x = x + 1
		end -- XFIX
	end
	GUI_history_Count = #GUI_history_data
	NumScannedChars = #GUI_history_data

	local sortfunc
	if GUI_history_SortBy == 1 then -- 1 = class
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.class == b.class then
					if a.name < b.name then
						return true
					end
				elseif a.class < b.class then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.class == b.class then
					if a.name < b.name then
						return true
					end
				elseif a.class > b.class then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 2 then -- 2 = name
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.name < b.name then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.name > b.name then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 3 then -- 3 = time
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.time == b.time then
					if a.name < b.name then
						return true
					end
				elseif a.time < b.time then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.time == b.time then
					if a.name < b.name then
						return true
					end
				elseif a.time > b.time then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 4 then -- 4 = mainrole
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.mainrole == b.mainrole then
					if a.class == b.class then
						if a.name < b.name then
							return true
						end
					elseif a.class < b.class then
						return true
					end
				elseif a.mainrole < b.mainrole then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.mainrole == b.mainrole then
					if a.class == b.class then
						if a.name < b.name then
							return true
						end
					elseif a.class < b.class then
						return true
					end
				elseif a.mainrole > b.mainrole then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 5 then -- 5 = talents1
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.talents1Num == b.talents1Num then
					if a.name < b.name then
						return true
					end
				elseif a.talents1Num < b.talents1Num then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.talents1Num == b.talents1Num then
					if a.name < b.name then
						return true
					end
				elseif a.talents1Num > b.talents1Num then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 6 then -- 6 = talents2
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.talents2Num == b.talents2Num then
					if a.name < b.name then
						return true
					end
				elseif a.talents2Num < b.talents2Num then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.talents2Num == b.talents2Num then
					if a.name < b.name then
						return true
					end
				elseif a.talents2Num > b.talents2Num then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 7 then -- 7 = EnchantmentCount
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.enchcount == b.enchcount then
					if a.itemlevelNum == b.itemlevelNum then
						if a.name < b.name then
							return true
						end
					elseif a.itemlevelNum > b.itemlevelNum then
						return true
					end
				elseif a.enchcount > b.enchcount then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.enchcount == b.enchcount then
					if a.itemlevelNum == b.itemlevelNum then
						if a.name < b.name then
							return true
						end
					elseif a.itemlevelNum > b.itemlevelNum then
						return true
					end
				elseif a.enchcount < b.enchcount then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 8 then -- 8 = PvPCount
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.pvpcount == b.pvpcount then
					if a.itemlevelNum == b.itemlevelNum then
						if a.name < b.name then
							return true
						end
					elseif a.itemlevelNum > b.itemlevelNum then
						return true
					end
				elseif a.pvpcount > b.pvpcount then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.pvpcount == b.pvpcount then
					if a.itemlevelNum == b.itemlevelNum then
						if a.name < b.name then
							return true
						end
					elseif a.itemlevelNum > b.itemlevelNum then
						return true
					end
				elseif a.pvpcount < b.pvpcount then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 9 then -- 9 = itemlevel (itemlevelNum)
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.itemlevelNum == b.itemlevelNum then
					if a.class == b.class then
						if a.name < b.name then
							return true
						end
					elseif a.class < b.class then
						return true
					end
				elseif a.itemlevelNum > b.itemlevelNum then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.itemlevelNum == b.itemlevelNum then
					if a.class == b.class then
						if a.name < b.name then
							return true
						end
					elseif a.class < b.class then
						return true
					end
				elseif a.itemlevelNum < b.itemlevelNum then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 10 then -- 10 = achievement (achievementNum)
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.achievementNum == b.achievementNum then
					if a.bossNum == b.bossNum then
						if a.kills == b.kills then
							if a.name < b.name then
								return true
							end
						elseif a.kills > b.kills then
							return true
						end
					elseif a.bossNum > b.bossNum then
						return true
					end
				elseif a.achievementNum > b.achievementNum then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.achievementNum == b.achievementNum then
					if a.bossNum == b.bossNum then
						if a.kills == b.kills then
							if a.name < b.name then
								return true
							end
						elseif a.kills < b.kills then
							return true
						end
					elseif a.bossNum < b.bossNum then
						return true
					end
				elseif a.achievementNum < b.achievementNum then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 11 then -- 11 = boss (bossNum)
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.bossNum == b.bossNum then
					if a.achievementNum == b.achievementNum then
						if a.kills == b.kills then
							if a.name < b.name then
								return true
							end
						elseif a.kills > b.kills then
							return true
						end
					elseif a.achievementNum > b.achievementNum then
						return true
					end
				elseif a.bossNum > b.bossNum then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.bossNum == b.bossNum then
					if a.achievementNum == b.achievementNum then
						if a.kills == b.kills then
							if a.name < b.name then
								return true
							end
						elseif a.kills < b.kills then
							return true
						end
					elseif a.achievementNum < b.achievementNum then
						return true
					end
				elseif a.bossNum < b.bossNum then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 12 then -- 12 = kills (kills)
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.kills == b.kills then
					if a.bossNum == b.bossNum then
						if a.achievementNum == b.achievementNum then
							if a.name < b.name then
								return true
							end
						elseif a.achievementNum > b.achievementNum then
							return true
						end
					elseif a.bossNum > b.bossNum then
						return true
					end
				elseif a.kills > b.kills then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.kills == b.kills then
					if a.bossNum == b.bossNum then
						if a.achievementNum == b.achievementNum then
							if a.name < b.name then
								return true
							end
						elseif a.achievementNum < b.achievementNum then
							return true
						end
					elseif a.bossNum < b.bossNum then
						return true
					end
				elseif a.kills < b.kills then
					return true
				end
			end
		end
	elseif GUI_history_SortBy == 13 then -- 13 = achievementCompare (SingleAchievementCompare)
		if GUI_history_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.achievementCompare == b.achievementCompare then
					if a.name < b.name then
						return true
					end
				elseif a.achievementCompare < b.achievementCompare then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.achievementCompare == b.achievementCompare then
					if a.name < b.name then
						return true
					end
				elseif a.achievementCompare > b.achievementCompare then
					return true
				end
			end
		end

	end

	table.sort(GUI_history_data, sortfunc)
end

function CharScanner:History_GUISort(what, sortBy, l1, l2, l3)
	if what == "RAID" then

		CharScanner_HistoryFrame_SortButton_Achievement:Show()
		CharScanner_HistoryFrame_SortButton_Boss:Show()
		CharScanner_HistoryFrame_SortButton_Kills:Show()
		CharScanner_HistoryFrame_SortButton_SingleAchievementCompare:Hide()
		for i = 1, GUI_history_maxRows do
			_G["CharScanner_HistoryFrame_Button"..i.."Achievement"]:Show()
			_G["CharScanner_HistoryFrame_Button"..i.."Boss"]:Show()
			_G["CharScanner_HistoryFrame_Button"..i.."Kills"]:Show()
			_G["CharScanner_HistoryFrame_Button"..i.."SingleAchievementCompare"]:Hide()
		end

		GUI_history_RaidL1 = l1
		GUI_history_RaidL2 = l2
		GUI_history_RaidL3 = l3
		local name
		if GUI_history_RaidL3 then
			name = L[PVEDUMP[l1][l2][l3].name].." ("..L[PVEDUMP[l1][l2].name]..")"
		elseif GUI_history_RaidL2 then
			name = L[PVEDUMP[l1][l2].name]
		else
			if l1 == 0 then
				name = L["Total"]
			else
				name = L[PVEDUMP[l1].name]
			end
		end
		CharScanner_HistoryFrame_Sort_RaidSelectMainButtonName:SetText(name)
		CharScanner_HistoryFrame_Sort_RaidSelectMainButton_Expand:SetTexCoord(unpack(Textures.Expand.coords))
		CharScanner_HistoryFrame_Sort_RaidSelectFrame:Hide()
		GUI_history_SortOrder = 1
		GUI_history_SortBy = 11 -- Boss
	elseif what == "ACHIEVEMENT" then
		GUI_achievementdetail_Compare = l1
		CharScanner_HistoryFrame_SortButton_Achievement:Hide()
		CharScanner_HistoryFrame_SortButton_Boss:Hide()
		CharScanner_HistoryFrame_SortButton_Kills:Hide()
		CharScanner_HistoryFrame_SortButton_SingleAchievementCompare:Show()
		for i = 1, GUI_history_maxRows do
			_G["CharScanner_HistoryFrame_Button"..i.."Achievement"]:Hide()
			_G["CharScanner_HistoryFrame_Button"..i.."Boss"]:Hide()
			_G["CharScanner_HistoryFrame_Button"..i.."Kills"]:Hide()
			_G["CharScanner_HistoryFrame_Button"..i.."SingleAchievementCompare"]:Show()
		end

		CharScanner_HistoryFrame_Sort_RaidSelectMainButtonName:SetText(AchievementIDTable[l1])
		CharScanner_HistoryFrame_Sort_RaidSelectMainButton_Expand:SetTexCoord(unpack(Textures.Expand.coords))
		CharScanner_HistoryFrame_Sort_RaidSelectFrame:Hide()
		GUI_history_SortOrder = 1
		GUI_history_SortBy = 13 -- SingleAchievementCompare
	elseif what == "PVP" then
		--print(what, sortBy, l1, l2, l3)
	else
		if GUI_history_SortBy == sortBy then
			if GUI_history_SortOrder == 1 then
				GUI_history_SortOrder = 0
			else
				GUI_history_SortOrder = 1
			end
		else
			GUI_history_SortOrder = 1
		end
		GUI_history_SortBy = sortBy
	end
	CharScanner:History_SortButtonUpdate()
	CharScanner:History_GUICreate()
	CharScanner:History_FillList()
end

function CharScanner:History_SortButtonUpdate()
	local arrow
	for i=1, 13 do
		if i == 1 then     -- 1 = class
			arrow = "CharScanner_HistoryFrame_SortButton_IconArrow"
		elseif i == 2 then -- 2 = name
			arrow = "CharScanner_HistoryFrame_SortButton_NameArrow"
		elseif i == 3 then -- 3 = time
			arrow = "CharScanner_HistoryFrame_SortButton_TimeArrow"
		elseif i == 4 then -- 4 = mainrole
			arrow = "CharScanner_HistoryFrame_SortButton_MainRoleArrow"
		elseif i == 5 then -- 5 = talents1
			arrow = "CharScanner_HistoryFrame_SortButton_Talents1Arrow"
		elseif i == 6 then -- 6 = talents2
			arrow = "CharScanner_HistoryFrame_SortButton_Talents2Arrow"
		elseif i == 7 then -- 7 = EnchantmentCount
			arrow = "CharScanner_HistoryFrame_SortButton_EnchantmentCountArrow"
		elseif i == 8 then -- 8 = PvPCount
			arrow = "CharScanner_HistoryFrame_SortButton_PvPCountArrow"
		elseif i == 9 then -- 9 = ItemLevel
			arrow = "CharScanner_HistoryFrame_SortButton_ItemLevelArrow"
		elseif i == 10 then -- 10 = Achievement
			arrow = "CharScanner_HistoryFrame_SortButton_AchievementArrow"
		elseif i == 11 then -- 11 = Boss
			arrow = "CharScanner_HistoryFrame_SortButton_BossArrow"
		elseif i == 12 then -- 12 = Kills
			arrow = "CharScanner_HistoryFrame_SortButton_KillsArrow"
		elseif i == 13 then -- 13 = SingleAchievementCompare
			arrow = "CharScanner_HistoryFrame_SortButton_SingleAchievementCompareArrow"
		end
		if GUI_history_SortBy == i then
			if GUI_history_SortOrder == 1 then _G[arrow]:SetTexCoord(unpack(Textures.SortArrow.coords1)) end
			if GUI_history_SortOrder == 0 then _G[arrow]:SetTexCoord(unpack(Textures.SortArrow.coords2)) end
		else
			_G[arrow]:SetTexCoord(0, 0, 0, 0)
		end
	end
end

function CharScanner:History_OnVerticalScroll(value)
	CharScanner_HistoryFrameScrollFrameScrollBar:SetValue(value)
	CharScanner_HistoryFrameScrollFrame.offset = math_floor((value / sizeGeneral.BarHeight) + 0.5)
	CharScanner:History_FillList()
end

function CharScanner:History_FillList()
	local frameName = "CharScanner_HistoryFrameScrollFrame"
	local frame = _G[frameName]
	local scrollBar = _G[frameName.."ScrollBar"]
	local scrollUpButton = _G[frameName.."ScrollBarScrollUpButton"]
	local scrollDownButton = _G[frameName.."ScrollBarScrollDownButton"]
	local scrollFrameHeight = 0
	local numToDisplay = GUI_history_curRows
	local valueStep = 12

	if GUI_history_Count < numToDisplay then
		numToDisplay = GUI_history_Count
	end

	if GUI_history_Count > 0 then
		scrollFrameHeight = (GUI_history_Count - numToDisplay) * valueStep
		if scrollFrameHeight < 0 then
			scrollFrameHeight = 0
		end
	end
	scrollBar:SetMinMaxValues(0, scrollFrameHeight)
	scrollBar:SetValueStep(valueStep)

	if GUI_history_Count <= GUI_history_curRows then
		ScrollFrameDisable(frameName)
	else
		ScrollFrameEnable(frameName)
	end

	if scrollBar:GetValue() == 0 then
		scrollUpButton:Disable()
	else
		scrollUpButton:Enable()
	end
	if ((scrollBar:GetValue() - scrollFrameHeight) == 0) then
		scrollDownButton:Disable()
	else
		scrollDownButton:Enable()
	end

	for i=1, GUI_history_maxRows do
		_G["CharScanner_HistoryFrame_Button"..i]:Hide()
	end

	local j = 1
	local sliderPos = CharScanner_HistoryFrameScrollFrame.offset
	for i=sliderPos+1, sliderPos+numToDisplay do
		if sliderPos < GUI_history_Count then
			_G["CharScanner_HistoryFrame_Button"..j.."Icon"]:SetTexCoord(classes[GUI_history_data[i].classEN].icon[1], classes[GUI_history_data[i].classEN].icon[2], classes[GUI_history_data[i].classEN].icon[3], classes[GUI_history_data[i].classEN].icon[4])
			_G["CharScanner_HistoryFrame_Button"..j.."Name"]:SetText(GUI_history_data[i].name)
			_G["CharScanner_HistoryFrame_Button"..j.."Name"]:SetTextColor(classcolors[GUI_history_data[i].classEN].r, classcolors[GUI_history_data[i].classEN].g, classcolors[GUI_history_data[i].classEN].b, 1)
			_G["CharScanner_HistoryFrame_Button"..j.."Time"]:SetText(GUI_history_data[i].time)

			_G["CharScanner_HistoryFrame_Button"..j.."MainRole"]:SetTexture(Textures.RoleIcon[GUI_history_data[i].mainrole].path)
			_G["CharScanner_HistoryFrame_Button"..j.."MainRole"]:SetTexCoord(Textures.RoleIcon[GUI_history_data[i].mainrole].coords[1], Textures.RoleIcon[GUI_history_data[i].mainrole].coords[2], Textures.RoleIcon[GUI_history_data[i].mainrole].coords[3], Textures.RoleIcon[GUI_history_data[i].mainrole].coords[4])

			_G["CharScanner_HistoryFrame_Button"..j.."Talents1"]:SetText(GUI_history_data[i].talents1)
			_G["CharScanner_HistoryFrame_Button"..j.."Talents2"]:SetText(GUI_history_data[i].talents2)

			_G["CharScanner_HistoryFrame_Button"..j.."EnchantmentCount"]:SetText(GUI_history_data[i].enchcount)
			_G["CharScanner_HistoryFrame_Button"..j.."PvPCount"]:SetText(GUI_history_data[i].pvpcount)
			_G["CharScanner_HistoryFrame_Button"..j.."ItemLevel"]:SetText(GUI_history_data[i].itemlevel)

			_G["CharScanner_HistoryFrame_Button"..j.."Achievement"]:SetText(GUI_history_data[i].achievement)
			_G["CharScanner_HistoryFrame_Button"..j.."Boss"]:SetText(GUI_history_data[i].boss)
			_G["CharScanner_HistoryFrame_Button"..j.."Kills"]:SetText(GUI_history_data[i].kills)

			_G["CharScanner_HistoryFrame_Button"..j.."SingleAchievementCompare"]:SetText(GUI_history_data[i].achievementCompare)

			_G["CharScanner_HistoryFrame_Button"..j]:Show()
			_G["CharScanner_HistoryFrame_Button"..j].guid = GUI_history_data[i].guid

			_G["CharScanner_HistoryFrame_Button"..j.."Arrow"]:Hide()
			if GUIselectedGUID == GUI_history_data[i].guid then
				_G["CharScanner_HistoryFrame_Button"..j.."Arrow"]:Show()
			end

			j = j + 1
		end
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:AchievementDetail_Update(mode, l1, l2, l3, sel)
	-- do not show AchievementDetail if last scan is incomplete
	if not LatestScanIsComplete then
		CharScanner:AchievementDetail_HideAll()
		return
	end

	if sel then
		GUI_achievementdetail_Select = sel
		local name
		local col
		if sel == 1 then
			name = ALL
			col = "White"
		elseif sel == 2 then
			name = COMPLETE
			col = "Green"
		elseif sel == 3 then
			name = INCOMPLETE
			col = "Red"
		end
		CharScanner_AchievementDetailFrame_Sort_AchievementDetailSelectMainButtonName:SetText(name)
		CharScanner_AchievementDetailFrame_Sort_AchievementDetailSelectMainButtonName:SetTextColor(color[col].rgba[1], color[col].rgba[2], color[col].rgba[3], color[col].rgba[4])
		CharScanner_AchievementDetailFrame_Sort_AchievementDetailSelectMainButton_Expand:SetTexCoord(unpack(Textures.Expand.coords))
		CharScanner_AchievementDetailFrame_Sort_AchievementDetailSelectFrame:Hide()
	else
		if mode == "PvE" then
			GUI_achievementdetail_L1 = l1
			GUI_achievementdetail_L2 = l2
			GUI_achievementdetail_L3 = l3
		else
			GUI_PvPachievementdetail_L1 = l1
			GUI_PvPachievementdetail_L2 = l2
			GUI_PvPachievementdetail_L3 = l3
		end
	end

	CharScanner_AchievementDetailFrame:Hide() -- for SetToplevel
	CharScanner_AchievementDetailFrame:Show()
	CharScanner:AchievementDetail_GUICreate(mode)
	CharScanner:AchievementDetail_FillList()
	CharScanner:AchievementDetail_StatusUpdate(mode)
end

function CharScanner:AchievementDetail_StatusUpdate(mode)
	local all = 0
	local complete = 0
	local incomplete = 0

	if mode == "PvE" then
		if GUI_achievementdetail_L3 then
			all        = PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievementTotal
			complete   = SCAN[GUIselectedGUID].PVEDATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievementCount
			incomplete = all-complete
		elseif GUI_achievementdetail_L2 then
			all        = PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievementTotal
			complete   = SCAN[GUIselectedGUID].PVEDATA[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievementCount
			incomplete = all-complete
		else
			all        = PVEDUMP[GUI_achievementdetail_L1].achievementTotal
			complete   = SCAN[GUIselectedGUID].PVEDATA[GUI_achievementdetail_L1].achievementCount
			incomplete = all-complete
		end
	else
		if GUI_PvPachievementdetail_L3 then
			all        = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievementTotal
			complete   = SCAN[GUIselectedGUID].PVPDATA[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievementCount
			incomplete = all-complete
		elseif GUI_PvPachievementdetail_L2 then
			all        = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievementTotal
			complete   = SCAN[GUIselectedGUID].PVPDATA[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievementCount
			incomplete = all-complete
		else
			all        = PVPDUMP[GUI_PvPachievementdetail_L1].achievementTotal
			complete   = SCAN[GUIselectedGUID].PVPDATA[GUI_PvPachievementdetail_L1].achievementCount
			incomplete = all-complete
		end
	end

	CharScanner_AchievementDetailFrame_Status_Text:SetText("|cff"..color.Green.hex..complete.."|r | |cff"..color.Red.hex..incomplete.."|r | |cff"..color.White.hex..all.."|r")
end

function CharScanner:AchievementDetail_GUICreate(mode)
	if mode == "PvE" then
		-- PvE START  ----------------------
		-- DUMMY_FactionFix START
		if SCAN[GUIselectedGUID].faction == "Horde" then
			PVEDUMP[3][1].achievement[54] = 4297
		else
			PVEDUMP[3][1].achievement[54] = 4298
		end
		-- DUMMY_FactionFix END

		table.wipe(GUI_achievementdetail_data)
		local x = 1
		if GUI_achievementdetail_L3 then

			if GUI_achievementdetail_Select == 1 then
				for i = 1, #PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement do
					GUI_achievementdetail_data[x] = {}
					GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i]
					GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i] ]
					if SCAN[GUIselectedGUID].PVEDATA.achievementID and SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i] ] then
						GUI_achievementdetail_data[x].color = 1
					end
					x = x + 1
				end
				GUI_achievementdetail_Count = x - 1
			elseif GUI_achievementdetail_Select == 2 then
				for i = 1, #PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement do
					if SCAN[GUIselectedGUID].PVEDATA.achievementID and SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i] ] then
						GUI_achievementdetail_data[x] = {}
						GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i]
						GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i] ]
						GUI_achievementdetail_data[x].color = 1
						x = x + 1
					end
				end
				GUI_achievementdetail_Count = x - 1
			elseif GUI_achievementdetail_Select == 3 then
				for i = 1, #PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement do
					if not SCAN[GUIselectedGUID].PVEDATA.achievementID or not SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i] ] then
						GUI_achievementdetail_data[x] = {}
						GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i]
						GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][GUI_achievementdetail_L3].achievement[i] ]
						x = x + 1
					end
				end
				GUI_achievementdetail_Count = x - 1
			end

		elseif GUI_achievementdetail_L2 then

			if GUI_achievementdetail_Select == 1 then
				if not PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement then
					for k = 1, #PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2] do
						if PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement then
							for i = 1, #PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement do
								GUI_achievementdetail_data[x] = {}
								GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i]
								GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i] ]
								if SCAN[GUIselectedGUID].PVEDATA.achievementID and SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i] ] then
									GUI_achievementdetail_data[x].color = 1
								end
								x = x + 1
							end
						end
					end
				else
					for i = 1, #PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement do
						GUI_achievementdetail_data[x] = {}
						GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i]
						GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i] ]
						if SCAN[GUIselectedGUID].PVEDATA.achievementID and SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i] ] then
							GUI_achievementdetail_data[x].color = 1
						end
						x = x + 1
					end
				end
				GUI_achievementdetail_Count = x - 1
			elseif GUI_achievementdetail_Select == 2 then
				if not PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement then
					for k = 1, #PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2] do
						if PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement then
							for i = 1, #PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement do
								if SCAN[GUIselectedGUID].PVEDATA.achievementID and SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i] ] then
									GUI_achievementdetail_data[x] = {}
									GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i]
									GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i] ]
									GUI_achievementdetail_data[x].color = 1
									x = x + 1
								end
							end
						end
					end
				else
					for i = 1, #PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement do
						if SCAN[GUIselectedGUID].PVEDATA.achievementID and SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i] ] then
							GUI_achievementdetail_data[x] = {}
							GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i]
							GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i] ]
							GUI_achievementdetail_data[x].color = 1
							x = x + 1
						end
					end
				end
				GUI_achievementdetail_Count = x - 1
			elseif GUI_achievementdetail_Select == 3 then
				if not PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement then
					for k = 1, #PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2] do
						if PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement then
							for i = 1, #PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement do
								if not SCAN[GUIselectedGUID].PVEDATA.achievementID or not SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i] ] then
									GUI_achievementdetail_data[x] = {}
									GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i]
									GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2][k].achievement[i] ]
									x = x + 1
								end
							end
						end
					end
				else
					for i = 1, #PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement do
						if not SCAN[GUIselectedGUID].PVEDATA.achievementID or not SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i] ] then
							GUI_achievementdetail_data[x] = {}
							GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i]
							GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][GUI_achievementdetail_L2].achievement[i] ]
							x = x + 1
						end
					end
				end
				GUI_achievementdetail_Count = x - 1
			end

		else

			if GUI_achievementdetail_Select == 1 then
				if not PVEDUMP[GUI_achievementdetail_L1].achievement then
					for j = 1, #PVEDUMP[GUI_achievementdetail_L1] do
						if PVEDUMP[GUI_achievementdetail_L1][j].achievement then
							for i = 1, #PVEDUMP[GUI_achievementdetail_L1][j].achievement do
								GUI_achievementdetail_data[x] = {}
								GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][j].achievement[i]
								GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][j].achievement[i] ]
								if SCAN[GUIselectedGUID].PVEDATA.achievementID and SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][j].achievement[i] ] then
									GUI_achievementdetail_data[x].color = 1
								end
								x = x + 1
							end
						end
						for k = 1, #PVEDUMP[GUI_achievementdetail_L1][j] do
							if PVEDUMP[GUI_achievementdetail_L1][j][k].achievement then
								for i = 1, #PVEDUMP[GUI_achievementdetail_L1][j][k].achievement do
									GUI_achievementdetail_data[x] = {}
									GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][j][k].achievement[i]
									GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][j][k].achievement[i] ]
									if SCAN[GUIselectedGUID].PVEDATA.achievementID and SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][j][k].achievement[i] ] then
										GUI_achievementdetail_data[x].color = 1
									end
									x = x + 1
								end
							end
						end
					end
				else
					for i = 1, #PVEDUMP[GUI_achievementdetail_L1].achievement do
						GUI_achievementdetail_data[x] = {}
						GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1].achievement[i]
						GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1].achievement[i] ]
						if SCAN[GUIselectedGUID].PVEDATA.achievementID and SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1].achievement[i] ] then
							GUI_achievementdetail_data[x].color = 1
						end
						x = x + 1
					end
				end
				GUI_achievementdetail_Count = x - 1
			elseif GUI_achievementdetail_Select == 2 then
				if not PVEDUMP[GUI_achievementdetail_L1].achievement then
					for j = 1, #PVEDUMP[GUI_achievementdetail_L1] do
						if PVEDUMP[GUI_achievementdetail_L1][j].achievement then
							for i = 1, #PVEDUMP[GUI_achievementdetail_L1][j].achievement do
								if SCAN[GUIselectedGUID].PVEDATA.achievementID and SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][j].achievement[i] ] then
									GUI_achievementdetail_data[x] = {}
									GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][j].achievement[i]
									GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][j].achievement[i] ]
									GUI_achievementdetail_data[x].color = 1
									x = x + 1
								end
							end
						end
						for k = 1, #PVEDUMP[GUI_achievementdetail_L1][j] do
							if PVEDUMP[GUI_achievementdetail_L1][j][k].achievement then
								for i = 1, #PVEDUMP[GUI_achievementdetail_L1][j][k].achievement do
									if SCAN[GUIselectedGUID].PVEDATA.achievementID and SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][j][k].achievement[i] ] then
										GUI_achievementdetail_data[x] = {}
										GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][j][k].achievement[i]
										GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][j][k].achievement[i] ]
										GUI_achievementdetail_data[x].color = 1
										x = x + 1
									end
								end
							end
						end
					end
				else
					for i = 1, #PVEDUMP[GUI_achievementdetail_L1].achievement do
						if SCAN[GUIselectedGUID].PVEDATA.achievementID and SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1].achievement[i] ] then
							GUI_achievementdetail_data[x] = {}
							GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1].achievement[i]
							GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1].achievement[i] ]
							GUI_achievementdetail_data[x].color = 1
							x = x + 1
						end
					end
				end
				GUI_achievementdetail_Count = x - 1
			elseif GUI_achievementdetail_Select == 3 then
				if not PVEDUMP[GUI_achievementdetail_L1].achievement then
					for j = 1, #PVEDUMP[GUI_achievementdetail_L1] do
						if PVEDUMP[GUI_achievementdetail_L1][j].achievement then
							for i = 1, #PVEDUMP[GUI_achievementdetail_L1][j].achievement do
								if not SCAN[GUIselectedGUID].PVEDATA.achievementID or not SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][j].achievement[i] ] then
									GUI_achievementdetail_data[x] = {}
									GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][j].achievement[i]
									GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][j].achievement[i] ]
									x = x + 1
								end
							end
						end
						for k = 1, #PVEDUMP[GUI_achievementdetail_L1][j] do
							if PVEDUMP[GUI_achievementdetail_L1][j][k].achievement then
								for i = 1, #PVEDUMP[GUI_achievementdetail_L1][j][k].achievement do
									if not SCAN[GUIselectedGUID].PVEDATA.achievementID or not SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1][j][k].achievement[i] ] then
										GUI_achievementdetail_data[x] = {}
										GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1][j][k].achievement[i]
										GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1][j][k].achievement[i] ]
										x = x + 1
									end
								end
							end
						end
					end
				else
					for i = 1, #PVEDUMP[GUI_achievementdetail_L1].achievement do
						if not SCAN[GUIselectedGUID].PVEDATA.achievementID or not SCAN[GUIselectedGUID].PVEDATA.achievementID[ PVEDUMP[GUI_achievementdetail_L1].achievement[i] ] then
							GUI_achievementdetail_data[x] = {}
							GUI_achievementdetail_data[x].id = PVEDUMP[GUI_achievementdetail_L1].achievement[i]
							GUI_achievementdetail_data[x].name = AchievementIDTable[ PVEDUMP[GUI_achievementdetail_L1].achievement[i] ]
							x = x + 1
						end
					end
				end
				GUI_achievementdetail_Count = x - 1
			end

		end
		-- PvE END -------------------------
	else
		-- PvP START -----------------------
		table.wipe(GUI_achievementdetail_data)

		local x = 1
		if GUI_PvPachievementdetail_L3 then
			if GUI_achievementdetail_Select == 1 then
				for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement do
					GUI_achievementdetail_data[x] = {}
					local valID
					if PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].id then
						valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].id
					elseif PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].Aid then
						if SCAN[GUIselectedGUID].faction == "Horde" then
							valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].Hid
						else
							valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].Aid
						end
					end
					GUI_achievementdetail_data[x].id = valID
					GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
					if SCAN[GUIselectedGUID].PVPDATA.achievementID and SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
						GUI_achievementdetail_data[x].color = 1
					end
					x = x + 1
				end
				GUI_achievementdetail_Count = x - 1
			elseif GUI_achievementdetail_Select == 2 then
				for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement do

					local valID
					if PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].id then
						valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].id
					elseif PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].Aid then
						if SCAN[GUIselectedGUID].faction == "Horde" then
							valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].Hid
						else
							valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].Aid
						end
					end

					if SCAN[GUIselectedGUID].PVPDATA.achievementID and SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
						GUI_achievementdetail_data[x] = {}

						GUI_achievementdetail_data[x].id = valID
						GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
						GUI_achievementdetail_data[x].color = 1
						x = x + 1
					end
				end
				GUI_achievementdetail_Count = x - 1
			elseif GUI_achievementdetail_Select == 3 then
				for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement do

					local valID
					if PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].id then
						valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].id
					elseif PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].Aid then
						if SCAN[GUIselectedGUID].faction == "Horde" then
							valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].Hid
						else
							valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][GUI_PvPachievementdetail_L3].achievement[i].Aid
						end
					end

					if not SCAN[GUIselectedGUID].PVPDATA.achievementID or not SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
						GUI_achievementdetail_data[x] = {}

						GUI_achievementdetail_data[x].id = valID
						GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
						x = x + 1
					end
				end
				GUI_achievementdetail_Count = x - 1
			end

		elseif GUI_PvPachievementdetail_L2 then

			if GUI_achievementdetail_Select == 1 then
				if not PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement then
					for k = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2] do
						if PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement then
							for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement do
								GUI_achievementdetail_data[x] = {}
								local valID
								if PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].id then
									valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].id
								elseif PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].Aid then
									if SCAN[GUIselectedGUID].faction == "Horde" then
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].Hid
									else
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].Aid
									end
								end
								GUI_achievementdetail_data[x].id = valID
								GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
								if SCAN[GUIselectedGUID].PVPDATA.achievementID and SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
									GUI_achievementdetail_data[x].color = 1
								end
								x = x + 1
							end
						end
					end
				else
					for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement do
						GUI_achievementdetail_data[x] = {}
						local valID
						if PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].id then
							valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].id
						elseif PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].Aid then
							if SCAN[GUIselectedGUID].faction == "Horde" then
								valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].Hid
							else
								valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].Aid
							end
						end
						GUI_achievementdetail_data[x].id = valID
						GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
						if SCAN[GUIselectedGUID].PVPDATA.achievementID and SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
							GUI_achievementdetail_data[x].color = 1
						end
						x = x + 1
					end
				end
				GUI_achievementdetail_Count = x - 1
			elseif GUI_achievementdetail_Select == 2 then
				if not PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement then
					for k = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2] do
						if PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement then
							for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement do

								local valID
								if PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].id then
									valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].id
								elseif PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].Aid then
									if SCAN[GUIselectedGUID].faction == "Horde" then
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].Hid
									else
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].Aid
									end
								end

								if SCAN[GUIselectedGUID].PVPDATA.achievementID and SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
									GUI_achievementdetail_data[x] = {}

									GUI_achievementdetail_data[x].id = valID
									GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
									GUI_achievementdetail_data[x].color = 1
									x = x + 1
								end
							end
						end
					end
				else
					for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement do

						local valID
						if PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].id then
							valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].id
						elseif PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].Aid then
							if SCAN[GUIselectedGUID].faction == "Horde" then
								valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].Hid
							else
								valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].Aid
							end
						end

						if SCAN[GUIselectedGUID].PVPDATA.achievementID and SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
							GUI_achievementdetail_data[x] = {}

							GUI_achievementdetail_data[x].id = valID
							GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
							GUI_achievementdetail_data[x].color = 1
							x = x + 1
						end
					end
				end
				GUI_achievementdetail_Count = x - 1
			elseif GUI_achievementdetail_Select == 3 then
				if not PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement then
					for k = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2] do
						if PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement then
							for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement do

								local valID
								if PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].id then
									valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].id
								elseif PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].Aid then
									if SCAN[GUIselectedGUID].faction == "Horde" then
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].Hid
									else
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2][k].achievement[i].Aid
									end
								end

								if not SCAN[GUIselectedGUID].PVPDATA.achievementID or not SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
									GUI_achievementdetail_data[x] = {}

									GUI_achievementdetail_data[x].id = valID
									GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
									x = x + 1
								end
							end
						end
					end
				else
					for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement do

						local valID
						if PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].id then
							valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].id
						elseif PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].Aid then
							if SCAN[GUIselectedGUID].faction == "Horde" then
								valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].Hid
							else
								valID = PVPDUMP[GUI_PvPachievementdetail_L1][GUI_PvPachievementdetail_L2].achievement[i].Aid
							end
						end

						if not SCAN[GUIselectedGUID].PVPDATA.achievementID or not SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
							GUI_achievementdetail_data[x] = {}

							GUI_achievementdetail_data[x].id = valID
							GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
							x = x + 1
						end
					end
				end
				GUI_achievementdetail_Count = x - 1
			end

		else

			if GUI_achievementdetail_Select == 1 then
				if not PVPDUMP[GUI_PvPachievementdetail_L1].achievement then
					for j = 1, #PVPDUMP[GUI_PvPachievementdetail_L1] do
						if PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement then
							for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement do
								GUI_achievementdetail_data[x] = {}
								local valID
								if PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].id then
									valID = PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].id
								elseif PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].Aid then
									if SCAN[GUIselectedGUID].faction == "Horde" then
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].Hid
									else
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].Aid
									end
								end
								GUI_achievementdetail_data[x].id = valID
								GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
								if SCAN[GUIselectedGUID].PVPDATA.achievementID and SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
									GUI_achievementdetail_data[x].color = 1
								end
								x = x + 1
							end
						end
						for k = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][j] do
							if PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement then
								for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement do
									GUI_achievementdetail_data[x] = {}
									local valID
									if PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].id then
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].id
									elseif PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].Aid then
										if SCAN[GUIselectedGUID].faction == "Horde" then
											valID = PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].Hid
										else
											valID = PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].Aid
										end
									end
									GUI_achievementdetail_data[x].id = valID
									GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
									if SCAN[GUIselectedGUID].PVPDATA.achievementID and SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
										GUI_achievementdetail_data[x].color = 1
									end
									x = x + 1
								end
							end
						end
					end
				else
					for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1].achievement do
						GUI_achievementdetail_data[x] = {}
						local valID
						if PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].id then
							valID = PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].id
						elseif PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].Aid then
							if SCAN[GUIselectedGUID].faction == "Horde" then
								valID = PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].Hid
							else
								valID = PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].Aid
							end
						end
						GUI_achievementdetail_data[x].id = valID
						GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
						if SCAN[GUIselectedGUID].PVPDATA.achievementID and SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
							GUI_achievementdetail_data[x].color = 1
						end
						x = x + 1
					end
				end
				GUI_achievementdetail_Count = x - 1
			elseif GUI_achievementdetail_Select == 2 then
				if not PVPDUMP[GUI_PvPachievementdetail_L1].achievement then
					for j = 1, #PVPDUMP[GUI_PvPachievementdetail_L1] do
						if PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement then
							for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement do

								local valID
								if PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].id then
									valID = PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].id
								elseif PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].Aid then
									if SCAN[GUIselectedGUID].faction == "Horde" then
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].Hid
									else
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].Aid
									end
								end

								if SCAN[GUIselectedGUID].PVPDATA.achievementID and SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
									GUI_achievementdetail_data[x] = {}

									GUI_achievementdetail_data[x].id = valID
									GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
									GUI_achievementdetail_data[x].color = 1
									x = x + 1
								end
							end
						end
						for k = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][j] do
							if PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement then
								for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement do

									local valID
									if PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].id then
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].id
									elseif PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].Aid then
										if SCAN[GUIselectedGUID].faction == "Horde" then
											valID = PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].Hid
										else
											valID = PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].Aid
										end
									end

									if SCAN[GUIselectedGUID].PVPDATA.achievementID and SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
										GUI_achievementdetail_data[x] = {}

										GUI_achievementdetail_data[x].id = valID
										GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
										GUI_achievementdetail_data[x].color = 1
										x = x + 1
									end
								end
							end
						end
					end
				else
					for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1].achievement do

						local valID
						if PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].id then
							valID = PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].id
						elseif PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].Aid then
							if SCAN[GUIselectedGUID].faction == "Horde" then
								valID = PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].Hid
							else
								valID = PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].Aid
							end
						end

						if SCAN[GUIselectedGUID].PVPDATA.achievementID and SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
							GUI_achievementdetail_data[x] = {}

							GUI_achievementdetail_data[x].id = valID
							GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
							GUI_achievementdetail_data[x].color = 1
							x = x + 1
						end
					end
				end
				GUI_achievementdetail_Count = x - 1
			elseif GUI_achievementdetail_Select == 3 then
				if not PVPDUMP[GUI_PvPachievementdetail_L1].achievement then
					for j = 1, #PVPDUMP[GUI_PvPachievementdetail_L1] do
						if PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement then
							for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement do

								local valID
								if PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].id then
									valID = PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].id
								elseif PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].Aid then
									if SCAN[GUIselectedGUID].faction == "Horde" then
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].Hid
									else
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][j].achievement[i].Aid
									end
								end

								if not SCAN[GUIselectedGUID].PVPDATA.achievementID or not SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
									GUI_achievementdetail_data[x] = {}

									GUI_achievementdetail_data[x].id = valID
									GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
									x = x + 1
								end
							end
						end
						for k = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][j] do
							if PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement then
								for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement do

									local valID
									if PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].id then
										valID = PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].id
									elseif PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].Aid then
										if SCAN[GUIselectedGUID].faction == "Horde" then
											valID = PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].Hid
										else
											valID = PVPDUMP[GUI_PvPachievementdetail_L1][j][k].achievement[i].Aid
										end
									end

									if not SCAN[GUIselectedGUID].PVPDATA.achievementID or not SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
										GUI_achievementdetail_data[x] = {}

										GUI_achievementdetail_data[x].id = valID
										GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
										x = x + 1
									end
								end
							end
						end
					end
				else
					for i = 1, #PVPDUMP[GUI_PvPachievementdetail_L1].achievement do

						local valID
						if PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].id then
							valID = PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].id
						elseif PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].Aid then
							if SCAN[GUIselectedGUID].faction == "Horde" then
								valID = PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].Hid
							else
								valID = PVPDUMP[GUI_PvPachievementdetail_L1].achievement[i].Aid
							end
						end

						if not SCAN[GUIselectedGUID].PVPDATA.achievementID or not SCAN[GUIselectedGUID].PVPDATA.achievementID[ valID ] then
							GUI_achievementdetail_data[x] = {}

							GUI_achievementdetail_data[x].id = valID
							GUI_achievementdetail_data[x].name = AchievementIDTable[ valID ]
							x = x + 1
						end
					end
				end
				GUI_achievementdetail_Count = x - 1
			end

		end
		-- PvP END -------------------------
	end

	local sortfunc
	if GUI_achievementdetail_SortBy == 1 then -- 1 = id
		if GUI_achievementdetail_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.id < b.id then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.id > b.id then
					return true
				end
			end
		end
	elseif GUI_achievementdetail_SortBy == 2 then -- 2 = name
		if GUI_achievementdetail_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.name < b.name then
					return true
				end
			end
		else
			sortfunc = function(a, b)
				if a.name > b.name then
					return true
				end
			end
		end
	end

	table.sort(GUI_achievementdetail_data, sortfunc)
end

function CharScanner:AchievementDetail_GUISort(sortBy)
	if GUI_achievementdetail_SortBy == sortBy then
		if GUI_achievementdetail_SortOrder == 1 then
			GUI_achievementdetail_SortOrder = 0
		else
			GUI_achievementdetail_SortOrder = 1
		end
	else
		GUI_achievementdetail_SortOrder = 1
	end
	GUI_achievementdetail_SortBy = sortBy
	CharScanner:AchievementDetail_SortButtonUpdate()
	CharScanner:AchievementDetail_GUICreate(AchievementDetailFrameMode)
	CharScanner:AchievementDetail_FillList()
end

function CharScanner:AchievementDetail_SortButtonUpdate()
	local arrow
	for i=1, 2 do
		if i == 1 then -- 1 = id
			arrow = "CharScanner_AchievementDetailFrame_SortButton_IdArrow"
		elseif i == 2 then -- 2 = name
			arrow = "CharScanner_AchievementDetailFrame_SortButton_NameArrow"
		end
		if GUI_achievementdetail_SortBy == i then
			if GUI_achievementdetail_SortOrder == 1 then _G[arrow]:SetTexCoord(unpack(Textures.SortArrow.coords1)) end
			if GUI_achievementdetail_SortOrder == 0 then _G[arrow]:SetTexCoord(unpack(Textures.SortArrow.coords2)) end
		else
			_G[arrow]:SetTexCoord(0, 0, 0, 0)
		end
	end
end

function CharScanner:AchievementDetail_OnVerticalScroll(value)
	CharScanner_AchievementDetailFrameScrollFrameScrollBar:SetValue(value)
	CharScanner_AchievementDetailFrameScrollFrame.offset = math_floor((value / sizeGeneral.BarHeight) + 0.5)
	CharScanner:AchievementDetail_FillList()
end

function CharScanner:AchievementDetail_FillList()
	local frameName = "CharScanner_AchievementDetailFrameScrollFrame"
	local frame = _G[frameName]
	local scrollBar = _G[frameName.."ScrollBar"]
	local scrollUpButton = _G[frameName.."ScrollBarScrollUpButton"]
	local scrollDownButton = _G[frameName.."ScrollBarScrollDownButton"]
	local scrollFrameHeight = 0
	local numToDisplay = GUI_achievementdetail_curRows
	local valueStep = 12

	if GUI_achievementdetail_Count < numToDisplay then
		numToDisplay = GUI_achievementdetail_Count
	end

	if GUI_achievementdetail_Count > 0 then
		scrollFrameHeight = (GUI_achievementdetail_Count - numToDisplay) * valueStep
		if scrollFrameHeight < 0 then
			scrollFrameHeight = 0
		end
	end
	scrollBar:SetMinMaxValues(0, scrollFrameHeight)
	scrollBar:SetValueStep(valueStep)

	if GUI_achievementdetail_Count <= GUI_achievementdetail_curRows then
		ScrollFrameDisable(frameName)
	else
		ScrollFrameEnable(frameName)
	end

	if scrollBar:GetValue() == 0 then
		scrollUpButton:Disable()
	else
		scrollUpButton:Enable()
	end
	if ((scrollBar:GetValue() - scrollFrameHeight) == 0) then
		scrollDownButton:Disable()
	else
		scrollDownButton:Enable()
	end

	for i=1, GUI_achievementdetail_maxRows do
		_G["CharScanner_AchievementDetailFrame_Button"..i]:Hide()
	end

	local j = 1
	local sliderPos = CharScanner_AchievementDetailFrameScrollFrame.offset
	for i=sliderPos+1, sliderPos+numToDisplay do
		if sliderPos < GUI_achievementdetail_Count then
			_G["CharScanner_AchievementDetailFrame_Button"..j.."Id"]:SetText(GUI_achievementdetail_data[i].id)
			_G["CharScanner_AchievementDetailFrame_Button"..j.."Name"]:SetText(GUI_achievementdetail_data[i].name)
			if GUI_achievementdetail_data[i].color then
				_G["CharScanner_AchievementDetailFrame_Button"..j.."Name"]:SetTextColor(color.Green.rgba[1], color.Green.rgba[2], color.Green.rgba[3], color.Green.rgba[4])
			else
				_G["CharScanner_AchievementDetailFrame_Button"..j.."Name"]:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
			end
			_G["CharScanner_AchievementDetailFrame_Button"..j].achievementID = GUI_achievementdetail_data[i].id

			_G["CharScanner_AchievementDetailFrame_Button"..j]:Show()
			j = j + 1
		end
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:PreShowSavedData(guid, switch)
	CharScanner:ClearData()
	CharScanner:ScanFrameToggle(true)
	if CompareFrameIsShown and switch then
		CharScanner:HistoryFrameToggle()
	end
	SCAN[guid].DataStep = 0
	CharScanner:ShowData(1, guid, false)
	CharScanner:ShowData(2, guid, false)
	CharScanner:ShowData(3, guid, false)
	CharScanner:ShowData(4, guid, false)
end

function CharScanner:DeleteCharacter(guid, id)
	if guid then -- delete one GUID
		NumScannedChars = NumScannedChars - 1
		if NumScannedChars == 0 then
			if GroupScanIsActive then
				CharScanner:StopGroupScan()
			end
			table.wipe(SCAN)
			GUIselectedGUID = nil
			CharScanner:DeleteButtonToggle(false)
			CharScanner:ScanFrameToggle(nil, true)
			CharScanner:AchievementDetail_HideAll()
			CharScanner:ClearData()
			CharScanner:History_Update()
		elseif NumScannedChars == 1 then
			SCAN[guid] = nil
			if guid == GUIselectedGUID then
				if id == 1 then
					CharScanner:PreShowSavedData(CharScanner_HistoryFrame_Button2.guid)
				else
					CharScanner:PreShowSavedData(CharScanner_HistoryFrame_Button1.guid)
				end
			end
			CharScanner:History_Update()
		elseif NumScannedChars > 1 then
			SCAN[guid] = nil
			if guid == GUIselectedGUID then
				if id == 1 then
					CharScanner:PreShowSavedData(CharScanner_HistoryFrame_Button2.guid)
				else
					CharScanner:PreShowSavedData(_G["CharScanner_HistoryFrame_Button"..(id-1)].guid)
				end
			end
			CharScanner:History_Update()
		end
	else -- delete all GUIDs
		if GroupScanIsActive then
			CharScanner:StopGroupScan()
		end
		table.wipe(SCAN)
		GUIselectedGUID = nil
		NumScannedChars = 0
		CharScanner:DeleteButtonToggle(false)
		CharScanner:ScanFrameToggle(nil, true)
		CharScanner:AchievementDetail_HideAll()
		CharScanner:ClearData()
		CharScanner:History_Update()
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:History_StatusUpdate()
	local text = string.format("%s %s", NumScannedChars, L["|4Char:Chars;"])
	-- average scan time START
	local avgScanTime = 0
	for guid, value in pairs(SCAN) do
		if value.scanTime then
			avgScanTime = avgScanTime + value.scanTime
		end
	end
	local avgScanTime = math_floor((avgScanTime/NumScannedChars) * mult4 + 0.5) / mult4
	-- average scan time END
	CharScanner_HistoryFrame_Status_Text:SetText(text.." - "..L["Avg"].." "..avgScanTime)

	local roleTable = {}

	for guid, value in pairs(SCAN) do
		if value.time then -- XFIX
			if SCAN[guid].talentGroup1role then
				if SCAN[guid].activeTalentGroup == 1 then
					--checker START
					if not roleTable[ SCAN[guid].talentGroup1role ] then
						roleTable[ SCAN[guid].talentGroup1role ] = {}
						roleTable[ SCAN[guid].talentGroup1role ].active = 1
						roleTable[ SCAN[guid].talentGroup1role ].count = 1
					else
						roleTable[ SCAN[guid].talentGroup1role ].active = roleTable[ SCAN[guid].talentGroup1role ].active + 1
						roleTable[ SCAN[guid].talentGroup1role ].count = roleTable[ SCAN[guid].talentGroup1role ].count + 1
					end
					--checker END
				else
					--checker START
					if not roleTable[ SCAN[guid].talentGroup1role ] then
						roleTable[ SCAN[guid].talentGroup1role ] = {}
						roleTable[ SCAN[guid].talentGroup1role ].active = 0
						roleTable[ SCAN[guid].talentGroup1role ].count = 1
					else
						--roleTable[ SCAN[guid].talentGroup1role ].active = roleTable[ SCAN[guid].talentGroup1role ].active
						roleTable[ SCAN[guid].talentGroup1role ].count = roleTable[ SCAN[guid].talentGroup1role ].count + 1
					end
					--checker END
				end

				if SCAN[guid].talentGroup2role then
					if SCAN[guid].activeTalentGroup == 2 then
						--checker START
						if not roleTable[ SCAN[guid].talentGroup2role ] then
							roleTable[ SCAN[guid].talentGroup2role ] = {}
							roleTable[ SCAN[guid].talentGroup2role ].active = 1
							roleTable[ SCAN[guid].talentGroup2role ].count = 1
						else
							roleTable[ SCAN[guid].talentGroup2role ].active = roleTable[ SCAN[guid].talentGroup2role ].active + 1
							if SCAN[guid].talentGroup1role ~= SCAN[guid].talentGroup2role then
								roleTable[ SCAN[guid].talentGroup2role ].count = roleTable[ SCAN[guid].talentGroup2role ].count + 1
							end
						end
						--checker END
					else
						--checker START
						if not roleTable[ SCAN[guid].talentGroup2role ] then
							roleTable[ SCAN[guid].talentGroup2role ] = {}
							roleTable[ SCAN[guid].talentGroup2role ].active = 0
							roleTable[ SCAN[guid].talentGroup2role ].count = 1
						else
							--roleTable[ SCAN[guid].talentGroup2role ].active = roleTable[ SCAN[guid].talentGroup2role ].active
							if SCAN[guid].talentGroup1role ~= SCAN[guid].talentGroup2role then
								roleTable[ SCAN[guid].talentGroup2role ].count = roleTable[ SCAN[guid].talentGroup2role ].count + 1
							end
						end
						--checker END
					end
				end
			else
				--checker START
				if not roleTable[ SCAN[guid].mainRole ] then
					roleTable[ SCAN[guid].mainRole ] = {}
					roleTable[ SCAN[guid].mainRole ].active = 1
					roleTable[ SCAN[guid].mainRole ].count = 1
				else
					roleTable[ SCAN[guid].mainRole ].active = roleTable[ SCAN[guid].mainRole ].active + 1
					roleTable[ SCAN[guid].mainRole ].count = roleTable[ SCAN[guid].mainRole ].count + 1
				end
				--checker END
			end
		end -- XFIX
	end

	for i = 1, 4 do
		if roleTable[i] then
			_G["CharScanner_HistoryFrame_Status_Role"..i.."Texture"]:SetAlpha(1)
			_G["CharScanner_HistoryFrame_Status_Role"..i.."Text"]:SetText(roleTable[i].active.." |cff"..color.Grey.hex.."("..roleTable[i].count..")|r")
			_G["CharScanner_HistoryFrame_Status_Role"..i.."Text"]:SetTextColor(color.Role[ i ].rgba[1], color.Role[ i ].rgba[2], color.Role[ i ].rgba[3], color.Role[ i ].rgba[4])
		else
			_G["CharScanner_HistoryFrame_Status_Role"..i.."Texture"]:SetAlpha(0.25)
			_G["CharScanner_HistoryFrame_Status_Role"..i.."Text"]:SetText("- (-)")
			_G["CharScanner_HistoryFrame_Status_Role"..i.."Text"]:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
		end
	end
	table.wipe(roleTable)
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ScanGroup(delData)
	if inCombat then
		CharScanner_MainFrame_Status_Text:SetText(L["Scan is deactivated in combat!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
		CharScanner_MainFrame_Status:Show()
		return
	end

	-- ScanGroup()-is-only-available-if-we-are-in-group,-but-a-paranoia-check-is-good! START
	local partymember = GetNumPartyMembers()
	local raidmember  = GetNumRaidMembers()
	local inParty
	local inRaid
	if partymember > 0 then
		inRaid  = false
		inParty = true
	end
	if raidmember > 1 then -- 1, not 0. Think the only situation where a 1-member-raid can exist is at the end of a battleground.
		inRaid  = true
		inParty = false
	end
	if not inRaid and not inParty then return end
	-- ScanGroup()-is-only-available-if-we-are-in-group,-but-a-paranoia-check-is-good! END

	if delData then
		CharScanner:DeleteCharacter()
	end

	GroupScanTimerStart = GetTime()
	GroupScanIsActive = true
	GroupScanNext = false
	CharScanner:SingleScanHide()
	CharScanner:GroupScanHide()
	CharScanner:ProgressDisplay("GROUP", true, false)

	-- temp membertable START
	-- we build our own membertable. so we are independent from PARTY_MEMBERS_CHANGED/RAID_ROSTER_UPDATE changes
	-- raid roster changes between button click and end of groupscan are discarded!
	local t = {}
	local x = 1
	if inParty then
		for i = 1, partymember do
			local unit = "party"..i
			local guid = UnitGUID(unit)
			if guid then
				t[x] = {}
				t[x].guid = guid
				t[x].unit = unit
				x = x + 1
			end
		end
		-- add player to temp db
		t[x] = {}
		t[x].guid = playerGUID
		t[x].unit = "player"
	end
	if inRaid then
		for i = 1, raidmember do
			local unit = "raid"..i
			local guid = UnitGUID(unit)
			if guid then
				t[x] = {}
				t[x].guid = guid
				t[x].unit = unit
				x = x + 1
			end
		end
	end
	local numMember = #t
	-- temp membertable END

	local gsStart = GetTime()
	local currentScanNumber = 1
	GroupScanFirstGUID = nil

	local function scanNextMember()
		local guid = UnitGUID(t[currentScanNumber].unit)
		if guid and guid == t[currentScanNumber].guid then
			CharScanner_MainFrame_Status_ProgressText:SetText(currentScanNumber.."/"..numMember)
			CharScanner_MainFrame_Status_ProgressDone:SetWidth(currentScanNumber*sizeGeneral.StatusProgressBarWidth/numMember)
			CharScanner:ScanUnit(t[currentScanNumber].unit)
			currentScanNumber = currentScanNumber + 1
		else
			CharScanner_MainFrame_Status_ProgressText:SetText(currentScanNumber.."/"..numMember)
			CharScanner_MainFrame_Status_ProgressDone:SetWidth(currentScanNumber*sizeGeneral.StatusProgressBarWidth/numMember)
			currentScanNumber = currentScanNumber + 1
		end
	end

	local function gsTimeCheck()
		if inCombat then
			CharScanner:StopGroupScan()
			return
		end
		if not inPartyOrRaid then
			CharScanner:StopGroupScan()
			return
		end
		if RestrictionTalNum >= RestrictionCountMax or RestrictionAchNum >= RestrictionCountMax then
			if currentScanNumber > numMember then
				CharScanner:StopGroupScan()
				return
			end
			local elapsedTime = GetTime() - RestrictionTime
			local throttlewidth = (sizeGeneral.StatusProgressBarWidth/RestrictionMaxTime)*elapsedTime
			if throttlewidth > sizeGeneral.StatusProgressBarWidth then
				throttlewidth = sizeGeneral.StatusProgressBarWidth
			end
			CharScanner:ProgressDisplay("GROUP", true, true)
			CharScanner_MainFrame_Status_ThrottleText2:SetText( format(L["Please wait %.2f seconds."], (RestrictionMaxTime-elapsedTime)) )
			CharScanner_MainFrame_Status_ThrottleDone:SetWidth(throttlewidth)
			if elapsedTime >= RestrictionMaxTime then
				CharScanner:ProgressDisplay("GROUP", true, false)
				RestrictionAchNum = 0
				RestrictionTalNum = 0
			end
		end
		if RestrictionTalNum < RestrictionCountMax and RestrictionAchNum < RestrictionCountMax and
		   ( (GetTime() - gsStart) >= CharScanner_Options.scanDuration or GroupScanNext ) then
			CharScanner:ProgressDisplay("GROUP", true, false)
			if currentScanNumber > numMember then
				CharScanner:StopGroupScan()
				return
			end
			-- next scan
			gsStart = GetTime()
			GroupScanNext = false
			scanNextMember()
		end
	end

	CharScanner.OnUpdateGroupScan:SetScript("OnUpdate", gsTimeCheck)
	if RestrictionTalNum < RestrictionCountMax and RestrictionAchNum < RestrictionCountMax then
		scanNextMember()
	end
end

function CharScanner:StopGroupScan()
	CharScanner.OnUpdateGroupScan:SetScript("OnUpdate", nil)
	CharScanner:Unregister_INSPECT_READY()
	CharScanner:Unregister_INSPECT_ACHIEVEMENT_READY()
	GroupScanNext = false
	GroupScanIsActive = false
	CleanTempData()

	if not GUIselectedGUID then
		CharScanner:ClearData()
		CharScanner:History_Update()
		if NumScannedChars == 0 then
			table.wipe(SCAN)
			CharScanner:DeleteButtonToggle(false)
			CharScanner:ScanFrameToggle(nil, true)
		end
	end

	if CharScanner_Options.chatoutput then
		local time = 0
		local TimeElapsed = math_floor((GetTime()-GroupScanTimerStart) * mult4 + 0.5) / mult4
		if type(TimeElapsed) == "number" and TimeElapsed > 0 then
			time = TimeElapsed
		end
		if time == 1 then
			time = string.format("%s %s", time, SECONDS)
		else
			time = time.." "..SECONDS
		end
		Print(L["GroupScan finished!"].." |cff"..color.Green.hex.."("..time..")|r")
	end

	CharScanner:SingleScanShow()
	CharScanner:GroupScanShow()
	CharScanner:ProgressDisplay("GROUP", false, false)
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ScanUnit(unit)
	if inCombat then
		CharScanner_MainFrame_Status_Text:SetText(L["Scan is deactivated in combat!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
		CharScanner_MainFrame_Status:Show()
		GroupScanNext = true
		return
	end
	if AchievementFrame and AchievementFrame:IsVisible() then
		CharScanner_MainFrame_Status_Text:SetText(L["Please close AchievementFrame!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
		CharScanner_MainFrame_Status:Show()
		GroupScanNext = true
		return
	end
	if not UnitExists(unit) or not UnitIsPlayer(unit) then
		CharScanner_MainFrame_Status_Text:SetText(L["Target is NOT a player!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
		CharScanner_MainFrame_Status:Show()
		GroupScanNext = true
		return
	end
	if not CheckInteractDistance(unit, 1) then
		CharScanner_MainFrame_Status_Text:SetText(L["Target is NOT in inspect range!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
		CharScanner_MainFrame_Status:Show()
		GroupScanNext = true
		return
	end
	if not CanInspect(unit) then
		CharScanner_MainFrame_Status_Text:SetText(L["Can NOT inspect target!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
		CharScanner_MainFrame_Status:Show()
		GroupScanNext = true
		return
	end

	if not GroupScanIsActive then
		CharScanner:ProgressDisplay("SINGLE", true, false)
		if RestrictionTalNum >= RestrictionCountMax or RestrictionAchNum >= RestrictionCountMax then
			if (GetTime() - RestrictionTime) < RestrictionMaxTime then
				CharScanner_MainFrame_Status_Text:SetText( L["Achievement/Talent scan restriction from Blizzard:"].." "..format(L["Please wait %.2f seconds."], RestrictionMaxTime) )
				CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
				CharScanner_MainFrame_Status:Show()

				local function restrictionTimeCheck()
					if inCombat then
						CharScanner.OnUpdateRestriction:SetScript("OnUpdate", nil)
						CharScanner:StopSingleScan(true)
						return
					end
					local elapsedWaitTime = GetTime() - RestrictionTime
					if elapsedWaitTime < RestrictionMaxTime then
						CharScanner:ProgressDisplay("SINGLE", false, false)
						CharScanner_MainFrame_Status_Text:SetText( L["Achievement/Talent scan restriction from Blizzard:"].." "..format(L["Please wait %.2f seconds."], (RestrictionMaxTime-elapsedWaitTime)) )
					else
						CharScanner_MainFrame_Status_Text:SetText("")
						CharScanner.OnUpdateRestriction:SetScript("OnUpdate", nil)
					end
				end
				CharScanner.OnUpdateRestriction:SetScript("OnUpdate", restrictionTimeCheck)
				return
			end
			RestrictionTalNum = 0
			RestrictionAchNum = 0
		end
	end

	-- DUMMY_FactionFix START
	local faction
	if UnitFactionGroup(unit) == "Horde" then
		PVEDUMP[3][1].achievement[54] = 4297
		faction = "Horde"
	else
		PVEDUMP[3][1].achievement[54] = 4298
		faction = "Alliance"
	end
	-- DUMMY_FactionFix END

	SingleScanTimerStart = GetTime()
	LatestScanIsComplete = false

	if not GroupScanIsActive then
		CharScanner:ProgressDisplay("SINGLE", true, false)
		CharScanner:ClearData()
	end
	CharScanner:ScanFrameToggle(true)

	if UnitIsUnit("player", unit) then
		IsUnitInpected = false -- "Character" -> it's me
		unit = "player"
	else
		IsUnitInpected = true -- "Inspect"
	end

	SCANGUID = UnitGUID(unit) --..GetTime() -- TEST
	if not SCANGUID then Print("GUID ERROR!", SCANGUID) return end -- TEST
	SCANGUID = "_"..SCANGUID -- previewGUID -> if a succesfuly scanned unit is scanned again with an uncomplete scan the original (complete) scan is untouched
	CreateScanDataByGUID(SCANGUID)
	SCAN[SCANGUID].DataStep = 0

	if not GroupScanIsActive then
		SingleScanIsActive = true
		CharScanner:SingleScanHide()
		CharScanner:GroupScanHide()

		local function ssTimeCheck()
			if inCombat then
				CharScanner:StopSingleScan(true)
				return
			end
			if (GetTime() - SingleScanTimerStart) >= CharScanner_Options.scanDuration then
				CharScanner:StopSingleScan(true)
				return
			end
		end

		CharScanner.OnUpdateSingleScan:SetScript("OnUpdate", ssTimeCheck)
	end

	local name, realm = UnitName(unit)
	if realm then
		name = name.."-"..realm
	end
	SCAN[SCANGUID].name = name
	SCAN[SCANGUID].level = UnitLevel(unit)
	SCAN[SCANGUID].race = UnitRace(unit)
	SCAN[SCANGUID].faction = faction
	SCAN[SCANGUID].class, SCAN[SCANGUID].classEN = UnitClass(unit)
	SCAN[SCANGUID].guildName, SCAN[SCANGUID].guildRankName, SCAN[SCANGUID].guildRankIndex = GetGuildInfo(unit)

	CharScanner:ShowData(1, SCANGUID, true)

	-- achievementFunctions workaround START
	-- hack to avoid error messages because of achievementFunctions.selectedCategory
	if achievementFunctions and achievementFunctions.selectedCategory then
		bugfix = achievementFunctions.selectedCategory
		achievementFunctions.selectedCategory = "summary"
	end
	-- achievementFunctions workaround END
	
	ClearAchievementComparisonUnit()
	SetAchievementComparisonUnit(unit)
	Event_INSPECT_ACHIEVEMENT_READY = true

	NotifyInspect(unit)
	InspectFrame.unit = unit
	SCANUNIT = unit

	CharScanner:GearScanPost(unit)
	Event_INSPECT_READY = true
end

function CharScanner:StopSingleScan(statusupdate)
	CharScanner.OnUpdateSingleScan:SetScript("OnUpdate", nil)
	CharScanner:Unregister_INSPECT_READY()
	CharScanner:Unregister_INSPECT_ACHIEVEMENT_READY()
	SingleScanIsActive = false
	CleanTempData()
	CharScanner:SingleScanShow()
	CharScanner:GroupScanShow()
	CharScanner:ProgressDisplay("SINGLE", false, false)
	if statusupdate then
		CharScanner:AchievementDetail_HideAll()
		CharScanner_MainFrame_Status_Text:SetText(L["Scan incomplete!"])
		CharScanner_MainFrame_Status_Text:SetTextColor(color.Red.rgba[1], color.Red.rgba[2], color.Red.rgba[3], color.Red.rgba[4])
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ShowData(step, guid, scan)
	SCAN[guid].DataStep = SCAN[guid].DataStep + 1

	if (GroupScanIsActive and not GroupScanFirstGUID) or SingleScanIsActive or not scan then

		if step == 1 then

			for i = 1, GUI_history_maxRows do
				_G["CharScanner_HistoryFrame_Button"..i.."Arrow"]:Hide()
			end
			EnableIconButton("CharScanner_MainFrame_Headline_ButtonCollapse", 1)
			CharScanner:DeleteButtonToggle(true)

			-- ClassIcon
			CharScanner_MainFrame_Base_ClassIcon:SetTexCoord(classes[SCAN[guid].classEN].icon[1], classes[SCAN[guid].classEN].icon[2], classes[SCAN[guid].classEN].icon[3], classes[SCAN[guid].classEN].icon[4])
			-- FactionIcon
			if SCAN[guid].faction == "Horde" then
				CharScanner_MainFrame_Base_FactionIcon:SetTexture("Interface\\TargetingFrame\\UI-PVP-Horde")
			else
				CharScanner_MainFrame_Base_FactionIcon:SetTexture("Interface\\TargetingFrame\\UI-PVP-Alliance")
			end
			CharScanner_MainFrame_Base_FactionIcon:SetTexCoord(0.046875, 0.609375, 0.03125, 0.59375) -- (3/64, 39/64, 2/64, 38/64)
			-- Name
			CharScanner_MainFrame_Base_Name:SetText(SCAN[guid].name)
			CharScanner_MainFrame_Base_Name:SetTextColor(classcolors[SCAN[guid].classEN].r, classcolors[SCAN[guid].classEN].g, classcolors[SCAN[guid].classEN].b, 1)
			-- GUID
			if scan then
				CharScanner_MainFrame_Base_GUID:SetText("(GUID: "..string_sub(guid, 2)..")") -- save preview guid (with leading '_')
			else
				CharScanner_MainFrame_Base_GUID:SetText("(GUID: "..guid..")") -- real guid (without leading '_')
			end
			-- Guild
			local guild = ""
			if SCAN[guid].guildName then
				guild = GUILD..": "..SCAN[guid].guildName
				if SCAN[guid].guildRankIndex and SCAN[guid].guildRankIndex < 5 then
					guild = guild.." | "..RANK..": "..guildRank[SCAN[guid].guildRankIndex+1]
				else
					guild = guild.." | "..RANK..": -"
				end
			else
				guild = GUILD..": - "
			end
			CharScanner_MainFrame_Base_Guild:SetText(guild)
			CharScanner_MainFrame_Base_Guild:SetTextColor(color.Level2.rgba[1], color.Level2.rgba[2], color.Level2.rgba[3], color.Level2.rgba[4])
			-- Level/Race/Class
			CharScanner_MainFrame_Base_Level:SetText(LEVEL.." "..SCAN[guid].level.." - "..SCAN[guid].race..", "..SCAN[guid].class)
			CharScanner_MainFrame_Base_Level:SetTextColor(classcolors[SCAN[guid].classEN].r, classcolors[SCAN[guid].classEN].g, classcolors[SCAN[guid].classEN].b, 1)

		elseif step == 2 then

			-- Equipment
			CharScanner_MainFrame_Equipment_Title:SetText(L["Equipment"]..":")

			local miss = ""
			if SCAN[guid].itemlevelMiss > 0 then
				miss = " - |cff"..color.Red.hex..string.format(L["%s %s not used"], SCAN[guid].itemlevelMiss, L["|4Slot:Slots;"]).."|r"
			end
			local enchcount = L["Enchantments"]..": |cff"..color.White.hex..SCAN[guid].gearEnchCount.."|r"
			local pvpcount = " - "..L["PvP Items"]..": |cff"..color.White.hex..SCAN[guid].gearPvPCount.."|r"
			local txt = L["Avg"].." "..cnum(SCAN[guid].itemlevelAvg).." ("..L["min."].." "..SCAN[guid].itemlevelMin.." - "..L["max."].." "..SCAN[guid].itemlevelMax..")"..miss

			CharScanner_MainFrame_Equipment_Text1:SetText(txt)
			CharScanner_MainFrame_Equipment_Text2:SetText(SCAN[guid].itemlevelQuality)
			CharScanner_MainFrame_Equipment_Text3:SetText(enchcount..pvpcount)

			local w1 = ceil(CharScanner_MainFrame_Equipment_Title:GetStringWidth())
			local w2 = ceil(CharScanner_MainFrame_Equipment_Text2:GetStringWidth())
			local maxwidth = sizeGeneral.MainWidth - (sizeGeneral.Offset + w1 + sizeGeneral.Offset + sizeGeneral.Offset + 25) -- ~25px for buttonwidth

			if w2 >= maxwidth then
				CharScanner_MainFrame_Equipment_Text2:SetWidth(maxwidth)
				CharScanner_MainFrame_Equipment:SetHeight(sizeEquipmentFrame.BaseHeightXXL)
			else
				CharScanner_MainFrame_Equipment_Text2:SetWidth(maxwidth)
				CharScanner_MainFrame_Equipment:SetHeight(sizeEquipmentFrame.BaseHeight)
			end

			-- Gear
			for i = 1, #SCAN[guid].gear do
				for j = 1, #slotButton do

					if i == slotButton[j].id then
						local linkID = select(3, SCAN[guid].gear[i].iLink:find("item:(%d+)"))
						local itemTexture = GetItemIcon(linkID)
						if itemTexture then
							local iLink = SCAN[guid].gear[i].iLink
							_G["CharScanner_MainFrame_Gear_Slot"..j.."IconTexture"]:SetTexture(itemTexture)
							if SCAN[guid].gear[i].iLvl == 0 then
								_G["CharScanner_MainFrame_Gear_Slot"..j.."Count"]:SetText("|cff"..itemQuality[SCAN[guid].gear[i].iRar+1].color.."?|r")
							else
								_G["CharScanner_MainFrame_Gear_Slot"..j.."Count"]:SetText("|cff"..itemQuality[SCAN[guid].gear[i].iRar+1].color..SCAN[guid].gear[i].iLvl.."|r")
							end
							_G["CharScanner_MainFrame_Gear_Slot"..j.."Count"]:Show()
							if SCAN[guid].gear[i].iEnch > 0 then
								_G["CharScanner_MainFrame_Gear_Slot"..j.."EnchTexture"]:Show()
							--else
							--	_G["CharScanner_MainFrame_Gear_Slot"..j.."EnchTexture"]:Hide()
							end
							if SCAN[guid].gear[i].iPvP == 1 then
								_G["CharScanner_MainFrame_Gear_Slot"..j.."PvPTexture"]:Show()
							--else
							--	_G["CharScanner_MainFrame_Gear_Slot"..j.."PvPTexture"]:Hide()
							end
							_G["CharScanner_MainFrame_Gear_Slot"..j]:SetScript("OnEnter",
								function(self)
									GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
									GameTooltip:SetHyperlink(iLink)
									_G["CharScanner_MainFrame_Gear_Slot"..j.."Border"]:SetTexture(1, 1, 0.49, 1)
									self.UpdateTooltip = GearTooltipHandler
								end)
							_G["CharScanner_MainFrame_Gear_Slot"..j]:SetScript("OnLeave",
								function(self)
									_G["CharScanner_MainFrame_Gear_Slot"..j.."Border"]:SetTexture(0.2, 0.2, 0.2, 1)
									GameTooltip:Hide()
									ResetCursor()
									self.comparing = false
								end)
							_G["CharScanner_MainFrame_Gear_Slot"..j]:SetScript("OnClick",
								function()
									HandleModifiedItemClick(iLink)
								end)
						end
						break
					end

				end
			end

			if SCAN[guid].gearTwoHandWeapon then
				CharScanner_MainFrame_Gear_Slot10IconTexture:SetTexture("Interface\\PaperDollInfoFrame\\UI-GearManager-LeaveItem-Opaque")
				CharScanner_MainFrame_Gear_Slot10Border:SetTexture(0, 0, 0, 0)
				CharScanner_MainFrame_Gear_Slot10IconTexture:SetAlpha(0.4)
			else
				CharScanner_MainFrame_Gear_Slot10Border:SetTexture(0.2, 0.2, 0.2, 1)
				CharScanner_MainFrame_Gear_Slot10IconTexture:SetAlpha(1)
			end

			-- GearStats START
			--local statsText = ""
			--for key, value in pairs(SCAN[guid].gearStats) do
			--	statsText = statsText.."|cff"..color.LightYellow.hex.._G[key]..":|r "..(math_floor(value * mult2 + 0.5) / mult2).."\n"
				--TEST code for future GearStats updates START
				--if not CharScanner_Options.statsTest then CharScanner_Options.statsTest = {} end
				--if not CharScanner_Options.statsTest[key] then
				--	CharScanner_Options.statsTest[key] = _G[key]
				--end
				--TEST code for future GearStats updates END
			--end
			--statsText = string_sub(statsText, 0, -2) -- last '\n'
			--CharScanner_MainFrame_Equipment_StatsText:SetText(statsText)
			--local h = ceil(CharScanner_MainFrame_Equipment_StatsText:GetStringHeight())
			--local w = ceil(CharScanner_MainFrame_Equipment_StatsText:GetStringWidth())
			--CharScanner_MainFrame_Equipment_StatsFrame:SetHeight(h+20)
			--CharScanner_MainFrame_Equipment_StatsFrame:SetWidth(w+20)
			-- GearStats END

		elseif step == 3 then

			-- Talent
			local col
			CharScanner_MainFrame_Talent_Title:SetText(TALENTS..":")
			if SCAN[guid].talentGroups then

				local text = RoleTextShort[SCAN[guid].talentGroup1role].." "..SCAN[guid].talentGroup1name.." "..SCAN[guid].talentGroup1pointsString
				local unspent = ""
				if SCAN[guid].talentGroup1pointsUnspent > 0 then
					if SCAN[guid].activeTalentGroup == 1 then
						col = "|cff"..color.Red.hex
					else
						col = "|cff"..color.Grey.hex
					end
					unspent = col.."("..string.format(UNSPENT_TALENT_POINTS, SCAN[guid].talentGroup1pointsUnspent)..")|r"
				end

				local texture = SCAN[guid].talentGroup1icon
				if SCAN[guid].talentGroup1pointsString == "0/0/0" then
					texture = nil
					text = RoleTextShort[SCAN[guid].talentGroup1role].." "..unspent
				else
					if unspent ~= "" then
						text = text.." "..unspent
					end
				end

				CharScanner_MainFrame_Talent_TalentGroup1Icon:SetTexture(texture)
				CharScanner_MainFrame_Talent_TalentGroup1Text:SetText(text)
				CharScanner_MainFrame_Talent_TalentGroup1Text:SetPoint("LEFT", CharScanner_MainFrame_Talent_TalentGroup1Icon, "RIGHT", sizeGeneral.Offset, 0)

				if SCAN[guid].activeTalentGroup == 1 then
					CharScanner_MainFrame_Talent_TalentMainRoleActive:SetTexture(Textures.CharScannerIcons.path)
					CharScanner_MainFrame_Talent_TalentMainRoleActive:SetPoint("RIGHT", CharScanner_MainFrame_Talent_TalentGroup1Icon, "LEFT", -sizeGeneral.Offset, 0)
					CharScanner_MainFrame_Talent_TalentMainRoleActive:SetVertexColor(color.Role[ SCAN[guid].talentGroup1role ].rgba[1], color.Role[ SCAN[guid].talentGroup1role ].rgba[2], color.Role[ SCAN[guid].talentGroup1role ].rgba[3], color.Role[ SCAN[guid].talentGroup1role ].rgba[4])
					CharScanner_MainFrame_Talent_TalentGroup1Text:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
				else
					CharScanner_MainFrame_Talent_TalentGroup1Text:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
				end

				if SCAN[guid].talentGroups > 1 then

					local text = RoleTextShort[SCAN[guid].talentGroup2role].." "..SCAN[guid].talentGroup2name.." "..SCAN[guid].talentGroup2pointsString
					local unspent = ""
					if SCAN[guid].talentGroup2pointsUnspent > 0 then
						if SCAN[guid].activeTalentGroup == 2 then
							col = "|cff"..color.Red.hex
						else
							col = "|cff"..color.Grey.hex
						end
						unspent = col.."("..string.format(UNSPENT_TALENT_POINTS, SCAN[guid].talentGroup2pointsUnspent)..")|r"
					end

					local texture = SCAN[guid].talentGroup2icon
					if SCAN[guid].talentGroup2pointsString == "0/0/0" then
						texture = nil
						text = RoleTextShort[SCAN[guid].talentGroup2role].." "..unspent
					else
						if unspent ~= "" then
							text = text.." "..unspent
						end
					end

					CharScanner_MainFrame_Talent_TalentGroup2Icon:SetTexture(texture)
					CharScanner_MainFrame_Talent_TalentGroup2Text:SetText(text)

					if SCAN[guid].activeTalentGroup == 2 then
						CharScanner_MainFrame_Talent_TalentMainRoleActive:SetTexture(Textures.CharScannerIcons.path)
						CharScanner_MainFrame_Talent_TalentMainRoleActive:SetPoint("RIGHT", CharScanner_MainFrame_Talent_TalentGroup2Icon, "LEFT", -sizeGeneral.Offset, 0)
						CharScanner_MainFrame_Talent_TalentMainRoleActive:SetVertexColor(color.Role[ SCAN[guid].talentGroup2role ].rgba[1], color.Role[ SCAN[guid].talentGroup2role ].rgba[2], color.Role[ SCAN[guid].talentGroup2role ].rgba[3], color.Role[ SCAN[guid].talentGroup2role ].rgba[4])
						CharScanner_MainFrame_Talent_TalentGroup2Text:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
					else
						CharScanner_MainFrame_Talent_TalentGroup2Text:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
					end

				else
					CharScanner_MainFrame_Talent_TalentGroup2Icon:SetTexture(nil)
					CharScanner_MainFrame_Talent_TalentGroup2Text:SetText("")
				end

				CharScanner_MainFrame_Talent_TalentMainRole:SetTexture(Textures.RoleIcon[SCAN[guid].mainRole].path)
				CharScanner_MainFrame_Talent_TalentMainRole:SetTexCoord(Textures.RoleIcon[SCAN[guid].mainRole].coords[1], Textures.RoleIcon[SCAN[guid].mainRole].coords[2], Textures.RoleIcon[SCAN[guid].mainRole].coords[3], Textures.RoleIcon[SCAN[guid].mainRole].coords[4])
				CharScanner:TalentGroup_Switch(guid, SCAN[guid].activeTalentGroup)
				
			else
				CharScanner_MainFrame_Talent_TalentMainRoleActive:SetTexture(nil)
				CharScanner_MainFrame_Talent_TalentGroup1Icon:SetTexture(nil)
				CharScanner_MainFrame_Talent_TalentGroup1Text:SetText(L["No Talents"])
				CharScanner_MainFrame_Talent_TalentGroup1Text:SetPoint("LEFT", CharScanner_MainFrame_Talent_Title, "RIGHT", sizeGeneral.Offset, 0)
				CharScanner_MainFrame_Talent_TalentGroup1Text:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
				CharScanner_MainFrame_Talent_TalentGroup2Icon:SetTexture(nil)
				CharScanner_MainFrame_Talent_TalentGroup2Text:SetText("")

				CharScanner_MainFrame_Talent_TalentMainRole:SetTexture(nil)
				CharScanner:TalentGroup_Switch(guid, 0)
			end

		elseif step == 4 then

			CharScanner_MainFrame_PvEStats_TitleName:SetText("")

			-- PvE START
			if BossSearchQuery then
				CharScanner:ABKsearch_Update(guid)
			else

				CharScanner_MainFrame_PvEStats_TitleBossesTotal:SetText(L["Kills"])
				CharScanner_MainFrame_PvEStats_TitleAchievement:SetText(L["Achievements"])
				CharScanner_MainFrame_PvEStats_TitleStatistic:SetText(L["Bosses"])

				local col
				-- level 1 START
				for i = 1, #PVEDUMP do
					_G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name.."Name"]:SetText(L[PVEDUMP[i].name])
					if PVEDUMP[i].achievementTotal == SCAN[guid].PVEDATA[i].achievementCount then
						col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
					end
					_G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name.."Achievement"]:SetText(col..SCAN[guid].PVEDATA[i].achievementCount.."|r / "..PVEDUMP[i].achievementTotal)

					if PVEDUMP[i].statisticTotal == SCAN[guid].PVEDATA[i].statisticBoss then
						col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
					end
					if PVEDUMP[i].statisticTotal == 0 then --NOBOSS
						_G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name.."Statistic"]:SetText("- / -")
						_G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name.."BossesTotal"]:SetText("-")
					else
						_G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name.."Statistic"]:SetText(col..SCAN[guid].PVEDATA[i].statisticBoss.."|r / "..PVEDUMP[i].statisticTotal)
						_G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name.."BossesTotal"]:SetText(SCAN[guid].PVEDATA[i].statisticBossAll)
					end
				--end
				-- level 1 END

				-- level 2 START
				--for i = 1, #PVEDUMP do
					if PVEDUMP[i].statistic then
						for j = 1, #PVEDUMP[i].statistic do
							if SCAN[guid].PVEDATA[i].statisticBossEach[j] > 0 then
								col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
							end
							_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..j.."Name"]:SetText(col..StatisticIDTable[ PVEDUMP[i].statistic[j] ].."|r")
							_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..j.."BossesTotal"]:SetText(SCAN[guid].PVEDATA[i].statisticBossEach[j])
						end
					else
						for j = 1, #PVEDUMP[i] do
							_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name.."Name"]:SetText(L[PVEDUMP[i][j].name])
							if PVEDUMP[i][j].achievementTotal == SCAN[guid].PVEDATA[i][j].achievementCount then
								col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
							end
							_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name.."Achievement"]:SetText(col..SCAN[guid].PVEDATA[i][j].achievementCount.."|r / "..PVEDUMP[i][j].achievementTotal)
							if PVEDUMP[i][j].statisticTotal == SCAN[guid].PVEDATA[i][j].statisticBoss then
								col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
							end
							if PVEDUMP[i][j].statisticTotal == 0 then --NOBOSS
								_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name.."Statistic"]:SetText("- / -")
								_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name.."BossesTotal"]:SetText("-")
							else
								_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name.."Statistic"]:SetText(col..SCAN[guid].PVEDATA[i][j].statisticBoss.."|r / "..PVEDUMP[i][j].statisticTotal)
								_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name.."BossesTotal"]:SetText(SCAN[guid].PVEDATA[i][j].statisticBossAll)
							end
						--end
					--end
				--end
				-- level 2 END

				-- level 3 START
				--for i = 1, #PVEDUMP do
					--if not PVEDUMP[i].statistic then
						--for j = 1, #PVEDUMP[i] do

							if PVEDUMP[i][j].statistic then
								for k = 1, #PVEDUMP[i][j].statistic do
									if SCAN[guid].PVEDATA[i][j].statisticBossEach[k] > 0 then
										col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
									end
									_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."Name"]:SetText(col..StatisticIDTable[ PVEDUMP[i][j].statistic[k] ].."|r")
									_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."BossesTotal"]:SetText(SCAN[guid].PVEDATA[i][j].statisticBossEach[k])
								end
							else
								for k = 1, #PVEDUMP[i][j] do
									_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."Name"]:SetText(L[PVEDUMP[i][j][k].name])
									if PVEDUMP[i][j][k].achievementTotal == SCAN[guid].PVEDATA[i][j][k].achievementCount then
										col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
									end
									_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."Achievement"]:SetText(col..SCAN[guid].PVEDATA[i][j][k].achievementCount.."|r / "..PVEDUMP[i][j][k].achievementTotal)
									if PVEDUMP[i][j][k].statisticTotal == SCAN[guid].PVEDATA[i][j][k].statisticBoss then
										col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
									end
									if PVEDUMP[i][j][k].statisticTotal == 0 then --NOBOSS
										_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."Statistic"]:SetText("- / -")
										_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."BossesTotal"]:SetText("-")
									else
										_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."Statistic"]:SetText(col..SCAN[guid].PVEDATA[i][j][k].statisticBoss.."|r / "..PVEDUMP[i][j][k].statisticTotal)
										_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."BossesTotal"]:SetText(SCAN[guid].PVEDATA[i][j][k].statisticBossAll)
									end
								--end
							--end

						--end
					--end
				--end
				-- level 3 END

				-- level 4 START
				--for i = 1, #PVEDUMP do
					--if not PVEDUMP[i].statistic then
						--for j = 1, #PVEDUMP[i] do

							--if not PVEDUMP[i][j].statistic then
								--for k = 1, #PVEDUMP[i][j] do

									if PVEDUMP[i][j][k].statistic then
										for l = 1, #PVEDUMP[i][j][k].statistic do
											if SCAN[guid].PVEDATA[i][j][k].statisticBossEach[l] > 0 then
												col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
											end
											_G["CharScanner_MainFrame_PvEStats_StatsLevel4_"..PVEDUMP[i].name..PVEDUMP[i][j].name..PVEDUMP[i][j][k].name..l.."Name"]:SetText(col..StatisticIDTable[ PVEDUMP[i][j][k].statistic[l] ].."|r")
											_G["CharScanner_MainFrame_PvEStats_StatsLevel4_"..PVEDUMP[i].name..PVEDUMP[i][j].name..PVEDUMP[i][j][k].name..l.."BossesTotal"]:SetText(SCAN[guid].PVEDATA[i][j][k].statisticBossEach[l])
										end
									end

								end
							end

						end
					end
				end
				-- level 4 END

				-- summary
				CharScanner_MainFrame_SummaryButton1PvEAchievement:SetText(SCAN[guid].PvEAchievementCount.." / "..SCAN[guid].PvEAchievementTotal)
				CharScanner_MainFrame_SummaryButton1PvEAchievement:SetTextColor(color.Level1.rgba[1], color.Level1.rgba[2], color.Level1.rgba[3], color.Level1.rgba[4])
				local achievement = "0.00%"
				if SCAN[guid].PvEAchievementTotal > 0 then
					achievement = cnum(math_floor((SCAN[guid].PvEAchievementCount/SCAN[guid].PvEAchievementTotal*100) * mult2 + 0.5) / mult2) .. "%"
				end
				CharScanner_MainFrame_SummaryButton2PvEAchievement:SetText(achievement)
				CharScanner_MainFrame_SummaryButton2PvEAchievement:SetTextColor(color.LightYellow.rgba[1], color.LightYellow.rgba[2], color.LightYellow.rgba[3], color.LightYellow.rgba[4])

				CharScanner_MainFrame_SummaryButton1PvEStatistic:SetText(SCAN[guid].PvEStatisticCount.." / "..SCAN[guid].PvEStatisticTotal)
				CharScanner_MainFrame_SummaryButton1PvEStatistic:SetTextColor(color.Level1.rgba[1], color.Level1.rgba[2], color.Level1.rgba[3], color.Level1.rgba[4])
				local statistic = "0.00%"
				if SCAN[guid].PvEStatisticTotal > 0 then
					statistic = cnum(math_floor((SCAN[guid].PvEStatisticCount/SCAN[guid].PvEStatisticTotal*100) * mult2 + 0.5) / mult2) .. "%"
				end
				CharScanner_MainFrame_SummaryButton2PvEStatistic:SetText(statistic)
				CharScanner_MainFrame_SummaryButton2PvEStatistic:SetTextColor(color.LightYellow.rgba[1], color.LightYellow.rgba[2], color.LightYellow.rgba[3], color.LightYellow.rgba[4])

				CharScanner_MainFrame_SummaryButton1PvEBossesTotal:SetText(SCAN[guid].PvEStatisticBossKills)
				CharScanner_MainFrame_SummaryButton1PvEBossesTotal:SetTextColor(color.Level1.rgba[1], color.Level1.rgba[2], color.Level1.rgba[3], color.Level1.rgba[4])
			end
			-- PvE END



			-- PvP START
			CharScanner_MainFrame_PvPStats_TitleName:SetText("")
			CharScanner_MainFrame_PvPStats_TitleAchievement:SetText(L["Achievements"])
			--CharScanner_MainFrame_PvPStats_TitleWinLoss:SetText("Win / Matches") -- not needed WinLoss+WinLossPercent=title
			CharScanner_MainFrame_PvPStats_TitleWinLossPercent:SetText(L["Victories / Battles / %"])
			CharScanner_MainFrame_PvPStats_TitleHonorableKills:SetText(L["HK"])

			local col
			-- level 1 START
			for i = 1, #PVPDUMP do
				_G["CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[i].name.."Name"]:SetText(L[PVPDUMP[i].name])
				if PVPDUMP[i].achievementTotal == SCAN[guid].PVPDATA[i].achievementCount then
					col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
				end
				_G["CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[i].name.."Achievement"]:SetText(col..SCAN[guid].PVPDATA[i].achievementCount.."|r / "..PVPDUMP[i].achievementTotal)
				local percent = ""
				local winloss = ""
				local honorkills = ""
				if PVPDUMP[i].name == "Battlegrounds" then
					percent = "0.00%"
					if SCAN[guid].PvPBattlegroundMatches > 0 then
						percent = cnum( math_floor((SCAN[guid].PvPBattlegroundWins/SCAN[guid].PvPBattlegroundMatches*100) * mult2 + 0.5) / mult2 ) .. "%"
					end
					winloss = SCAN[guid].PvPBattlegroundWins.." / "..SCAN[guid].PvPBattlegroundMatches
					honorkills = SCAN[guid].PvPBattlegroundKills
				elseif PVPDUMP[i].name == "Arena" then
					percent = "0.00%"
					if SCAN[guid].PvPArenaMatches > 0 then
						percent = cnum( math_floor((SCAN[guid].PvPArenaWins/SCAN[guid].PvPArenaMatches*100) * mult2 + 0.5) / mult2 ) .. "%"
					end
					winloss = SCAN[guid].PvPArenaWins.." / "..SCAN[guid].PvPArenaMatches
					honorkills = SCAN[guid].PvPArenaKills
				elseif PVPDUMP[i].name == "PvPGeneral" then
					percent = ""
					winloss = ""
					honorkills = SCAN[guid].PvPStatisticTotalKills
				end
				_G["CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[i].name.."WinLoss"]:SetText(winloss)
				_G["CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[i].name.."WinLossPercent"]:SetText(percent)
				_G["CharScanner_MainFrame_PvPStats_StatsLevel1_"..PVPDUMP[i].name.."HonorableKills"]:SetText(honorkills)
			--end
			-- level 1 END

			-- level 2 START
			--for i = 1, #PVPDUMP do
				if PVPDUMP[i].statistic then
					for j = 1, PVPDUMP[i].statisticTotal do
						local nid = PVPDUMP[i].statistic[j].matches or PVPDUMP[i].statistic[j].wins or PVPDUMP[i].statistic[j].kills or PVPDUMP[i].statistic[j].misc
						_G["CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..j.."Name"]:SetText(StatisticIDTable[ nid ])
						_G["CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..j.."HonorableKills"]:SetText("") -- **
					end
				else
					for j = 1, #PVPDUMP[i] do
						_G["CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..PVPDUMP[i][j].name.."Name"]:SetText(L[PVPDUMP[i][j].name])
						if PVPDUMP[i][j].achievementTotal == SCAN[guid].PVPDATA[i][j].achievementCount then
							col="|cff"..color.Green.hex else col="|cff"..color.Red.hex
						end
						_G["CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..PVPDUMP[i][j].name.."Achievement"]:SetText(col..SCAN[guid].PVPDATA[i][j].achievementCount.."|r / "..PVPDUMP[i][j].achievementTotal)
						local percent = ""
						local winloss = ""
						local honorkills = ""
						if PVPDUMP[i].name == "Battlegrounds" then
							if PVPDUMP[i][j].name == "RatedBG" then
								percent = ""
								winloss = ""
								honorkills = ""
							else
								percent = "0.00%"
								if SCAN[guid].PVPDATA[i][j].matches > 0 then
									percent = cnum( math_floor((SCAN[guid].PVPDATA[i][j].wins/SCAN[guid].PVPDATA[i][j].matches*100) * mult2 + 0.5) / mult2 ) .. "%"
								end
								winloss = SCAN[guid].PVPDATA[i][j].wins.." / "..SCAN[guid].PVPDATA[i][j].matches
								honorkills = SCAN[guid].PVPDATA[i][j].kills
							end
						elseif PVPDUMP[i].name == "Arena" and PVPDUMP[i][j].name ~= "ArenaGeneral" then
							percent = "0.00%"
							if SCAN[guid].PVPDATA[i][j].matches > 0 then
								percent = cnum( math_floor((SCAN[guid].PVPDATA[i][j].wins/SCAN[guid].PVPDATA[i][j].matches*100) * mult2 + 0.5) / mult2 ) .. "%"
							end
							winloss = SCAN[guid].PVPDATA[i][j].wins.." / "..SCAN[guid].PVPDATA[i][j].matches
							honorkills = SCAN[guid].PVPDATA[i][j].kills
						end
						_G["CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..PVPDUMP[i][j].name.."WinLoss"]:SetText(winloss)
						_G["CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..PVPDUMP[i][j].name.."WinLossPercent"]:SetText(percent)
						_G["CharScanner_MainFrame_PvPStats_StatsLevel2_"..PVPDUMP[i].name..PVPDUMP[i][j].name.."HonorableKills"]:SetText(honorkills)
					--end
				--end
			--end
			-- level 2 END

			-- level 3 START
			--for i = 1, #PVPDUMP do
				--if not PVPDUMP[i].statistic then
					--for j = 1, #PVPDUMP[i] do

						if PVPDUMP[i][j].statistic then
							for k = 1, PVPDUMP[i][j].statisticTotal do
								local nid = PVPDUMP[i][j].statistic[k].matches or PVPDUMP[i][j].statistic[k].wins or PVPDUMP[i][j].statistic[k].kills or PVPDUMP[i][j].statistic[k].misc
								_G["CharScanner_MainFrame_PvPStats_StatsLevel3_"..PVPDUMP[i].name..PVPDUMP[i][j].name..k.."Name"]:SetText(StatisticIDTable[ nid ])
								local text = ""
								if PVPDUMP[i][j].statistic[k].misc and SCAN[guid].PVPDATA[i][j].misc[k] then
									text = SCAN[guid].PVPDATA[i][j].misc[k]
								end
								_G["CharScanner_MainFrame_PvPStats_StatsLevel3_"..PVPDUMP[i].name..PVPDUMP[i][j].name..k.."HonorableKills"]:SetText(text) -- **
							end
						end

					end
				end
			end
			-- level 3 END

			-- summary
			CharScanner_MainFrame_SummaryButton1PvPAchievement:SetText(SCAN[guid].PvPAchievementCount.." / "..SCAN[guid].PvPAchievementTotal)
			CharScanner_MainFrame_SummaryButton1PvPAchievement:SetTextColor(color.Level1.rgba[1], color.Level1.rgba[2], color.Level1.rgba[3], color.Level1.rgba[4])
			local achievement = "0.00%"
			if SCAN[guid].PvPAchievementTotal > 0 then
				achievement = cnum(math_floor((SCAN[guid].PvPAchievementCount/SCAN[guid].PvPAchievementTotal*100) * mult2 + 0.5) / mult2) .. "%"
			end
			CharScanner_MainFrame_SummaryButton2PvPAchievement:SetText(achievement)
			CharScanner_MainFrame_SummaryButton2PvPAchievement:SetTextColor(color.LightYellow.rgba[1], color.LightYellow.rgba[2], color.LightYellow.rgba[3], color.LightYellow.rgba[4])

			CharScanner_MainFrame_SummaryButton1PvPWinLoss:SetText(SCAN[guid].PvPStatisticWins.." / "..SCAN[guid].PvPStatisticMatches)
			CharScanner_MainFrame_SummaryButton1PvPWinLoss:SetTextColor(color.Level1.rgba[1], color.Level1.rgba[2], color.Level1.rgba[3], color.Level1.rgba[4])
			local statistic = "0.00%"
			if SCAN[guid].PvPStatisticMatches > 0 then
				statistic = cnum(math_floor((SCAN[guid].PvPStatisticWins/SCAN[guid].PvPStatisticMatches*100) * mult2 + 0.5) / mult2) .. "%"
			end
			CharScanner_MainFrame_SummaryButton1PvPWinLossPercent:SetText(statistic)
			CharScanner_MainFrame_SummaryButton1PvPWinLossPercent:SetTextColor(color.Level1.rgba[1], color.Level1.rgba[2], color.Level1.rgba[3], color.Level1.rgba[4])

			CharScanner_MainFrame_SummaryButton1PvPHonorableKills:SetText(SCAN[guid].PvPStatisticTotalKills)
			CharScanner_MainFrame_SummaryButton1PvPHonorableKills:SetTextColor(color.Level1.rgba[1], color.Level1.rgba[2], color.Level1.rgba[3], color.Level1.rgba[4])
			CharScanner_MainFrame_SummaryButton2PvPHonorableKills:SetText("("..L["Total"]..")")
			CharScanner_MainFrame_SummaryButton2PvPHonorableKills:SetTextColor(color.ButtonBGActive.rgba[1], color.ButtonBGActive.rgba[2], color.ButtonBGActive.rgba[3], color.ButtonBGActive.rgba[4])
			-- PvP END

		end

		if scan and not GroupScanIsActive then
			CharScanner_MainFrame_Status_ProgressText:SetText(L["Step"].." "..SCAN[guid].DataStep.."/4")
			CharScanner_MainFrame_Status_ProgressDone:SetWidth(SCAN[guid].DataStep*sizeGeneral.StatusProgressBarWidth/4)
		end
	end

	if SCAN[guid].DataStep == 4 then

		if scan then
			-- scan complete: copy&delete preview scandata START
			local realGUID = string_sub(guid, 2)
			CreateScanDataByGUID(realGUID)
			SCAN[realGUID] = SCAN[guid] -- copy preview
			for k, v in pairs(SCAN) do
				if string_sub(k, 0, 1) == "_" then
					SCAN[k] = nil -- delete preview
				end
			end
			guid = realGUID
			-- scan complete: copy&delete preview scandata END

			-- talent group guid adjustment
			if (GroupScanIsActive and not GroupScanFirstGUID) or SingleScanIsActive or not scan then
				CharScanner_MainFrame_TalentTree_GroupSelection_Button1:SetScript("OnClick", function() CharScanner:TalentGroup_Switch(guid, 1) end)
				CharScanner_MainFrame_TalentTree_GroupSelection_Button2:SetScript("OnClick", function() CharScanner:TalentGroup_Switch(guid, 2) end)
			end

			SCAN[guid].scanTime = math_floor((GetTime()-SingleScanTimerStart) * mult4 + 0.5) / mult4

			if (GroupScanIsActive and not GroupScanFirstGUID) or SingleScanIsActive or not scan then
				local sTime = SCAN[guid].scanTime
				if sTime == 1 then
					sTime = string.format("%s %s", sTime, SECONDS)
				else
					sTime = sTime.." "..SECONDS
				end
				CharScanner_MainFrame_Status_Text:SetText(sTime)
				CharScanner_MainFrame_Status_Text:SetTextColor(color.Green.rgba[1], color.Green.rgba[2], color.Green.rgba[3], color.Green.rgba[4])
			end

			local hour, minute = GetGameTime()
			if hour < 10 then
				hour = "0"..hour
			end
			if minute < 10 then
				minute = "0"..minute
			end
			SCAN[guid].time = hour..":"..minute

			if (CharScanner_Options.chatoutput and not GroupScanIsActive) or
			   (CharScanner_Options.chatoutput and GroupScanIsActive and not CharScanner_Options.chatoutputGrp) then
				local col = ClassHexColor(SCAN[guid].classEN)
				local txt= "|cff"..col..SCAN[guid].name.."|r"				
				txt = txt.." |cff"..color.Level2.hex..LEVEL.." "..SCAN[guid].level.." - "..SCAN[guid].race..",|r"
				txt = txt.." |cff"..col..SCAN[guid].class.."|r"
				if SCAN[guid].guildName then
					txt = txt.." |cff"..color.Level2.hex.."("..GUILD..": "..SCAN[guid].guildName..")|r"
				end
				Print(txt)
			end

			if (GroupScanIsActive and not GroupScanFirstGUID) or SingleScanIsActive or not scan then
				GUIselectedGUID = guid
			end

			LatestScanIsComplete = true
			CharScanner:History_Update()

			if GroupScanIsActive and not GroupScanFirstGUID then
				if not HistoryFrameIsShown and not CompareFrameIsShown then
					CharScanner:HistoryFrameShow()
				end
			end

			if not GroupScanIsActive then
				CharScanner:StopSingleScan(false)
			end
		else
			GUIselectedGUID = guid
			if not GroupScanIsActive and not SingleScanIsActive then -- XFIX
				LatestScanIsComplete = true
			end -- XFIX
			local sTime = SCAN[guid].scanTime
			if sTime == 1 then
				sTime = string.format("%s %s", sTime, SECONDS)
			else
				sTime = sTime.." "..SECONDS
			end
			CharScanner_MainFrame_Status_Text:SetText(L["SCANNED AT"]..": "..SCAN[guid].time.."  |cff"..color.Green.hex.."("..sTime..")|r")
			CharScanner_MainFrame_Status_Text:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
		end

		if AchievementDetailFrameMode == "PvE" then
			if CharScanner_MainFrame_PvEStats_Title.AchievementDetailShow then
				CharScanner:AchievementDetail_Update("PvE", GUI_achievementdetail_L1, GUI_achievementdetail_L2, GUI_achievementdetail_L3)
			end
		else
			if CharScanner_MainFrame_PvPStats_Title.AchievementDetailShow then
				CharScanner:AchievementDetail_Update("PvP", GUI_PvPachievementdetail_L1, GUI_PvPachievementdetail_L2, GUI_PvPachievementdetail_L3)
			end
		end

		if (GroupScanIsActive and not GroupScanFirstGUID) or SingleScanIsActive or not scan then
			for i = 1, GUI_history_maxRows do
				if _G["CharScanner_HistoryFrame_Button"..i].guid == guid then
					_G["CharScanner_HistoryFrame_Button"..i.."Arrow"]:Show()
					break
				end
			end
		end

		EnableIconButton("CharScanner_MainFrame_Headline_ButtonHistory")
		EnableIconButton("CharScanner_MainFrame_Headline_ButtonCompare")

		GroupScanFirstGUID = guid
		GroupScanNext = true

	end

end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
local GTT = CreateFrame("GameTooltip", "CharScanner_ScanningTooltip", nil, "GameTooltipTemplate")
GTT:SetOwner(_G["CharScanner"], "ANCHOR_NONE")

local function ItemData(unit, index)
	local slotID = GetInventorySlotInfo(slots[index])
	local itemLink = GetInventoryItemLink(unit, slotID)
	if not SCAN[SCANGUID].gear[index] then SCAN[SCANGUID].gear[index] = {} end

	if itemLink then
		local baseID, enchantID, gem1ID, gem2ID, gem3ID, gem4ID, suffixID, uniqueID, linkLevel, reforgeID = itemLink:match("item:(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+)") -- show more data TODO
		local _, _, itemRarity, iLevel = GetItemInfo(baseID)
		--print("baseID:", baseID, type(baseID), "enchantID:", enchantID, type(enchantID), "gem1ID:", gem1ID, type(gem1ID), "gem2ID:", gem2ID, type(gem2ID), "gem3ID:", gem3ID, type(gem3ID), "gem4ID:", gem4ID, type(gem4ID), "suffixID:", suffixID, type(suffixID), "uniqueID:", uniqueID, type(uniqueID), "linkLevel:", linkLevel, type(linkLevel), "reforgeID:", reforgeID, type(reforgeID), "itemRarity:", itemRarity, type(itemRarity), "iLevel:", iLevel, type(iLevel))

		SCAN[SCANGUID].gear[index].iLink = itemLink or ""
		SCAN[SCANGUID].gear[index].iLvl  = tonumber(iLevel) or 0
		SCAN[SCANGUID].gear[index].iRar  = tonumber(itemRarity) or 0
		SCAN[SCANGUID].gear[index].iEnch = tonumber(enchantID) or 0
		SCAN[SCANGUID].gear[index].iRefo = tonumber(reforgeID) or 0

--[[ --GEAR
		print(itemLink)
		for key, value in pairs(GetItemStats(itemLink)) do
			if key and value and type(value) == "number" then
				if not SCAN[SCANGUID].gearStats[key] then
					SCAN[SCANGUID].gearStats[key] = value
				else
					SCAN[SCANGUID].gearStats[key] = SCAN[SCANGUID].gearStats[key] + value
				end
			end
		end
--]]

		SCAN[SCANGUID].gear[index].iPvP = 0
		for key in pairs(GetItemStats(itemLink)) do
			if key == "ITEM_MOD_RESILIENCE_RATING" or key == "ITEM_MOD_RESILIENCE_RATING_SHORT" then
				SCAN[SCANGUID].gear[index].iPvP = 1
			end
		end
	else
		SCAN[SCANGUID].gear[index].iLink = ""
		SCAN[SCANGUID].gear[index].iLvl  = -1
		SCAN[SCANGUID].gear[index].iRar  = -1
		SCAN[SCANGUID].gear[index].iEnch = -1
		SCAN[SCANGUID].gear[index].iPvP  = -1
		SCAN[SCANGUID].gear[index].iRefo = -1

		local textureName = GetInventoryItemTexture(unit, slotID);
		if textureName then
			GTT:ClearLines()
			GTT:SetInventoryItem(unit, slotID)
			local itemName, itemLink = GTT:GetItem()
			local baseID, enchantID, gem1ID, gem2ID, gem3ID, gem4ID, suffixID, uniqueID, linkLevel, reforgeID = itemLink:match("item:(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+)") -- show more data TODO
			local _, _, itemRarity, iLevel = GetItemInfo(baseID)
			if baseID and enchantID then
				SCAN[SCANGUID].gear[index].iLink = "item:"..baseID..":"..enchantID
				SCAN[SCANGUID].gear[index].iLvl  = tonumber(iLevel) or 0
				SCAN[SCANGUID].gear[index].iRar  = tonumber(itemRarity) or 8
				SCAN[SCANGUID].gear[index].iEnch = tonumber(enchantID) or 0
				SCAN[SCANGUID].gear[index].iPvP  = 0
				SCAN[SCANGUID].gear[index].iRefo = tonumber(reforgeID) or 0
			end
		end

	end
	return SCAN[SCANGUID].gear[index].iLvl, SCAN[SCANGUID].gear[index].iRar, SCAN[SCANGUID].gear[index].iEnch, SCAN[SCANGUID].gear[index].iPvP
end

local function TwoHandWeaponCheck(unit, who)
	local id = GetInventoryItemLink(unit, _G[who.."MainHandSlot"]:GetID())
	local twohandweapon = false
	if id then
		local linkID = select(3, id:find("item:(%d+)"))
		local itemEquipLoc = select(9, GetItemInfo(linkID))
		if itemEquipLoc == "INVTYPE_2HWEAPON" then
			local id2 = GetInventoryItemLink(unit, _G[who.."SecondaryHandSlot"]:GetID())
			if not id2 then
				twohandweapon = true
			end
		end
	end
	return twohandweapon
end

function CharScanner:GearScanPost(unit)
	for i = 1, slotsNum do
		local slotID = GetInventorySlotInfo(slots[i])
		local textureName = GetInventoryItemTexture(unit, slotID);
		if textureName then
			GTT:ClearLines()
			GTT:SetInventoryItem(unit, slotID)
			local itemName, itemLink = GTT:GetItem()
			--local baseID, enchantID, gem1ID, gem2ID, gem3ID, gem4ID, suffixID, uniqueID, linkLevel, reforgeID = itemLink:match("item:(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+):(%-?%d+)")
		end
	end
end

function CharScanner:ItemLevelCalc(unit, who)
	CharScanner:GearScanPost(unit)

	local SlotCount = 17
	if TwoHandWeaponCheck(unit, who) then
		SlotCount = 16
	end
	local ItemLevel = 0
	local ItemCount = 0
	local minItem = 0
	local maxItem = 0
	local EnchCount = 0
	local PvPCount = 0

	for i = 1, itemQualityNum do
		itemQuality[i].count = 0
	end

	for i = 1, slotsNum do
		local iLvL, iRarity, iEnch, iPvP = ItemData(unit, i)
		if iLvL >= 0 then
			ItemLevel = ItemLevel + iLvL
			ItemCount = ItemCount + 1
		end
		if iLvL > 0 then
			if minItem == 0 then
				minItem = iLvL
			end
			if maxItem == 0 then
				maxItem = iLvL
			end
			if iLvL < minItem then
				minItem = iLvL
			end
			if iLvL > maxItem then
				maxItem = iLvL
			end
		end
		if iRarity >= 0 then
			itemQuality[iRarity+1].count = itemQuality[iRarity+1].count + 1
		end
		if iEnch > 0 then
			EnchCount = EnchCount + 1
		end
		if iPvP == 1 then
			PvPCount = PvPCount + 1
		end
	end

	local quality = ""
	local x = 0
	for i = 1, itemQualityNum do
		if itemQuality[i].count > 0 then
			if x == 1 then
				quality = quality..", "
				x = 0
			end
			quality = quality.."|cff"..color.White.hex..itemQuality[i].count.."x|r |cff"..itemQuality[i].color..itemQuality[i].name.."|r"
			x = 1
		end
	end

	SCAN[SCANGUID].itemlevelMin     = minItem
	SCAN[SCANGUID].itemlevelMax     = maxItem
	SCAN[SCANGUID].itemlevelQuality = quality
	SCAN[SCANGUID].gearEnchCount    = EnchCount
	SCAN[SCANGUID].gearPvPCount     = PvPCount
	if SlotCount == 16 then
		SCAN[SCANGUID].gearTwoHandWeapon = true
	end

	local ItemIQcount = SlotCount-ItemCount
	if ItemIQcount > 0 then
		SCAN[SCANGUID].itemlevelMiss = ItemIQcount
	end

	local ItemIQ = math_floor((ItemLevel/SlotCount) * mult2 + 0.5) / mult2
	if type(ItemIQ) == "number" and ItemIQ > 0 then
		SCAN[SCANGUID].itemlevelAvg = ItemIQ
	end

	CharScanner:ShowData(2, SCANGUID, true)
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:TalentGroup_Switch(guid, group)
	if SCAN[guid] and SCAN[guid].talents[group] then
		local activgroup = SCAN[guid].activeTalentGroup
		local button1name = "CharScanner_MainFrame_TalentTree_GroupSelection_Button1"
		local button2name = "CharScanner_MainFrame_TalentTree_GroupSelection_Button2"

		_G[button1name]:SetScript("OnClick", function() CharScanner:TalentGroup_Switch(guid, 1) end)
		_G[button2name]:SetScript("OnClick", function() CharScanner:TalentGroup_Switch(guid, 2) end)


		if SCAN[guid].talentGroups == 1 then
			_G[button1name]:Show()
			_G[button2name]:Hide()
		else
			_G[button1name]:Show()
			_G[button2name]:Show()
		end

		if activgroup == 1 then
			_G[button1name.."Text"]:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
			_G[button2name.."Text"]:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
		else
			_G[button1name.."Text"]:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
			_G[button2name.."Text"]:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
		end


		if group == 1 then
			_G[button1name]:Disable()
			if group == activgroup then
				_G[button1name.."Text"]:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
			else
				_G[button1name.."Text"]:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
			end
			local border1 = _G[button1name.."Border"]
			border1:SetPoint("TOPLEFT"     ,sizeGeneral.Border, -sizeGeneral.Border)
			border1:SetPoint("BOTTOMLEFT"  ,sizeGeneral.Border, 0)
			border1:SetPoint("TOPRIGHT"    ,-sizeGeneral.Border, -sizeGeneral.Border)
			border1:SetPoint("BOTTOMRIGHT" ,-sizeGeneral.Border, 0)

			_G[button2name]:Enable()
			local border2 = _G[button2name.."Border"]
			border2:SetPoint("TOPLEFT"     ,sizeGeneral.Border, -sizeGeneral.Border)
			border2:SetPoint("BOTTOMLEFT"  ,0, sizeGeneral.Border)
			border2:SetPoint("TOPRIGHT"    ,-sizeGeneral.Border, -sizeGeneral.Border)
			border2:SetPoint("BOTTOMRIGHT" ,-sizeGeneral.Border, sizeGeneral.Border)
		else
			_G[button1name]:Enable()
			local border1 = _G[button1name.."Border"]
			border1:SetPoint("TOPLEFT"     ,sizeGeneral.Border, -sizeGeneral.Border)
			border1:SetPoint("BOTTOMLEFT"  ,0, sizeGeneral.Border)
			border1:SetPoint("TOPRIGHT"    ,-sizeGeneral.Border, -sizeGeneral.Border)
			border1:SetPoint("BOTTOMRIGHT" ,-sizeGeneral.Border, sizeGeneral.Border)

			_G[button2name]:Disable()
			if group == activgroup then
				_G[button2name.."Text"]:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
			else
				_G[button2name.."Text"]:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
			end

			local border2 = _G[button2name.."Border"]
			border2:SetPoint("TOPLEFT"     ,sizeGeneral.Border, -sizeGeneral.Border)
			border2:SetPoint("BOTTOMLEFT"  ,sizeGeneral.Border, 0)
			border2:SetPoint("TOPRIGHT"    ,-sizeGeneral.Border, -sizeGeneral.Border)
			border2:SetPoint("BOTTOMRIGHT" ,-sizeGeneral.Border, 0)
		end

		local tier = 1
		local column = 1
		for tab = 1, myMAX_TALENT_TABS do

			local point
			if group == 1 then
				point = select(tab, string.match(SCAN[guid].talentGroup1pointsString, "(.+)/(.+)/(.+)"))
			else
				point = select(tab, string.match(SCAN[guid].talentGroup2pointsString, "(.+)/(.+)/(.+)"))
			end
			local bName   = _G["CharScanner_MainFrame_TalentTree"..tab.."Name"]
			local bPoints = _G["CharScanner_MainFrame_TalentTree"..tab.."Points"]

			bName:SetText(classes[SCAN[guid].classEN].talentTab[tab].tabName)
			bPoints:SetText(point)

			if group == activgroup then
				bName:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
				bPoints:SetTextColor(color.White.rgba[1], color.White.rgba[2], color.White.rgba[3], color.White.rgba[4])
			else
				bName:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
				bPoints:SetTextColor(color.Grey.rgba[1], color.Grey.rgba[2], color.Grey.rgba[3], color.Grey.rgba[4])
			end

			for talent = 1, myMAX_NUM_TALENTS do

				if talent == 1 then
					tier = 1
					column = 1
				elseif mod(talent, myMAX_COLUMN) == 1 then
					tier = tier + 1
					column = 1
				else
					column = column + 1
				end

				local buttonName = "CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column

				if
					classes[SCAN[guid].classEN].talents[tab] and
					classes[SCAN[guid].classEN].talents[tab][tier] and
					classes[SCAN[guid].classEN].talents[tab][tier][column]
					then

					local spellRank    = SCAN[guid].talents[group][tab][tier][column]
					local spellIcon    = classes[SCAN[guid].classEN].talents[tab][tier][column].icon
					local spellID      = classes[SCAN[guid].classEN].talents[tab][tier][column].id
					local spellName    = classes[SCAN[guid].classEN].talents[tab][tier][column].name
					local talentString = "talent:"..spellID..":"..(spellRank-1)
					local talentLink   = "|cff4e96f7|H"..talentString.."|h["..spellName.."]|h|r"
					_G[buttonName.."IconTexture"]:SetTexture(spellIcon)
					_G[buttonName]:Show()
					_G[buttonName]:SetScript("OnEnter", function(self)
						_G[buttonName.."Border"]:SetTexture(1, 1, 0.49, 1)
						GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
						GameTooltip:SetHyperlink(talentString)
					end)
					_G[buttonName]:SetScript("OnLeave", function()
						_G[buttonName.."Border"]:SetTexture(0.2, 0.2, 0.2, 1)
						GameTooltip:Hide()
					end)
					_G[buttonName]:SetScript("OnClick", function()
						HandleModifiedItemClick(talentLink)
					end)
					if spellRank == 0 then
						SetItemButtonDesaturated(_G[buttonName], 1, 0.65, 0.65, 0.65)
						_G[buttonName]:SetAlpha(0.5)
						_G[buttonName.."IconTexture"]:SetVertexColor(0.5, 0.5, 0.5)
						_G[buttonName.."RankBorder"]:Hide()
						_G[buttonName.."RankNumber"]:Hide()
					else
						SetItemButtonDesaturated(_G[buttonName], nil)
						_G[buttonName]:SetAlpha(1)
						_G[buttonName.."IconTexture"]:SetVertexColor(1, 1, 1)
						_G[buttonName.."RankBorder"]:Show()
						_G[buttonName.."RankNumber"]:SetText(spellRank)
						_G[buttonName.."RankNumber"]:Show()
					end
				else
					_G[buttonName]:Hide()
					_G[buttonName]:SetScript("OnEnter", nil)
					_G[buttonName]:SetScript("OnLeave", nil)
				end

			end
		end
	else
		for tab = 1, myMAX_TALENT_TABS do
			_G["CharScanner_MainFrame_TalentTree"..tab.."Name"]:SetText("")
			_G["CharScanner_MainFrame_TalentTree"..tab.."Points"]:SetText("")
		end
		CharScanner_MainFrame_TalentTree_GroupSelection_Button1:Hide()
		CharScanner_MainFrame_TalentTree_GroupSelection_Button1:SetScript("OnClick", nil)
		CharScanner_MainFrame_TalentTree_GroupSelection_Button2:Hide()
		CharScanner_MainFrame_TalentTree_GroupSelection_Button2:SetScript("OnClick", nil)
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:ClearData()
	-- Base
	CharScanner_MainFrame_Base_ClassIcon:SetTexCoord(0, 0, 0, 0)
	CharScanner_MainFrame_Base_FactionIcon:SetTexCoord(0, 0, 0, 0)
	CharScanner_MainFrame_Base_Name:SetText("")
	CharScanner_MainFrame_Base_GUID:SetText("")
	CharScanner_MainFrame_Base_Guild:SetText("")
	CharScanner_MainFrame_Base_Level:SetText("")

	-- Equipment
	CharScanner_MainFrame_Equipment_Title:SetText("")
	CharScanner_MainFrame_Equipment_Text1:SetText("")
	CharScanner_MainFrame_Equipment_Text2:SetText("")
	CharScanner_MainFrame_Equipment_Text3:SetText("")

	-- Gear
	for i = 1, #slotButton do
		_G["CharScanner_MainFrame_Gear_Slot"..i.."IconTexture"]:SetTexture(slotButton[i].texture)
		_G["CharScanner_MainFrame_Gear_Slot"..i.."Count"]:SetText("")
		_G["CharScanner_MainFrame_Gear_Slot"..i.."Count"]:Hide()
		_G["CharScanner_MainFrame_Gear_Slot"..i.."EnchTexture"]:Hide()
		_G["CharScanner_MainFrame_Gear_Slot"..i.."PvPTexture"]:Hide()
		_G["CharScanner_MainFrame_Gear_Slot"..i]:SetScript("OnEnter", nil)
		_G["CharScanner_MainFrame_Gear_Slot"..i]:SetScript("OnLeave", nil)
	end

	-- Talent
	CharScanner_MainFrame_Talent_Title:SetText("")
	CharScanner_MainFrame_Talent_TalentMainRole:SetTexture(nil)
	CharScanner_MainFrame_Talent_TalentMainRoleActive:SetTexture(nil)
	CharScanner_MainFrame_Talent_TalentGroup1Icon:SetTexture(nil)
	CharScanner_MainFrame_Talent_TalentGroup1Text:SetText("")
	CharScanner_MainFrame_Talent_TalentGroup2Icon:SetTexture(nil)
	CharScanner_MainFrame_Talent_TalentGroup2Text:SetText("")

	-- TalentTree
	CharScanner:TalentGroup_Switch(0, 0)

	local tier = 1
	local column = 1
	for tab = 1, myMAX_TALENT_TABS do
		for talent = 1, myMAX_NUM_TALENTS do
			if talent == 1 then
				tier = 1
				column = 1
			elseif mod(talent, myMAX_COLUMN) == 1 then
				tier = tier + 1
				column = 1
			else
				column = column + 1
			end
			_G["CharScanner_MainFrame_TalentTree_Slot"..tab.."_"..tier.."_"..column]:Hide()
		end
	end

	-- PvE Stats
	CharScanner_MainFrame_PvEStats_TitleName:SetText("")
	CharScanner_MainFrame_PvEStats_TitleAchievement:SetText("")
	CharScanner_MainFrame_PvEStats_TitleStatistic:SetText("")
	CharScanner_MainFrame_PvEStats_TitleBossesTotal:SetText("")

	for i = 1, #PVEDUMP do
		_G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name.."Name"]:SetText("")
		_G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name.."Achievement"]:SetText("")
		_G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name.."Statistic"]:SetText("")
		_G["CharScanner_MainFrame_PvEStats_StatsLevel1_"..PVEDUMP[i].name.."BossesTotal"]:SetText("")

		if PVEDUMP[i].statistic then
			for j = 1, #PVEDUMP[i].statistic do
				_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..j.."Name"]:SetText("")
				_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..j.."BossesTotal"]:SetText("")
			end
		else
			for j = 1, #PVEDUMP[i] do
				_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name.."Name"]:SetText("")
				_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name.."Achievement"]:SetText("")
				_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name.."Statistic"]:SetText("")
				_G["CharScanner_MainFrame_PvEStats_StatsLevel2_"..PVEDUMP[i].name..PVEDUMP[i][j].name.."BossesTotal"]:SetText("")

				if PVEDUMP[i][j].statistic then
					for k = 1, #PVEDUMP[i][j].statistic do
						_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."Name"]:SetText("")
						_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."BossesTotal"]:SetText("")
					end
				else
					for k = 1, #PVEDUMP[i][j] do
						_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."Name"]:SetText("")
						_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."Achievement"]:SetText("")
						_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."Statistic"]:SetText("")
						_G["CharScanner_MainFrame_PvEStats_StatsLevel3_"..PVEDUMP[i].name..PVEDUMP[i][j].name..k.."BossesTotal"]:SetText("")

						if PVEDUMP[i][j][k].statistic then
							for l = 1, #PVEDUMP[i][j][k].statistic do
								_G["CharScanner_MainFrame_PvEStats_StatsLevel4_"..PVEDUMP[i].name..PVEDUMP[i][j].name..PVEDUMP[i][j][k].name..l.."Name"]:SetText("")
								_G["CharScanner_MainFrame_PvEStats_StatsLevel4_"..PVEDUMP[i].name..PVEDUMP[i][j].name..PVEDUMP[i][j][k].name..l.."BossesTotal"]:SetText("")
							end
						end

					end
				end

			end
		end
	end

	-- PvE summary
	CharScanner_MainFrame_SummaryButton1PvEAchievement:SetText("")
	CharScanner_MainFrame_SummaryButton2PvEAchievement:SetText("")

	CharScanner_MainFrame_SummaryButton1PvEStatistic:SetText("")
	CharScanner_MainFrame_SummaryButton2PvEStatistic:SetText("")

	CharScanner_MainFrame_SummaryButton1PvEBossesTotal:SetText("")
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function CharScanner:Unregister_INSPECT_READY()
	Event_INSPECT_READY = false
	ClearInspectPlayer()
end

function CharScanner:Unregister_INSPECT_ACHIEVEMENT_READY()
	Event_INSPECT_ACHIEVEMENT_READY = false
	-- achievementFunctions workaround START
	-- hack to avoid error messages because of achievementFunctions.selectedCategory
	if achievementFunctions and achievementFunctions.selectedCategory then
		achievementFunctions.selectedCategory = bugfix or "summary"
	end
	-- achievementFunctions workaround END
	ClearAchievementComparisonUnit()
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
local function OnEvent(self, event, ...)
	if event == "PLAYER_REGEN_DISABLED" then
		inCombat = true
		CharScanner_ScanButtonFrame:Hide()
	elseif event == "PLAYER_REGEN_ENABLED" then
		inCombat = false
		if CharScanner_Options.scanButton and CharScanner_Options.scanButtonLock then
			if UnitIsPlayer("target") then
				CharScanner_ScanButtonFrame:Show()
			end
		end
		if inCombatGrpChk then
			if GetNumRaidMembers() > 1 or GetNumPartyMembers() > 0 then
				inPartyOrRaid = true
			else
				inPartyOrRaid = false
			end
			CharScanner:GroupScanToggle()
			inCombatGrpChk = false
		end
	elseif event == "PLAYER_TARGET_CHANGED" then
		if inCombat then return end
		if CharScanner_Options.scanButton and CharScanner_Options.scanButtonLock then
			if UnitIsPlayer("target") then
				CharScanner_ScanButtonFrame:Show()
			else
				CharScanner_ScanButtonFrame:Hide()
			end
		end
	elseif event == "INSPECT_READY" then
		-- Achievement/Talent scan restriction START
		local currTime = GetTime()
		if (currTime - RestrictionTime) > RestrictionMaxTime or not RestrictionFirstTime then
			RestrictionAchNum = 0
			RestrictionTalNum = 0
		end
		RestrictionTime = currTime
		RestrictionTalNum = RestrictionTalNum + 1
		RestrictionFirstTime = true
		-- Achievement/Talent scan restriction END

		if not Event_INSPECT_READY then return end

		local talentGroup1pointsTotal = nil
		local talentGroup2pointsTotal = nil
		SCAN[SCANGUID].talentGroup1name = nil
		SCAN[SCANGUID].talentGroup1role = nil
		SCAN[SCANGUID].talentGroup1pointsString = nil
		SCAN[SCANGUID].talentGroup2name = nil
		SCAN[SCANGUID].talentGroup2role = nil
		SCAN[SCANGUID].talentGroup2pointsString = nil
		SCAN[SCANGUID].talentGroups = GetNumTalentGroups(IsUnitInpected, nil)
		SCAN[SCANGUID].activeTalentGroup = GetActiveTalentGroup(IsUnitInpected, nil)

		-- save talent information for each class START
		-- this data is saved to the classes table if a class is scanned the first time
		if not classes[SCAN[SCANGUID].classEN].talentsSaved then
			for tab = 1, myMAX_TALENT_TABS do

				local _, tabName = GetTalentTabInfo(tab, IsUnitInpected, nil)
				if not classes[SCAN[SCANGUID].classEN].talentTab then classes[SCAN[SCANGUID].classEN].talentTab = {} end
				if not classes[SCAN[SCANGUID].classEN].talentTab[tab] then
					classes[SCAN[SCANGUID].classEN].talentTab[tab] = {}
					classes[SCAN[SCANGUID].classEN].talentTab[tab].tabName = tabName
				end

				for talent=1, myMAX_NUM_TALENTS do
					local name, iconTexture, tier, column, rank = GetTalentInfo(tab, talent, IsUnitInpected, nil, nil)
					local spellLink = GetTalentLink(tab, talent, IsUnitInpected, nil, nil, nil)

					if iconTexture and spellLink then
						local spellID = select(3, spellLink:find("talent:(%d+)"))

						if not classes[SCAN[SCANGUID].classEN].talents then classes[SCAN[SCANGUID].classEN].talents = {} end
						if not classes[SCAN[SCANGUID].classEN].talents[tab] then classes[SCAN[SCANGUID].classEN].talents[tab] = {} end
						if not classes[SCAN[SCANGUID].classEN].talents[tab][tier] then classes[SCAN[SCANGUID].classEN].talents[tab][tier] = {} end
						if not classes[SCAN[SCANGUID].classEN].talents[tab][tier][column] then classes[SCAN[SCANGUID].classEN].talents[tab][tier][column] = {} end

						classes[SCAN[SCANGUID].classEN].talents[tab][tier][column].id   = spellID
						classes[SCAN[SCANGUID].classEN].talents[tab][tier][column].icon = iconTexture
						classes[SCAN[SCANGUID].classEN].talents[tab][tier][column].name = name or "CharScanner_TalentScan_ERROR"
					end
				end
			end
			classes[SCAN[SCANGUID].classEN].talentsSaved = true
		end
		-- save talent information for each class END

		-- talent scan START
		for group = 1, SCAN[SCANGUID].talentGroups do
			if not SCAN[SCANGUID].talents[group] then SCAN[SCANGUID].talents[group] = {} end
			if group == 1 then
				SCAN[SCANGUID].talentGroup1pointsUnspent = GetUnspentTalentPoints(IsUnitInpected, nil, group)
			else
				SCAN[SCANGUID].talentGroup2pointsUnspent = GetUnspentTalentPoints(IsUnitInpected, nil, group)
			end
			local points = 0
			for tab = 1, myMAX_TALENT_TABS do
				if not SCAN[SCANGUID].talents[group][tab] then SCAN[SCANGUID].talents[group][tab] = {} end

				-- --------------------
				for talent=1, myMAX_NUM_TALENTS do
					local _, iconTexture, tier, column, rank = GetTalentInfo(tab, talent, IsUnitInpected, nil, group)
					if not SCAN[SCANGUID].talents[group][tab][tier] then SCAN[SCANGUID].talents[group][tab][tier] = {} end
					if iconTexture then
						SCAN[SCANGUID].talents[group][tab][tier][column] = rank
					end
				end
				-- --------------------

				-- --------------------
				local _, name, _, icon, pointsSpent = GetTalentTabInfo(tab, IsUnitInpected, nil, group)
				if name and icon and pointsSpent then
					if tab == 1 then
						if group == 1 then
							points = pointsSpent
							talentGroup1pointsTotal = pointsSpent
							SCAN[SCANGUID].talentGroup1pointsString = pointsSpent
							SCAN[SCANGUID].talentGroup1name         = name
							SCAN[SCANGUID].talentGroup1role         = tab
							SCAN[SCANGUID].talentGroup1icon         = icon
						else
							points = pointsSpent
							talentGroup2pointsTotal = pointsSpent
							SCAN[SCANGUID].talentGroup2pointsString = pointsSpent
							SCAN[SCANGUID].talentGroup2name         = name
							SCAN[SCANGUID].talentGroup2role         = tab
							SCAN[SCANGUID].talentGroup2icon         = icon
						end
					else
						if group == 1 then
							talentGroup1pointsTotal = talentGroup1pointsTotal + pointsSpent
							SCAN[SCANGUID].talentGroup1pointsString = SCAN[SCANGUID].talentGroup1pointsString.."/"..pointsSpent
							if pointsSpent > points then
								points = pointsSpent
								SCAN[SCANGUID].talentGroup1name = name
								SCAN[SCANGUID].talentGroup1role = tab
								SCAN[SCANGUID].talentGroup1icon = icon
							end
						else
							talentGroup2pointsTotal = talentGroup2pointsTotal + pointsSpent
							SCAN[SCANGUID].talentGroup2pointsString = SCAN[SCANGUID].talentGroup2pointsString.."/"..pointsSpent
							if pointsSpent > points then
								points = pointsSpent
								SCAN[SCANGUID].talentGroup2name = name
								SCAN[SCANGUID].talentGroup2role = tab
								SCAN[SCANGUID].talentGroup2icon = icon
							end
						end
					end
				end
				-- --------------------

			end
		end
		-- talent scan END

		-- opposite faction level (lvl ??/skull/-1 only!!!) check START
		if SCAN[SCANGUID].level < 0 then
			-- - Since Patch 4.0.1 there is no precise method to get the exact level for chars from the opposite faction that are more
			--   than 9 levels above your level.
			-- - If we scan a char with a skull we can get the exact level for 1,2,3,4,5,6,7,8,9,10 and 81,82,83,84,85.
			-- - All other possible levels between 11/12 and 79/80 are listed below.
			-- - If we are level 1 we can get the level from chars <= 8.
			-- - If we can not see a level and totalPoints = 0 the scanned char is level 9.
			-- NOTE: There is a possibility to get more precise results in conjunction with the gearscan (linkLevel from itemLink). This
			--       method is not 100% precise too...anyway, maybe I extend this check
			local realLevel = "??"
			local totalPoints = talentGroup1pointsTotal + SCAN[SCANGUID].talentGroup1pointsUnspent

			if     totalPoints ==  0 then realLevel = "9"
			elseif totalPoints ==  1 then realLevel = "10"
			elseif totalPoints ==  2 then realLevel = "11/12"
			elseif totalPoints ==  3 then realLevel = "13/14"
			elseif totalPoints ==  4 then realLevel = "15/16"
			elseif totalPoints ==  5 then realLevel = "17/18"
			elseif totalPoints ==  6 then realLevel = "19/20"
			elseif totalPoints ==  7 then realLevel = "21/22"
			elseif totalPoints ==  8 then realLevel = "23/24"
			elseif totalPoints ==  9 then realLevel = "25/26"
			elseif totalPoints == 10 then realLevel = "27/28"
			elseif totalPoints == 11 then realLevel = "29/30"
			elseif totalPoints == 12 then realLevel = "31/32"
			elseif totalPoints == 13 then realLevel = "33/34"
			elseif totalPoints == 14 then realLevel = "35/36"
			elseif totalPoints == 15 then realLevel = "37/38"
			elseif totalPoints == 16 then realLevel = "39/40"
			elseif totalPoints == 17 then realLevel = "41/42"
			elseif totalPoints == 18 then realLevel = "43/44"
			elseif totalPoints == 19 then realLevel = "45/46"
			elseif totalPoints == 20 then realLevel = "47/48"
			elseif totalPoints == 21 then realLevel = "49/50"
			elseif totalPoints == 22 then realLevel = "51/52"
			elseif totalPoints == 23 then realLevel = "53/54"
			elseif totalPoints == 24 then realLevel = "55/56"
			elseif totalPoints == 25 then realLevel = "57/58"
			elseif totalPoints == 26 then realLevel = "59/60"
			elseif totalPoints == 27 then realLevel = "61/62"
			elseif totalPoints == 28 then realLevel = "63/64"
			elseif totalPoints == 29 then realLevel = "65/66"
			elseif totalPoints == 30 then realLevel = "67/68"
			elseif totalPoints == 31 then realLevel = "69/70"
			elseif totalPoints == 32 then realLevel = "71/72"
			elseif totalPoints == 33 then realLevel = "73/74"
			elseif totalPoints == 34 then realLevel = "75/76"
			elseif totalPoints == 35 then realLevel = "77/78"
			elseif totalPoints == 36 then realLevel = "79/80"
			elseif totalPoints == 37 then realLevel = "81"
			elseif totalPoints == 38 then realLevel = "82"
			elseif totalPoints == 39 then realLevel = "83"
			elseif totalPoints == 40 then realLevel = "84"
			elseif totalPoints == 41 then realLevel = "85" end
	
			SCAN[SCANGUID].level = realLevel
			CharScanner_MainFrame_Base_Level:SetText(LEVEL.." "..SCAN[SCANGUID].level.." - "..SCAN[SCANGUID].race..", "..SCAN[SCANGUID].class)
		end
		-- opposite faction level (lvl ??/skull/-1 only!!!) check END

		-- low level (level <= 9) check START
		if (talentGroup1pointsTotal + SCAN[SCANGUID].talentGroup1pointsUnspent) == 0 then
			SCAN[SCANGUID].talentGroup1name = nil
			SCAN[SCANGUID].talentGroup1role = nil
			SCAN[SCANGUID].talentGroup1pointsString = nil
			SCAN[SCANGUID].talentGroup2name = nil
			SCAN[SCANGUID].talentGroup2role = nil
			SCAN[SCANGUID].talentGroup2pointsString = nil
			SCAN[SCANGUID].talentGroups = nil
			SCAN[SCANGUID].activeTalentGroup = nil
		end
		-- low level (level <= 9) check END

		-- main role START
		SCAN[SCANGUID].mainRole = 5
		local role
		if SCAN[SCANGUID].talentGroup1name then
			if talentGroup1pointsTotal > 0 then
				role = classes[SCAN[SCANGUID].classEN].roleNum[SCAN[SCANGUID].talentGroup1role]
				if SCAN[SCANGUID].classEN == "DRUID" then
					if role == 1 and SCAN[SCANGUID].talents[1][2][3][4]+SCAN[SCANGUID].talents[1][2][5][4] < 4 then -- Thick Hide + Natural Reaction
						role = 3 -- DAMAGE
					end
				end
				SCAN[SCANGUID].mainRole = role
				SCAN[SCANGUID].talentGroup1role = role
			else
				SCAN[SCANGUID].mainRole = 4
				SCAN[SCANGUID].talentGroup1role = 4
			end
		end

		if SCAN[SCANGUID].talentGroups and SCAN[SCANGUID].talentGroups > 1 then
			if talentGroup2pointsTotal > 0 then
				role = classes[SCAN[SCANGUID].classEN].roleNum[SCAN[SCANGUID].talentGroup2role]
				if SCAN[SCANGUID].classEN == "DRUID" then
					if role == 1 and SCAN[SCANGUID].talents[2][2][3][4]+SCAN[SCANGUID].talents[2][2][5][4] < 4 then -- Thick Hide + Natural Reaction
						role = 3 -- DAMAGE
					end
				end
				SCAN[SCANGUID].talentGroup2role = role
			else
				SCAN[SCANGUID].talentGroup2role = 4
			end
			if SCAN[SCANGUID].activeTalentGroup == 2 then
				SCAN[SCANGUID].mainRole = SCAN[SCANGUID].talentGroup2role
			end
		end
		-- main role END

		if IsUnitInpected then
			CharScanner:ItemLevelCalc(SCANUNIT, "Inspect")
		else
			CharScanner:ItemLevelCalc("player", "Character")
		end

		CharScanner:ShowData(3, SCANGUID, true)
		CharScanner:Unregister_INSPECT_READY()
	elseif event == "INSPECT_ACHIEVEMENT_READY" then
		-- Achievement/Talent scan restriction START
		local currTime = GetTime()
		if (currTime - RestrictionTime) > RestrictionMaxTime or not RestrictionFirstTime then
			RestrictionAchNum = 0
			RestrictionTalNum = 0
		end
		RestrictionTime = currTime
		RestrictionAchNum = RestrictionAchNum + 1
		RestrictionFirstTime = true
		-- Achievement/Talent scan restriction END

		if not Event_INSPECT_ACHIEVEMENT_READY then return end

		-- PvE achievement START
		SCAN[SCANGUID].PVEDATA.achievementID = {}
		for i = 1, #PVEDUMP do
			SCAN[SCANGUID].PVEDATA[i] = {}
			SCAN[SCANGUID].PVEDATA[i].achievementCount = 0
			if PVEDUMP[i].achievement then
				for key, value in pairs(PVEDUMP[i].achievement) do
					if GetAchievementComparisonInfo(value, 1) then
						SCAN[SCANGUID].PVEDATA[i].achievementCount = SCAN[SCANGUID].PVEDATA[i].achievementCount + 1
						SCAN[SCANGUID].PVEDATA.achievementID[value] = 1
					end
				end
				SCAN[SCANGUID].PvEAchievementTotal = SCAN[SCANGUID].PvEAchievementTotal + PVEDUMP[i].achievementTotal
				SCAN[SCANGUID].PvEAchievementCount = SCAN[SCANGUID].PvEAchievementCount + SCAN[SCANGUID].PVEDATA[i].achievementCount
			else

				for j = 1, #PVEDUMP[i] do
					SCAN[SCANGUID].PVEDATA[i][j] = {}
					SCAN[SCANGUID].PVEDATA[i][j].achievementCount = 0
					if PVEDUMP[i][j].achievement then
						for key, value in pairs(PVEDUMP[i][j].achievement) do
							if GetAchievementComparisonInfo(value, 1) then
								SCAN[SCANGUID].PVEDATA[i].achievementCount = SCAN[SCANGUID].PVEDATA[i].achievementCount + 1
								SCAN[SCANGUID].PVEDATA[i][j].achievementCount = SCAN[SCANGUID].PVEDATA[i][j].achievementCount + 1
								SCAN[SCANGUID].PVEDATA.achievementID[value] = 1
							end
						end
						SCAN[SCANGUID].PvEAchievementTotal = SCAN[SCANGUID].PvEAchievementTotal + PVEDUMP[i][j].achievementTotal
						SCAN[SCANGUID].PvEAchievementCount = SCAN[SCANGUID].PvEAchievementCount + SCAN[SCANGUID].PVEDATA[i][j].achievementCount
					else

						for k = 1, #PVEDUMP[i][j] do
							SCAN[SCANGUID].PVEDATA[i][j][k] = {}
							SCAN[SCANGUID].PVEDATA[i][j][k].achievementCount = 0
							if PVEDUMP[i][j][k].achievement then
								for key, value in pairs(PVEDUMP[i][j][k].achievement) do
									if GetAchievementComparisonInfo(value, 1) then
										SCAN[SCANGUID].PVEDATA[i].achievementCount = SCAN[SCANGUID].PVEDATA[i].achievementCount + 1
										SCAN[SCANGUID].PVEDATA[i][j].achievementCount = SCAN[SCANGUID].PVEDATA[i][j].achievementCount + 1
										SCAN[SCANGUID].PVEDATA[i][j][k].achievementCount = SCAN[SCANGUID].PVEDATA[i][j][k].achievementCount + 1
										SCAN[SCANGUID].PVEDATA.achievementID[value] = 1
									end
								end
								SCAN[SCANGUID].PvEAchievementTotal = SCAN[SCANGUID].PvEAchievementTotal + PVEDUMP[i][j][k].achievementTotal
								SCAN[SCANGUID].PvEAchievementCount = SCAN[SCANGUID].PvEAchievementCount + SCAN[SCANGUID].PVEDATA[i][j][k].achievementCount
							--else
							end
						end

					end

				end

			end
		end
		-- PvE achievement END

		-- PvP achievement START
		SCAN[SCANGUID].PVPDATA.achievementID = {}
		for i = 1, #PVPDUMP do
			SCAN[SCANGUID].PVPDATA[i] = {}
			SCAN[SCANGUID].PVPDATA[i].achievementCount = 0
			if PVPDUMP[i].achievement then
				for key, value in pairs(PVPDUMP[i].achievement) do
  				local valID
					if value.id then
						valID = value.id
					else
						if SCAN[SCANGUID].faction == "Horde" then
							valID = value.Hid
						else
							valID = value.Aid
						end
					end
					if GetAchievementComparisonInfo(valID, 1) then
						SCAN[SCANGUID].PVPDATA[i].achievementCount = SCAN[SCANGUID].PVPDATA[i].achievementCount + 1
						SCAN[SCANGUID].PVPDATA.achievementID[valID] = 1
					end
				end
				SCAN[SCANGUID].PvPAchievementTotal = SCAN[SCANGUID].PvPAchievementTotal + PVPDUMP[i].achievementTotal
				SCAN[SCANGUID].PvPAchievementCount = SCAN[SCANGUID].PvPAchievementCount + SCAN[SCANGUID].PVPDATA[i].achievementCount
			else

				for j = 1, #PVPDUMP[i] do
					SCAN[SCANGUID].PVPDATA[i][j] = {}
					SCAN[SCANGUID].PVPDATA[i][j].achievementCount = 0
					if PVPDUMP[i][j].achievement then
						for key, value in pairs(PVPDUMP[i][j].achievement) do
  						local valID
							if value.id then
								valID = value.id
							else
								if SCAN[SCANGUID].faction == "Horde" then
									valID = value.Hid
								else
									valID = value.Aid
								end
							end
							if GetAchievementComparisonInfo(valID, 1) then
								SCAN[SCANGUID].PVPDATA[i].achievementCount = SCAN[SCANGUID].PVPDATA[i].achievementCount + 1
								SCAN[SCANGUID].PVPDATA[i][j].achievementCount = SCAN[SCANGUID].PVPDATA[i][j].achievementCount + 1
								SCAN[SCANGUID].PVPDATA.achievementID[valID] = 1
							end
						end
						SCAN[SCANGUID].PvPAchievementTotal = SCAN[SCANGUID].PvPAchievementTotal + PVPDUMP[i][j].achievementTotal
						SCAN[SCANGUID].PvPAchievementCount = SCAN[SCANGUID].PvPAchievementCount + SCAN[SCANGUID].PVPDATA[i][j].achievementCount
					end
				end

			end

		end
		-- PvP achievement END

		-- PvE statistic START
		for i = 1, #PVEDUMP do
			SCAN[SCANGUID].PVEDATA[i].statisticBoss = 0
			SCAN[SCANGUID].PVEDATA[i].statisticBossAll = 0
			if PVEDUMP[i].statistic then
				SCAN[SCANGUID].PVEDATA[i].statisticBossEach = {}
				for key, value in pairs(PVEDUMP[i].statistic) do
					local q = BossStatistic(value, IsUnitInpected)
					SCAN[SCANGUID].PVEDATA[i].statisticBossEach[key] = q
					if q > 0 then
						SCAN[SCANGUID].PVEDATA[i].statisticBoss = SCAN[SCANGUID].PVEDATA[i].statisticBoss + 1
						SCAN[SCANGUID].PVEDATA[i].statisticBossAll = SCAN[SCANGUID].PVEDATA[i].statisticBossAll + q
					end
				end
				SCAN[SCANGUID].PvEStatisticTotal = SCAN[SCANGUID].PvEStatisticTotal + PVEDUMP[i].statisticTotal
				SCAN[SCANGUID].PvEStatisticCount = SCAN[SCANGUID].PvEStatisticCount + SCAN[SCANGUID].PVEDATA[i].statisticBoss
				SCAN[SCANGUID].PvEStatisticBossKills = SCAN[SCANGUID].PvEStatisticBossKills + SCAN[SCANGUID].PVEDATA[i].statisticBossAll
			else

				for j = 1, #PVEDUMP[i] do
					SCAN[SCANGUID].PVEDATA[i][j].statisticBoss = 0
					SCAN[SCANGUID].PVEDATA[i][j].statisticBossAll = 0
					if PVEDUMP[i][j].statistic then
						SCAN[SCANGUID].PVEDATA[i][j].statisticBossEach = {}
						for key, value in pairs(PVEDUMP[i][j].statistic) do
							local q = BossStatistic(value, IsUnitInpected)
							SCAN[SCANGUID].PVEDATA[i][j].statisticBossEach[key] = q
							if q > 0 then
								SCAN[SCANGUID].PVEDATA[i].statisticBoss = SCAN[SCANGUID].PVEDATA[i].statisticBoss + 1
								SCAN[SCANGUID].PVEDATA[i].statisticBossAll = SCAN[SCANGUID].PVEDATA[i].statisticBossAll + q
								SCAN[SCANGUID].PVEDATA[i][j].statisticBoss = SCAN[SCANGUID].PVEDATA[i][j].statisticBoss + 1
								SCAN[SCANGUID].PVEDATA[i][j].statisticBossAll = SCAN[SCANGUID].PVEDATA[i][j].statisticBossAll + q
							end
						end
						SCAN[SCANGUID].PvEStatisticTotal = SCAN[SCANGUID].PvEStatisticTotal + PVEDUMP[i][j].statisticTotal
						SCAN[SCANGUID].PvEStatisticCount = SCAN[SCANGUID].PvEStatisticCount + SCAN[SCANGUID].PVEDATA[i][j].statisticBoss
						SCAN[SCANGUID].PvEStatisticBossKills = SCAN[SCANGUID].PvEStatisticBossKills + SCAN[SCANGUID].PVEDATA[i][j].statisticBossAll
					else

						for k = 1, #PVEDUMP[i][j] do
							SCAN[SCANGUID].PVEDATA[i][j][k].statisticBoss = 0
							SCAN[SCANGUID].PVEDATA[i][j][k].statisticBossAll = 0
							if PVEDUMP[i][j][k].statistic then
								SCAN[SCANGUID].PVEDATA[i][j][k].statisticBossEach = {}
								for key, value in pairs(PVEDUMP[i][j][k].statistic) do
									local q = BossStatistic(value, IsUnitInpected)
									SCAN[SCANGUID].PVEDATA[i][j][k].statisticBossEach[key] = q
									if q > 0 then
										SCAN[SCANGUID].PVEDATA[i].statisticBoss = SCAN[SCANGUID].PVEDATA[i].statisticBoss + 1
										SCAN[SCANGUID].PVEDATA[i].statisticBossAll = SCAN[SCANGUID].PVEDATA[i].statisticBossAll + q
										SCAN[SCANGUID].PVEDATA[i][j].statisticBoss = SCAN[SCANGUID].PVEDATA[i][j].statisticBoss + 1
										SCAN[SCANGUID].PVEDATA[i][j].statisticBossAll = SCAN[SCANGUID].PVEDATA[i][j].statisticBossAll + q
										SCAN[SCANGUID].PVEDATA[i][j][k].statisticBoss = SCAN[SCANGUID].PVEDATA[i][j][k].statisticBoss + 1
										SCAN[SCANGUID].PVEDATA[i][j][k].statisticBossAll = SCAN[SCANGUID].PVEDATA[i][j][k].statisticBossAll + q
									end
								end
								SCAN[SCANGUID].PvEStatisticTotal = SCAN[SCANGUID].PvEStatisticTotal + PVEDUMP[i][j][k].statisticTotal
								SCAN[SCANGUID].PvEStatisticCount = SCAN[SCANGUID].PvEStatisticCount + SCAN[SCANGUID].PVEDATA[i][j][k].statisticBoss
								SCAN[SCANGUID].PvEStatisticBossKills = SCAN[SCANGUID].PvEStatisticBossKills + SCAN[SCANGUID].PVEDATA[i][j][k].statisticBossAll
							--else
							end
						end

					end

				end

			end
		end
		-- PvE statistic END

		-- PvP statistic START
		for i = 1, #PVPDUMP do
			SCAN[SCANGUID].PVPDATA[i].wins = 0
			SCAN[SCANGUID].PVPDATA[i].matches = 0
			SCAN[SCANGUID].PVPDATA[i].kills = 0
			if PVPDUMP[i].statistic then
				for key, value in pairs(PVPDUMP[i].statistic) do
					if value.wins then
						local q = BossStatistic(value.wins, IsUnitInpected)
						SCAN[SCANGUID].PVPDATA[i].wins = SCAN[SCANGUID].PVPDATA[i].wins + q
					elseif value.matches then
						local q = BossStatistic(value.matches, IsUnitInpected)
						SCAN[SCANGUID].PVPDATA[i].matches = SCAN[SCANGUID].PVPDATA[i].matches + q
					elseif value.kills then
						local q = BossStatistic(value.kills, IsUnitInpected)
						SCAN[SCANGUID].PVPDATA[i].kills = SCAN[SCANGUID].PVPDATA[i].kills + q
					elseif value.totalHK then
						local q = BossStatistic(value.totalHK, IsUnitInpected)
						SCAN[SCANGUID].PvPStatisticTotalKills = q
					end
				end
				SCAN[SCANGUID].PvPStatisticWins    = SCAN[SCANGUID].PvPStatisticWins    + SCAN[SCANGUID].PVPDATA[i].wins
				SCAN[SCANGUID].PvPStatisticMatches = SCAN[SCANGUID].PvPStatisticMatches + SCAN[SCANGUID].PVPDATA[i].matches
				SCAN[SCANGUID].PvPStatisticKills   = SCAN[SCANGUID].PvPStatisticKills   + SCAN[SCANGUID].PVPDATA[i].kills
			else

				for j = 1, #PVPDUMP[i] do
					SCAN[SCANGUID].PVPDATA[i][j].wins = 0
					SCAN[SCANGUID].PVPDATA[i][j].matches = 0
					SCAN[SCANGUID].PVPDATA[i][j].kills = 0
					if PVPDUMP[i][j].statistic then
						for key, value in pairs(PVPDUMP[i][j].statistic) do
							if value.wins then
								local q = BossStatistic(value.wins, IsUnitInpected)
								SCAN[SCANGUID].PVPDATA[i][j].wins = SCAN[SCANGUID].PVPDATA[i][j].wins + q
							elseif value.matches then
								local q = BossStatistic(value.matches, IsUnitInpected)
								SCAN[SCANGUID].PVPDATA[i][j].matches = SCAN[SCANGUID].PVPDATA[i][j].matches + q
							elseif value.kills then
								local q = BossStatistic(value.kills, IsUnitInpected)
								SCAN[SCANGUID].PVPDATA[i][j].kills = SCAN[SCANGUID].PVPDATA[i][j].kills + q
							elseif value.misc then
								local q = BossStatistic(value.misc, IsUnitInpected)
								if not SCAN[SCANGUID].PVPDATA[i][j].misc then
									SCAN[SCANGUID].PVPDATA[i][j].misc = {}
								end
								SCAN[SCANGUID].PVPDATA[i][j].misc[key] = q
							end
						end
						if PVPDUMP[i].name == "Battlegrounds" then
							SCAN[SCANGUID].PvPBattlegroundWins    = SCAN[SCANGUID].PvPBattlegroundWins    + SCAN[SCANGUID].PVPDATA[i][j].wins
							SCAN[SCANGUID].PvPBattlegroundMatches = SCAN[SCANGUID].PvPBattlegroundMatches + SCAN[SCANGUID].PVPDATA[i][j].matches
							SCAN[SCANGUID].PvPBattlegroundKills   = SCAN[SCANGUID].PvPBattlegroundKills   + SCAN[SCANGUID].PVPDATA[i][j].kills
						elseif PVPDUMP[i].name == "Arena" then
							SCAN[SCANGUID].PvPArenaWins    = SCAN[SCANGUID].PvPArenaWins    + SCAN[SCANGUID].PVPDATA[i][j].wins
							SCAN[SCANGUID].PvPArenaMatches = SCAN[SCANGUID].PvPArenaMatches + SCAN[SCANGUID].PVPDATA[i][j].matches
							SCAN[SCANGUID].PvPArenaKills   = SCAN[SCANGUID].PvPArenaKills   + SCAN[SCANGUID].PVPDATA[i][j].kills
						end
						SCAN[SCANGUID].PvPStatisticWins    = SCAN[SCANGUID].PvPStatisticWins    + SCAN[SCANGUID].PVPDATA[i][j].wins
						SCAN[SCANGUID].PvPStatisticMatches = SCAN[SCANGUID].PvPStatisticMatches + SCAN[SCANGUID].PVPDATA[i][j].matches
						SCAN[SCANGUID].PvPStatisticKills   = SCAN[SCANGUID].PvPStatisticKills   + SCAN[SCANGUID].PVPDATA[i][j].kills
					end
				end

			end
		end
		-- PvP statistic END

		CharScanner:ShowData(4, SCANGUID, true)
		CharScanner:Unregister_INSPECT_ACHIEVEMENT_READY()
	elseif event == "RAID_ROSTER_UPDATE" then
		if inCombat then
			inCombatGrpChk = true
			return
		end
		local raidmembers = GetNumRaidMembers()
		if rosterhack ~= raidmembers then -- avoid execute this block if raidmembers are just moved around
			if raidmembers > 1 then -- 1, not 0. Think the only situation where a 1-member-raid can exist is at the end of a battleground.
				inPartyOrRaid = true
			else
				inPartyOrRaid = false
			end	
			rosterhack = raidmembers
			CharScanner:GroupScanToggle()
		end
	elseif event == "PARTY_MEMBERS_CHANGED" then
		if inCombat then
			inCombatGrpChk = true
			return
		end
		if GetNumRaidMembers() > 0 then return end -- avoid execute this block if we are in raid
		if GetNumPartyMembers() > 0 then
			inPartyOrRaid = true
		else
			inPartyOrRaid = false
		end
		CharScanner:GroupScanToggle()
	elseif event == "PLAYER_ENTERING_WORLD" then
		if GroupScanIsActive then
			CharScanner:StopGroupScan()
		end
		if SingleScanIsActive then
			CharScanner:StopSingleScan(true)
		end
		if not CharScanner_Options.StartMessage then
			Print(L["Please read CharScanner.lua for more information."])
			Print(L["Scan access via SCAN-Button, Minimap-Button, LDB, Blizzard Interface Options or '/scan' resp. '/scan options'."])
			CharScanner_Options.StartMessage = true
		end
	elseif event == "PLAYER_LOGIN" then
		LoadAddOn("Blizzard_InspectUI") -- we need this
		playerGUID = UnitGUID("player")
		playerFaction = UnitFactionGroup("player")
		BuildIDTable()
		CharScanner:InitializeOptions()
		CharScanner:CreateInterfaceOptions()
		CharScanner:LDBcheck()
		CharScanner:CreateMinimapButton()
		CharScanner:CreateMainFrame()
		CharScanner:CreateScanButtonFrame()
		CharScanner:CreateMiniOptionsFrame()
		CharScanner:CreateOptionsFrame()
		CharScanner:InitializeFrame()
		table.insert(UISpecialFrames, "CharScanner_MainFrame")
		table.insert(UISpecialFrames, "CharScanner_OptionsFrame")
		CharScanner:UnregisterEvent("PLAYER_LOGIN")
	end
end
-- ---------------------------------------------------------------------------------------------------------------------
CharScanner:RegisterEvent("PLAYER_LOGIN")
CharScanner:RegisterEvent("PLAYER_ENTERING_WORLD")
CharScanner:RegisterEvent("PLAYER_REGEN_DISABLED")
CharScanner:RegisterEvent("PLAYER_REGEN_ENABLED")
CharScanner:RegisterEvent("PLAYER_TARGET_CHANGED")
CharScanner:RegisterEvent("RAID_ROSTER_UPDATE")
CharScanner:RegisterEvent("PARTY_MEMBERS_CHANGED")
CharScanner:RegisterEvent("INSPECT_ACHIEVEMENT_READY")
CharScanner:RegisterEvent("INSPECT_READY")
CharScanner:SetScript("OnEvent", OnEvent) -- start the Mystical Machine